@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2010-2014 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2015 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -9,20 +9,28 @@ import os
 import sys
 import socket
 import binascii
+import traceback
 import string
 from xpra.gtk_common.gobject_compat import import_gobject, import_glib
 gobject = import_gobject()
 
 from xpra.log import Logger
 log = Logger("client")
+printlog = Logger("printing")
+filelog = Logger("file")
+netlog = Logger("network")
+authlog = Logger("auth")
 
-from xpra.net.protocol import Protocol, get_network_caps
-from xpra.scripts.config import ENCRYPTION_CIPHERS
-from xpra.version_util import version_compat_check, get_version_info, get_platform_info, local_version
-from xpra.platform.features import GOT_PASSWORD_PROMPT_SUGGESTION
+from xpra.scripts.config import InitExit
+from xpra.child_reaper import getChildReaper, reaper_cleanup
+from xpra.net.protocol import Protocol, get_network_caps, sanity_checks
+from xpra.net.crypto import get_iterations, get_iv, get_salt, choose_padding, \
+    ENCRYPTION_CIPHERS, ENCRYPT_FIRST_PACKET, DEFAULT_IV, DEFAULT_SALT, DEFAULT_ITERATIONS, INITIAL_PADDING, DEFAULT_PADDING, ALL_PADDING_OPTIONS, PADDING_OPTIONS
+from xpra.version_util import version_compat_check, get_version_info, local_version
 from xpra.platform.info import get_name
 from xpra.os_util import get_hex_uuid, get_machine_id, get_user_uuid, load_binary_file, SIGNAMES, strtobytes, bytestostr
-from xpra.util import b, ss, typedict, updict, xor, repr_ellipsized, nonl, disconnect_is_an_error
+from xpra.util import typedict, updict, xor, repr_ellipsized, nonl, disconnect_is_an_error
+from xpra.net.file_transfer import HAS_PRINTING, FileTransferHandler
 
 EXIT_OK = 0
 EXIT_CONNECTION_LOST = 1
@@ -39,14 +47,16 @@ EXIT_NO_AUTHENTICATION = 11
 EXIT_UNSUPPORTED = 12
 EXIT_REMOTE_ERROR = 13
 EXIT_INTERNAL_ERROR = 14
+EXIT_FILE_TOO_BIG = 15
 
 
 EXTRA_TIMEOUT = 10
 ALLOW_UNENCRYPTED_PASSWORDS = os.environ.get("XPRA_ALLOW_UNENCRYPTED_PASSWORDS", "0")=="1"
 DETECT_LEAKS = os.environ.get("XPRA_DETECT_LEAKS", "0")=="1"
+DELETE_PRINTER_FILE = os.environ.get("XPRA_DELETE_PRINTER_FILE", "1")=="1"
 
 
-class XpraClientBase(object):
+class XpraClientBase(FileTransferHandler):
     """ Base class for Xpra clients.
         Provides the glue code for:
         * sending packets via Protocol
@@ -58,13 +68,21 @@ class XpraClientBase(object):
     """
 
     def __init__(self):
+        FileTransferHandler.__init__(self)
         #this may be called more than once,
         #skip doing internal init again:
         if not hasattr(self, "exit_code"):
             self.defaults_init()
 
     def defaults_init(self):
+        getChildReaper()
+        log("XpraClientBase.defaults_init() os.environ:")
+        for k,v in os.environ.items():
+            log(" %s=%s", k, nonl(v))
+        #client state:
         self.exit_code = None
+        self.exit_on_signal = False
+        #connection attributes:
         self.compression_level = 0
         self.display = None
         self.username = None
@@ -72,10 +90,14 @@ class XpraClientBase(object):
         self.password_sent = False
         self.encryption = None
         self.encryption_keyfile = None
+        self.server_padding_options = [DEFAULT_PADDING]
         self.quality = -1
         self.min_quality = 0
         self.speed = 0
         self.min_speed = -1
+        self.printer_attributes = []
+        self.send_printers_pending = False
+        self.exported_printers = None
         #protocol stuff:
         self._protocol = None
         self._priority_packets = []
@@ -95,6 +117,7 @@ class XpraClientBase(object):
         self._remote_platform_linux_distribution = None
         self.uuid = get_user_uuid()
         self.init_packet_handlers()
+        sanity_checks()
 
     def init(self, opts):
         self.compression_level = opts.compression_level
@@ -107,6 +130,8 @@ class XpraClientBase(object):
         self.min_quality = opts.min_quality
         self.speed = opts.speed
         self.min_speed = opts.min_speed
+        #printing and file transfer:
+        FileTransferHandler.init(self, opts)
 
         if DETECT_LEAKS:
             from xpra.util import detect_leaks
@@ -142,23 +167,53 @@ class XpraClientBase(object):
             sys.stderr.flush()
             signal.signal(signal.SIGINT, deadly_signal)
             signal.signal(signal.SIGTERM, deadly_signal)
-            self.timeout_add(0, self.disconnect_and_quit, 128 + signum, "exit on signal %s" % SIGNAMES.get(signum, signum))
-        signal.signal(signal.SIGINT, app_signal)
+            self.signal_cleanup()
+            self.timeout_add(0, self.signal_disconnect_and_quit, 128 + signum, "exit on signal %s" % SIGNAMES.get(signum, signum))
+        if sys.version_info[0]<3:
+            #breaks GTK3..
+            signal.signal(signal.SIGINT, app_signal)
         signal.signal(signal.SIGTERM, app_signal)
 
+    def signal_disconnect_and_quit(self, exit_code, reason):
+        log("signal_disconnect_and_quit(%s, %s) exit_on_signal=%s", exit_code, reason, self.exit_on_signal)
+        if not self.exit_on_signal:
+            #if we get another signal, we'll try to exit without idle_add...
+            self.exit_on_signal = True
+            self.idle_add(self.disconnect_and_quit, exit_code, reason)
+            self.idle_add(self.quit, exit_code)
+            self.idle_add(self.exit)
+            return
+        #warning: this will run cleanup code from the signal handler
+        self.disconnect_and_quit(exit_code, reason)
+        self.quit(exit_code)
+        self.exit()
+        os._exit(exit_code)
+
+    def signal_cleanup(self):
+        #placeholder for stuff that can be cleaned up from the signal handler
+        #(non UI thread stuff)
+        pass
+
     def disconnect_and_quit(self, exit_code, reason):
+        #make sure that we set the exit code early,
+        #so the protocol shutdown won't set a different one:
+        if self.exit_code is None:
+            self.exit_code = exit_code
         #try to tell the server we're going, then quit
         log("disconnect_and_quit(%s, %s)", exit_code, reason)
         p = self._protocol
         if p is None or p._closed:
             self.quit(exit_code)
             return
-        def do_quit():
-            log("disconnect_and_quit: do_quit()")
+        def protocol_closed():
+            log("disconnect_and_quit: protocol_closed()")
             self.idle_add(self.quit, exit_code)
         if p:
-            p.flush_then_close(["disconnect", reason], done_callback=do_quit)
-        self.timeout_add(1000, do_quit)
+            p.flush_then_close(["disconnect", reason], done_callback=protocol_closed)
+        self.timeout_add(1000, self.quit, exit_code)
+
+    def exit(self):
+        sys.exit()
 
 
     def client_type(self):
@@ -169,34 +224,58 @@ class XpraClientBase(object):
         raise NotImplementedError()
 
     def setup_connection(self, conn):
-        log.debug("setup_connection(%s)", conn)
+        netlog("setup_connection(%s) timeout=%s", conn, conn.timeout)
         self._protocol = Protocol(self.get_scheduler(), conn, self.process_packet, self.next_packet)
         self._protocol.large_packets.append("keymap-changed")
         self._protocol.large_packets.append("server-settings")
+        self._protocol.large_packets.append("logging")
         self._protocol.set_compression_level(self.compression_level)
         self._protocol.receive_aliases.update(self._aliases)
         self._protocol.enable_default_encoder()
         self._protocol.enable_default_compressor()
+        if self.encryption and ENCRYPT_FIRST_PACKET:
+            password = self.get_encryption_key()
+            self._protocol.set_cipher_out(self.encryption, DEFAULT_IV, password, DEFAULT_SALT, DEFAULT_ITERATIONS, INITIAL_PADDING)
         self.have_more = self._protocol.source_has_more
         if conn.timeout>0:
             self.timeout_add((conn.timeout + EXTRA_TIMEOUT) * 1000, self.verify_connected)
+        netlog("setup_connection(%s) protocol=%s", conn, self._protocol)
+
+
+    def remove_packet_handlers(self, *keys):
+        for k in keys:
+            for d in (self._packet_handlers, self._ui_packet_handlers):
+                try:
+                    del d[k]
+                except:
+                    pass
+
+    def set_packet_handlers(self, to, defs):
+        """ configures the given packet handlers,
+            and make sure we remove any existing ones with the same key
+            (which can be useful for subclasses, not here)
+        """
+        log("set_packet_handlers(%s, %s)", to, defs)
+        self.remove_packet_handlers(*defs.keys())
+        for k,v in defs.items():
+            to[k] = v
 
     def init_packet_handlers(self):
-        self._packet_handlers = {
-            "hello": self._process_hello,
-            }
-        self._ui_packet_handlers = {
+        self._packet_handlers = {}
+        self._ui_packet_handlers = {}
+        self.set_packet_handlers(self._packet_handlers, {"hello" : self._process_hello})
+        self.set_packet_handlers(self._ui_packet_handlers, {
             "challenge":                self._process_challenge,
             "disconnect":               self._process_disconnect,
             "set_deflate":              self._process_set_deflate,
+            "startup-complete":         self._process_startup_complete,
             Protocol.CONNECTION_LOST:   self._process_connection_lost,
             Protocol.GIBBERISH:         self._process_gibberish,
             Protocol.INVALID:           self._process_invalid,
-            }
+            })
 
     def init_authenticated_packet_handlers(self):
-        #for subclasses to override
-        pass
+        self.set_packet_handlers(self._packet_handlers, {"send-file" : self._process_send_file})
 
 
     def init_aliases(self):
@@ -208,25 +287,31 @@ class XpraClientBase(object):
             self._reverse_aliases[key] = i
             i += 1
 
+
     def send_hello(self, challenge_response=None, client_salt=None):
         try:
             hello = self.make_hello_base()
-            if self.password_file and not challenge_response:
+            if (self.password_file or os.environ.get('XPRA_PASSWORD')) and not challenge_response:
                 #avoid sending the full hello: tell the server we want
                 #a packet challenge first
                 hello["challenge"] = True
             else:
                 hello.update(self.make_hello())
-        except Exception, e:
-            log.error("error preparing connection: %s", e)
+        except InitExit as e:
+            log.error("error preparing connection:")
+            log.error(" %s", e)
+            self.quit(EXIT_INTERNAL_ERROR)
+            return
+        except Exception as e:
+            log.error("error preparing connection: %s", e, exc_info=True)
             self.quit(EXIT_INTERNAL_ERROR)
             return
         if challenge_response:
-            assert self.password_file
+            assert self.password_file or os.environ.get('XPRA_PASSWORD')
             hello["challenge_response"] = challenge_response
             if client_salt:
                 hello["challenge_client_salt"] = client_salt
-        log.debug("send_hello(%s) packet=%s", binascii.hexlify(b(challenge_response or "")), hello)
+        log("send_hello(%s) packet=%s", binascii.hexlify(strtobytes(challenge_response or "")), hello)
         self.send("hello", hello)
 
     def verify_connected(self):
@@ -248,36 +333,43 @@ class XpraClientBase(object):
                 "client_type"           : self.client_type(),
                 "python.version"        : sys.version_info[:3],
                 "compression_level"     : self.compression_level,
+                "argv"                  : sys.argv,
                 })
+        capabilities.update(self.get_file_transfer_features())
         if self.display:
             capabilities["display"] = self.display
         def up(prefix, d):
             updict(capabilities, prefix, d)
-        up("platform",  get_platform_info())
-        up("build",     get_version_info())
+        up("build",     self.get_version_info())
         mid = get_machine_id()
         if mid:
             capabilities["machine_id"] = mid
 
         if self.encryption:
             assert self.encryption in ENCRYPTION_CIPHERS
-            iv = get_hex_uuid()[:16]
-            key_salt = get_hex_uuid()+get_hex_uuid()
-            iterations = 1000
-            capabilities.update({
-                        "cipher"                       : self.encryption,
-                        "cipher.iv"                    : iv,
-                        "cipher.key_salt"              : key_salt,
-                        "cipher.key_stretch_iterations": iterations,
-                        })
+            iv = get_iv()
+            key_salt = get_salt()
+            iterations = get_iterations()
+            padding = choose_padding(self.server_padding_options)
+            up("cipher", {
+                    ""                      : self.encryption,
+                    "iv"                    : iv,
+                    "key_salt"              : key_salt,
+                    "key_stretch_iterations": iterations,
+                    "padding"               : padding,
+                    "padding.options"       : PADDING_OPTIONS,
+                    })
             key = self.get_encryption_key()
             if key is None:
                 self.warn_and_quit(EXIT_ENCRYPTION, "encryption key is missing")
                 return
-            self._protocol.set_cipher_in(self.encryption, iv, key, key_salt, iterations)
-            log("encryption capabilities: %s", [(k,v) for k,v in capabilities.items() if k.startswith("cipher")])
+            self._protocol.set_cipher_in(self.encryption, iv, key, key_salt, iterations, padding)
+            netlog("encryption capabilities: %s", dict((k,v) for k,v in capabilities.items() if k.startswith("cipher")))
         return capabilities
 
+    def get_version_info(self):
+        return get_version_info()
+
     def make_hello(self):
         capabilities = {
                         "randr_notify"        : False,        #only client.py cares about this
@@ -327,10 +419,22 @@ class XpraClientBase(object):
 
 
     def cleanup(self):
-        log("XpraClientBase.cleanup() protocol=%s", self._protocol)
-        if self._protocol:
-            self._protocol.close()
+        reaper_cleanup()
+        p = self._protocol
+        log("XpraClientBase.cleanup() protocol=%s", p)
+        if p:
+            log("calling %s", p.close)
+            p.close()
             self._protocol = None
+        self.cleanup_printing()
+        frames = sys._current_frames()
+        log("after cleanup, found %s frames:", len(frames))
+        for i,(fid,frame) in enumerate(frames.items()):
+            log("%i: %s - %s:", i, fid, frame)
+            for x in traceback.format_stack(frame):
+                for l in x.splitlines():
+                    log("%s", l)
+
 
     def glib_init(self):
         try:
@@ -349,17 +453,18 @@ class XpraClientBase(object):
     def quit(self, exit_code=0):
         raise Exception("override me!")
 
-    def warn_and_quit(self, exit_code, warning):
-        log.warn(warning)
+    def warn_and_quit(self, exit_code, message):
+        log.warn(message)
         self.quit(exit_code)
 
+
     def _process_disconnect(self, packet):
         #ie: ("disconnect", "version error", "incompatible version")
-        reason = ss(packet[1])
+        reason = bytestostr(packet[1])
         info = packet[2:]
         s = nonl(reason)
         if len(info):
-            s += " (%s)" % (", ".join([nonl(ss(x)) for x in info]))
+            s += " (%s)" % (", ".join([nonl(bytestostr(x)) for x in info]))
         if self.server_capabilities is None or len(self.server_capabilities)==0:
             #server never sent hello to us - so disconnect is an error
             #(but we don't know which one - the info message may help)
@@ -369,7 +474,12 @@ class XpraClientBase(object):
             log.warn("server failure: %s", reason)
             e = EXIT_FAILURE
         else:
-            e = EXIT_OK
+            if self.exit_code is None:
+                #we're not in the process of exiting already,
+                #tell the user why the server is disconnecting us
+                log.info("server requested disconnect: %s", s)
+            self.quit(EXIT_OK)
+            return
         self.warn_and_quit(e, "server requested disconnect: %s" % s)
 
     def _process_connection_lost(self, packet):
@@ -378,101 +488,136 @@ class XpraClientBase(object):
             props = p.get_info()
             c = props.get("compression", "unknown")
             e = props.get("encoder", "unknown")
-            log.warn("disconnected without receiving a single packet, not an xpra server?")
-            log.warn("(maybe it does not support '%s' compression or '%s' packet encoding)", c, e)
-        self.warn_and_quit(EXIT_CONNECTION_LOST, "Connection lost")
+            netlog.warn("failed to receive anything, not an xpra server?")
+            netlog.warn("  could also be the wrong username, password or port")
+            if c!="unknown" or e!="unknown":
+                netlog.warn("  or maybe this server does not support '%s' compression or '%s' packet encoding?", c, e)
+        if self.exit_code!=0:
+            self.warn_and_quit(EXIT_CONNECTION_LOST, "Connection lost")
 
     def _process_challenge(self, packet):
-        log("processing challenge: %s", packet[1:])
-        if not self.password_file:
-            self.warn_and_quit(EXIT_PASSWORD_REQUIRED, "server requires authentication, please provide a password")
+        authlog("processing challenge: %s", packet[1:])
+        def warn_server_and_exit(code, message, server_message="authentication failed"):
+            authlog.error("Error: authentication failed:")
+            authlog.error(" %s", message)
+            self.disconnect_and_quit(code, server_message)
+        if not self.password_file and not os.environ.get('XPRA_PASSWORD'):
+            warn_server_and_exit(EXIT_PASSWORD_REQUIRED, "this server requires authentication, please provide a password", "no password available")
             return
         password = self.load_password()
         if not password:
-            self.warn_and_quit(EXIT_PASSWORD_FILE_ERROR, "failed to load password from file %s" % self.password_file)
+            warn_server_and_exit(EXIT_PASSWORD_FILE_ERROR, "failed to load password from file %s" % self.password_file, "no password available")
             return
         salt = packet[1]
         if self.encryption:
             assert len(packet)>=3, "challenge does not contain encryption details to use for the response"
-            server_cipher = packet[2]
+            server_cipher = typedict(packet[2])
             key = self.get_encryption_key()
             if key is None:
-                self.warn_and_quit(EXIT_ENCRYPTION, "encryption key is missing")
+                warn_server_and_exit(EXIT_ENCRYPTION, "the server does not use any encryption", "client requires encryption")
                 return
             if not self.set_server_encryption(server_cipher, key):
                 return
-        digest = "hmac"
-        client_can_salt = len(packet)>=4
-        client_salt = None
-        if client_can_salt:
-            #server supports client salt, and tells us which digest to use:
-            digest = packet[3]
-            client_salt = get_hex_uuid()+get_hex_uuid()
-            #TODO: use some key stretching algorigthm? (meh)
+        #all server versions support a client salt,
+        #they also tell us which digest to use:
+        digest = packet[3]
+        client_salt = get_hex_uuid()+get_hex_uuid()
+        #TODO: use some key stretching algorigthm? (meh)
+        try:
+            from xpra.codecs.xor.cyxor import xor_str           #@UnresolvedImport
+            salt = xor_str(salt, client_salt)
+        except:
             salt = xor(salt, client_salt)
-        if digest=="hmac":
-            import hmac
-            challenge_response = hmac.HMAC(password, salt).hexdigest()
-        elif digest=="xor":
+        if digest==b"hmac":
+            import hmac, hashlib
+            def s(v):
+                try:
+                    return v.encode()
+                except:
+                    return str(v)
+            password = s(password)
+            salt = s(salt)
+            challenge_response = hmac.HMAC(password, salt, digestmod=hashlib.md5).hexdigest()
+        elif digest==b"xor":
             #don't send XORed password unencrypted:
             if not self._protocol.cipher_out and not ALLOW_UNENCRYPTED_PASSWORDS:
-                self.warn_and_quit(EXIT_ENCRYPTION, "server requested digest %s, cowardly refusing to use it without encryption" % digest)
+                warn_server_and_exit(EXIT_ENCRYPTION, "server requested digest %s, cowardly refusing to use it without encryption" % digest, "invalid digest")
                 return
             challenge_response = xor(password, salt)
         else:
-            self.warn_and_quit(EXIT_PASSWORD_REQUIRED, "server requested an unsupported digest: %s" % digest)
+            warn_server_and_exit(EXIT_PASSWORD_REQUIRED, "server requested an unsupported digest: %s" % digest, "invalid digest")
             return
         if digest:
-            log("%s(%s, %s)=%s", digest, password, salt, challenge_response)
+            authlog("%s(%s, %s)=%s", digest, binascii.hexlify(password), binascii.hexlify(salt), challenge_response)
         self.password_sent = True
+        self.remove_packet_handlers("challenge")
         self.send_hello(challenge_response, client_salt)
 
-    def set_server_encryption(self, capabilities, key):
-        def get(key, default=None):
-            return capabilities.get(strtobytes(key), default)
-        cipher = get("cipher")
-        cipher_iv = get("cipher.iv")
-        key_salt = get("cipher.key_salt")
-        iterations = get("cipher.key_stretch_iterations")
+    def set_server_encryption(self, caps, key):
+        cipher = caps.strget("cipher")
+        cipher_iv = caps.strget("cipher.iv")
+        key_salt = caps.strget("cipher.key_salt")
+        iterations = caps.intget("cipher.key_stretch_iterations")
+        padding = caps.strget("cipher.padding", DEFAULT_PADDING)
+        #server may tell us what it supports,
+        #either from hello response or from challenge packet:
+        self.server_padding_options = caps.strlistget("cipher.padding.options", [DEFAULT_PADDING])
         if not cipher or not cipher_iv:
             self.warn_and_quit(EXIT_ENCRYPTION, "the server does not use or support encryption/password, cannot continue with %s cipher" % self.encryption)
             return False
         if cipher not in ENCRYPTION_CIPHERS:
             self.warn_and_quit(EXIT_ENCRYPTION, "unsupported server cipher: %s, allowed ciphers: %s" % (cipher, ", ".join(ENCRYPTION_CIPHERS)))
             return False
-        self._protocol.set_cipher_out(cipher, cipher_iv, key, key_salt, iterations)
+        if padding not in ALL_PADDING_OPTIONS:
+            self.warn_and_quit(EXIT_ENCRYPTION, "unsupported server cipher padding: %s, allowed ciphers: %s" % (padding, ", ".join(ALL_PADDING_OPTIONS)))
+            return False
+        p = self._protocol
+        if not p:
+            return False
+        p.set_cipher_out(cipher, cipher_iv, key, key_salt, iterations, padding)
         return True
 
 
     def get_encryption_key(self):
         key = load_binary_file(self.encryption_keyfile)
-        if key is None and self.password_file:
+        if not key:
+            key = os.environ.get('XPRA_ENCRYPTION_KEY')
+        if not key and self.password_file:
             key = load_binary_file(self.password_file)
             if key:
-                log("used password file as encryption key")
+                netlog("used password file as encryption key")
+        if not key:
+            key = os.environ.get('XPRA_PASSWORD')
+            if key:
+                netlog("used XPRA_PASSWORD as encryption key")
         if key is None:
-            raise Exception("failed to load encryption keyfile %s" % self.encryption_keyfile)
+            raise InitExit(1, "no encryption key")
         return key.strip("\n\r")
 
     def load_password(self):
+        if not self.password_file:
+            return os.environ.get('XPRA_PASSWORD')
         filename = os.path.expanduser(self.password_file)
         password = load_binary_file(filename)
         if password is None:
             return None
-        password = password.strip("\n\r")
-        log("password read from file %s is %s", self.password_file, "".join(["*" for _ in password]))
+        try:
+            password = password.decode("utf8").strip("\n\r")
+        except:
+            password = str(password)
+        netlog("password read from file %s is %s", self.password_file, "".join(["*" for _ in password]))
         return password
 
     def _process_hello(self, packet):
-        if not self.password_sent and self.password_file:
+        if not self.password_sent and (self.password_file or os.environ.get('XPRA_PASSWORD')):
             self.warn_and_quit(EXIT_NO_AUTHENTICATION, "the server did not request our password")
             return
         try:
             self.server_capabilities = typedict(packet[1])
-            log("processing hello from server: %s", self.server_capabilities)
+            netlog("processing hello from server: %s", self.server_capabilities)
             self.server_connection_established()
-        except Exception, e:
-            log.info("error in hello packet", exc_info=True)
+        except Exception as e:
+            netlog.info("error in hello packet", exc_info=True)
             self.warn_and_quit(EXIT_FAILURE, "error processing hello packet from server: %s" % e)
 
     def capsget(self, capabilities, key, default):
@@ -483,32 +628,124 @@ class XpraClientBase(object):
 
 
     def server_connection_established(self):
-        log("server_connection_established()")
+        netlog("server_connection_established()")
         if not self.parse_version_capabilities():
-            log("server_connection_established() failed version capabilities")
+            netlog("server_connection_established() failed version capabilities")
             return False
         if not self.parse_server_capabilities():
-            log("server_connection_established() failed server capabilities")
+            netlog("server_connection_established() failed server capabilities")
             return False
         if not self.parse_network_capabilities():
-            log("server_connection_established() failed network capabilities")
+            netlog("server_connection_established() failed network capabilities")
             return False
         if not self.parse_encryption_capabilities():
-            log("server_connection_established() failed encryption capabilities")
+            netlog("server_connection_established() failed encryption capabilities")
             return False
-        log("server_connection_established() adding authenticated packet handlers")
+        self.parse_printing_capabilities()
+        netlog("server_connection_established() adding authenticated packet handlers")
         self.init_authenticated_packet_handlers()
         return True
 
 
+    def parse_printing_capabilities(self):
+        if self.printing:
+            if self.server_capabilities.boolget("printing"):
+                self.printer_attributes = self.server_capabilities.strlistget("printer.attributes", ["printer-info", "device-uri"])
+                self.timeout_add(1000, self.init_printing)
+
+    def init_printing(self):
+        if not HAS_PRINTING:
+            return
+        try:
+            from xpra.platform.printing import init_printing
+            printlog("init_printing=%s", init_printing)
+            init_printing(self.send_printers)
+            self.do_send_printers()
+        except Exception:
+            log.warn("failed to send printers", exc_info=True)
+
+    def cleanup_printing(self):
+        if not HAS_PRINTING:
+            return
+        try:
+            from xpra.platform.printing import cleanup_printing
+            printlog("cleanup_printing=%s", cleanup_printing)
+            cleanup_printing()
+        except Exception:
+            log.warn("failed to cleanup printing subsystem", exc_info=True)
+
+    def send_printers(self, *args):
+        #dbus can fire dozens of times for a single printer change
+        #so we wait a bit and fire via a timer to try to batch things together:
+        if self.send_printers_pending:
+            return
+        self.send_printers_pending = True
+        self.timeout_add(500, self.do_send_printers)
+
+    def do_send_printers(self):
+        try:
+            self.send_printers_pending = False
+            from xpra.platform.printing import get_printers, get_mimetypes
+            printers = get_printers()
+            printlog("do_send_printers() found printers=%s", printers)
+            #remove xpra-forwarded printers to avoid loops and multi-forwards,
+            #also ignore stopped printers
+            #and only keep the attributes that the server cares about (self.printer_attributes)
+            exported_printers = {}
+            def used_attrs(d):
+                #filter attributes so that we only compare things that are actually used
+                if not d:
+                    return d
+                return dict((k,v) for k,v in d.items() if k in self.printer_attributes)
+            for k,v in printers.items():
+                device_uri = v.get("device-uri", "")
+                if device_uri:
+                    #this is cups specific.. oh well
+                    printlog("do_send_printers() device-uri(%s)=%s", k, device_uri)
+                    if device_uri.startswith("xpraforwarder"):
+                        printlog("do_send_printers() skipping xpra forwarded printer=%s", k)
+                        continue
+                state = v.get("printer-state")
+                #"3" if the destination is idle,
+                #"4" if the destination is printing a job,
+                #"5" if the destination is stopped.
+                if state==5:
+                    printlog("do_send_printers() skipping stopped printer=%s", k)
+                    continue
+                attrs = used_attrs(v)
+                #add mimetypes:
+                attrs["mimetypes"] = get_mimetypes()
+                exported_printers[k.encode("utf8")] = attrs
+            if self.exported_printers is None:
+                #not been sent yet, ensure we can use the dict below:
+                self.exported_printers = {}
+            elif exported_printers==self.exported_printers:
+                printlog("do_send_printers() exported printers unchanged: %s", self.exported_printers)
+                return
+            #show summary of what has changed:
+            added = [k for k in exported_printers.keys() if k not in self.exported_printers]
+            if added:
+                printlog("do_send_printers() new printers: %s", added)
+            removed = [k for k in self.exported_printers.keys() if k not in exported_printers]
+            if removed:
+                printlog("do_send_printers() printers removed: %s", removed)
+            modified = [k for k,v in exported_printers.items() if self.exported_printers.get(k)!=v and k not in added]
+            if modified:
+                printlog("do_send_printers() printers modified: %s", modified)
+            printlog("do_send_printers() printers=%s", exported_printers.keys())
+            printlog("do_send_printers() exported printers=%s", ", ".join(str(x) for x in exported_printers.keys()))
+            self.exported_printers = exported_printers
+            self.send("printers", self.exported_printers)
+        except:
+            log.error("do_send_printers()", exc_info=True)
+
+
     def parse_version_capabilities(self):
         c = self.server_capabilities
         self._remote_machine_id = c.strget("machine_id")
         self._remote_uuid = c.strget("uuid")
-        self._remote_version = c.strget("version")
-        self._remote_version = c.strget("build.version", self._remote_version)
-        self._remote_revision = c.strget("revision")
-        self._remote_revision = c.strget("build.revision", self._remote_revision)
+        self._remote_version = c.strget("build.version", c.strget("version"))
+        self._remote_revision = c.strget("build.revision", c.strget("revision"))
         self._remote_platform = c.strget("platform")
         self._remote_platform_release = c.strget("platform.release")
         self._remote_platform_platform = c.strget("platform.platform")
@@ -531,55 +768,57 @@ class XpraClientBase(object):
 
     def parse_network_capabilities(self):
         c = self.server_capabilities
-        if not self._protocol.enable_encoder_from_caps(c):
+        p = self._protocol
+        if not p or not p.enable_encoder_from_caps(c):
             return False
-        self._protocol.enable_compressor_from_caps(c)
+        p.enable_compressor_from_caps(c)
         return True
 
     def parse_encryption_capabilities(self):
         c = self.server_capabilities
+        p = self._protocol
+        if not p:
+            return False
+        p.send_aliases = c.dictget("aliases", {})
         if self.encryption:
             #server uses a new cipher after second hello:
             key = self.get_encryption_key()
             assert key, "encryption key is missing"
             if not self.set_server_encryption(c, key):
                 return False
-        self._protocol.send_aliases = c.dictget("aliases", {})
         return True
 
     def _process_set_deflate(self, packet):
         #legacy, should not be used for anything
         pass
 
+    def _process_startup_complete(self, packet):
+        #can be received if we connect with "xpra stop" or other command line client
+        #as the server is starting up
+        pass
+
+
     def _process_gibberish(self, packet):
         (_, message, data) = packet
         p = self._protocol
-        if (p and p.input_packetcount>0) or len([c for c in data if c not in string.printable])>0:
-            log.info("Received uninterpretable nonsense: %s", message)
-            log.info(" packet no %i data: %s", p.input_packetcount, repr_ellipsized(data))
-        else:
+        show_as_text = p and p.input_packetcount==0 and all(c in string.printable for c in bytestostr(data))
+        if show_as_text:
             #looks like the first packet back is just text, print it:
             data = bytestostr(data)
             if data.find("Traceback "):
                 for x in data.split("\n"):
-                    log.warn(x.strip("\r"))
+                    netlog.warn(x.strip("\r"))
             else:
-                log.warn("Failed to connect, received: %s", repr_ellipsized(data.strip("\n").strip("\r")))
-        if str(data).find("assword")>0:
-            self.warn_and_quit(EXIT_SSH_FAILURE,
-                              "Your ssh program appears to be asking for a password."
-                             + GOT_PASSWORD_PROMPT_SUGGESTION)
-        elif str(data).find("login")>=0:
-            self.warn_and_quit(EXIT_SSH_FAILURE,
-                             "Your ssh program appears to be asking for a username.\n"
-                             "Perhaps try using something like 'ssh:USER@host:display'?")
+                netlog.warn("Failed to connect, received: %s", repr_ellipsized(data.strip("\n").strip("\r")))
         else:
-            self.quit(EXIT_PACKET_FAILURE)
+            netlog.warn("Received uninterpretable nonsense: %s", message)
+            netlog.warn(" packet no %i data: %s", p.input_packetcount, repr_ellipsized(data))
+        self.quit(EXIT_PACKET_FAILURE)
 
     def _process_invalid(self, packet):
         (_, message, data) = packet
-        log.info("Received invalid packet: %s", message)
-        log(" data: %s", repr_ellipsized(data))
+        netlog.info("Received invalid packet: %s", message)
+        netlog(" data: %s", repr_ellipsized(data))
         self.quit(EXIT_PACKET_FAILURE)
 
 
@@ -595,10 +834,10 @@ class XpraClientBase(object):
                 return
             handler = self._ui_packet_handlers.get(packet_type)
             if not handler:
-                log.error("unknown packet type: %s", packet_type)
+                netlog.error("unknown packet type: %s", packet_type)
                 return
             self.idle_add(handler, packet)
         except KeyboardInterrupt:
             raise
         except:
-            log.error("Unhandled error while processing a '%s' packet from peer using %s", packet_type, handler, exc_info=True)
+            netlog.error("Unhandled error while processing a '%s' packet from peer using %s", packet_type, handler, exc_info=True)
