@@ -1,33 +1,37 @@
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2014 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2015 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-import sys
+import sys, os
 import re
 
 from xpra.client.client_widget_base import ClientWidgetBase
-from xpra.util import typedict, bytestostr, WORKSPACE_UNSET
+from xpra.util import typedict, bytestostr, WORKSPACE_UNSET, WORKSPACE_NAMES
 from xpra.log import Logger
 log = Logger("window")
 plog = Logger("paint")
 focuslog = Logger("focus")
 mouselog = Logger("mouse")
 workspacelog = Logger("workspace")
+keylog = Logger("keyboard")
 metalog = Logger("metadata")
 
 
+REPAINT_ALL = os.environ.get("XPRA_REPAINT_ALL", "")
+
+
 class ClientWindowBase(ClientWidgetBase):
 
-    def __init__(self, client, group_leader, wid, x, y, w, h, metadata, override_redirect, client_properties, border, max_window_size):
-        log("%s%s", type(self), (client, group_leader, wid, x, y, w, h, metadata, override_redirect, client_properties, max_window_size))
+    def __init__(self, client, group_leader, wid, x, y, ww, wh, bw, bh, metadata, override_redirect, client_properties, border, max_window_size):
+        log("%s%s", type(self), (client, group_leader, wid, x, y, ww, wh, bw, bh, metadata, override_redirect, client_properties, max_window_size))
         ClientWidgetBase.__init__(self, client, wid, metadata.boolget("has-alpha"))
         self._override_redirect = override_redirect
         self.group_leader = group_leader
         self._pos = (x, y)
-        self._size = (w, h)
+        self._size = (ww, wh)
         self._client_properties = client_properties
         self._set_initial_position = False
         self.size_constraints = typedict()
@@ -42,12 +46,14 @@ class ClientWindowBase(ClientWidgetBase):
         self._skip_taskbar = False
         self._sticky = False
         self._iconified = False
+        self._focused = False
         self.border = border
         self.max_window_size = max_window_size
         self.button_state = {}
 
         self.init_window(metadata)
-        self.setup_window()
+        self.setup_window(bw, bh)
+        self.update_metadata(metadata)
 
     def __repr__(self):
         return "ClientWindow(%s)" % self._id
@@ -66,11 +72,12 @@ class ClientWindowBase(ClientWidgetBase):
                 metadata["workspace"] = int(workspace)
         self._window_workspace = WORKSPACE_UNSET        #will get set in set_metadata if present
         self._desktop_workspace = self.get_desktop_workspace()
-        workspacelog("init_window(..) workspace=%s, current workspace=%s", self._window_workspace, self._desktop_workspace)
+        def wn(w):
+            return WORKSPACE_NAMES.get(w, w)
+        workspacelog("init_window(..) workspace=%s, current workspace=%s", wn(self._window_workspace), wn(self._desktop_workspace))
         if self.max_window_size and b"size-constraints" not in metadata:
             #this ensures that we will set size-constraints and honour max_window_size:
             metadata[b"size-constraints"] = {}
-        self.update_metadata(metadata)
 
 
     def get_desktop_workspace(self):
@@ -80,12 +87,15 @@ class ClientWindowBase(ClientWidgetBase):
         return None
 
 
-    def new_backing(self, w, h):
+    def new_backing(self, bw, bh):
         backing_class = self.get_backing_class()
-        log("new_backing(%s, %s) backing_class=%s", w, h, backing_class)
+        log("new_backing(%s, %s) backing_class=%s", bw, bh, backing_class)
         assert backing_class is not None
-        self._backing = self.make_new_backing(backing_class, w, h)
+        w, h = self._size
+        self._backing = self.make_new_backing(backing_class, w, h, bw, bh)
         self._backing.border = self.border
+        return self._backing._backing
+
 
     def destroy(self):
         #ensure we clear reference to other windows:
@@ -97,11 +107,25 @@ class ClientWindowBase(ClientWidgetBase):
             self._backing = None
 
 
-    def setup_window(self):
-        self.new_backing(*self._size)
+    def setup_window(self, bw, bh):
+        self.new_backing(bw, bh)
         #tell the server about the encoding capabilities of this backing instance:
-        #"rgb_formats", "full_csc_modes", "csc_modes":
-        self._client_properties.update(self._backing.get_encoding_properties())
+        #but don't bother if they're the same as what we sent as defaults
+        #(with a bit of magic to collapse the missing namespace from encoding_defaults)
+        backing_props = self._backing.get_encoding_properties()
+        encoding_defaults = self._client.encoding_defaults
+        for k in list(backing_props.keys()):
+            v = backing_props[k]
+            try:
+                #ie: "encodings.rgb_formats" -> "rgb_formats"
+                #ie: "encoding.full_csc_modes" -> "full_csc_modes"
+                ek = k.split(".", 1)[1]
+            except:
+                ek = k
+            dv = encoding_defaults.get(ek)
+            if dv is not None and dv==v:
+                del backing_props[k]
+        self._client_properties.update(backing_props)
 
 
     def send(self, *args):
@@ -136,11 +160,6 @@ class ClientWindowBase(ClientWidgetBase):
     def update_metadata(self, metadata):
         metalog("update_metadata(%s)", metadata)
         self._metadata.update(metadata)
-        if not self.is_realized():
-            #Warning: window managers may ignore the icons we try to set
-            #if the wm_class value is set and matches something somewhere undocumented
-            #(if the default is used, you cannot override the window icon)
-            self.set_wmclass(*self._metadata.strlistget("class-instance", ("xpra", "Xpra")))
         try:
             self.set_metadata(metadata)
         except Exception:
@@ -148,6 +167,11 @@ class ClientWindowBase(ClientWidgetBase):
 
     def set_metadata(self, metadata):
         metalog("set_metadata(%s)", metadata)
+        #WARNING: "class-instance" needs to go first because others may realize the window
+        #(and GTK doesn't set the "class-instance" once the window is realized)
+        if b"class-instance" in metadata:
+            self.set_class_instance(*self._metadata.strlistget("class-instance", ("xpra", "Xpra")))
+
         if b"title" in metadata:
             try:
                 title = bytestostr(self._client.title).replace("\0", "")
@@ -298,6 +322,27 @@ class ClientWindowBase(ClientWidgetBase):
         if b"fullscreen-monitors" in metadata:
             self.set_fullscreen_monitors(metadata.intlistget("fullscreen-monitors"))
 
+        if b"shape" in metadata:
+            self.set_shape(metadata.dictget("shape"))
+
+        if b"command" in metadata:
+            self.set_command(metadata.strget("command"))
+
+        if b"menu" in metadata:
+            self.set_menu(metadata.dictget("menu"))
+
+
+    def set_menu(self, menu):
+        pass
+
+    def set_command(self, command):
+        pass
+
+    def set_class_instance(self, wmclass_name, wmclass_class):
+        pass
+
+    def set_shape(self, shape):
+        log("set_shape(%s) not implemented", shape)
 
     def set_bypass_compositor(self, v):
         pass        #see gtk client window base
@@ -313,32 +358,33 @@ class ClientWindowBase(ClientWidgetBase):
 
 
     def set_size_constraints(self, size_constraints, max_window_size):
+        metalog("set_size_constraints(%s, %s)", size_constraints, max_window_size)
         self._set_initial_position = size_constraints.get("set-initial-position")
-        hints = {}
+        hints = typedict()
         for (a, h1, h2) in [
-            ("maximum-size", "max_width", "max_height"),
-            ("minimum-size", "min_width", "min_height"),
-            ("base-size", "base_width", "base_height"),
-            ("increment", "width_inc", "height_inc"),
+            (b"maximum-size", b"max_width", b"max_height"),
+            (b"minimum-size", b"min_width", b"min_height"),
+            (b"base-size", b"base_width", b"base_height"),
+            (b"increment", b"width_inc", b"height_inc"),
             ]:
-            v = size_constraints.intlistget(a)
+            v = size_constraints.intpair(a)
             if v:
                 v1, v2 = v
-                hints[h1], hints[h2] = int(v1), int(v2)
+                hints[h1], hints[h2] = self._client.sp(v1, v2)
         for (a, h) in [
-            ("minimum-aspect-ratio", "min_aspect"),
-            ("maximum-aspect-ratio", "max_aspect"),
+            (b"minimum-aspect-ratio", b"min_aspect"),
+            (b"maximum-aspect-ratio", b"max_aspect"),
             ]:
-            v = size_constraints.intlistget(a)
+            v = size_constraints.intpair(a)
             if v:
                 v1, v2 = v
-                hints[h] = float(v1)/float(v2)
+                hints[h] = (v1*self._client.xscale)/(v2*self._client.yscale)
         #apply max-size override if needed:
         w,h = max_window_size
         if w>0 and h>0 and not self._fullscreen:
             #get the min size, if there is one:
-            minw = max(1, hints.get("min_width", 1))
-            minh = max(1, hints.get("min_height", 1))
+            minw = max(1, hints.intget(b"min_width", 1))
+            minh = max(1, hints.intget(b"min_height", 1))
             #the actual max size is:
             # * greater than the min-size
             # * the lowest of the max-size set by the application and the one we have
@@ -346,12 +392,12 @@ class ClientWindowBase(ClientWidgetBase):
             #according to the GTK docs:
             #allowed window widths are base_width + width_inc * N where N is any integer
             #allowed window heights are base_height + width_inc * N where N is any integer
-            maxw = hints.get("max_width", 32768)
-            maxh = hints.get("max_height", 32768)
+            maxw = hints.intget(b"max_width", 32768)
+            maxh = hints.intget(b"max_height", 32768)
             maxw = max(minw, min(w, maxw))
             maxh = max(minh, min(h, maxh))
-            rw = (maxw - hints.get("base_width", 0)) % max(hints.get("width_inc", 1), 1)
-            rh = (maxh - hints.get("base_height", 0)) % max(hints.get("height_inc", 1), 1)
+            rw = (maxw - hints.intget(b"base_width", 0)) % max(hints.intget(b"width_inc", 1), 1)
+            rh = (maxh - hints.intget(b"base_height", 0)) % max(hints.intget(b"height_inc", 1), 1)
             maxw -= rw
             maxh -= rh
             #if the hints combination is invalid, it's possible that we'll end up
@@ -359,8 +405,16 @@ class ClientWindowBase(ClientWidgetBase):
             maxw = max(minw, maxw)
             maxh = max(minh, maxh)
             metalog("modified hints for max window size %s: %s (rw=%s, rh=%s) -> max=%sx%s", max_window_size, hints, rw, rh, maxw, maxh)
-            hints["max_width"] = maxw
-            hints["max_height"] = maxh
+            #ensure we don't have duplicates with bytes / strings,
+            #and that keys are always "bytes":
+            #(in practice this code should never fire, just here as a reminder)
+            for x in ("max_width", "max_height"):
+                try:
+                    del hints[x]
+                except:
+                    pass
+            hints[b"max_width"] = maxw
+            hints[b"max_height"] = maxh
         try:
             metalog("calling: %s(%s)", self.apply_geometry_hints, hints)
             #save them so the window hooks can use the last value used:
@@ -392,6 +446,60 @@ class ClientWindowBase(ClientWidgetBase):
     def set_xid(self, xid):
         pass
 
+
+    def increase_quality(self, *args):
+        if self._client.quality>0:
+            #change fixed quality:
+            self._client.quality = min(100, self._client.quality + 10)
+            self._client.send_quality()
+            log("new quality=%s", self._client.quality)
+        else:
+            self._client.min_quality = min(100, self._client.min_quality + 10)
+            self._client.send_min_quality()
+            log("new min-quality=%s", self._client.min_quality)
+
+    def decrease_quality(self, *args):
+        if self._client.quality>0:
+            #change fixed quality:
+            self._client.quality = max(1, self._client.quality - 10)
+            self._client.send_quality()
+            log("new quality=%s", self._client.quality)
+        else:
+            self._client.min_quality = max(0, self._client.min_quality - 10)
+            self._client.send_min_quality()
+            log("new min-quality=%s", self._client.min_quality)
+
+    def increase_speed(self, *args):
+        if self._client.speed>0:
+            #change fixed speed:
+            self._client.speed = min(100, self._client.speed + 10)
+            self._client.send_speed()
+            log("new speed=%s", self._client.speed)
+        else:
+            self._client.min_speed = min(100, self._client.min_speed + 10)
+            self._client.send_min_speed()
+            log("new min-speed=%s", self._client.min_speed)
+
+    def decrease_speed(self, *args):
+        if self._client.speed>0:
+            #change fixed speed:
+            self._client.speed = max(1, self._client.speed - 10)
+            self._client.send_speed()
+            log("new speed=%s", self._client.speed)
+        else:
+            self._client.min_speed = max(0, self._client.min_speed - 10)
+            self._client.send_min_speed()
+            log("new min-speed=%s", self._client.min_speed)
+
+    def scaleup(self, *args):
+        self._client.scaleup()
+
+    def scaledown(self, *args):
+        self._client.scaledown()
+
+    def scalereset(self, *args):
+        self._client.scalereset()
+
     def magic_key(self, *args):
         log.info("magic_key(%s) not handled in %s", args, type(self))
 
@@ -399,7 +507,7 @@ class ClientWindowBase(ClientWidgetBase):
         log("refresh_window(%s) wid=%s", args, self._id)
         self._client.send_refresh(self._id)
 
-    def refresh_all_windows(self):
+    def refresh_all_windows(self, *args):
         #this method is only here because we may want to fire it
         #from a --key-shortcut action and the event is delivered to
         #the "ClientWindow"
@@ -411,19 +519,23 @@ class ClientWindowBase(ClientWidgetBase):
         if not backing:
             log("draw_region: window %s has no backing, gone?", self._id)
             from xpra.client.window_backing_base import fire_paint_callbacks
-            fire_paint_callbacks(callbacks, False)
+            fire_paint_callbacks(callbacks, -1, "no backing")
             return
-        def after_draw_refresh(success):
-            plog("after_draw_refresh(%s) %sx%s at %sx%s encoding=%s, options=%s", success, width, height, x, y, coding, options)
-            if not success:
+        def after_draw_refresh(success, message=""):
+            plog("after_draw_refresh(%s, %s) %sx%s at %sx%s encoding=%s, options=%s", success, message, width, height, x, y, coding, options)
+            if success<=0:
                 return
             backing = self._backing
             if backing and backing.draw_needs_refresh:
-                self.queue_draw(x, y, width, height)
+                if REPAINT_ALL=="1" or self._client.xscale!=1 or self._client.yscale!=1:
+                    w, h = self.get_size()
+                    self.queue_draw(0, 0, w, h)
+                else:
+                    self.queue_draw(*self._client.srect(x, y, width, height))
         #only register this callback if we actually need it:
         if backing.draw_needs_refresh:
             callbacks.append(after_draw_refresh)
-        self._backing.draw_region(x, y, width, height, coding, img_data, rowstride, options, callbacks)
+        backing.draw_region(x, y, width, height, coding, img_data, rowstride, options, callbacks)
 
     def spinner(self, ok):
         if not self.can_have_spinner():
@@ -459,16 +571,37 @@ class ClientWindowBase(ClientWidgetBase):
     def show_session_info(self, *args):
         self._client.show_session_info(*args)
 
+    def show_menu(self, *args):
+        self._client.show_menu(*args)
+
     def show_start_new_command(self, *args):
         self._client.show_start_new_command(*args)
 
+    def show_bug_report(self, *args):
+        self._client.show_bug_report(*args)
+
+    def show_file_upload(self, *args):
+        self._client.show_file_upload(*args)
+
 
     def log(self, message=""):
         log.info(message)
 
+
+    def keyboard_layout_changed(self, *args):
+        #used by win32 hooks to tell us about keyboard layout changes for this window
+        keylog("keyboard_layout_changed%s", args)
+        self._client.window_keyboard_layout_changed(self)
+
+
     def dbus_call(self, *args, **kwargs):
-        #see UIXpraClient.dbus_call
-        return self._client.dbus_call(self._id, *args, **kwargs)
+        #alias for rpc_call using dbus as rpc_type, see UIXpraClient.dbus_call
+        if not self._client.server_dbus_proxy:
+            log.error("Error: cannot send remote dbus call:")
+            log.error(" this server does not support dbus-proxying")
+            return
+        rpc_args = [self._id]+args
+        return self._client.rpc_call("dbus", rpc_args, **kwargs)
 
 
     def get_mouse_event_wid(self):
