@@ -15,20 +15,28 @@ required_extensions = ["GL_ARB_texture_rectangle", "GL_ARB_vertex_program"]
 
 WHITELIST = {}
 GREYLIST = {}
-BLACKLIST = {"vendor" : ["nouveau", "Humper", "VMware, Inc."],
+VERSION_REQ = {"nouveau" : [3, 0],      #older versions have issues
+               }
+BLACKLIST = {"vendor" : ["Humper", "VMware, Inc."],
              "renderer" : ["Software Rasterizer"]}
 
+#the Intel driver causes too many problems:
+GREYLIST.setdefault("vendor", []).append("Intel")
+GREYLIST.setdefault("vendor", []).append("Intel Inc.")
+GREYLIST.setdefault("vendor", []).append("Intel Open Source Technology Center")
+from xpra.os_util import getUbuntuVersion
+uv = getUbuntuVersion()
+if uv and uv<[15]:
+    #Ubuntu 14.x drivers are just too old
+    GREYLIST.setdefault("vendor", []).append("X.Org")
 if False:
     #for testing:
     BLACKLIST["vendor"].append("NVIDIA Corporation")
     WHITELIST["renderer"] = ["GeForce GTX 760/PCIe/SSE2"]
 
-if sys.platform.startswith("darwin") or sys.platform.startswith("win"):
-    #crashes were reported with the Intel driver on OSX and win32:
-    GREYLIST.setdefault("vendor", []).append("Intel Inc.")
-if sys.platform.startswith("darwin"):
-    #frequent crashes on osx with GT 650M: (see ticket #808)
-    GREYLIST.setdefault("vendor", []).append("NVIDIA Corporation")
+    if sys.platform.startswith("darwin"):
+        #frequent crashes on osx with GT 650M: (see ticket #808)
+        GREYLIST.setdefault("vendor", []).append("NVIDIA Corporation")
 
 
 #alpha requires gtk3 or *nix only for gtk2:
@@ -104,8 +112,12 @@ def raise_error(msg):
 gl_check_error = raise_error
 
 
+_version_warning_shown = False
 #support for memory views requires Python 2.7 and PyOpenGL 3.1
-def is_pyopengl_memoryview_safe(pyopengl_version):
+def is_pyopengl_memoryview_safe(pyopengl_version, accel_version):
+    if accel_version is not None and pyopengl_version!=accel_version:
+        #mismatch is not safe!
+        return False
     if sys.version_info[:2]<(2,7):
         #memoryview type requires python 2.7
         return False
@@ -174,7 +186,7 @@ def do_check_GL_support(force_enable):
         import OpenGL
         props["pyopengl"] = OpenGL.__version__
         from OpenGL.GL import GL_VERSION, GL_EXTENSIONS
-        from OpenGL.GL import glGetString, glGetInteger
+        from OpenGL.GL import glGetString, glGetInteger, glGetIntegerv
         gl_version_str = glGetString(GL_VERSION)
         if gl_version_str is None:
             gl_check_error("OpenGL version is missing - cannot continue")
@@ -190,11 +202,30 @@ def do_check_GL_support(force_enable):
             log("found valid OpenGL version: %s.%s", gl_major, gl_minor)
 
         from OpenGL import version as OpenGL_version
+        pyopengl_version = OpenGL_version.__version__
         try:
             import OpenGL_accelerate
+            accel_version = OpenGL_accelerate.__version__
+            props["accelerate"] = accel_version
+            log("OpenGL_accelerate version %s", accel_version)
         except:
+            log("OpenGL_accelerate not found")
             OpenGL_accelerate = None
-        props["zerocopy"] = is_pyopengl_memoryview_safe(OpenGL_version.__version__) and bool(OpenGL_accelerate)
+            accel_version = None
+
+        if accel_version is not None and pyopengl_version!=accel_version:
+            global _version_warning_shown
+            if not _version_warning_shown:
+                log.warn("Warning: version mismatch between PyOpenGL and PyOpenGL-accelerate")
+                log.warn(" this may cause crashes")
+                _version_warning_shown = True
+        vsplit = pyopengl_version.split('.')
+        #we now require PyOpenGL 3.1 or later
+        if vsplit[:2]<['3','1'] and not force_enable:
+            gl_check_error("PyOpenGL version 3.1 or later is required (found version %s)" % pyopengl_version)
+            return {}
+
+        props["zerocopy"] = bool(OpenGL_accelerate) and is_pyopengl_memoryview_safe(pyopengl_version, accel_version)
 
         try:
             extensions = glGetString(GL_EXTENSIONS).decode().split(" ")
@@ -212,12 +243,17 @@ def do_check_GL_support(force_enable):
             pass
 
         from OpenGL.GL import GL_RENDERER, GL_VENDOR, GL_SHADING_LANGUAGE_VERSION
+        def fixstring(v):
+            try:
+                return str(v).strip()
+            except:
+                return str(v)
         for d,s,fatal in (("vendor",     GL_VENDOR,      True),
                           ("renderer",   GL_RENDERER,    True),
                           ("shading language version", GL_SHADING_LANGUAGE_VERSION, False)):
             try:
                 v = glGetString(s)
-                v = v.decode()
+                v = fixstring(v.decode())
                 log("%s: %s", d, v)
             except:
                 if fatal:
@@ -226,6 +262,16 @@ def do_check_GL_support(force_enable):
                     log("OpenGL property '%s' is missing", d)
                 v = ""
             props[d] = v
+        vendor = props["vendor"]
+        version_req = VERSION_REQ.get(vendor)
+        if version_req:
+            req_maj, req_min = version_req
+            if gl_major<req_maj or (gl_major==req_maj and gl_minor<req_min):
+                if force_enable:
+                    log.warn("Warning: '%s' OpenGL driver requires version %i.%i", vendor, req_maj, req_min)
+                    log.warn(" version %i.%i was found", gl_major, gl_minor)
+                else:
+                    gl_check_error("OpenGL version %i.%i is too old, %i.%i is required for %s" % (gl_major, gl_minor, req_maj, req_min, vendor))
 
         from OpenGL.GLU import gluGetString, GLU_VERSION, GLU_EXTENSIONS
         for d,s in {"GLU version": GLU_VERSION, "GLU extensions":GLU_EXTENSIONS}.items():
@@ -292,6 +338,7 @@ def do_check_GL_support(force_enable):
             log.warn("OpenGL glEnablei is not available, disabling transparency")
             global GL_ALPHA_SUPPORTED
             GL_ALPHA_SUPPORTED = False
+        props["transparency"] = GL_ALPHA_SUPPORTED
 
         #check for framebuffer functions we need:
         from OpenGL.GL.ARB.framebuffer_object import GL_FRAMEBUFFER, \
@@ -341,8 +388,20 @@ def do_check_GL_support(force_enable):
             gl_check_error("unable to query max texture size: %s" % emsg)
             return props
 
-            log("Texture size GL_MAX_RECTANGLE_TEXTURE_SIZE=%s, GL_MAX_TEXTURE_SIZE=%s", rect_texture_size, texture_size)
-        props["texture-size-limit"] = min(rect_texture_size, texture_size)
+        log("Texture size GL_MAX_RECTANGLE_TEXTURE_SIZE=%s, GL_MAX_TEXTURE_SIZE=%s", rect_texture_size, texture_size)
+        texture_size_limit = min(rect_texture_size, texture_size)
+        props["texture-size-limit"] = texture_size_limit
+
+        try:
+            from OpenGL.GL import GL_MAX_VIEWPORT_DIMS
+            v = glGetIntegerv(GL_MAX_VIEWPORT_DIMS)
+            max_viewport_dims = v[0], v[1]
+            assert max_viewport_dims[0]>=texture_size_limit and max_viewport_dims[1]>=texture_size_limit
+            log("GL_MAX_VIEWPORT_DIMS=%s", max_viewport_dims)
+        except ImportError as e:
+            log.error("Error querying max viewport dims: %s", e)
+            max_viewport_dims = texture_size_limit, texture_size_limit
+        props["max-viewport-dims"] = max_viewport_dims
         return props
     finally:
         for x in alogger.handlers[0].records:
@@ -389,7 +448,10 @@ def do_check_GL_support(force_enable):
                 m = m.replace("accelerators", "").replace("accelerator", "").strip()
                 missing_accelerators.append(m)
                 continue
-            log.info(msg)
+            elif msg.startswith("Using accelerated"):
+                log(msg)
+            else:
+                log.info(msg)
         if missing_accelerators:
             log.info("OpenGL accelerate missing: %s", ", ".join(missing_accelerators))
 
@@ -415,7 +477,11 @@ def check_support(force_enable=False, check_colormap=False):
 
     props = {}
     #this will import gtk.gtkgl / gdkgl or gi.repository.GtkGLExt / GdkGLExt:
-    from xpra.client.gl.gtk_compat import get_info, gdkgl, Config_new_by_mode, MODE_DOUBLE, GLDrawingArea
+    try:
+        from xpra.client.gl.gtk_compat import get_info, gdkgl, Config_new_by_mode, MODE_DOUBLE, GLDrawingArea
+    except RuntimeError as e:
+        gl_check_error(str(e))
+        return {}
     props.update(get_info())
     display_mode = get_DISPLAY_MODE()
     glconfig = Config_new_by_mode(display_mode)
@@ -482,9 +548,11 @@ def main():
     from xpra.platform import init,clean
     from xpra.platform.gui import init as gui_init
     from xpra.util import pver
+    from xpra.log import enable_color
     try:
         init("OpenGL-Check")
         gui_init()
+        enable_color()
         verbose = "-v" in sys.argv or "--verbose" in sys.argv
         if verbose:
             log.enable_debug()
