@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2013 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2012-2014 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2012-2015 Antoine Martin <antoine@devloop.org.uk>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -9,11 +9,11 @@ import time, math
 
 from xpra.log import Logger
 log = Logger("opengl", "paint")
+fpslog = Logger("opengl", "fps")
 OPENGL_DEBUG = os.environ.get("XPRA_OPENGL_DEBUG", "0")=="1"
-OPENGL_PAINT_BOX = os.environ.get("XPRA_OPENGL_PAINT_BOX", "0")=="1"
+OPENGL_PAINT_BOX = int(os.environ.get("XPRA_OPENGL_PAINT_BOX", "0"))
 
-from xpra.gtk_common.gtk_util import import_gobject, color_parse
-idle_add = import_gobject().idle_add
+from xpra.gtk_common.gtk_util import color_parse, is_realized
 
 
 _DEFAULT_BOX_COLORS = {
@@ -28,7 +28,6 @@ _DEFAULT_BOX_COLORS = {
               "png/L"   : "teal",
               "h265"    : "khaki",
               "vp9"     : "lavender",
-              "expose"  : "violet",
               }
 
 def get_fcolor(encoding):
@@ -56,7 +55,6 @@ for x in _DEFAULT_BOX_COLORS.keys():
     BOX_COLORS[x] = get_fcolor(x)
 
 
-from xpra.os_util import memoryview_to_bytes
 from xpra.codecs.codec_constants import get_subsampling_divs
 from xpra.client.window_backing_base import fire_paint_callbacks
 from xpra.gtk_common.gtk_util import POINTER_MOTION_MASK, POINTER_MOTION_HINT_MASK
@@ -66,6 +64,7 @@ from xpra.client.gl.gtk_compat import Config_new_by_mode, MODE_DOUBLE, GLContext
 from xpra.client.gl.gl_check import get_DISPLAY_MODE, GL_ALPHA_SUPPORTED, CAN_DOUBLE_BUFFER, is_pyopengl_memoryview_safe
 from xpra.client.gl.gl_colorspace_conversions import YUV2RGB_shader, RGBP2RGB_shader
 from OpenGL import version as OpenGL_version
+from OpenGL.error import GLError
 from OpenGL.GL import \
     GL_PROJECTION, GL_MODELVIEW, \
     GL_UNPACK_ROW_LENGTH, GL_UNPACK_ALIGNMENT, \
@@ -73,10 +72,10 @@ from OpenGL.GL import \
     GL_UNSIGNED_BYTE, GL_LUMINANCE, GL_LINEAR, \
     GL_TEXTURE0, GL_TEXTURE1, GL_TEXTURE2, GL_QUADS, GL_POLYGON, GL_LINE_LOOP, GL_COLOR_BUFFER_BIT, \
     GL_TEXTURE_WRAP_S, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER, \
-    GL_DONT_CARE, GL_TRUE, GL_DEPTH_TEST, \
+    GL_DONT_CARE, GL_TRUE, GL_DEPTH_TEST, GL_SCISSOR_TEST, GL_LIGHTING, GL_DITHER, \
     GL_RGB, GL_RGBA, GL_BGR, GL_BGRA, \
     GL_BLEND, GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA, \
-    GL_TEXTURE_MAX_LEVEL, GL_TEXTURE_2D, \
+    GL_TEXTURE_MAX_LEVEL, GL_TEXTURE_BASE_LEVEL, \
     GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST, \
     glLineStipple, GL_LINE_STIPPLE, \
     glTexEnvi, GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE, \
@@ -148,10 +147,10 @@ if OPENGL_DEBUG:
 from ctypes import c_char_p
 
 try:
-    import OpenGL_accelerate
+    import OpenGL_accelerate            #@UnresolvedImport
 except:
     OpenGL_accelerate = None
-zerocopy_upload = os.environ.get("XPRA_ZEROCOPY_OPENGL_UPLOAD", "1")=="1" and is_pyopengl_memoryview_safe(OpenGL_version.__version__) and OpenGL_accelerate
+zerocopy_upload = bool(OpenGL_accelerate) and os.environ.get("XPRA_ZEROCOPY_OPENGL_UPLOAD", "1")=="1" and is_pyopengl_memoryview_safe(OpenGL_version.__version__, OpenGL_accelerate.__version__)
 try:
     memoryview_type = memoryview
 except:
@@ -164,18 +163,13 @@ except:
 
 
 # Texture number assignment
-#  1 = Y plane
-#  2 = U plane
-#  3 = V plane
-#  4 = RGB updates
-#  5 = FBO texture (guaranteed up-to-date window contents)
 # The first four are used to update the FBO,
 # the FBO is what is painted on screen.
 TEX_Y = 0
 TEX_U = 1
 TEX_V = 2
 TEX_RGB = 3
-TEX_FBO = 4
+TEX_FBO = 4         #FBO texture (guaranteed up-to-date window contents)
 
 # Shader number assignment
 YUV2RGB_SHADER = 0
@@ -195,17 +189,16 @@ class GLWindowBackingBase(GTKWindowBacking):
     RGB_MODES = ["YUV420P", "YUV422P", "YUV444P", "GBRP", "BGRA", "BGRX", "RGBA", "RGBX", "RGB", "BGR"]
     HAS_ALPHA = GL_ALPHA_SUPPORTED
 
-    def __init__(self, wid, w, h, window_alpha):
+    def __init__(self, wid, window_alpha):
         self.wid = wid
         self.size = 0, 0
-        self.pixel_format = None
+        self.render_size = 0, 0
         self.texture_pixel_format = None
         #this is the pixel format we are currently updating the fbo with
         #can be: "YUV420P", "YUV422P", "YUV444P", "GBRP" or None when not initialized yet.
         self.pixel_format = None
         self.textures = None # OpenGL texture IDs
         self.shaders = None
-        self.size = 0, 0
         self.texture_size = 0, 0
         self.gl_setup = False
         self.debug_setup = False
@@ -214,6 +207,7 @@ class GLWindowBackingBase(GTKWindowBacking):
         self.paint_spinner = False
         self.draw_needs_refresh = False
         self.offscreen_fbo = None
+        self.pending_fbo_paint = []
 
         GTKWindowBacking.__init__(self, wid, window_alpha)
         self.init_gl_config(window_alpha)
@@ -262,12 +256,13 @@ class GLWindowBackingBase(GTKWindowBacking):
     def __repr__(self):
         return "GLWindowBacking(%s, %s, %s)" % (self.wid, self.size, self.pixel_format)
 
-    def init(self, w, h):
+    def init(self, ww, wh, bw, bh):
         #re-init gl projection with new dimensions
         #(see gl_init)
-        if self.size!=(w, h):
+        if self.size!=(bw, bh):
             self.gl_setup = False
-            self.size = w, h
+            self.size = bw, bh
+        self.render_size = ww, wh
 
     def gl_marker(self, *msg):
         log(*msg)
@@ -331,18 +326,29 @@ class GLWindowBackingBase(GTKWindowBacking):
         # Create and assign fragment programs
         self.shaders = [ 1, 2 ]
         glGenProgramsARB(2, self.shaders)
-        for progid, progstr in ((YUV2RGB_SHADER, YUV2RGB_shader), (RGBP2RGB_SHADER, RGBP2RGB_shader)):
+        for name, progid, progstr in (("YUV2RGB", YUV2RGB_SHADER, YUV2RGB_shader), ("RGBP2RGB", RGBP2RGB_SHADER, RGBP2RGB_shader)):
             glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, self.shaders[progid])
             glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, len(progstr), progstr)
             err = glGetString(GL_PROGRAM_ERROR_STRING_ARB)
             if err:
                 #FIXME: maybe we should do something else here?
-                log.error(err)
+                log.error("OpenGL shader %s failed:", name)
+                log.error(" %s", err)
+                raise Exception("OpenGL shader %s setup failure: %s" % (name, err))
+            else:
+                log("%s shader initialized", name)
 
     def gl_context(self):
         if not self._backing:
+            log("Error: no OpenGL backing")
+            return None
+        if not is_realized(self._backing):
+            log.error("Error: OpenGL backing is not realized")
             return None
         w, h = self.size
+        if w<=0 or h<=0:
+            log.error("Error: invalid OpenGL backing size: %ix%i", w, h)
+            return None
         context = GLContextManager(self._backing)
         log("%s.gl_context() GL Pixmap backing size: %d x %d, context=%s", self, w, h, context)
         return context
@@ -356,7 +362,7 @@ class GLWindowBackingBase(GTKWindowBacking):
 
         if not self.gl_setup:
             w, h = self.size
-            self.gl_marker("Initializing GL context for window size %d x %d", w, h)
+            self.gl_marker("Initializing GL context for window size %s, backing size %s", self.render_size, self.size)
             # Initialize viewport and matrices for 2D rendering
             glViewport(0, 0, w, h)
             glMatrixMode(GL_PROJECTION)
@@ -376,6 +382,9 @@ class GLWindowBackingBase(GTKWindowBacking):
 
             # we don't use the depth (2D only):
             glDisable(GL_DEPTH_TEST)
+            glDisable(GL_SCISSOR_TEST)
+            glDisable(GL_LIGHTING)
+            glDisable(GL_DITHER)
             # only do alpha blending in present_fbo:
             glDisable(GL_BLEND)
 
@@ -390,8 +399,15 @@ class GLWindowBackingBase(GTKWindowBacking):
             glEnable(GL_FRAGMENT_PROGRAM_ARB)
             glBindTexture(GL_TEXTURE_RECTANGLE_ARB, self.textures[TEX_FBO])
             # nvidia needs this even though we don't use mipmaps (repeated through this file):
-            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0)
-            glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, self.texture_pixel_format, w, h, 0, self.texture_pixel_format, GL_UNSIGNED_BYTE, None)
+            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_BASE_LEVEL, 0)
+            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAX_LEVEL, 0)
+            try:
+                glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, self.texture_pixel_format, w, h, 0, self.texture_pixel_format, GL_UNSIGNED_BYTE, None)
+            except Exception as e:
+                log.error("Error: cannot initialize %ix%i %s texture", w, h, CONSTANT_TO_PIXEL_FORMAT.get(self.texture_pixel_format, self.texture_pixel_format))
+                log.error(" %r", e)
+                log("%s", self.gl_init, exc_info=True)
+                raise Exception("cannot initialize %ix%i %s texture: %r" % (w, h, CONSTANT_TO_PIXEL_FORMAT.get(self.texture_pixel_format, self.texture_pixel_format), e))
             glBindFramebuffer(GL_FRAMEBUFFER, self.offscreen_fbo)
             glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_RECTANGLE_ARB, self.textures[TEX_FBO], 0)
             glClear(GL_COLOR_BUFFER_BIT)
@@ -445,9 +461,16 @@ class GLWindowBackingBase(GTKWindowBacking):
         #   change fragment program
         glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, self.shaders[YUV2RGB_SHADER])
 
-    def present_fbo(self, encoding, is_delta, x, y, w, h):
+    def present_fbo(self, x, y, w, h, flush=0):
+        log("present_fbo: adding %s to pending paint list (size=%i), flush=%s, paint_screen=%s", (x, y, w, h), len(self.pending_fbo_paint), flush, self.paint_screen)
+        self.pending_fbo_paint.append((x, y, w, h))
         if not self.paint_screen:
             return
+        #flush>0 means we should wait for the final flush=0 paint
+        if flush==0:
+            self.do_present_fbo()
+
+    def do_present_fbo(self):
         self.gl_marker("Presenting FBO on screen")
         # Change state to target screen instead of our FBO
         glBindFramebuffer(GL_FRAMEBUFFER, 0)
@@ -461,53 +484,55 @@ class GLWindowBackingBase(GTKWindowBacking):
 
         # Draw FBO texture on screen
         self.set_rgb_paint_state()
-        ww, wh = self.size
-        if self.glconfig.is_double_buffered():
-            #refresh the whole window:
-            x, y = 0, 0
-            w, h = ww, wh
+        bw, bh = self.size
+        ww, wh = self.render_size
 
+        if self.glconfig.is_double_buffered() or bw!=ww or bh!=wh:
+            #refresh the whole window:
+            rectangles = ((0, 0, bw, bh), )
+        else:
+            #paint just the rectangles we have accumulated:
+            rectangles = self.pending_fbo_paint
+        self.pending_fbo_paint = []
+        log("do_present_fbo: painting %s", rectangles)
+
+        glEnable(GL_TEXTURE_RECTANGLE_ARB)
         glBindTexture(GL_TEXTURE_RECTANGLE_ARB, self.textures[TEX_FBO])
         if self._alpha_enabled:
             # support alpha channel if present:
             glEnablei(GL_BLEND, self.textures[TEX_FBO])
             glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
         glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
+
+        #viewport for painting to window:
+        glViewport(0, 0, ww, wh)
+        if ww!=bw or wh!=bh:
+            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
+            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
+
         glBegin(GL_QUADS)
-        #note how we invert coordinates..
-        glTexCoord2i(x, wh-y)
-        glVertex2i(x, y)            #top-left of window viewport
-        glTexCoord2i(x, wh-y-h)
-        glVertex2i(x, y+h)          #bottom-left of window viewport
-        glTexCoord2i(x+w, wh-y-h)
-        glVertex2i(x+w, y+h)        #bottom-right of window viewport
-        glTexCoord2i(x+w, wh-y)
-        glVertex2i(x+w, y)          #top-right of window viewport
+        for x,y,w,h in rectangles:
+            #note how we invert coordinates..
+            tx1, ty1, tx2, ty2 = x, bh-y,  x+w, bh-y-h
+            vx1, vy1, vx2, vy2 = x, y,     x+w, y+h
+            glTexCoord2i(tx1, ty1)
+            glVertex2i(vx1, vy1)        #top-left of window viewport
+            glTexCoord2i(tx1, ty2)
+            glVertex2i(vx1, vy2)        #bottom-left of window viewport
+            glTexCoord2i(tx2, ty2)
+            glVertex2i(vx2, vy2)        #bottom-right of window viewport
+            glTexCoord2i(tx2, ty1)
+            glVertex2i(vx2, vy1)        #top-right of window viewport
         glEnd()
         glDisable(GL_TEXTURE_RECTANGLE_ARB)
 
-        #show region being painted:
-        if OPENGL_PAINT_BOX:
-            glLineWidth(1)
-            if is_delta:
-                glLineStipple(1, 0xf0f0)
-                glEnable(GL_LINE_STIPPLE)
-            glBegin(GL_LINE_LOOP)
-            color = BOX_COLORS.get(encoding, _DEFAULT_BOX_COLOR)
-            glColor4f(*color)
-            for px,py in ((x, y), (x+w, y), (x+w, y+h), (x, y+h)):
-                glVertex2i(px, py)
-            glEnd()
-            if is_delta:
-                glDisable(GL_LINE_STIPPLE)
-
         if self.paint_spinner:
             #add spinner:
-            dim = min(ww/3.0, wh/3.0)
+            dim = min(bw/3.0, bh/3.0)
             t = time.time()
             count = int(t*4.0)
-            bx = ww//2
-            by = wh//2
+            bx = bw//2
+            by = bh//2
             for i in range(8):      #8 lines
                 glBegin(GL_POLYGON)
                 c = cv.trs[count%8][i]
@@ -526,7 +551,7 @@ class GLWindowBackingBase(GTKWindowBacking):
             glLineWidth(self.border.size*2)
             glBegin(GL_LINE_LOOP)
             glColor4f(self.border.red, self.border.green, self.border.blue, self.border.alpha)
-            for px,py in ((0, 0), (ww, 0), (ww, wh), (0, wh)):
+            for px,py in ((0, 0), (bw, 0), (bw, bh), (0, bh)):
                 glVertex2i(px, py)
             glEnd()
 
@@ -534,12 +559,18 @@ class GLWindowBackingBase(GTKWindowBacking):
         self.gl_show()
         self.gl_frame_terminator()
 
+        #restore pbo viewport
+        glViewport(0, 0, bw, bh)
+        glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
+        glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
+
         self.unset_rgb_paint_state()
         log("%s(%s, %s)", glBindFramebuffer, GL_FRAMEBUFFER, self.offscreen_fbo)
         glBindFramebuffer(GL_FRAMEBUFFER, self.offscreen_fbo)
-        log("%s.present_fbo() done", self)
+        log("%s.do_present_fbo() done", self)
 
     def gl_show(self):
+        start = time.time()
         if self.glconfig.is_double_buffered():
             # Show the backbuffer on screen
             log("%s.gl_show() swapping buffers now", self)
@@ -549,6 +580,50 @@ class GLWindowBackingBase(GTKWindowBacking):
             #just ensure stuff gets painted:
             log("%s.gl_show() flushing", self)
             glFlush()
+        end = time.time()
+        fpslog("gl_show took %ims", (end-start)*1000)
+
+
+    def paint_box(self, encoding, is_delta, x, y, w, h):
+        #show region being painted if debug paint box is enabled only:
+        if not OPENGL_PAINT_BOX>0:
+            return
+        glDisable(GL_TEXTURE_RECTANGLE_ARB)
+        glDisable(GL_FRAGMENT_PROGRAM_ARB)
+        glLineWidth(OPENGL_PAINT_BOX+0.5)
+        if is_delta:
+            glLineStipple(1, 0xaaaa)
+            glEnable(GL_LINE_STIPPLE)
+        glBegin(GL_LINE_LOOP)
+        color = BOX_COLORS.get(encoding, _DEFAULT_BOX_COLOR)
+        log("Painting colored box around %s screen update using: %s (delta=%s)", encoding, color, is_delta)
+        glColor4f(*color)
+        for px,py in ((x, y), (x+w, y), (x+w, y+h), (x, y+h)):
+            glVertex2i(px, py)
+        glEnd()
+        if is_delta:
+            glDisable(GL_LINE_STIPPLE)
+
+
+    def pixels_for_upload(self, img_data):
+        #prepare the pixel buffer for upload:
+        t = type(img_data)
+        if t==memoryview_type:
+            if not zerocopy_upload:
+                #not safe, make a copy :(
+                return "copy:memoryview.tobytes", img_data.tobytes()
+            return "zerocopy:memoryview", img_data
+        elif t in (str, buffer_type) and zerocopy_upload:
+            #we can zerocopy if we wrap it:
+            return "zerocopy:buffer-as-memoryview", memoryview_type(img_data)
+        elif t!=str:
+            if hasattr(img_data, "raw"):
+                return "zerocopy:mmap", img_data.raw
+            #everything else.. copy to bytes (aka str):
+            return "copy:str(%s)" % t, str(img_data)
+        else:
+            #str already
+            return  "copy:str", img_data
 
 
     def _do_paint_rgb32(self, img_data, x, y, width, height, rowstride, options):
@@ -565,29 +640,7 @@ class GLWindowBackingBase(GTKWindowBacking):
             return False
 
         #TODO: move this code up to the decode thread section
-        #prepare the pixel buffer for upload:
-        t = type(img_data)
-        if t==memoryview_type:
-            if not zerocopy_upload:
-                #not safe, make a copy :(
-                img_data = memoryview_to_bytes(img_data)
-                upload = "copy:memoryview_to_bytes"
-            else:
-                upload = "zerocopy:memoryview"
-        elif t in (str, buffer_type) and zerocopy_upload:
-            #we can zerocopy if we wrap it:
-            img_data = memoryview_type(img_data)
-            upload = "zerocopy:memoryview", t
-        elif t!=str:
-            if hasattr(img_data, "raw"):
-                img_data = img_data.raw
-                upload = "zerocopy:mmap"
-            else:
-                #everything else.. copy to bytes (aka str):
-                img_data = str(img_data)
-                upload = "copy:str", t
-        else:
-            upload = "copy:str"
+        upload, img_data = self.pixels_for_upload(img_data)
 
         with context:
             self.gl_init()
@@ -630,9 +683,10 @@ class GLWindowBackingBase(GTKWindowBacking):
             glPixelStorei(GL_UNPACK_ALIGNMENT, alignment)
             glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
             glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
-            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0)
-            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER)
-            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER)
+            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_BASE_LEVEL, 0)
+            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAX_LEVEL, 0)
+            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER)
+            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER)
             glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, self.texture_pixel_format, width, height, 0, pformat, GL_UNSIGNED_BYTE, img_data)
 
             # Draw textured RGB quad at the right coordinates
@@ -647,27 +701,29 @@ class GLWindowBackingBase(GTKWindowBacking):
             glVertex2i(x+width, y)
             glEnd()
 
+            self.paint_box(options.get("encoding"), options.get("delta", -1)>=0, x, y, width, height)
+
             # Present update to screen
-            self.present_fbo(options.get("encoding"), options.get("delta", -1)>=0, x, y, width, height)
+            self.present_fbo(x, y, width, height, options.get("flush", 0))
             # present_fbo has reset state already
         return True
 
     def do_video_paint(self, img, x, y, enc_width, enc_height, width, height, options, callbacks):
         #copy so the data will be usable (usually a str)
         img.clone_pixel_data()
-        idle_add(self.gl_paint_planar, options.get("encoding"), img, x, y, enc_width, enc_height, width, height, callbacks)
+        self.idle_add(self.gl_paint_planar, options.get("flush", 0), options.get("encoding"), img, x, y, enc_width, enc_height, width, height, callbacks)
 
-    def gl_paint_planar(self, encoding, img, x, y, enc_width, enc_height, width, height, callbacks):
+    def gl_paint_planar(self, flush, encoding, img, x, y, enc_width, enc_height, width, height, callbacks):
         #this function runs in the UI thread, no video_decoder lock held
-        log("gl_paint_planar%s", (img, x, y, enc_width, enc_height, width, height, callbacks))
+        log("gl_paint_planar%s", (flush, encoding, img, x, y, enc_width, enc_height, width, height, callbacks))
         try:
             pixel_format = img.get_pixel_format()
             assert pixel_format in ("YUV420P", "YUV422P", "YUV444P", "GBRP"), "sorry the GL backing does not handle pixel format '%s' yet!" % (pixel_format)
 
             context = self.gl_context()
             if not context:
-                log("%s._do_paint_rgb(..) not context!", self)
-                fire_paint_callbacks(callbacks, False)
+                log("%s._do_paint_rgb(..) no context!", self)
+                fire_paint_callbacks(callbacks, False, "failed to get a gl context")
                 return
             with context:
                 self.gl_init()
@@ -680,12 +736,17 @@ class GLWindowBackingBase(GTKWindowBacking):
                     x_scale = float(width)/enc_width
                     y_scale = float(height)/enc_height
                 self.render_planar_update(x, y, enc_width, enc_height, x_scale, y_scale)
+                self.paint_box(encoding, False, x, y, width, height)
                 # Present it on screen
-                self.present_fbo(encoding, False, x, y, width, height)
+                self.present_fbo(x, y, width, height, flush)
             fire_paint_callbacks(callbacks, True)
+            return
+        except GLError as e:
+            message = "gl_paint_planar error: %r" % e
         except Exception as e:
-            log.error("%s.gl_paint_planar(..) error: %s", self, e, exc_info=True)
-            fire_paint_callbacks(callbacks, False)
+            message = "gl_paint_planar error: %s" % e
+        log.error("%s.gl_paint_planar(..) error: %s", self, e, exc_info=True)
+        fire_paint_callbacks(callbacks, False, message)
 
     def update_planar_textures(self, x, y, width, height, img, pixel_format, scaling=False):
         assert self.textures is not None, "no OpenGL textures!"
@@ -709,10 +770,10 @@ class GLWindowBackingBase(GTKWindowBacking):
                     mag_filter = GL_LINEAR
                 glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, mag_filter)
                 glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
-                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0)
+                glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_BASE_LEVEL, 0)
+                glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAX_LEVEL, 0)
                 glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_LUMINANCE, width//div_w, height//div_h, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, None)
 
-
         self.gl_marker("updating planar textures: %sx%s %s", width, height, pixel_format)
         rowstrides = img.get_rowstride()
         img_data = img.get_pixels()
@@ -722,8 +783,10 @@ class GLWindowBackingBase(GTKWindowBacking):
             glActiveTexture(texture)
             glBindTexture(GL_TEXTURE_RECTANGLE_ARB, self.textures[index])
             glPixelStorei(GL_UNPACK_ROW_LENGTH, rowstrides[index])
-            pixel_data = img_data[index]
-            log("texture %s: div=%s, rowstride=%s, %sx%s, data=%s bytes", index, divs[index], rowstrides[index], width//div_w, height//div_h, len(pixel_data))
+            upload, pixel_data = self.pixels_for_upload(img_data[index])
+            log("texture %s: div=%s, rowstride=%s, %sx%s, data=%s bytes, upload=%s", index, divs[index], rowstrides[index], width//div_w, height//div_h, len(pixel_data), upload)
+            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_BASE_LEVEL, 0)
+            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAX_LEVEL, 0)
             glTexSubImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, 0, 0, width//div_w, height//div_h, GL_LUMINANCE, GL_UNSIGNED_BYTE, pixel_data)
 
     def render_planar_update(self, rx, ry, rw, rh, x_scale=1, y_scale=1):
