@@ -4,8 +4,9 @@
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from xpra.gtk_common.gobject_compat import import_gobject
+from xpra.gtk_common.gobject_compat import import_gobject, import_glib
 gobject = import_gobject()
+glib = import_glib()
 
 from xpra.log import Logger
 log = Logger("gobject", "client")
@@ -34,21 +35,21 @@ class GObjectXpraClient(XpraClientBase, gobject.GObject):
         self.gobject_init()
 
     def timeout_add(self, *args):
-        return gobject.timeout_add(*args)
+        return glib.timeout_add(*args)
 
     def idle_add(self, *args):
-        return gobject.idle_add(*args)
+        return glib.idle_add(*args)
 
     def source_remove(self, *args):
-        return gobject.source_remove(*args)
+        return glib.source_remove(*args)
 
     def get_scheduler(self):
-        return gobject
+        return glib
 
 
     def client_type(self):
         #overriden in subclasses!
-        return "Python/GObject"
+        return "Python%s/GObject" % sys.version_info[0]
 
     def timeout(self, *args):
         log.warn("timeout!")
@@ -77,13 +78,13 @@ class GObjectXpraClient(XpraClientBase, gobject.GObject):
     def connect_with_timeout(self, conn):
         self.setup_connection(conn)
         if conn.timeout>0:
-            gobject.timeout_add((conn.timeout + EXTRA_TIMEOUT) * 1000, self.timeout)
-        gobject.idle_add(self.send_hello)
+            glib.timeout_add((conn.timeout + EXTRA_TIMEOUT) * 1000, self.timeout)
+        glib.idle_add(self.send_hello)
 
     def run(self):
         XpraClientBase.run(self)
-        self.gobject_mainloop = gobject.MainLoop()
-        self.gobject_mainloop.run()
+        self.glib_mainloop = glib.MainLoop()
+        self.glib_mainloop.run()
         return  self.exit_code
 
     def make_hello(self):
@@ -96,7 +97,7 @@ class GObjectXpraClient(XpraClientBase, gobject.GObject):
         if self.exit_code is None:
             self.exit_code = exit_code
         self.cleanup()
-        gobject.timeout_add(50, self.gobject_mainloop.quit)
+        glib.timeout_add(50, self.glib_mainloop.quit)
 
 
 
@@ -116,7 +117,6 @@ class CommandConnectClient(GObjectXpraClient):
         self.file_transfer = False
         self.printing = False
 
-
     def make_hello(self):
         capabilities = GObjectXpraClient.make_hello(self)
         #don't bother with many of these things for one-off commands:
@@ -295,14 +295,15 @@ class ControlXpraClient(CommandConnectClient):
         self.warn_and_quit(EXIT_TIMEOUT, "timeout: server did not respond")
 
     def do_command(self):
-        cr = self.server_capabilities.get("command_response")
+        cr = self.server_capabilities.listget("command_response")
         if cr is None:
             self.warn_and_quit(EXIT_UNSUPPORTED, "server does not support control command")
             return
         code, text = cr
+        text = bytestostr(text)
         if code!=0:
             log.warn("server returned error code %s", code)
-            self.warn_and_quit(EXIT_REMOTE_ERROR, text)
+            self.warn_and_quit(EXIT_REMOTE_ERROR, " %s" % text)
             return
         self.warn_and_quit(EXIT_OK, text)
 
@@ -319,6 +320,8 @@ class PrintClient(SendCommandConnectClient):
     def set_command_args(self, command):
         log("set_command_args(%s)", command)
         self.filename = command[0]
+        #print command arguments:
+        #filename, file_data, mimetype, source_uuid, title, printer, no_copies, print_options_str = packet[1:9]
         self.command = command[1:]
         #FIXME: load as needed...
         from xpra.os_util import load_binary_file
@@ -330,6 +333,10 @@ class PrintClient(SendCommandConnectClient):
             log("read %i bytes from stdin", len(self.file_data))
         else:
             self.file_data = load_binary_file(self.filename)
+            log("read %i bytes from %s", len(self.file_data), self.filename)
+        if len(self.file_data)>=self.file_size_limit*1024*1024:
+            self.warn_and_quit(EXIT_FILE_TOO_BIG, "the file is too large: %iMB (the file size limit is %iMB)" % (len(self.file_data)//1024//1024, self.file_size_limit))
+            return
         assert self.file_data, "no data found for '%s'" % self.filename
 
     def client_type(self):
@@ -348,7 +355,7 @@ class PrintClient(SendCommandConnectClient):
         blob = Compressed("print", self.file_data)
         self.send("print", self.filename, blob, *self.command)
         log("print: sending %s as %s for printing", self.filename, blob)
-        gobject.idle_add(self.send, "disconnect", DONE, "detaching")
+        self.idle_add(self.send, "disconnect", DONE, "detaching")
 
     def make_hello(self):
         capabilities = SendCommandConnectClient.make_hello(self)
@@ -372,10 +379,7 @@ class ExitXpraClient(SendCommandConnectClient):
         return capabilities
 
     def do_command(self):
-        if not self.server_capabilities.get("exit_server"):
-            self.warn_and_quit(EXIT_UNSUPPORTED, "server does not support exit command")
-            return
-        gobject.idle_add(self.send, "exit-server")
+        self.idle_add(self.send, "exit-server")
 
 
 class StopXpraClient(SendCommandConnectClient):
@@ -392,7 +396,7 @@ class StopXpraClient(SendCommandConnectClient):
         self.warn_and_quit(EXIT_TIMEOUT, "timeout: server did not disconnect us")
 
     def do_command(self):
-        gobject.idle_add(self.send, "shutdown-server")
+        self.idle_add(self.send, "shutdown-server")
         #not exiting the client here,
         #the server should send us the shutdown disconnection message anyway
         #and if not, we will then hit the timeout to tell us something went wrong
@@ -412,6 +416,6 @@ class DetachXpraClient(SendCommandConnectClient):
         self.warn_and_quit(EXIT_TIMEOUT, "timeout: server did not disconnect us")
 
     def do_command(self):
-        gobject.idle_add(self.send, "disconnect", DONE, "detaching")
+        self.idle_add(self.send, "disconnect", DONE, "detaching")
         #not exiting the client here,
         #the server should disconnect us with the response
