@@ -5,8 +5,6 @@
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-import sys
-import os
 import gobject
 try:
     #we *have to* do this as early as possible on win32..
@@ -18,9 +16,6 @@ from gtk import gdk
 gdk.threads_init()
 
 
-from xpra.platform.ui_thread_watcher import get_UI_watcher
-UI_watcher = get_UI_watcher(gobject.timeout_add)
-
 from xpra.gtk_common.gtk_util import gtk_main, color_parse
 from xpra.client.gtk_base.gtk_client_base import GTKXpraClient
 from xpra.client.gtk2.tray_menu import GTK2TrayMenu
@@ -33,39 +28,28 @@ clipboardlog = Logger("gtk", "client", "clipboard")
 grablog = Logger("gtk", "client", "grab")
 
 from xpra.client.gtk2.border_client_window import BorderClientWindow
-from xpra.client.gtk2.client_window import ClientWindow
-from xpra.client.gtk2.custom_client_window import CustomClientWindow
-WINDOW_LAYOUTS = {
-                  "border"  : BorderClientWindow,
-                  "default" : ClientWindow,
-                  "custom"  : CustomClientWindow,
-                  }
-
-FAKE_UI_LOCKUPS = int(os.environ.get("XPRA_FAKE_UI_LOCKUPS", "0"))
 
 
 class XpraClient(GTKXpraClient):
 
     def __init__(self):
         GTKXpraClient.__init__(self)
+        self.UI_watcher = None
         self.border = None
         self.local_clipboard_requests = 0
         self.remote_clipboard_requests = 0
-
-        self._ref_to_group_leader = {}
-        self._group_leader_wids = {}
+        #only used with the translated clipboard class:
+        self.local_clipboard = ""
+        self.remote_clipboard = ""
 
     def init(self, opts):
         GTKXpraClient.init(self, opts)
-        if opts.window_layout:
-            assert opts.window_layout in WINDOW_LAYOUTS
-            self.ClientWindowClass = WINDOW_LAYOUTS.get(opts.window_layout)
-        if self.ClientWindowClass:
-            log.info("window layout '%s' specified, disabling OpenGL", opts.window_layout)
-            opts.opengl = False
-        else:
-            self.ClientWindowClass = BorderClientWindow
-            log("init(..) ClientWindowClass=%s", self.ClientWindowClass)
+        self.ClientWindowClass = BorderClientWindow
+        self.remote_clipboard = opts.remote_clipboard
+        self.local_clipboard = opts.local_clipboard
+        log("init(..) ClientWindowClass=%s", self.ClientWindowClass)
+        from xpra.platform.ui_thread_watcher import get_UI_watcher
+        self.UI_watcher = get_UI_watcher(self.timeout_add)
 
 
     def parse_border(self, border_str, extra_args):
@@ -114,6 +98,9 @@ class XpraClient(GTKXpraClient):
             if size<=0:
                 log("border size is %s, disabling it", size)
                 return
+            if size>=45:
+                log.warn("border size is too high: %s, clipping it", size)
+                size = 45
         self.border = WindowBorder(True, color.red/65536.0, color.green/65536.0, color.blue/65536.0, alpha, size)
         log("parse_border(%s)=%s", border_str, self.border)
 
@@ -122,8 +109,10 @@ class XpraClient(GTKXpraClient):
         gtk_main()
 
     def cleanup(self):
-        global UI_watcher
-        UI_watcher.stop()
+        uw = self.UI_watcher
+        if uw:
+            self.UI_watcher = None
+            uw.stop()
         GTKXpraClient.cleanup(self)
 
     def __repr__(self):
@@ -150,9 +139,6 @@ class XpraClient(GTKXpraClient):
             w.border = self.border
         return w
 
-    def get_supported_window_layouts(self):
-        return  WINDOW_LAYOUTS
-
     def do_get_core_encodings(self):
         encodings = GTKXpraClient.do_get_core_encodings(self)
         #we can handle rgb32 format (but not necessarily transparency!)
@@ -185,23 +171,55 @@ class XpraClient(GTKXpraClient):
             return
         self.idle_add(self.clipboard_helper.process_clipboard_packet, packet)
 
+
+    def get_clipboard_helper_classes(self):
+        from xpra.scripts.config import TRUE_OPTIONS, FALSE_OPTIONS
+        ct = self.client_clipboard_type
+        if ct and ct.lower() in FALSE_OPTIONS:
+            return []
+        from xpra.platform.features import CLIPBOARD_NATIVE_CLASS
+        from xpra.scripts.main import CLIPBOARD_CLASS
+        #first add the platform specific one, (may be None):
+        clipboard_options = []
+        if CLIPBOARD_CLASS:
+            clipboard_options.append(CLIPBOARD_CLASS)
+        if CLIPBOARD_NATIVE_CLASS:
+            clipboard_options.append(CLIPBOARD_NATIVE_CLASS)
+        clipboard_options.append("xpra.clipboard.gdk_clipboard.GDKClipboardProtocolHelper")
+        clipboard_options.append("xpra.clipboard.clipboard_base.DefaultClipboardProtocolHelper")
+        clipboard_options.append("xpra.clipboard.translated_clipboard.TranslatedClipboardProtocolHelper")
+        clipboardlog("get_clipboard_helper_classes() unfiltered list=%s", clipboard_options)
+        if ct and ct.lower()!="auto" and ct.lower() not in TRUE_OPTIONS:
+            #try to match the string specified:
+            filtered = [x for x in clipboard_options if x.lower().find(self.client_clipboard_type)>=0]
+            if len(filtered)>0:
+                clipboardlog(" found %i clipboard types matching '%s'", len(filtered), self.client_clipboard_type)
+            else:
+                clipboardlog.warn("Warning: no clipboard types matching '%s'", self.client_clipboard_type)
+            clipboard_options = filtered
+        clipboardlog("get_clipboard_helper_classes()=%s", clipboard_options)
+        return clipboard_options
+
     def make_clipboard_helper(self):
         """
             Try the various clipboard classes until we find one
             that loads ok. (some platforms have more options than others)
         """
-        from xpra.platform.features import CLIPBOARDS, CLIPBOARD_NATIVE_CLASS
-        clipboards = [x for x in CLIPBOARDS if x in self.server_clipboards]
-        clipboardlog("make_clipboard_helper() server_clipboards=%s, local clipboards=%s, common=%s", self.server_clipboards, CLIPBOARDS, clipboards)
+        from xpra.platform.features import CLIPBOARDS
+        clipboard_options = self.get_clipboard_helper_classes()
+        clipboardlog("make_clipboard_helper() options=%s, server_clipboards=%s, local clipboards=%s", clipboard_options, self.server_clipboards, CLIPBOARDS)
         #first add the platform specific one, (may be None):
-        clipboard_options = []
-        if CLIPBOARD_NATIVE_CLASS:
-            clipboard_options.append(CLIPBOARD_NATIVE_CLASS)
-        clipboard_options.append(("xpra.clipboard.gdk_clipboard", "GDKClipboardProtocolHelper", {"clipboards" : clipboards}))
-        clipboard_options.append(("xpra.clipboard.clipboard_base", "DefaultClipboardProtocolHelper", {"clipboards" : clipboards}))
-        clipboardlog("make_clipboard_helper() clipboard_options=%s", clipboard_options)
-        for module_name, classname, kwargs in clipboard_options:
-            c = self.try_load_clipboard_helper(module_name, classname, kwargs)
+        kwargs= {"clipboards.local"     : CLIPBOARDS,                   #all the local clipboards supported
+                 "clipboards.remote"    : self.server_clipboards}       #all the remote clipboards supported
+        #only allow translation overrides if we have a way of telling the server about them:
+        if self.server_supports_clipboard_enable_selections:
+            kwargs.update({
+                 "clipboard.local"      : self.local_clipboard,         #the local clipboard we want to sync to (with the translated clipboard only)
+                 "clipboard.remote"     : self.remote_clipboard})       #the remote clipboard we want to we sync to (with the translated clipboard only)
+        clipboardlog("make_clipboard_helper() clipboard_options=%s, kwargs=%s", clipboard_options, kwargs)
+        for classname in clipboard_options:
+            module_name, _class = classname.rsplit(".", 1)
+            c = self.try_load_clipboard_helper(module_name, _class, kwargs)
             if c:
                 return c
         return None
@@ -216,6 +234,10 @@ class XpraClient(GTKXpraClient):
                 c = getattr(m, classname)
                 if c:
                     return self.setup_clipboard_helper(c, **kwargs)
+        except ImportError as e:
+            clipboardlog.error("Error: cannot load %s.%s:", module, classname)
+            clipboardlog.error(" %s", e)
+            return None
         except:
             clipboardlog.error("cannot load %s.%s", module, classname, exc_info=True)
             return None
@@ -260,7 +282,7 @@ class XpraClient(GTKXpraClient):
                 self.remote_clipboard_requests = 0
                 self.clipboard_notify(0)
             self.connect("clipboard-toggled", clipboard_toggled)
-        self.connect("handshake-complete", register_clipboard_toggled)
+        self.after_handshake(register_clipboard_toggled)
         return helperClass(clipboard_send, clipboard_progress, *args, **kwargs)
 
     def clipboard_notify(self, n):
@@ -284,7 +306,10 @@ class XpraClient(GTKXpraClient):
         lz4 = "lz4" in self.server_compressors and compression.use_lz4
         lzo = "lzo" in self.server_compressors and compression.use_lzo
         if zlib or lz4 or lzo:
-            return compression.compressed_wrapper(datatype, data, zlib=zlib, lz4=lz4, lzo=lzo, can_inline=False)
+            cw = compression.compressed_wrapper(datatype, data, zlib=zlib, lz4=lz4, lzo=lzo, can_inline=False)
+            if len(cw)<len(data):
+                #the compressed version is smaller, use it:
+                return cw
         #we can't compress, so at least avoid warnings in the protocol layer:
         return compression.Compressed("raw %s" % datatype, data, can_inline=True)
 
@@ -303,23 +328,15 @@ class XpraClient(GTKXpraClient):
 
     def process_ui_capabilities(self):
         GTKXpraClient.process_ui_capabilities(self)
-        global UI_watcher
-        UI_watcher.start()
+        self.UI_watcher.start()
         #if server supports it, enable UI thread monitoring workaround when needed:
         if self.suspend_resume:
             def UI_resumed():
                 self.send("resume", True, self._id_to_window.keys())
             def UI_failed():
                 self.send("suspend", True, self._id_to_window.keys())
-            UI_watcher.add_resume_callback(UI_resumed)
-            UI_watcher.add_fail_callback(UI_failed)
-
-
-    def get_root_window(self):
-        return gdk.get_default_root_window()
-
-    def get_root_size(self):
-        return gdk.get_default_root_window().get_size()
+            self.UI_watcher.add_resume_callback(UI_resumed)
+            self.UI_watcher.add_fail_callback(UI_failed)
 
 
     def _process_raise_window(self, packet):
@@ -344,89 +361,4 @@ class XpraClient(GTKXpraClient):
         gtk.gdk.keyboard_ungrab()
 
 
-    def get_group_leader(self, wid, metadata, override_redirect):
-        transient_for = metadata.intget("transient-for", -1)
-        log("get_group_leader: transient_for=%s", transient_for)
-        if transient_for>0:
-            client_window = self._id_to_window.get(transient_for)
-            if client_window:
-                gdk_window = client_window.get_window()
-                if gdk_window:
-                    return gdk_window
-        pid = metadata.intget("pid", -1)
-        leader_xid = metadata.intget("group-leader-xid", -1)
-        leader_wid = metadata.intget("group-leader-wid", -1)
-        group_leader_window = self._id_to_window.get(leader_wid)
-        if group_leader_window:
-            #leader is another managed window
-            log("found group leader window %s for wid=%s", group_leader_window, leader_wid)
-            return group_leader_window
-        log("get_group_leader: leader pid=%s, xid=%s, wid=%s", pid, leader_xid, leader_wid)
-        reftype = "xid"
-        ref = leader_xid
-        if ref<0:
-            reftype = "leader-wid"
-            ref = leader_wid
-        if ref<0:
-            ci = metadata.strlistget("class-instance")
-            if ci:
-                reftype = "class"
-                ref = "|".join(ci)
-            elif pid>0:
-                reftype = "pid"
-                ref = pid
-            elif transient_for>0:
-                #this should have matched a client window above..
-                #but try to use it anyway:
-                reftype = "transient-for"
-                ref = transient_for
-            else:
-                #no reference to use
-                return None
-        refkey = "%s:%s" % (reftype, ref)
-        group_leader_window = self._ref_to_group_leader.get(refkey)
-        if group_leader_window:
-            log("found existing group leader window %s using ref=%s", group_leader_window, refkey)
-            return group_leader_window
-        #we need to create one:
-        title = "%s group leader for %s" % (self.session_name or "Xpra", pid)
-        group_leader_window = gdk.Window(None, 1, 1, gdk.WINDOW_TOPLEVEL, 0, gdk.INPUT_ONLY, title)
-        self._ref_to_group_leader[refkey] = group_leader_window
-        #avoid warning on win32...
-        if not sys.platform.startswith("win"):
-            #X11 spec says window should point to itself:
-            group_leader_window.set_group(group_leader_window)
-        log("new hidden group leader window %s for ref=%s", group_leader_window, refkey)
-        self._group_leader_wids.setdefault(group_leader_window, []).append(wid)
-        return group_leader_window
-
-    def destroy_window(self, wid, window):
-        #override so we can cleanup the group-leader if needed,
-        GTKXpraClient.destroy_window(self, wid, window)
-        group_leader = window.group_leader
-        if group_leader is None or len(self._group_leader_wids)==0:
-            return
-        wids = self._group_leader_wids.get(group_leader)
-        if wids is None:
-            #not recorded any window ids on this group leader
-            #means it is another managed window, leave it alone
-            return
-        if wid in wids:
-            wids.remove(wid)
-        if len(wids)>0:
-            #still has another window pointing to it
-            return
-        #the last window has gone, we can remove the group leader,
-        #find all the references to this group leader:
-        del self._group_leader_wids[group_leader]
-        refs = []
-        for ref, gl in self._ref_to_group_leader.items():
-            if gl==group_leader:
-                refs.append(ref)
-        for ref in refs:
-            del self._ref_to_group_leader[ref]
-        log("last window for refs %s is gone, destroying the group leader %s", refs, group_leader)
-        group_leader.destroy()
-
-
 gobject.type_register(XpraClient)
