@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2014 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2015 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -16,13 +16,13 @@ workspacelog = Logger("workspace")
 
 
 from xpra.client.gtk_base.gtk_client_window_base import GTKClientWindowBase, HAS_X11_BINDINGS
-from xpra.util import WORKSPACE_UNSET
-from xpra.platform.gui import add_window_hooks, remove_window_hooks
+from xpra.gtk_common.gtk_util import WINDOW_NAME_TO_HINT, WINDOW_EVENT_MASK, BUTTON_MASK
+from xpra.util import WORKSPACE_UNSET, WORKSPACE_NAMES
+
+def wn(w):
+    return WORKSPACE_NAMES.get(w, w)
 
 
-GTK2_WINDOW_EVENT_MASK = gdk.STRUCTURE_MASK | gdk.KEY_PRESS_MASK | gdk.KEY_RELEASE_MASK \
-            | gdk.POINTER_MOTION_MASK | gdk.BUTTON_PRESS_MASK | gdk.BUTTON_RELEASE_MASK \
-            | gdk.PROPERTY_CHANGE_MASK
 GTK2_OR_TYPE_HINTS = (gdk.WINDOW_TYPE_HINT_DIALOG,
                 gdk.WINDOW_TYPE_HINT_MENU, gdk.WINDOW_TYPE_HINT_TOOLBAR,
                 #gdk.WINDOW_TYPE_HINT_SPLASHSCREEN,
@@ -35,27 +35,6 @@ GTK2_OR_TYPE_HINTS = (gdk.WINDOW_TYPE_HINT_DIALOG,
                 #gdk.WINDOW_TYPE_HINT_NOTIFICATION,
                 gdk.WINDOW_TYPE_HINT_COMBO,
                 gdk.WINDOW_TYPE_HINT_DND)
-GTK2_NAME_TO_HINT = {
-                "NORMAL"        : gdk.WINDOW_TYPE_HINT_NORMAL,
-                "DIALOG"        : gdk.WINDOW_TYPE_HINT_DIALOG,
-                "MENU"          : gdk.WINDOW_TYPE_HINT_MENU,
-                "TOOLBAR"       : gdk.WINDOW_TYPE_HINT_TOOLBAR,
-                "SPLASH"        : gdk.WINDOW_TYPE_HINT_SPLASHSCREEN,
-                "UTILITY"       : gdk.WINDOW_TYPE_HINT_UTILITY,
-                "DOCK"          : gdk.WINDOW_TYPE_HINT_DOCK,
-                "DESKTOP"       : gdk.WINDOW_TYPE_HINT_DESKTOP,
-                "DROPDOWN_MENU" : gdk.WINDOW_TYPE_HINT_DROPDOWN_MENU,
-                "POPUP_MENU"    : gdk.WINDOW_TYPE_HINT_POPUP_MENU,
-                "TOOLTIP"       : gdk.WINDOW_TYPE_HINT_TOOLTIP,
-                "NOTIFICATION"  : gdk.WINDOW_TYPE_HINT_NOTIFICATION,
-                "COMBO"         : gdk.WINDOW_TYPE_HINT_COMBO,
-                "DND"           : gdk.WINDOW_TYPE_HINT_DND
-                }
-GTK2_BUTTON_MASK = {gtk.gdk.BUTTON1_MASK : 1,
-                    gtk.gdk.BUTTON2_MASK : 2,
-                    gtk.gdk.BUTTON3_MASK : 3,
-                    gtk.gdk.BUTTON4_MASK : 4,
-                    gtk.gdk.BUTTON5_MASK : 5}
 # Map scroll directions back to mouse buttons.  Mapping is taken from
 # gdk/x11/gdkevents-x11.c.
 GTK2_SCROLL_MAP = {
@@ -72,11 +51,11 @@ GTK2 version of the ClientWindow class
 """
 class GTK2WindowBase(GTKClientWindowBase):
 
-    WINDOW_EVENT_MASK   = GTK2_WINDOW_EVENT_MASK
+    WINDOW_EVENT_MASK   = WINDOW_EVENT_MASK
     OR_TYPE_HINTS       = GTK2_OR_TYPE_HINTS
-    NAME_TO_HINT        = GTK2_NAME_TO_HINT
+    NAME_TO_HINT        = WINDOW_NAME_TO_HINT
     SCROLL_MAP          = GTK2_SCROLL_MAP
-    BUTTON_MASK         = GTK2_BUTTON_MASK
+    BUTTON_MASK         = BUTTON_MASK
 
     WINDOW_STATE_FULLSCREEN = gdk.WINDOW_STATE_FULLSCREEN
     WINDOW_STATE_MAXIMIZED  = gdk.WINDOW_STATE_MAXIMIZED
@@ -86,41 +65,25 @@ class GTK2WindowBase(GTKClientWindowBase):
     WINDOW_STATE_STICKY     = gdk.WINDOW_STATE_STICKY
 
 
-    def init_window(self, metadata):
-        if self._is_popup(metadata):
-            gtk.Window.__init__(self, gtk.WINDOW_POPUP)
-        else:
-            gtk.Window.__init__(self, gtk.WINDOW_TOPLEVEL)
-        self.set_decorated(self._is_decorated(metadata))
-        GTKClientWindowBase.init_window(self, metadata)
+    def do_init_window(self, window_type=gtk.WINDOW_TOPLEVEL):
+        gtk.Window.__init__(self, window_type)
         # tell KDE/oxygen not to intercept clicks
         # see: https://bugs.kde.org/show_bug.cgi?id=274485
         self.set_data("_kde_no_window_grab", 1)
 
 
-    def setup_window(self):
+    def setup_window(self, *args):
         #preserve screen:
         if not self._override_redirect:
             display = gtk.gdk.display_get_default()
             screen_num = self._client_properties.get("screen", -1)
             n = display.get_n_screens()
-            log("setup_window() screen=%s, nscreens=%s", screen_num, n)
+            log("setup_window%s screen=%s, nscreens=%s", args, screen_num, n)
             if screen_num>=0 and screen_num<n:
                 screen = display.get_screen(screen_num)
                 if screen:
                     self.set_screen(screen)
-        GTKClientWindowBase.setup_window(self)
-        #add platform hooks
-        self.connect("realize", self.on_realize)
-        self.connect('unrealize', self.on_unrealize)
-
-    def on_realize(self, widget):
-        eventslog("on_realize(%s)", widget)
-        add_window_hooks(self)
-
-    def on_unrealize(self, widget):
-        eventslog("on_unrealize(%s)", widget)
-        remove_window_hooks(self)
+        GTKClientWindowBase.setup_window(self, *args)
 
 
     def enable_alpha(self):
@@ -166,12 +129,16 @@ class GTK2WindowBase(GTKClientWindowBase):
 
     def do_get_workspace(self, target, prop, default_value=None):
         if not self._can_set_workspace:
-            return  None              #windows and OSX do not have workspaces
+            workspacelog("do_get_workspace: not supported, returning %s", wn(default_value))
+            return default_value        #windows and OSX do not have workspaces
+        if target is None:
+            workspacelog("do_get_workspace: target is None, returning %s", wn(default_value))
+            return default_value        #window is not realized yet
         value = self.xget_u32_property(target, prop)
         if value is not None:
-            workspacelog("do_get_workspace() found value=%s from %s / %s", value, target, prop)
+            workspacelog("do_get_workspace %s=%s on window %#x", prop, wn(value), target.xid)
             return value
-        workspacelog("do_get_workspace() value not found!")
+        workspacelog("do_get_workspace %s unset on window %#x, returning default value=%s", prop, target.xid, wn(default_value))
         return  default_value
 
 
@@ -197,7 +164,8 @@ class GTK2WindowBase(GTKClientWindowBase):
     def do_expose_event(self, event):
         #cannot use self
         eventslog("do_expose_event(%s) area=%s", event, event.area)
-        if not (self.flags() & gtk.MAPPED) or self._backing is None:
+        backing = self._backing
+        if not (self.flags() & gtk.MAPPED) or backing is None:
             return
         w,h = self.window.get_size()
         if w>=32768 or h>=32768:
@@ -206,6 +174,6 @@ class GTK2WindowBase(GTKClientWindowBase):
         context = self.window.cairo_create()
         context.rectangle(event.area)
         context.clip()
-        self._backing.cairo_draw(context)
+        backing.cairo_draw(context)
         if not self._client.server_ok():
             self.paint_spinner(context, event.area)
