@@ -11,21 +11,13 @@ import sys
 from xpra.platform.gui import init as gui_init
 gui_init()
 
-from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, import_gobject, import_pango, is_gtk3
-
+from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, import_pango, is_gtk3
 gtk = import_gtk()
 gdk = import_gdk()
-gobject = import_gobject()
-gobject.threads_init()
 pango = import_pango()
 
-
 from xpra.gtk_common.gtk_util import gtk_main, add_close_accel, scaled_image, pixbuf_new_from_file, get_display_info, \
                                     JUSTIFY_LEFT, WIN_POS_CENTER, STATE_NORMAL, FILE_CHOOSER_ACTION_SAVE, choose_file, get_gtk_version_info
-from xpra.scripts.config import read_xpra_defaults
-from xpra.client.gtk_base.about import about
-from xpra.platform.paths import get_icon_dir
-from xpra.platform.info import get_user_info
 from xpra.util import nonl, updict, strtobytes
 from xpra.log import Logger
 log = Logger("util")
@@ -60,6 +52,7 @@ class BugReport(object):
         hbox = gtk.HBox(False, 0)
         icon_pixbuf = self.get_icon("xpra.png")
         if icon_pixbuf and self.show_about:
+            from xpra.client.gtk_base.about import about
             logo_button = gtk.Button("")
             settings = logo_button.get_settings()
             settings.set_property('gtk-button-images', True)
@@ -91,6 +84,7 @@ class BugReport(object):
         self.description.set_justification(JUSTIFY_LEFT)
         self.description.set_border_width(2)
         self.description.set_size_request(300, 80)
+        self.description.modify_bg(STATE_NORMAL, gdk.Color(red=32768, green=32768, blue=32768))
         ibox.pack_start(self.description, expand=False, fill=False)
 
         # Toggles:
@@ -102,7 +96,9 @@ class BugReport(object):
         from xpra.codecs.loader import codec_versions, load_codecs
         load_codecs()
         try:
-            from xpra.sound.gstreamer_util import get_info as get_sound_info
+            from xpra.sound.wrapper import query_sound
+            def get_sound_info():
+                return query_sound()
         except:
             get_sound_info = None
         if self.opengl_info:
@@ -120,6 +116,8 @@ class BugReport(object):
         from xpra.platform.gui import get_info as get_gui_info
         from xpra.version_util import get_version_info, get_platform_info, get_host_info
         def get_sys_info():
+            from xpra.platform.info import get_user_info
+            from xpra.scripts.config import read_xpra_defaults
             d = {
                     "argv"          : sys.argv,
                     "path"          : sys.path,
@@ -147,10 +145,25 @@ class BugReport(object):
             take_screenshot_fn = take_screenshot
         except:
             log("failed to load platfrom specific screenshot code", exc_info=True)
+        if not take_screenshot_fn:
+            #try with Pillow:
+            try:
+                from PIL import ImageGrab           #@UnresolvedImport
+                from xpra.os_util import StringIOClass
+                def pillow_imagegrab_screenshot():
+                    img = ImageGrab.grab()
+                    out = StringIOClass()
+                    img.save(out, format="PNG")
+                    v = out.getvalue()
+                    out.close()
+                    return (img.width, img.height, "png", img.width*3, v)
+                take_screenshot_fn = pillow_imagegrab_screenshot
+            except Exception as e:
+                log("cannot use Pillow's ImageGrab: %s", e)
         if not take_screenshot_fn:
             #default: gtk screen capture
             try:
-                from xpra.server.gtk_root_window_model import GTKRootWindowModel
+                from xpra.server.shadow.gtk_root_window_model import GTKRootWindowModel
                 rwm = GTKRootWindowModel(gtk.gdk.get_default_root_window())
                 take_screenshot_fn = rwm.take_screenshot
             except:
@@ -241,6 +254,7 @@ class BugReport(object):
 
 
     def get_icon(self, icon_name):
+        from xpra.platform.paths import get_icon_dir
         icon_filename = os.path.join(get_icon_dir(), icon_name)
         if os.path.exists(icon_filename):
             return pixbuf_new_from_file(icon_filename)
@@ -303,7 +317,7 @@ class BugReport(object):
         basenoext, _ = os.path.splitext(os.path.basename(filename))
         data = self.get_text_data()
         import zipfile, time
-        zf = zipfile.ZipFile(filename, mode='w')
+        zf = zipfile.ZipFile(filename, mode='w', compression=zipfile.ZIP_DEFLATED)
         try:
             for title, tooltip, dtype, s in data:
                 cfile = os.path.join(basenoext, title.replace(" ", "_")+"."+dtype)
@@ -312,6 +326,6 @@ class BugReport(object):
                 #very poorly documented:
                 info.external_attr = 0o644 << 16
                 info.comment = strtobytes(tooltip)
-                zf.writestr(info, s, compress_type=zipfile.ZIP_DEFLATED)
+                zf.writestr(info, s)
         finally:
             zf.close()
