@@ -16,7 +16,8 @@ import shlex
 import signal
 import traceback
 
-from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, import_gobject, import_pango
+from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, import_gobject, import_pango, import_glib
+glib = import_glib()
 gobject = import_gobject()
 gobject.threads_init()
 gtk = import_gtk()
@@ -30,7 +31,7 @@ pango = import_pango()
 from xpra.platform.gui import init as gui_init
 gui_init()
 
-from xpra.scripts.config import read_config, make_defaults_struct, validate_config, save_config, ENCRYPTION_CIPHERS
+from xpra.scripts.config import read_config, make_defaults_struct, validate_config, save_config
 from xpra.codecs.loader import PREFERED_ENCODING_ORDER
 from xpra.gtk_common.gtk_util import gtk_main, add_close_accel, scaled_image, pixbuf_new_from_file, color_parse, \
                                     OptionMenu, choose_file, \
@@ -41,7 +42,8 @@ from xpra.os_util import thread
 from xpra.client.gtk_base.gtk_tray_menu_base import make_min_auto_menu, make_encodingsmenu, set_use_tray_workaround, \
                                     MIN_QUALITY_OPTIONS, QUALITY_OPTIONS, MIN_SPEED_OPTIONS, SPEED_OPTIONS
 from xpra.client.gtk_base.about import about
-from xpra.scripts.main import connect_to, make_client
+from xpra.net.crypto import ENCRYPTION_CIPHERS
+from xpra.scripts.main import connect_to, make_client, configure_network
 from xpra.platform.paths import get_icon_dir
 from xpra.log import Logger, enable_debug_for
 log = Logger("launcher")
@@ -111,11 +113,8 @@ class ApplicationWindow:
         # Default connection options
         self.config = make_defaults_struct(extras_defaults=LAUNCHER_DEFAULTS, extras_types=LAUNCHER_OPTION_TYPES, extras_validation=LAUNCHER_VALIDATION)
         #TODO: the fixup does not belong here?
-        from xpra.scripts.main import fixup_video_all_or_none, fixup_encodings, fixup_compression, fixup_packetencoding
-        fixup_video_all_or_none(self.config)
-        fixup_encodings(self.config)
-        fixup_compression(self.config)
-        fixup_packetencoding(self.config)
+        from xpra.scripts.main import fixup_options
+        fixup_options(self.config)
         #what we save by default:
         self.config_keys = set(SAVED_FIELDS)
         def raise_exception(*args):
@@ -152,6 +151,24 @@ class ApplicationWindow:
             image.set_from_pixbuf(icon_pixbuf)
             logo_button.set_image(image)
             hbox.pack_start(logo_button, expand=False, fill=False)
+        icon_pixbuf = self.get_icon("bugs.png")
+        self.bug_tool = None
+        if icon_pixbuf:
+            bug_button = gtk.Button("")
+            settings = bug_button.get_settings()
+            settings.set_property('gtk-button-images', True)
+            def bug(*args):
+                if self.bug_tool==None:
+                    from xpra.client.gtk_base.bug_report import BugReport
+                    self.bug_tool = BugReport()
+                    self.bug_tool.init(show_about=False)
+                self.bug_tool.show()
+            bug_button.connect("clicked", bug)
+            bug_button.set_tooltip_text("Bug Report")
+            image = gtk.Image()
+            image.set_from_pixbuf(icon_pixbuf)
+            bug_button.set_image(image)
+            hbox.pack_start(bug_button, expand=False, fill=False)
         label = gtk.Label("Connect to xpra server")
         label.modify_font(pango.FontDescription("sans 14"))
         hbox.pack_start(label, expand=True, fill=True)
@@ -417,14 +434,14 @@ class ApplicationWindow:
 
     def set_info_text(self, text):
         if self.info:
-            gobject.idle_add(self.info.set_text, text)
+            glib.idle_add(self.info.set_text, text)
 
     def set_info_color(self, is_error=False):
         self.set_widget_fg_color(self.info, is_error)
 
 
     def set_sensitive(self, s):
-        gobject.idle_add(self.window.set_sensitive, s)
+        glib.idle_add(self.window.set_sensitive, s)
 
     def connect_clicked(self, *args):
         self.update_options_from_gui()
@@ -451,7 +468,7 @@ class ApplicationWindow:
             self.set_info_color(True)
             self.set_info_text(t)
             self.window.show()
-        gobject.idle_add(ui_handle_exception)
+        glib.idle_add(ui_handle_exception)
 
     def do_connect(self):
         try:
@@ -528,8 +545,8 @@ class ApplicationWindow:
             log.error("failed to connect", exc_info=True)
             self.handle_exception(e)
             return
-        gobject.idle_add(self.window.hide)
-        gobject.idle_add(self.start_XpraClient, conn)
+        glib.idle_add(self.window.hide)
+        glib.idle_add(self.start_XpraClient, conn)
 
     def start_XpraClient(self, conn):
         try:
@@ -579,7 +596,7 @@ class ApplicationWindow:
                     self.client.quit = ignore_further_quit_events
                 self.set_sensitive(True)
                 self.reset_client()
-                gobject.idle_add(self.window.show)
+                glib.idle_add(self.window.show)
             else:
                 do_quit()
 
@@ -615,7 +632,7 @@ class ApplicationWindow:
         else:
             color_obj = white
         if color_obj:
-            gobject.idle_add(widget.modify_base, STATE_NORMAL, color_obj)
+            glib.idle_add(widget.modify_base, STATE_NORMAL, color_obj)
 
     def set_widget_fg_color(self, widget, is_error=False):
         if is_error:
@@ -623,7 +640,7 @@ class ApplicationWindow:
         else:
             color_obj = black
         if color_obj:
-            gobject.idle_add(widget.modify_fg, STATE_NORMAL, color_obj)
+            glib.idle_add(widget.modify_fg, STATE_NORMAL, color_obj)
 
 
     def update_options_from_gui(self):
@@ -723,69 +740,78 @@ def main():
     from xpra.gtk_common.quit import gtk_main_quit_on_fatal_exceptions_enable
     gtk_main_quit_on_fatal_exceptions_enable()
 
-    from xpra.platform import init as platform_init
+    from xpra.platform import init as platform_init, clean as platform_clean
     from xpra.platform.gui import ready as gui_ready
-    platform_init("Xpra-Launcher", "Xpra Connection Launcher")
-    gui_init()
-
+    from xpra.log import enable_color
     try:
-        from xpra.scripts.main import parse_cmdline, fixup_debug_option
-        options, args = parse_cmdline(sys.argv)
-        debug = fixup_debug_option(options.debug)
-        if debug:
-            for x in debug.split(","):
-                enable_debug_for(x)
-    except Exception:
-        exception_dialog("Error parsing command line")
-        return 1
-
-    try:
-        app = ApplicationWindow()
-        def app_signal(signum, frame):
-            print("")
-            log("got signal %s" % SIGNAMES.get(signum, signum))
-            def show_signal():
-                app.show()
-                app.client.cleanup()
-                gobject.timeout_add(1000, app.set_info_text, "got signal %s" % SIGNAMES.get(signum, signum))
-                gobject.timeout_add(1000, app.set_info_color, True)
-            #call from UI thread:
-            gobject.idle_add(show_signal)
-        signal.signal(signal.SIGINT, app_signal)
-        signal.signal(signal.SIGTERM, app_signal)
-        has_file = len(args) == 1
-        if has_file:
-            app.update_options_from_file(args[0])
-        debug = fixup_debug_option(app.config.debug)
-        if debug:
-            for x in debug.split(","):
-                enable_debug_for(x)
-        #suspend tray workaround for our window widgets:
+        platform_init("Xpra-Launcher", "Xpra Connection Launcher")
+        enable_color()
+        gui_init()
         try:
-            set_use_tray_workaround(False)
-            app.create_window()
-        finally:
-            set_use_tray_workaround(True)
-        app.update_gui_from_config()
-    except Exception:
-        exception_dialog("Error creating launcher form")
-        return 1
-    try:
-        if app.config.autoconnect:
-            #file says we should connect,
-            #do that only (not showing UI unless something goes wrong):
-            gobject.idle_add(app.do_connect)
-        if not has_file:
-            app.reset_errors()
-        gui_ready()
-        if not app.config.autoconnect or app.config.debug:
-            app.show()
-        app.run()
-    except KeyboardInterrupt:
-        pass
-    return 0
+            from xpra.scripts.main import parse_cmdline, fixup_debug_option
+            options, args = parse_cmdline(sys.argv)
+            debug = fixup_debug_option(options.debug)
+            if debug:
+                for x in debug.split(","):
+                    enable_debug_for(x)
+        except Exception:
+            exception_dialog("Error parsing command line")
+            return 1
+
+        try:
+            app = ApplicationWindow()
+            def app_signal(signum, frame):
+                print("")
+                log("got signal %s" % SIGNAMES.get(signum, signum))
+                def show_signal():
+                    app.show()
+                    app.client.cleanup()
+                    glib.timeout_add(1000, app.set_info_text, "got signal %s" % SIGNAMES.get(signum, signum))
+                    glib.timeout_add(1000, app.set_info_color, True)
+                #call from UI thread:
+                glib.idle_add(show_signal)
+            if sys.version_info[0]<3:
+                #breaks GTK3..
+                signal.signal(signal.SIGINT, app_signal)
+            signal.signal(signal.SIGTERM, app_signal)
+            has_file = len(args) == 1
+            if has_file:
+                app.update_options_from_file(args[0])
+                #the compressors and packet encoders cannot be changed from the UI
+                #so apply them now:
+                configure_network(app.config)
+            debug = fixup_debug_option(app.config.debug)
+            if debug:
+                for x in debug.split(","):
+                    enable_debug_for(x)
+            #suspend tray workaround for our window widgets:
+            try:
+                set_use_tray_workaround(False)
+                app.create_window()
+            finally:
+                set_use_tray_workaround(True)
+            app.update_gui_from_config()
+        except Exception:
+            exception_dialog("Error creating launcher form")
+            return 1
+        try:
+            if app.config.autoconnect:
+                #file says we should connect,
+                #do that only (not showing UI unless something goes wrong):
+                glib.idle_add(app.do_connect)
+            if not has_file:
+                app.reset_errors()
+            gui_ready()
+            if not app.config.autoconnect or app.config.debug:
+                app.show()
+            app.run()
+        except KeyboardInterrupt:
+            pass
+        return 0
+    finally:
+        platform_clean()
 
 
 if __name__ == "__main__":
     v = main()
-    sys.exit(v)
+    sys.exit(v)
\ No newline at end of file
