@@ -1,40 +1,48 @@
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2014 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2015 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-import os
+import os, sys
+import weakref
 from xpra.gtk_common.gobject_compat import import_gobject, import_gtk, import_gdk, is_gtk3
+from xpra.client.gtk_base.gtk_client_window_base import HAS_X11_BINDINGS
 gobject = import_gobject()
 gtk = import_gtk()
 gdk = import_gdk()
 
 
 from xpra.log import Logger
-log = Logger("gtk", "main")
+log = Logger("gtk", "client")
 opengllog = Logger("gtk", "opengl")
 cursorlog = Logger("gtk", "client", "cursor")
 screenlog = Logger("gtk", "client", "screen")
+framelog = Logger("gtk", "client", "frame")
+menulog = Logger("gtk", "client", "menu")
+filelog = Logger("gtk", "client", "file")
 
 from xpra.gtk_common.quit import (gtk_main_quit_really,
                            gtk_main_quit_on_fatal_exceptions_enable)
-from xpra.util import bytestostr, pver, DEFAULT_METADATA_SUPPORTED
+from xpra.util import bytestostr, updict, pver, iround, DEFAULT_METADATA_SUPPORTED
 from xpra.gtk_common.cursor_names import cursor_types
-from xpra.gtk_common.gtk_util import get_gtk_version_info, scaled_image, default_Cursor, \
+from xpra.gtk_common.gtk_util import get_gtk_version_info, scaled_image, get_default_cursor, \
             new_Cursor_for_display, new_Cursor_from_pixbuf, icon_theme_get_default, \
-            pixbuf_new_from_file, display_get_default, screen_get_default, get_pixbuf_from_data, INTERP_BILINEAR
+            pixbuf_new_from_file, display_get_default, screen_get_default, get_pixbuf_from_data, \
+            get_default_root_window, get_root_size, get_xwindow, \
+            INTERP_BILINEAR, WINDOW_TOPLEVEL
 from xpra.client.ui_client_base import UIXpraClient
 from xpra.client.gobject_client_base import GObjectXpraClient
 from xpra.client.gtk_base.gtk_keyboard_helper import GTKKeyboardHelper
 from xpra.client.gtk_base.session_info import SessionInfo
 from xpra.platform.paths import get_icon_filename
-from xpra.platform.gui import system_bell, get_workarea, get_workareas, get_fixed_cursor_size
+from xpra.platform.gui import get_window_frame_sizes, get_window_frame_size, system_bell, get_workarea, get_workareas, get_fixed_cursor_size, get_menu_support_function
 
 missing_cursor_names = set()
 
 METADATA_SUPPORTED = os.environ.get("XPRA_METADATA_SUPPORTED")
+USE_LOCAL_CURSORS = os.environ.get("XPRA_USE_LOCAL_CURSORS", "1")=="1"
 
 
 class GTKXpraClient(UIXpraClient, GObjectXpraClient):
@@ -53,13 +61,40 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
         self.client_supports_opengl = False
         self.opengl_enabled = False
         self.opengl_props = {}
+        self.gl_max_viewport_dims = 0, 0
         self.gl_texture_size_limit = 0
+        self._cursors = weakref.WeakKeyDictionary()
+        #frame request hidden window:
+        self.frame_request_window = None
+        #group leader bits:
+        self._ref_to_group_leader = {}
+        self._group_leader_wids = {}
+        self._set_window_menu = get_menu_support_function()
+        self.connect("scaling-changed", self.reset_windows_cursors)
+
 
     def init(self, opts):
         GObjectXpraClient.init(self, opts)
         UIXpraClient.init(self, opts)
 
+
+    def setup_frame_request_windows(self):
+        #query the window manager to get the frame size:
+        from xpra.gtk_common.error import xsync
+        from xpra.x11.gtk_x11.send_wm import send_wm_request_frame_extents
+        self.frame_request_window = gtk.Window(WINDOW_TOPLEVEL)
+        self.frame_request_window.set_title("Xpra-FRAME_EXTENTS")
+        root = self.get_root_window()
+        self.frame_request_window.realize()
+        with xsync:
+            win = self.frame_request_window.get_window()
+            framelog("setup_frame_request_windows() window=%#x", get_xwindow(win))
+            send_wm_request_frame_extents(root, win)
+
     def run(self):
+        log("run() HAS_X11_BINDINGS=%s", HAS_X11_BINDINGS)
+        if HAS_X11_BINDINGS:
+            self.setup_frame_request_windows()
         UIXpraClient.run(self)
         gtk_main_quit_on_fatal_exceptions_enable()
         self.gtk_main()
@@ -75,16 +110,17 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
             self.exit_code = exit_code
         if gtk.main_level()>0:
             #if for some reason cleanup() hangs, maybe this will fire...
-            gobject.timeout_add(4*1000, self.exit)
+            self.timeout_add(4*1000, self.exit)
             #try harder!:
             def force_quit():
                 from xpra import os_util
                 os_util.force_quit()
-            gobject.timeout_add(5*1000, force_quit)
+            self.timeout_add(5*1000, force_quit)
         self.cleanup()
+        log("GTKXpraClient.quit(%s) cleanup done, main_level=%s", exit_code, gtk.main_level())
         if gtk.main_level()>0:
             log("GTKXpraClient.quit(%s) main loop at level %s, calling gtk quit via timeout", exit_code, gtk.main_level())
-            gobject.timeout_add(500, self.exit)
+            self.timeout_add(500, self.exit)
 
     def exit(self):
         log("GTKXpraClient.exit() calling %s", gtk_main_quit_really)
@@ -107,12 +143,36 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
         log("show_start_new_command%s current start_new_command=%s, flag=%s", args, self.start_new_command, self.start_new_commands)
         if self.start_new_command is None:
             from xpra.client.gtk_base.start_new_command import getStartNewCommand
-            def run_command_cb(command):
-                self.send_start_command(command, command, False)
-            self.start_new_command = getStartNewCommand(run_command_cb)
+            def run_command_cb(command, sharing=True):
+                self.send_start_command(command, command, False, sharing)
+            self.start_new_command = getStartNewCommand(run_command_cb, self.server_supports_sharing and self.server_supports_window_filters)
         self.start_new_command.show()
         return self.start_new_command
 
+    def show_file_upload(self, *args):
+        filelog("show_file_upload%s can open=%s", args, self.server_open_files)
+        buttons = [gtk.STOCK_CANCEL,    gtk.RESPONSE_CANCEL]
+        if self.server_open_files:
+            buttons += [gtk.STOCK_OPEN,      gtk.RESPONSE_ACCEPT]
+        buttons += [gtk.STOCK_OK,        gtk.RESPONSE_OK]
+        dialog = gtk.FileChooserDialog("File to upload", parent=None, action=gtk.FILE_CHOOSER_ACTION_OPEN, buttons=tuple(buttons))
+        dialog.set_default_response(gtk.RESPONSE_OK)
+        v = dialog.run()
+        if v not in (gtk.RESPONSE_OK, gtk.RESPONSE_ACCEPT):
+            filelog("dialog response code %s", v)
+            dialog.destroy()
+            return
+        filename = dialog.get_filename()
+        gfile = dialog.get_file()
+        data, filesize, entity = gfile.load_contents()
+        filelog("load_contents: filename=%s, %i bytes, entity=%s, response=%s", filename, filesize, entity, v)
+        dialog.destroy()
+        self.send_file(filename, data, filesize, openit=(v==gtk.RESPONSE_ACCEPT))
+
+
+    def show_about(self, *args):
+        from xpra.client.gtk_base.about import about
+        about()
 
     def show_session_info(self, *args):
         if self.session_info and not self.session_info.is_closed:
@@ -128,10 +188,10 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
         self.session_info.show_all()
 
     def show_bug_report(self, *args):
+        self.send_info_request()
         if self.bug_report:
             self.bug_report.show()
             return
-        self.send_info_request()
         from xpra.client.gtk_base.bug_report import BugReport
         self.bug_report = BugReport()
         def init_bug_report():
@@ -175,6 +235,46 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
             return  None
 
 
+    def request_frame_extents(self, window):
+        from xpra.x11.gtk_x11.send_wm import send_wm_request_frame_extents
+        from xpra.gtk_common.error import xsync
+        root = self.get_root_window()
+        with xsync:
+            win = window.get_window()
+            framelog("request_frame_extents(%s) xid=%#x", window, get_xwindow(win))
+            send_wm_request_frame_extents(root, win)
+
+    def get_frame_extents(self, window):
+        #try native platform code first:
+        x, y = window.get_position()
+        w, h = window.get_size()
+        v = get_window_frame_size(x, y, w, h)
+        framelog("get_window_frame_size%s=%s", (x, y, w, h), v)
+        if v:
+            #(OSX does give us these values via Quartz API)
+            return v
+        if not HAS_X11_BINDINGS:
+            #nothing more we can do!
+            return None
+        from xpra.x11.gtk_x11.prop import prop_get
+        gdkwin = window.get_window()
+        assert gdkwin
+        v = prop_get(gdkwin, "_NET_FRAME_EXTENTS", ["u32"], ignore_errors=False)
+        framelog("get_frame_extents(%s)=%s", window.get_title(), v)
+        return v
+
+    def get_window_frame_sizes(self):
+        wfs = get_window_frame_sizes()
+        if self.frame_request_window:
+            v = self.get_frame_extents(self.frame_request_window)
+            if v:
+                l, r, t, b = v
+                wfs["frame"] = (l, r, t, b)
+                wfs["offset"] = (l, t)
+        framelog("get_window_frame_sizes()=%s", wfs)
+        return wfs
+
+
     def make_keyboard_helper(self, keyboard_sync, key_shortcuts):
         return GTKKeyboardHelper(self.send, keyboard_sync, key_shortcuts)
 
@@ -200,14 +300,33 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
         return True
 
 
+    def cook_metadata(self, new_window, metadata):
+        metadata = UIXpraClient.cook_metadata(self, new_window, metadata)
+        #ensures we will call set_window_menu for this window when we create it:
+        if new_window and b"menu" not in metadata and self._set_window_menu:
+            metadata[b"menu"] = {}
+        return metadata
+
+
+    def set_window_menu(self, add, wid, menus, application_action_callback=None, window_action_callback=None):
+        assert self._set_window_menu
+        model = self._id_to_window.get(wid)
+        window = None
+        if model:
+            window = model.get_window()
+        self._set_window_menu(add, wid, window, menus, application_action_callback, window_action_callback)
+
+
     def get_root_window(self):
-        raise Exception("override me!")
+        return get_default_root_window()
 
     def get_root_size(self):
-        raise Exception("override me!")
+        return get_root_size()
+
 
     def get_mouse_position(self):
-        return self.get_root_window().get_pointer()[:2]
+        p = self.get_root_window().get_pointer()
+        return self.cp(p[0], p[1])
 
     def get_current_modifiers(self):
         modifiers_mask = self.get_root_window().get_pointer()[-1]
@@ -249,17 +368,32 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
             #this is only really supported on X11, but posix is easier to check for..
             #"strut" and maybe even "fullscreen-monitors" could also be supported on other platforms I guess
             ms += ["shaded", "bypass-compositor", "strut", "fullscreen-monitors"]
+        if HAS_X11_BINDINGS:
+            ms += ["shape"]
+        if self._set_window_menu:
+            ms += ["menu"]
+        #figure out if we can handle the "global menu" stuff:
+        if os.name=="posix" and not sys.platform.startswith("darwin"):
+            try:
+                from xpra.dbus.helper import DBusHelper
+                assert DBusHelper
+            except:
+                pass
         log("metadata.supported: %s", ms)
         capabilities["metadata.supported"] = ms
         #we need the bindings to support initiate-moveresize (posix only for now):
-        from xpra.client.gtk_base.gtk_client_window_base import HAS_X11_BINDINGS
-        capabilities["window.initiate-moveresize"] = HAS_X11_BINDINGS
-        #window icon bits
-        capabilities["encoding.icons.greedy"] = True            #we don't set a default window icon any more
-        capabilities["encoding.icons.size"] = 64, 64            #size we want
-        capabilities["encoding.icons.max_size"] = 128, 128      #limit
+        updict(capabilities, "window", {
+               "initiate-moveresize"    : HAS_X11_BINDINGS,
+               "configure.pointer"      : True,
+               "frame_sizes"            : self.get_window_frame_sizes()
+               })
         from xpra.client.window_backing_base import DELTA_BUCKETS
-        capabilities["encoding.delta_buckets"] = DELTA_BUCKETS
+        updict(capabilities, "encoding", {
+                    "icons.greedy"      : True,         #we don't set a default window icon any more
+                    "icons.size"        : (64, 64),     #size we want
+                    "icons.max_size"    : (128, 128),   #limit
+                    "delta_buckets"     : DELTA_BUCKETS,
+                    })
         return capabilities
 
 
@@ -267,12 +401,18 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
         return screen_get_default().get_rgba_visual() is not None
 
 
-    def get_screen_sizes(self):
+    def get_screen_sizes(self, xscale=1, yscale=1):
+        def xs(v):
+            return iround(v/xscale)
+        def ys(v):
+            return iround(v/yscale)
+        def swork(*workarea):
+            return xs(workarea[0]), ys(workarea[1]), xs(workarea[2]), ys(workarea[3])
         display = display_get_default()
         i=0
         screen_sizes = []
         n_screens = display.get_n_screens()
-        screenlog("get_screen_sizes() found %s screens", n_screens)
+        screenlog("get_screen_sizes(%f, %f) found %s screens", xscale, yscale, n_screens)
         while i<n_screens:
             screen = display.get_screen(i)
             j = 0
@@ -282,11 +422,12 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
             #and it is only implemented on win32 right now
             #other platforms only implement "get_workarea()" instead, which is reported against the screen
             n_monitors = screen.get_n_monitors()
-            screenlog("get_screen_sizes() screen %s has %s monitors", i, n_monitors)
+            screenlog(" screen %s has %s monitors", i, n_monitors)
             if n_screens==1:
                 workareas = get_workareas()
                 if len(workareas)!=n_monitors:
-                    screenlog("number of monitors does not match number of workareas!")
+                    screenlog(" workareas: %s", workareas)
+                    screenlog(" number of monitors does not match number of workareas!")
                     workareas = []
             while j<screen.get_n_monitors():
                 geom = screen.get_monitor_geometry(j)
@@ -299,28 +440,34 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
                 hmm = -1
                 if hasattr(screen, "get_monitor_height_mm"):
                     hmm = screen.get_monitor_height_mm(j)
-                monitor = [plug_name, geom.x, geom.y, geom.width, geom.height, wmm, hmm]
-                screenlog("get_screen_sizes() monitor %s: %s", j, monitor)
+                monitor = [plug_name, xs(geom.x), ys(geom.y), xs(geom.width), ys(geom.height), wmm, hmm]
+                screenlog(" monitor %s: %s", j, monitor)
                 if workareas:
                     w = workareas[j]
-                    monitor += list(w)
+                    monitor += list(swork(*w))
                 monitors.append(tuple(monitor))
                 j += 1
-            work_x, work_y = 0, 0
-            work_width, work_height = screen.get_width(), screen.get_height()
+            work_x, work_y, work_width, work_height = swork(0, 0, screen.get_width(), screen.get_height())
             workarea = get_workarea()
             if workarea:
-                work_x, work_y, work_width, work_height = workarea
-            screenlog("get_screen_sizes() workarea=%s", workarea)
-            item = (screen.make_display_name(), screen.get_width(), screen.get_height(),
+                work_x, work_y, work_width, work_height = swork(*workarea)
+            screenlog(" workarea=%s", workarea)
+            item = (screen.make_display_name(), xs(screen.get_width()), ys(screen.get_height()),
                         screen.get_width_mm(), screen.get_height_mm(),
                         monitors,
                         work_x, work_y, work_width, work_height)
-            screenlog("get_screen_sizes() screen %s: %s", i, item)
+            screenlog(" screen %s: %s", i, item)
             screen_sizes.append(item)
             i += 1
         return screen_sizes
 
+
+    def reset_windows_cursors(self, *args):
+        cursorlog("reset_windows_cursors() resetting cursors for: %s", self._cursors.keys())
+        for w,cursor_data in list(self._cursors.items()):
+            self.set_windows_cursor([w], cursor_data)
+
+
     def set_windows_cursor(self, windows, cursor_data):
         cursorlog("set_windows_cursor(%s, ..)", windows)
         cursor = None
@@ -332,19 +479,22 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
                 log.warn("error creating cursor: %s (using default)", e, exc_info=True)
             if cursor is None:
                 #use default:
-                cursor = default_Cursor
+                cursor = get_default_cursor()
         for w in windows:
             gdkwin = w.get_window()
             #trays don't have a gdk window
             if gdkwin:
+                self._cursors[w] = cursor_data
                 gdkwin.set_cursor(cursor)
 
     def make_cursor(self, cursor_data):
         #if present, try cursor ny name:
         display = display_get_default()
-        if len(cursor_data)>=9 and cursor_types:
+        cursorlog("make_cursor: has-name=%s, has-cursor-types=%s, xscale=%s, yscale=%s, USE_LOCAL_CURSORS=%s", len(cursor_data)>=9, bool(cursor_types), self.xscale, self.yscale, USE_LOCAL_CURSORS)
+        #named cursors cannot be scaled (round to 10 to compare so 0.95 and 1.05 are considered the same as 1.0, no scaling):
+        if len(cursor_data)>=9 and cursor_types and iround(self.xscale*10)==10 and iround(self.yscale*10)==10:
             cursor_name = bytestostr(cursor_data[8])
-            if cursor_name:
+            if cursor_name and USE_LOCAL_CURSORS:
                 gdk_cursor = cursor_types.get(cursor_name.upper())
                 if gdk_cursor is not None:
                     cursorlog("setting new cursor by name: %s=%s", cursor_name, gdk_cursor)
@@ -381,15 +531,42 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
                 cursorlog("scaling cursor from %ix%i to fixed OS size %ix%i", w, h, fw, fh)
                 cursor_pixbuf = pixbuf.scale_simple(fw, fh, INTERP_BILINEAR)
                 xratio, yratio = float(w)/fw, float(h)/fh
-                x, y = int(x/xratio), int(y/yratio)
-        elif w>cmaxw or h>cmaxh or (csize>0 and (csize<w or csize<h)):
-            ratio = max(float(w)/cmaxw, float(h)/cmaxh, float(max(w,h))/csize)
-            x, y, w, h = int(x/ratio), int(y/ratio), int(w/ratio), int(h/ratio)
-            cursorlog("downscaling cursor %s by %.2f: %sx%s", pixbuf, ratio, w, h)
-            cursor_pixbuf = pixbuf.scale_simple(w, h, INTERP_BILINEAR)
+                x, y = iround(x/xratio), iround(y/yratio)
         else:
-            cursor_pixbuf = pixbuf
-        return new_Cursor_from_pixbuf(display, cursor_pixbuf, x, y)
+            sx, sy, sw, sh = x, y, w, h
+            #scale the cursors:
+            if self.xscale!=1 or self.yscale!=1:
+                sx, sy, sw, sh = self.srect(x, y, w, h)
+            sw = max(1, sw)
+            sh = max(1, sh)
+            #ensure we honour the max size if there is one:
+            if (cmaxw>0 and sw>cmaxw) or (cmaxh>0 and sh>cmaxh):
+                ratio = 1.0
+                if cmaxw>0:
+                    ratio = max(ratio, float(w)/cmaxw)
+                if cmaxh>0:
+                    ratio = max(ratio, float(h)/cmaxh)
+                cursorlog("clamping cursor size to %ix%i using ratio=%s", cmaxw, cmaxh, ratio)
+                sx, sy, sw, sh = iround(x/ratio), iround(y/ratio), min(cmaxw, iround(w/ratio)), min(cmaxh, iround(h/ratio))
+            if sw!=w or sh!=h:
+                cursorlog("scaling cursor from %ix%i hotspot at %ix%i to %ix%i hotspot at %ix%i", w, h, x, y, sw, sh, sx, sy)
+                cursor_pixbuf = pixbuf.scale_simple(sw, sh, INTERP_BILINEAR)
+                x, y = sx, sy
+            else:
+                cursor_pixbuf = pixbuf
+        #clamp to pixbuf size:
+        w = cursor_pixbuf.get_width()
+        h = cursor_pixbuf.get_height()
+        x = max(0, min(x, w-1))
+        y = max(0, min(y, h-1))
+        try:
+            c = new_Cursor_from_pixbuf(display, cursor_pixbuf, x, y)
+        except RuntimeError as e:
+            log.error("Error: failed to create cursor:")
+            log.error(" %s", e)
+            log.error(" using %s of size %ix%i with hotspot at %ix%i", cursor_pixbuf, w, h, x, y)
+            c = None
+        return c
 
 
     def process_ui_capabilities(self):
@@ -426,7 +603,9 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
         from xpra.platform.gui import gl_check as platform_gl_check
         warnings = []
         for check in (OpenGL_safety_check, platform_gl_check):
+            opengllog("checking with %s", check)
             warning = check()
+            opengllog("%s()=%s", check, warning)
             if warning:
                 warnings.append(warning)
         self.opengl_props["info"] = ""
@@ -458,10 +637,19 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
             #only enable opengl by default if force-enabled or if safe to do so:
             self.opengl_enabled = (enable_opengl is True) or self.opengl_props.get("safe", False)
             self.gl_texture_size_limit = self.opengl_props.get("texture-size-limit", 16*1024)
-            if self.gl_texture_size_limit<4*1024:
-                opengllog.warn("OpenGL disabled: the texture size limit is too low (%s)", self.gl_texture_size_limit)
+            self.gl_max_viewport_dims = self.opengl_props.get("max-viewport-dims", (self.gl_texture_size_limit, self.gl_texture_size_limit))
+            if min(self.gl_max_viewport_dims)<4*1024:
+                opengllog.warn("Warning: OpenGL is disabled:")
+                opengllog.warn(" the maximum viewport size is too low: %s", self.gl_max_viewport_dims)
                 self.opengl_enabled = False
-            self.GLClientWindowClass.MAX_TEXTURE_SIZE = self.gl_texture_size_limit
+            elif self.gl_texture_size_limit<4*1024:
+                opengllog.warn("Warning: OpenGL is disabled:")
+                opengllog.warn(" the texture size limit is too low: %s", self.gl_texture_size_limit)
+                self.opengl_enabled = False
+            self.GLClientWindowClass.MAX_VIEWPORT_DIMS = self.gl_max_viewport_dims
+            self.GLClientWindowClass.MAX_BACKING_DIMS = self.gl_texture_size_limit, self.gl_texture_size_limit
+            self.GLClientWindowClass.MAX_VIEWPORT_DIMS = 8192, 8192
+            self.GLClientWindowClass.MAX_BACKING_DIMS = 4096, 4096
             mww, mwh = self.max_window_size
             opengllog("OpenGL: enabled=%s, texture-size-limit=%s, max-window-size=%s", self.opengl_enabled, self.gl_texture_size_limit, self.max_window_size)
             if self.opengl_enabled and self.gl_texture_size_limit<16*1024 and (mww==0 or mwh==0 or self.gl_texture_size_limit<mww or self.gl_texture_size_limit<mwh):
@@ -479,7 +667,11 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
             elif self.client_supports_opengl:
                 opengllog("OpenGL supported with %s, but not enabled", driver_info)
         except ImportError as e:
-            opengllog.warn("OpenGL support could not be enabled:")
+            opengllog.warn("OpenGL support is missing:")
+            opengllog.warn(" %s", e)
+            self.opengl_props["info"] = str(e)
+        except RuntimeError as e:
+            opengllog.warn("OpenGL support could not be enabled on this hardware:")
             opengllog.warn(" %s", e)
             self.opengl_props["info"] = str(e)
         except Exception as e:
@@ -489,7 +681,8 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
 
     def get_client_window_classes(self, w, h, metadata, override_redirect):
         log("get_client_window_class(%i, %i, %s, %s) GLClientWindowClass=%s, opengl_enabled=%s, mmap_enabled=%s, encoding=%s", w, h, metadata, override_redirect, self.GLClientWindowClass, self.opengl_enabled, self.mmap_enabled, self.encoding)
-        if self.GLClientWindowClass is None or not self.opengl_enabled or w>self.gl_texture_size_limit or h>self.gl_texture_size_limit:
+        ms = min(self.sx(self.gl_texture_size_limit), *self.gl_max_viewport_dims)
+        if self.GLClientWindowClass is None or not self.opengl_enabled or w>ms or h>ms:
             return [self.ClientWindowClass]
         return [self.GLClientWindowClass, self.ClientWindowClass]
 
@@ -509,7 +702,11 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
             window.send = fake_send
             #copy attributes:
             x, y = window._pos
-            w, h = window._size
+            ww, wh = window._size
+            try:
+                bw, bh = window._backing.size
+            except:
+                bw, bh = ww, wh
             client_properties = window._client_properties
             metadata = window._metadata
             override_redirect = window._override_redirect
@@ -528,6 +725,7 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
                         backing._video_decoder = None
                         backing._csc_decoder = None
                         backing._decoder_lock = None
+                        backing.close()
 
                 #now we can unmap it:
                 self.destroy_window(wid, window)
@@ -543,7 +741,7 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
                 except:
                     pass
                 #create the new window, which should honour the new state of the opengl_enabled flag:
-                window = self.make_new_window(wid, x, y, w, h, metadata, override_redirect, client_properties)
+                window = self.make_new_window(wid, x, y, ww, wh, bw, bh, metadata, override_redirect, client_properties)
                 if video_decoder or csc_decoder:
                     backing = window._backing
                     backing._video_decoder = video_decoder
@@ -553,3 +751,103 @@ class GTKXpraClient(UIXpraClient, GObjectXpraClient):
                 if decoder_lock:
                     decoder_lock.release()
         opengllog("replaced all the windows with opengl=%s: %s", self.opengl_enabled, self._id_to_window)
+
+
+    def get_group_leader(self, wid, metadata, override_redirect):
+        transient_for = metadata.intget("transient-for", -1)
+        log("get_group_leader: transient_for=%s", transient_for)
+        if transient_for>0:
+            client_window = self._id_to_window.get(transient_for)
+            if client_window:
+                gdk_window = client_window.get_window()
+                if gdk_window:
+                    return gdk_window
+        pid = metadata.intget("pid", -1)
+        leader_xid = metadata.intget("group-leader-xid", -1)
+        leader_wid = metadata.intget("group-leader-wid", -1)
+        group_leader_window = self._id_to_window.get(leader_wid)
+        if group_leader_window:
+            #leader is another managed window
+            log("found group leader window %s for wid=%s", group_leader_window, leader_wid)
+            return group_leader_window
+        log("get_group_leader: leader pid=%s, xid=%s, wid=%s", pid, leader_xid, leader_wid)
+        reftype = "xid"
+        ref = leader_xid
+        if ref<0:
+            reftype = "leader-wid"
+            ref = leader_wid
+        if ref<0:
+            ci = metadata.strlistget("class-instance")
+            if ci:
+                reftype = "class"
+                ref = "|".join(ci)
+            elif pid>0:
+                reftype = "pid"
+                ref = pid
+            elif transient_for>0:
+                #this should have matched a client window above..
+                #but try to use it anyway:
+                reftype = "transient-for"
+                ref = transient_for
+            else:
+                #no reference to use
+                return None
+        refkey = "%s:%s" % (reftype, ref)
+        group_leader_window = self._ref_to_group_leader.get(refkey)
+        if group_leader_window:
+            log("found existing group leader window %s using ref=%s", group_leader_window, refkey)
+            return group_leader_window
+        #we need to create one:
+        title = "%s group leader for %s" % (self.session_name or "Xpra", pid)
+        #group_leader_window = gdk.Window(None, 1, 1, gdk.WINDOW_TOPLEVEL, 0, gdk.INPUT_ONLY, title)
+        #static new(parent, attributes, attributes_mask)
+        if is_gtk3():
+            #long winded and annoying
+            attributes = gdk.WindowAttr()
+            attributes.width = 1
+            attributes.height = 1
+            attributes.title = title
+            attributes.wclass = gdk.WindowWindowClass.INPUT_ONLY
+            attributes.event_mask = 0
+            attributes_mask = gdk.WindowAttributesType.TITLE | gdk.WindowAttributesType.WMCLASS
+            group_leader_window = gdk.Window(None, attributes, attributes_mask)
+            group_leader_window.resize(1, 1)
+        else:
+            #gtk2:
+            group_leader_window = gdk.Window(None, 1, 1, gdk.WINDOW_TOPLEVEL, 0, gdk.INPUT_ONLY, title)
+        self._ref_to_group_leader[refkey] = group_leader_window
+        #avoid warning on win32...
+        if not sys.platform.startswith("win"):
+            #X11 spec says window should point to itself:
+            group_leader_window.set_group(group_leader_window)
+        log("new hidden group leader window %s for ref=%s", group_leader_window, refkey)
+        self._group_leader_wids.setdefault(group_leader_window, []).append(wid)
+        return group_leader_window
+
+    def destroy_window(self, wid, window):
+        #override so we can cleanup the group-leader if needed,
+        UIXpraClient.destroy_window(self, wid, window)
+        group_leader = window.group_leader
+        if group_leader is None or len(self._group_leader_wids)==0:
+            return
+        wids = self._group_leader_wids.get(group_leader)
+        if wids is None:
+            #not recorded any window ids on this group leader
+            #means it is another managed window, leave it alone
+            return
+        if wid in wids:
+            wids.remove(wid)
+        if len(wids)>0:
+            #still has another window pointing to it
+            return
+        #the last window has gone, we can remove the group leader,
+        #find all the references to this group leader:
+        del self._group_leader_wids[group_leader]
+        refs = []
+        for ref, gl in self._ref_to_group_leader.items():
+            if gl==group_leader:
+                refs.append(ref)
+        for ref in refs:
+            del self._ref_to_group_leader[ref]
+        log("last window for refs %s is gone, destroying the group leader %s", refs, group_leader)
+        group_leader.destroy()
