@@ -6,11 +6,11 @@
 # later version. See the file COPYING for details.
 
 from xpra.client.keyboard_helper import KeyboardHelper, log
-from xpra.gtk_common.gobject_compat import import_gdk, import_gobject
+from xpra.gtk_common.gobject_compat import import_gdk, import_glib
 from xpra.gtk_common.keymap import get_gtk_keymap
 from xpra.gtk_common.gtk_util import display_get_default
 gdk = import_gdk()
-gobject = import_gobject()
+glib = import_glib()
 
 
 class GTKKeyboardHelper(KeyboardHelper):
@@ -27,18 +27,21 @@ class GTKKeyboardHelper(KeyboardHelper):
             self._keymap = None
         self.update()
         if self._keymap:
-            self._keymap_change_handler_id = self._keymap.connect("keys-changed", self._keys_changed)
+            self._keymap_change_handler_id = self._keymap.connect("keys-changed", self.keymap_changed)
 
-    def _keys_changed(self, *args):
-        log("keys_changed")
-        self._keymap.disconnect(self._keymap_change_handler_id)
+    def keymap_changed(self, *args):
+        log("keymap_changed%s", args)
+        if self._keymap_change_handler_id:
+            self._keymap.disconnect(self._keymap_change_handler_id)
+            self._keymap_change_handler_id = None
         self._keymap = gdk.keymap_get_default()
-        self._keymap_change_handler_id = self._keymap.connect("keys-changed", self._keys_changed)
         if self._keymap_changing:
             #timer due already
             return
         self._keymap_changing = True
         def do_keys_changed():
+            #re-register the change handler:
+            self._keymap_change_handler_id = self._keymap.connect("keys-changed", self.keymap_changed)
             self._keymap_changing = False
             if self.locked:
                 #automatic changes not allowed!
@@ -49,7 +52,7 @@ class GTKKeyboardHelper(KeyboardHelper):
                 if self.xkbmap_layout:
                     self.send_layout()
                 self.send_keymap()
-        gobject.timeout_add(500, do_keys_changed)
+        glib.timeout_add(500, do_keys_changed)
 
     def update(self):
         old_hash = self.hash
@@ -64,19 +67,31 @@ class GTKKeyboardHelper(KeyboardHelper):
     def get_full_keymap(self):
         return  get_gtk_keymap()
 
+    def cleanup(self):
+        KeyboardHelper.cleanup(self)
+        if self._keymap_change_handler_id:
+            try:
+                self._keymap.disconnect(self._keymap_change_handler_id)
+                self._keymap_change_handler_id = None
+            except Exception as e:
+                log.warn("failed to disconnect keymap change handler: %s", e)
+
 
 def main():
     #use gtk as display source:
-    from xpra.x11.gtk_x11 import gdk_display_source
+    from xpra.x11.gtk2 import gdk_display_source
     assert gdk_display_source
     from xpra.util import nonl
-    from xpra.platform import init as platform_init
-    platform_init("GTK-Keyboard", "GTK Keyboard")
+    from xpra.platform import init as platform_init, clean as platform_clean
+    try:
+        platform_init("GTK-Keyboard", "GTK Keyboard")
 
-    x = GTKKeyboardHelper(None, True, "")
-    x.query_xkbmap()
-    for k,v in x.get_keymap_properties().items():
-        print("%s=%s" % (k,nonl(v)))
+        x = GTKKeyboardHelper(None, True, "")
+        x.query_xkbmap()
+        for k,v in x.get_keymap_properties().items():
+            print("%s=%s" % (k,nonl(v)))
+    finally:
+        platform_clean()
 
 if __name__ == "__main__":
     main()
