@@ -6,18 +6,21 @@
 
 import sys
 import os
-from xpra.gtk_common.gobject_compat import import_gtk, import_gobject, is_gtk3
+from xpra.gtk_common.gobject_compat import import_gtk, import_glib, is_gtk3
 gtk = import_gtk()
-gobject = import_gobject()
+glib = import_glib()
 
-from xpra.util import CLIENT_EXIT
+from xpra.util import CLIENT_EXIT, iround
+from xpra.os_util import bytestostr
 from xpra.gtk_common.gtk_util import ensure_item_selected, menuitem, BUTTON_PRESS_MASK
 from xpra.client.client_base import EXIT_OK
 from xpra.client.gtk_base.about import about, close_about
 from xpra.codecs.loader import PREFERED_ENCODING_ORDER, ENCODINGS_HELP, ENCODINGS_TO_NAME
 from xpra.platform.gui import get_icon_size
+
 from xpra.log import Logger
-log = Logger("tray")
+log = Logger("menu")
+clipboardlog = Logger("menu", "clipboard")
 
 
 HIDE_DISABLED_MENU_ENTRIES = sys.platform.startswith("darwin")
@@ -76,15 +79,15 @@ class TrayCheckMenuItem(gtk.CheckMenuItem):
             if state is not None and state==self.get_active():
                 #toggle did not fire after the button release, so force it:
                 self.set_active(not state)
-        gobject.idle_add(recheck)
+        glib.idle_add(recheck)
 
-CheckMenuItem = TrayCheckMenuItem
 def set_use_tray_workaround(enabled):
     global CheckMenuItem
-    if enabled:
+    if enabled and sys.platform.startswith("win"):
         CheckMenuItem = TrayCheckMenuItem
     else:
         CheckMenuItem = gtk.CheckMenuItem
+set_use_tray_workaround(True)
 
 
 def set_sensitive(widget, sensitive):
@@ -255,26 +258,28 @@ class GTKTrayMenuBase(object):
         def set_menu_title(*args):
             #set the real name when available:
             self.menu.set_title(self.client.session_name)
-        self.client.connect("handshake-complete", set_menu_title)
+        self.client.after_handshake(set_menu_title)
 
         menu.append(self.make_aboutmenuitem())
         menu.append(self.make_sessioninfomenuitem())
         menu.append(self.make_bugreportmenuitem())
         menu.append(gtk.SeparatorMenuItem())
         menu.append(self.make_bellmenuitem())
+        menu.append(self.make_notificationsmenuitem())
         if self.client.windows_enabled:
             menu.append(self.make_cursorsmenuitem())
-        if not self.client.readonly and self.client.keyboard_helper:
-            menu.append(self.make_layoutsmenuitem())
-        if self.client.windows_enabled and not self.client.readonly:
-            menu.append(self.make_keyboardsyncmenuitem())
-        menu.append(self.make_notificationsmenuitem())
-        if not self.client.readonly:
-            menu.append(self.make_clipboardmenuitem())
         if self.client.client_supports_opengl:
             menu.append(self.make_openglmenuitem())
+        if self.client.windows_enabled and not self.client.readonly:
+            menu.append(self.make_keyboardsyncmenuitem())
+        if not self.client.readonly and self.client.keyboard_helper:
+            menu.append(self.make_layoutsmenuitem())
+        if not self.client.readonly:
+            menu.append(self.make_clipboardmenuitem())
         if self.client.windows_enabled and len(self.client.get_encodings())>1:
             menu.append(self.make_encodingsmenuitem())
+        if self.client.can_scale:
+            menu.append(self.make_scalingmenuitem())
         menu.append(self.make_qualitymenuitem())
         menu.append(self.make_speedmenuitem())
         if STARTSTOP_SOUND_MENU:
@@ -289,6 +294,7 @@ class GTKTrayMenuBase(object):
         #menu.append(item("Options", "configure", None, self.options))
         menu.append(gtk.SeparatorMenuItem())
         menu.append(self.make_startnewcommandmenuitem())
+        menu.append(self.make_uploadmenuitem())
         menu.append(self.make_disconnectmenuitem())
         if show_close:
             menu.append(self.make_closemenuitem())
@@ -329,9 +335,13 @@ class GTKTrayMenuBase(object):
         set_sensitive(mi, False)
         def enable_menuitem(*args):
             set_sensitive(mi, True)
-        self.client.connect("handshake-complete", enable_menuitem)
+        self.client.after_handshake(enable_menuitem)
         return mi
 
+
+    def make_menu(self):
+        return gtk.Menu()
+
     def menuitem(self, title, icon_name=None, tooltip=None, cb=None):
         """ Utility method for easily creating an ImageMenuItem """
         image = None
@@ -390,7 +400,7 @@ class GTKTrayMenuBase(object):
                 self.bell_menuitem.set_tooltip_text("Forward system bell")
             else:
                 self.bell_menuitem.set_tooltip_text("Cannot forward the system bell: the feature has been disabled")
-        self.client.connect("handshake-complete", set_bell_menuitem)
+        self.client.after_handshake(set_bell_menuitem)
         return  self.bell_menuitem
 
     def make_cursorsmenuitem(self):
@@ -415,7 +425,7 @@ class GTKTrayMenuBase(object):
                 self.cursors_menuitem.set_tooltip_text("Forward custom mouse cursors")
             else:
                 self.cursors_menuitem.set_tooltip_text("Cannot forward mouse cursors: the feature has been disabled")
-        self.client.connect("handshake-complete", set_cursors_menuitem)
+        self.client.after_handshake(set_cursors_menuitem)
         return  self.cursors_menuitem
 
     def make_notificationsmenuitem(self):
@@ -438,20 +448,21 @@ class GTKTrayMenuBase(object):
                 self.notifications_menuitem.set_tooltip_text("Forward system notifications")
             else:
                 self.notifications_menuitem.set_tooltip_text("Cannot forward system notifications: the feature has been disabled")
-        self.client.connect("handshake-complete", set_notifications_menuitem)
+        self.client.after_handshake(set_notifications_menuitem)
         return self.notifications_menuitem
 
     def make_clipboard_togglemenuitem(self):
+        clipboardlog("make_clipboard_togglemenuitem()")
         def menu_clipboard_toggled(*args):
             new_state = self.clipboard_menuitem.get_active()
-            log("clipboard_toggled(%s) clipboard_enabled=%s, new_state=%s", args, self.client.clipboard_enabled, new_state)
+            clipboardlog("clipboard_toggled(%s) clipboard_enabled=%s, new_state=%s", args, self.client.clipboard_enabled, new_state)
             if self.client.clipboard_enabled!=new_state:
                 self.client.clipboard_enabled = new_state
                 self.client.emit("clipboard-toggled")
         self.clipboard_menuitem = self.checkitem("Clipboard", menu_clipboard_toggled)
         set_sensitive(self.clipboard_menuitem, False)
         def set_clipboard_menuitem(*args):
-            log("set_clipboard_menuitem%s enabled=%s", args, self.client.clipboard_enabled)
+            clipboardlog("set_clipboard_menuitem%s enabled=%s", args, self.client.clipboard_enabled)
             self.clipboard_menuitem.set_active(self.client.clipboard_enabled)
             c = self.client
             can_clipboard = c.server_supports_clipboard and c.client_supports_clipboard
@@ -460,7 +471,7 @@ class GTKTrayMenuBase(object):
                 self.clipboard_menuitem.set_tooltip_text("Enable clipboard synchronization")
             else:
                 self.clipboard_menuitem.set_tooltip_text("Clipboard synchronization cannot be enabled: disabled by server")
-        self.client.connect("handshake-complete", set_clipboard_menuitem)
+        self.client.after_handshake(set_clipboard_menuitem)
         def clipboard_toggled(*args):
             #keep menu in sync with actual "clipboard_enabled" flag:
             if self.client.clipboard_enabled != self.clipboard_menuitem.get_active():
@@ -469,6 +480,7 @@ class GTKTrayMenuBase(object):
         return self.clipboard_menuitem
 
     def make_translatedclipboard_optionsmenuitem(self):
+        clipboardlog("make_translatedclipboard_optionsmenuitem()")
         clipboard_menu = self.menuitem("Clipboard", "clipboard.png", "Choose which remote clipboard to connect to", None)
         set_sensitive(clipboard_menu, False)
         def set_clipboard_menu(*args):
@@ -477,7 +489,7 @@ class GTKTrayMenuBase(object):
             self.popup_menu_workaround(clipboard_submenu)
             c = self.client
             can_clipboard = c.server_supports_clipboard and c.client_supports_clipboard and c.server_supports_clipboard
-            log("set_clipboard_menu(%s) can_clipboard=%s, server=%s, client=%s", args, can_clipboard, c.server_supports_clipboard, c.client_supports_clipboard)
+            clipboardlog("set_clipboard_menu(%s) can_clipboard=%s, server=%s, client=%s", args, can_clipboard, c.server_supports_clipboard, c.client_supports_clipboard)
             set_sensitive(clipboard_menu, can_clipboard)
             LABEL_TO_NAME = {"Disabled"  : None,
                             "Clipboard" : "CLIPBOARD",
@@ -492,20 +504,25 @@ class GTKTrayMenuBase(object):
                     label = item.get_label()
                     remote_clipboard = LABEL_TO_NAME.get(label)
                     old_state = self.client.clipboard_enabled
-                    log("remote_clipboard_changed(%s) remote_clipboard=%s, old_state=%s", item, remote_clipboard, old_state)
+                    clipboardlog("remote_clipboard_changed(%s) remote_clipboard=%s, old_state=%s", item, remote_clipboard, old_state)
                     send_tokens = False
                     if remote_clipboard is not None:
                         #clipboard is not disabled
                         if self.client.clipboard_helper is None:
                             self.client.setup_clipboard_helper(TranslatedClipboardProtocolHelper)
                         self.client.clipboard_helper.remote_clipboard = remote_clipboard
+                        self.client.clipboard_helper.remote_clipboards = [remote_clipboard]
                         send_tokens = True
                         new_state = True
+                        selections = [remote_clipboard]
                     else:
                         self.client.clipboard_helper = None
                         send_tokens = False
                         new_state = False
-                    log("remote_clipboard_changed(%s) label=%s, remote_clipboard=%s, old_state=%s, new_state=%s",
+                        selections = []
+                    #tell the server what to look for:
+                    self.client.send_clipboard_selections(selections)
+                    clipboardlog("remote_clipboard_changed(%s) label=%s, remote_clipboard=%s, old_state=%s, new_state=%s",
                              item, label, remote_clipboard, old_state, new_state)
                     if new_state!=old_state:
                         self.client.clipboard_enabled = new_state
@@ -521,17 +538,18 @@ class GTKTrayMenuBase(object):
                 clipboard_item.connect("toggled", remote_clipboard_changed)
                 clipboard_submenu.append(clipboard_item)
             clipboard_submenu.show_all()
-        self.client.connect("handshake-complete", set_clipboard_menu)
+        self.client.after_handshake(set_clipboard_menu)
         return clipboard_menu
 
     def make_clipboardmenuitem(self):
         try:
-            if self.client.clipboard_helper:
-                from xpra.clipboard.translated_clipboard import TranslatedClipboardProtocolHelper
-                if isinstance(self.client.clipboard_helper, TranslatedClipboardProtocolHelper):
+            copts = self.client.get_clipboard_helper_classes()
+            #ugly alert: the helper does not exist yet.. we just check the helper classnames:
+            for c in copts:
+                if c.find("translated_clipboard")>0:
                     return self.make_translatedclipboard_optionsmenuitem()
         except:
-            log.error("make_clipboardmenuitem()", exc_info=True)
+            clipboardlog.error("make_clipboardmenuitem()", exc_info=True)
         return self.make_clipboard_togglemenuitem()
 
 
@@ -557,7 +575,7 @@ class GTKTrayMenuBase(object):
             self.keyboard_sync_menuitem.set_active(self.client.keyboard_helper.keyboard_sync)
             set_sensitive(self.keyboard_sync_menuitem, self.client.toggle_keyboard_sync)
             set_keyboard_sync_tooltip()
-        self.client.connect("handshake-complete", set_keyboard_sync_menuitem)
+        self.client.after_handshake(set_keyboard_sync_menuitem)
         return self.keyboard_sync_menuitem
 
     def make_openglmenuitem(self):
@@ -574,7 +592,7 @@ class GTKTrayMenuBase(object):
                 log("opengl_toggled%s", args)
                 self.client.toggle_opengl()
             gl.connect("toggled", opengl_toggled)
-        self.client.connect("handshake-complete", gl_set)
+        self.client.after_handshake(gl_set)
         return gl
 
     def make_encodingsmenuitem(self):
@@ -589,7 +607,7 @@ class GTKTrayMenuBase(object):
                 encodings.set_tooltip_text("memory mapped transfers are in use so picture encoding is disabled")
             else:
                 encodings.set_submenu(self.make_encodingssubmenu())
-        self.client.connect("handshake-complete", set_encodingsmenuitem)
+        self.client.after_handshake(set_encodingsmenuitem)
         return encodings
 
     def make_encodingssubmenu(self, handshake_complete=True):
@@ -620,13 +638,61 @@ class GTKTrayMenuBase(object):
                 set_sensitive(x, encoding in self.client.server_encodings)
 
 
+    def make_scalingmenuitem(self):
+        self.scaling = self.menuitem("Scaling", "scaling.png", "Desktop Scaling")
+        def set_scalingmenuitem(*args):
+            log("set_scalingmenuitem%s", args)
+            set_sensitive(self.scaling, not self.client.mmap_enabled)
+        self.client.after_handshake(set_scalingmenuitem)
+        scaling_submenu = self.make_scalingmenu()
+        self.scaling.set_submenu(scaling_submenu)
+        return self.scaling
+
+    def make_scalingmenu(self):
+        scaling_submenu = gtk.Menu()
+        scaling_submenu.updating = False
+        self.popup_menu_workaround(scaling_submenu)
+        def scalecmp(v):
+            return abs(self.client.xscale-v)<0.1
+        from xpra.client.ui_client_base import SCALING_OPTIONS
+        def scalingitem(scalingvalue=1.0):
+            pct = iround(100.0*scalingvalue)
+            label = {100 : "None"}.get(pct, "%i%%" % pct)
+            c = CheckMenuItem(label)
+            c.scalingvalue = scalingvalue
+            c.set_draw_as_radio(True)
+            c.set_active(scalecmp(scalingvalue))
+            def scaling_activated(item):
+                if scaling_submenu.updating:
+                    return
+                ensure_item_selected(scaling_submenu, item)
+                self.client.scaleset(item.scalingvalue, item.scalingvalue)
+            c.connect('activate', scaling_activated)
+            return c
+        for x in SCALING_OPTIONS:
+            scaling_submenu.append(scalingitem(x))
+        def scaling_changed(*args):
+            log("scaling_changed%s updating selected tray menu item", args)
+            #find the nearest scaling option to show as current:
+            scaling = (self.client.xscale + self.client.yscale)/2.0
+            by_distance = dict((abs(scaling-x),x) for x in SCALING_OPTIONS)
+            closest = by_distance.get(sorted(by_distance)[0], 1)
+            scaling_submenu.updating = True
+            for x in scaling_submenu.get_children():
+                scalingvalue = getattr(x, "scalingvalue", -1)
+                x.set_active(scalingvalue==closest)
+            scaling_submenu.updating = False
+        self.client.connect("scaling-changed", scaling_changed)
+        return scaling_submenu
+
+
     def make_qualitymenuitem(self):
         self.quality = self.menuitem("Quality", "slider.png", "Picture quality", None)
         set_sensitive(self.quality, False)
         def may_enable_qualitymenu(*args):
             self.quality.set_submenu(self.make_qualitysubmenu())
             self.set_qualitymenu()
-        self.client.connect("handshake-complete", may_enable_qualitymenu)
+        self.client.after_handshake(may_enable_qualitymenu)
         return self.quality
 
     def make_qualitysubmenu(self):
@@ -672,7 +738,7 @@ class GTKTrayMenuBase(object):
         def may_enable_speedmenu(*args):
             self.speed.set_submenu(self.make_speedsubmenu())
             self.set_speedmenu()
-        self.client.connect("handshake-complete", may_enable_speedmenu)
+        self.client.after_handshake(may_enable_speedmenu)
         return self.speed
 
     def make_speedsubmenu(self):
@@ -699,12 +765,10 @@ class GTKTrayMenuBase(object):
 
     def set_speedmenu(self, *args):
         if self.speed:
-            can_use = not self.client.mmap_enabled and self.client.encoding in self.client.server_encodings_with_speed and self.client.change_speed
+            can_use = not self.client.mmap_enabled and self.client.encoding in self.client.server_encodings_with_speed
             set_sensitive(self.speed, can_use)
             if self.client.mmap_enabled:
                 self.speed.set_tooltip_text("Quality is always 100% with mmap")
-            elif not self.client.change_speed:
-                self.speed.set_tooltip_text("Server does not support changing speed")
             elif self.client.encoding!="h264":
                 self.speed.set_tooltip_text("Not supported with %s encoding" % self.client.encoding)
             else:
@@ -733,7 +797,7 @@ class GTKTrayMenuBase(object):
                 return
             set_sensitive(speaker, True)
             speaker.set_submenu(self.make_soundsubmenu(is_speaker_on, self.spk_on, self.spk_off, "speaker-changed"))
-        self.client.connect("handshake-complete", speaker_state)
+        self.client.after_handshake(speaker_state)
         return speaker
 
     def mic_on(self, *args):
@@ -758,7 +822,7 @@ class GTKTrayMenuBase(object):
                 return
             set_sensitive(microphone, True)
             microphone.set_submenu(self.make_soundsubmenu(is_microphone_on, self.mic_on, self.mic_off, "microphone-changed"))
-        self.client.connect("handshake-complete", microphone_state)
+        self.client.after_handshake(microphone_state)
         return microphone
 
     def make_soundsubmenu(self, is_on_cb, on_cb, off_cb, client_signal):
@@ -840,7 +904,12 @@ class GTKTrayMenuBase(object):
         layout,layouts,variant,variants = self.client.keyboard_helper.keyboard.get_layout_spec()
         full_layout_list = False
         if len(layouts)>1:
-            log.info("keyboard layouts: %s", ",".join(layouts))
+            log("keyboard layouts: %s", u",".join(bytestostr(x) for x in layouts))
+            #log after removing dupes:
+            def uniq(seq):
+                seen = set()
+                return [x for x in seq if not (x in seen or seen.add(x))]
+            log.info("keyboard layouts: %s", u",".join(bytestostr(x) for x in uniq(layouts)))
             auto = kbitem("Auto", "Auto", "", True)
             self.layout_submenu.append(auto)
             if layout:
@@ -848,7 +917,7 @@ class GTKTrayMenuBase(object):
             if variants:
                 for v in variants:
                     self.layout_submenu.append(kbitem("%s - %s" % (layout, v), layout, v))
-            for l in layouts:
+            for l in uniq(layouts):
                 if l!=layout:
                     self.layout_submenu.append(kbitem("%s" % l, l, ""))
         elif layout and variants and len(variants)>1:
@@ -888,7 +957,7 @@ class GTKTrayMenuBase(object):
                 keyboard.hide()
                 return
             set_sensitive(keyboard, True)
-        self.client.connect("handshake-complete", set_layout_enabled)
+        self.client.after_handshake(set_layout_enabled)
         return keyboard
 
     def make_compressionmenu(self):
@@ -913,7 +982,7 @@ class GTKTrayMenuBase(object):
         def enable_compressionmenu(self):
             set_sensitive(self.compression, True)
             self.compression_submenu.show_all()
-        self.client.connect("handshake-complete", enable_compressionmenu)
+        self.client.after_handshake(enable_compressionmenu)
         return self.compression
 
 
@@ -930,19 +999,27 @@ class GTKTrayMenuBase(object):
                     win.present()
         return self.handshake_menuitem("Raise Windows", "raise.png", None, raise_windows)
 
-    def make_startnewcommandmenuitem(self, handshake_done=False):
+    def make_startnewcommandmenuitem(self):
         self.startnewcommand = self.menuitem("Run Command", "forward.png", "Run a new command on the server", self.client.show_start_new_command)
         def enable_start_new_command(*args):
             log("enable_start_new_command%s start_new_command=%s", args, self.client.start_new_commands)
             set_sensitive(self.startnewcommand, self.client.start_new_commands)
             if not self.client.start_new_commands:
                 self.startnewcommand.set_tooltip_text("Not supported by the server")
-        if not handshake_done:
-            self.client.connect("handshake-complete", enable_start_new_command)
-        else:
-            enable_start_new_command()
+        self.client.after_handshake(enable_start_new_command)
         return self.startnewcommand
 
+    def make_uploadmenuitem(self):
+        self.upload = self.menuitem("Upload File", "upload.png", "Send a file to the server", self.client.show_file_upload)
+        def enable_upload(*args):
+            log("enable_upload%s server_file_transfer=%s", args, self.client.server_file_transfer)
+            set_sensitive(self.upload, self.client.server_file_transfer)
+            if not self.client.server_file_transfer:
+                self.upload.set_tooltip_text("Not supported by the server")
+        self.client.after_handshake(enable_upload)
+        return self.upload
+
+
     def make_disconnectmenuitem(self):
         def menu_quit(*args):
             self.client.disconnect_and_quit(EXIT_OK, CLIENT_EXIT)
@@ -977,7 +1054,7 @@ class GTKTrayMenuBase(object):
                 if expected_counter!=self.mouse_in_tray_menu_counter:
                     return    False            #counter has changed
                 self.close_menu()
-            gobject.timeout_add(500, check_menu_left, self.mouse_in_tray_menu_counter)
+            glib.timeout_add(500, check_menu_left, self.mouse_in_tray_menu_counter)
         self.mouse_in_tray_menu_counter = 0
         self.mouse_in_tray_menu = False
         log("popup_menu_workaround: adding events callbacks")
