@@ -5,19 +5,19 @@
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, import_gobject, is_gtk3
+from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, import_glib, is_gtk3
 gtk = import_gtk()
 gdk = import_gdk()
-gobject = import_gobject()
+glib = import_glib()
 import sys
 import time
 import datetime
 
-from xpra.os_util import os_info, bytestostr
-from xpra.util import prettify_plug_name
+from xpra.os_util import bytestostr
+from xpra.util import prettify_plug_name, typedict, csv, engs
 from xpra.gtk_common.graph import make_graph_pixmap
 from collections import deque
-from xpra.simple_stats import values_to_scaled_values, values_to_diff_scaled_values, to_std_unit, std_unit_dec
+from xpra.simple_stats import values_to_scaled_values, values_to_diff_scaled_values, to_std_unit, std_unit_dec, std_unit
 from xpra.scripts.config import python_platform
 from xpra.log import Logger
 from xpra.gtk_common.gtk_util import add_close_accel, label, title_box, \
@@ -29,6 +29,7 @@ log = Logger("info")
 N_SAMPLES = 20      #how many sample points to show on the graphs
 SHOW_PIXEL_STATS = True
 SHOW_SOUND_STATS = True
+SHOW_RECV = True
 
 
 def pixelstr(v):
@@ -105,8 +106,13 @@ class SessionInfo(gtk.Window):
         tb.attach(title_box("Server"), 2, xoptions=EXPAND|FILL, xpadding=0)
         tb.inc()
 
-        def make_os_str(*args):
-            s = os_info(*args)
+        def make_os_str(sys_platform, platform_release, platform_platform, platform_linux_distribution):
+            from xpra.os_util import platform_name
+            s = [platform_name(sys_platform, platform_release)]
+            if platform_linux_distribution and len(platform_linux_distribution)==3 and len(platform_linux_distribution[0])>0:
+                s.append(" ".join([str(x) for x in platform_linux_distribution]))
+            elif platform_platform:
+                s.append(platform_platform)
             return "\n".join(s)
         distro = ""
         if hasattr(python_platform, "linux_distribution"):
@@ -142,7 +148,7 @@ class SessionInfo(gtk.Window):
         def make_revision_str(rev, changes):
             if not changes:
                 return rev
-            return "%s (%s changes)" % (rev, changes)
+            return "%s (%s change%s)" % (rev, changes, engs(changes))
         def make_datetime(date, time):
             if not time:
                 return date
@@ -169,13 +175,16 @@ class SessionInfo(gtk.Window):
         tb.new_row("Pango",     client_vinfo("pango"),      server_vinfo("pango"))
         tb.new_row("Python", label(python_platform.python_version()), label(server_version_info("server.python.version", "python_version", "python.version")))
 
-        cl_gst_v, cl_pygst_v = "", ""
         try:
-            from xpra.sound.gstreamer_util import gst_version as cl_gst_v, pygst_version as cl_pygst_v
+            from xpra.sound.wrapper import query_sound
+            props = query_sound()
         except Exception as e:
-            log("cannot load gstreamer: %s", e)
-        tb.new_row("GStreamer", label(make_version_str(cl_gst_v)), label(server_version_info("sound.gst.version", "gst_version")))
-        tb.new_row("pygst", label(make_version_str(cl_pygst_v)), label(server_version_info("sound.pygst.version", "pygst_version")))
+            log("cannot load sound information: %s", e)
+            props = {}
+        gst_version = props.get("gst.version", "")
+        pygst_version = props.get("pygst.version", "")
+        tb.new_row("GStreamer", label(make_version_str(gst_version)), label(server_version_info("sound.gst.version", "gst_version")))
+        tb.new_row("pygst", label(make_version_str(pygst_version)), label(server_version_info("sound.pygst.version", "pygst_version")))
         tb.new_row("OpenGL", label(make_version_str(self.client.opengl_props.get("opengl", "n/a"))), label("n/a"))
         tb.new_row("OpenGL Vendor", label(make_version_str(self.client.opengl_props.get("vendor", ""))), label("n/a"))
         tb.new_row("PyOpenGL", label(make_version_str(self.client.opengl_props.get("pyopengl", "n/a"))), label("n/a"))
@@ -195,6 +204,8 @@ class SessionInfo(gtk.Window):
         randr_box.add(self.server_randr_icon)
         randr_box.add(self.server_randr_label)
         tb.new_row("RandR Support", randr_box)
+        self.client_display = label()
+        tb.new_row("Client Display", self.client_display)
         opengl_box = gtk.HBox(False, 20)
         self.client_opengl_label = label()
         self.client_opengl_label.set_line_wrap(True)
@@ -203,29 +214,17 @@ class SessionInfo(gtk.Window):
         opengl_box.add(self.client_opengl_label)
         tb.new_row("Client OpenGL", opengl_box)
         self.opengl_buffering = label()
-        tb.new_row("OpenGL Buffering", self.opengl_buffering)
+        tb.new_row("OpenGL Mode", self.opengl_buffering)
         self.server_mmap_icon = gtk.Image()
         tb.new_row("Memory Mapped Transfers", self.server_mmap_icon)
         self.server_clipboard_icon = gtk.Image()
         tb.new_row("Clipboard", self.server_clipboard_icon)
         self.server_notifications_icon = gtk.Image()
-        tb.new_row("Notification Forwarding", self.server_notifications_icon)
+        tb.new_row("Notifications", self.server_notifications_icon)
         self.server_bell_icon = gtk.Image()
-        tb.new_row("Bell Forwarding", self.server_bell_icon)
+        tb.new_row("Bell", self.server_bell_icon)
         self.server_cursors_icon = gtk.Image()
-        tb.new_row("Cursor Forwarding", self.server_cursors_icon)
-        speaker_box = gtk.HBox(False, 20)
-        self.server_speaker_icon = gtk.Image()
-        speaker_box.add(self.server_speaker_icon)
-        self.speaker_codec_label = label()
-        speaker_box.add(self.speaker_codec_label)
-        tb.new_row("Speaker Forwarding", speaker_box)
-        microphone_box = gtk.HBox(False, 20)
-        self.server_microphone_icon = gtk.Image()
-        microphone_box.add(self.server_microphone_icon)
-        self.microphone_codec_label = label()
-        microphone_box.add(self.microphone_codec_label)
-        tb.new_row("Microphone Forwarding", microphone_box)
+        tb.new_row("Cursors", self.server_cursors_icon)
         #add bottom table:
         tb = TableBuilder(rows=1, columns=3)
         table = tb.get_table()
@@ -373,13 +372,20 @@ class SessionInfo(gtk.Window):
             if SHOW_PIXEL_STATS:
                 bandwidth_label += ",\nand number of pixels rendered"
             self.bandwidth_graph = self.add_graph_button(bandwidth_label, self.save_graphs)
-            self.connect("realize", self.populate_graphs)
             self.latency_graph = self.add_graph_button(None, self.save_graphs)
+            if SHOW_SOUND_STATS:
+                self.sound_queue_graph = self.add_graph_button(None, self.save_graphs)
+            else:
+                self.sound_queue_graph = None
+            self.connect("realize", self.populate_graphs)
         self.pixel_in_data = deque(maxlen=N_SAMPLES+4)
         self.net_in_bytecount = deque(maxlen=N_SAMPLES+4)
         self.net_out_bytecount = deque(maxlen=N_SAMPLES+4)
-        self.sound_in_bytecount = deque(maxlen=N_SAMPLES+4)
-        self.sound_out_bytecount = deque(maxlen=N_SAMPLES+4)
+        self.sound_in_bitcount = deque(maxlen=N_SAMPLES+4)
+        self.sound_out_bitcount = deque(maxlen=N_SAMPLES+4)
+        self.sound_out_queue_min = deque(maxlen=N_SAMPLES*10+4)
+        self.sound_out_queue_max = deque(maxlen=N_SAMPLES*10+4)
+        self.sound_out_queue_cur  = deque(maxlen=N_SAMPLES*10+4)
 
         self.set_border_width(15)
         self.add(self.tab_box)
@@ -389,12 +395,14 @@ class SessionInfo(gtk.Window):
             self.is_closed = True
         self.connect('delete_event', window_deleted)
         self.show_tab(self.tabs[0][2])
-        self.set_size_request(-1, 480)
+        self.set_size_request(-1, -1)
         self.init_counters()
         self.populate()
         self.populate_all()
-        gobject.timeout_add(1000, self.populate)
-        gobject.timeout_add(100, self.populate_tab)
+        glib.timeout_add(1000, self.populate)
+        glib.timeout_add(100, self.populate_tab)
+        if SHOW_SOUND_STATS:
+            glib.timeout_add(100, self.populate_sound_stats)
         add_close_accel(self, self.destroy)
 
 
@@ -492,19 +500,35 @@ class SessionInfo(gtk.Window):
             icon = self.get_pixbuf("unticked-small.png")
         image.set_from_pixbuf(icon)
 
+    def populate_sound_stats(self, *args):
+        #runs every 100ms
+        if self.is_closed:
+            return False
+        ss = self.client.sound_sink
+        if SHOW_SOUND_STATS and ss:
+            info = ss.get_info()
+            if info:
+                info = typedict(info)
+                self.sound_out_queue_cur.append(info.intget("queue.cur"))
+                self.sound_out_queue_min.append(info.intget("queue.min"))
+                self.sound_out_queue_max.append(info.intget("queue.max"))
+        return not self.is_closed
+
     def populate(self, *args):
         if self.is_closed:
             return False
         self.client.send_ping()
         self.last_populate_time = time.time()
+
+        self.show_opengl_state()
         #record bytecount every second:
         self.net_in_bytecount.append(self.connection.input_bytecount)
         self.net_out_bytecount.append(self.connection.output_bytecount)
-        if SHOW_SOUND_STATS and self.client.sound_sink:
+        if SHOW_SOUND_STATS:
             if self.client.sound_in_bytecount>0:
-                self.sound_in_bytecount.append(self.client.sound_in_bytecount)
+                self.sound_in_bitcount.append(self.client.sound_in_bytecount * 8)
             if self.client.sound_out_bytecount>0:
-                self.sound_out_bytecount.append(self.client.sound_out_bytecount)
+                self.sound_out_bitcount.append(self.client.sound_out_bytecount * 8)
 
         #count pixels in the last second:
         since = time.time()-1
@@ -584,17 +608,8 @@ class SessionInfo(gtk.Window):
     def populate_package(self):
         return False
 
-    def populate_features(self):
-        size_info = ""
-        if self.client.server_actual_desktop_size:
-            w,h = self.client.server_actual_desktop_size
-            size_info = "%s*%s" % (w,h)
-            if self.client.server_randr and self.client.server_max_desktop_size:
-                size_info += " (max %s)" % ("x".join([str(x) for x in self.client.server_max_desktop_size]))
-        self.bool_icon(self.server_randr_icon, self.client.server_randr)
-        self.server_randr_label.set_text("%s" % size_info)
-        self.bool_icon(self.client_opengl_icon, self.client.client_supports_opengl)
-        buffering = "n/a"
+
+    def show_opengl_state(self):
         if self.client.opengl_enabled:
             glinfo = "%s / %s" % (self.client.opengl_props.get("vendor", ""), self.client.opengl_props.get("renderer", ""))
             display_mode = self.client.opengl_props.get("display_mode", [])
@@ -604,11 +619,37 @@ class SessionInfo(gtk.Window):
                 buffering = "single buffering"
             else:
                 buffering = "unknown"
+            if buffering!="unknown":
+                if "ALPHA" in display_mode:
+                    buffering += " with"
+                else:
+                    buffering += " without"
+                buffering += " transparency"
         else:
+            #info could be telling us that the gl bindings are missing:
             glinfo = self.client.opengl_props.get("info", "disabled")
+            buffering = "n/a"
         self.client_opengl_label.set_text(glinfo)
         self.opengl_buffering.set_text(buffering)
 
+    def populate_features(self):
+        size_info = ""
+        if self.client.server_actual_desktop_size:
+            w,h = self.client.server_actual_desktop_size
+            size_info = "%s*%s" % (w,h)
+            if self.client.server_randr and self.client.server_max_desktop_size:
+                size_info += " (max %s)" % ("x".join([str(x) for x in self.client.server_max_desktop_size]))
+        self.bool_icon(self.server_randr_icon, self.client.server_randr)
+        self.server_randr_label.set_text("%s" % size_info)
+        root_w, root_h = self.client.get_root_size()
+        if self.client.xscale!=1 or self.client.yscale!=1:
+            sw, sh = self.client.cp(root_w, root_h)
+            display_info = "%ix%i (scaled from %ix%i)" % (sw, sh, root_w, root_h)
+        else:
+            display_info = "%ix%i" % (root_w, root_h)
+        self.client_display.set_text(display_info)
+        self.bool_icon(self.client_opengl_icon, self.client.client_supports_opengl)
+
         scaps = self.client.server_capabilities
         self.bool_icon(self.server_mmap_icon, self.client.mmap_enabled)
         self.bool_icon(self.server_clipboard_icon,      scaps.boolget("clipboard", False))
@@ -625,39 +666,33 @@ class SessionInfo(gtk.Window):
                 return state
             s = "%s: %s" % (state, sound_pipeline.codec_description)
             try:
-                pid = sound_pipeline.get_info().get("pid")
+                info = sound_pipeline.get_info()
+                pid = info.get("pid")
                 if pid:
                     s += " (pid=%s)" % pid
             except:
-                pass
-            #if sound_pipeline.bitrate>0:
-            #    s += " / %sbit/s" % std_unit(sound_pipeline.bitrate)
+                log.warn("error getting sound info", exc_info=True)
             return s
         def codec_info(enabled, codecs):
             if not enabled:
                 return "n/a"
             return ", ".join(codecs or [])
-        def populate_speaker_info(*args):
-            can = scaps.boolget("sound.send", False) and self.client.speaker_allowed
-            self.bool_icon(self.server_speaker_icon, can)
-            self.speaker_codec_label.set_text(bytestostr(pipeline_info(can, self.client.sound_sink)))
-        populate_speaker_info()
-        self.client.connect("speaker-changed", populate_speaker_info)
-        def populate_microphone_info(*args):
-            can = scaps.boolget("sound.receive", False) and self.client.microphone_allowed
-            self.bool_icon(self.server_microphone_icon, can)
-            self.microphone_codec_label.set_text(bytestostr(pipeline_info(can, self.client.sound_source)))
-        populate_microphone_info()
-        self.client.connect("microphone-changed", populate_microphone_info)
 
         #sound/video codec table:
         self.server_speaker_codecs_label.set_text(codec_info(scaps.boolget("sound.send", False), scaps.strlistget("sound.encoders", [])))
         self.client_speaker_codecs_label.set_text(codec_info(self.client.speaker_allowed, self.client.speaker_codecs))
         self.server_microphone_codecs_label.set_text(codec_info(scaps.boolget("sound.receive", False), scaps.strlistget("sound.decoders", [])))
         self.client_microphone_codecs_label.set_text(codec_info(self.client.microphone_allowed, self.client.microphone_codecs))
+        def encliststr(v):
+            v = list(v)
+            try:
+                v.remove("rgb")
+            except:
+                pass
+            return csv(sorted(v))
         se = scaps.strlistget("encodings.core", scaps.strlistget("encodings"))
-        self.server_encodings_label.set_text(", ".join(sorted(se)))
-        self.client_encodings_label.set_text(", ".join(sorted(self.client.get_core_encodings())))
+        self.server_encodings_label.set_text(encliststr(se))
+        self.client_encodings_label.set_text(encliststr(self.client.get_core_encodings()))
 
         def get_encoder_list(caps):
             from xpra.net import packet_encoding
@@ -701,14 +736,14 @@ class SessionInfo(gtk.Window):
                 return {"state" : "inactive"}
             return prop.get_info()
         def set_sound_info(label, details, supported, prop):
-            p = get_sound_info(supported, prop)
-            label.set_text(p.get("state", ""))
+            d = typedict(get_sound_info(supported, prop))
+            label.set_text(d.strget("state", ""))
             if details:
-                d = p.get("queue.used", -1)
-                if d>=0:
-                    details.set_text(" (buffer: %sms)" % str(d).rjust(3))
-                else:
-                    details.set_text("")
+                s = ""
+                bitrate = d.intget("bitrate", 0)
+                if bitrate>0:
+                    s = "%sbit/s" % std_unit(bitrate)
+                details.set_text(s)
         set_sound_info(self.speaker_label, self.speaker_details, self.client.speaker_enabled, self.client.sound_sink)
         set_sound_info(self.microphone_label, None, self.client.microphone_enabled, self.client.sound_source)
 
@@ -870,23 +905,28 @@ class SessionInfo(gtk.Window):
                 #We are interested in data like:
                 #window[1].encoder=x264
                 #window[1].encoder.frames=1
+                #window[1].encoder.fps=25
                 for k,v in self.client.server_last_info.items():
                     k = bytestostr(k)
+                    if not k.startswith("window["):
+                        continue
                     pos = k.find("].encoder")
-                    if k.startswith("window[") and pos>0:
+                    if pos<=0:
+                        continue
+                    try:
                         wid_str = k[len("window["):pos]     #ie: "1"
-                        ekey = k[(pos+len("].encoder")):]   #ie: "" or ".frames"
-                        if ekey.startswith("."):
-                            ekey = ekey[1:]
-                        try:
-                            wid = int(wid_str)
-                            props = window_encoder_stats.setdefault(wid, {})
-                            props[ekey] = v
-                        except:
-                            #wid_str may be invalid, ie:
-                            #window[1].pipeline_option[1].encoder=codec_spec(xpra.codecs.enc_x264.encoder.Encoder)
-                            # -> wid_str= "1].pipeline_option[1"
-                            pass
+                        wid = int(wid_str)
+                    except:
+                        #wid_str may be invalid, ie:
+                        #window[1].pipeline_option[1].encoder=video_spec(xpra.codecs.enc_x264.encoder.Encoder)
+                        # -> wid_str= "1].pipeline_option[1"
+                        continue
+                    ekey = k[(pos+len("].encoder")):]   #ie: "" or ".frames"
+                    if ekey.startswith("."):
+                        ekey = ekey[1:]
+                    if ekey=="build_config":
+                        continue
+                    window_encoder_stats.setdefault(wid, {})[ekey] = v
                 #print("window_encoder_stats=%s" % window_encoder_stats)
                 for wid, props in window_encoder_stats.items():
                     l = label("%s (%s)" % (wid, bytestostr(props.get(""))))
@@ -906,8 +946,13 @@ class SessionInfo(gtk.Window):
             return True
         start_x_offset = min(1.0, (time.time()-self.last_populate_time)*0.95)
         rect = box.get_allocation()
-        h = max(200, h-bh-20, rect.height-bh-20)
-        w = max(360, rect.width-20)
+        maxw, maxh = self.client.get_root_size()
+        ngraphs = 2+int(SHOW_SOUND_STATS)
+        #the preferred size (which does not cause the window to grow too big):
+        W = 360
+        H = 160*3//ngraphs
+        w = min(maxw, max(W, rect.width-20))
+        h = min(maxh//ngraphs, max(H, (h-bh-20)//ngraphs, (rect.height-bh-20)//ngraphs))
         #bandwidth graph:
         labels, datasets = [], []
         if self.net_in_bytecount and self.net_out_bytecount:
@@ -921,55 +966,78 @@ class SessionInfo(gtk.Window):
                     return "x%s%s" % (int(value), unit)
             net_in_scale, net_in_data = values_to_diff_scaled_values(list(self.net_in_bytecount)[1:N_SAMPLES+3], scale_unit=1000, min_scaled_value=50)
             net_out_scale, net_out_data = values_to_diff_scaled_values(list(self.net_out_bytecount)[1:N_SAMPLES+3], scale_unit=1000, min_scaled_value=50)
-            labels += ["recv %sB/s" % unit(net_in_scale), "sent %sB/s" % unit(net_out_scale)]
-            datasets += [net_in_data, net_out_data]
+            if SHOW_RECV:
+                labels += ["recv %sB/s" % unit(net_in_scale), "sent %sB/s" % unit(net_out_scale)]
+                datasets += [net_in_data, net_out_data]
+            else:
+                labels += ["recv %sB/s" % unit(net_in_scale)]
+                datasets += [net_in_data]
         if SHOW_PIXEL_STATS and self.client.windows_enabled:
             pixel_scale, in_pixels = values_to_scaled_values(list(self.pixel_in_data)[3:N_SAMPLES+4], min_scaled_value=100)
             datasets.append(in_pixels)
             labels.append("%s pixels/s" % unit(pixel_scale))
-        if SHOW_SOUND_STATS and self.sound_in_bytecount:
-            sound_in_scale, sound_in_data =  values_to_diff_scaled_values(list(self.sound_in_bytecount)[1:N_SAMPLES+3], scale_unit=1000, min_scaled_value=50)
+        if SHOW_SOUND_STATS and self.sound_in_bitcount:
+            sound_in_scale, sound_in_data =  values_to_diff_scaled_values(list(self.sound_in_bitcount)[1:N_SAMPLES+3], scale_unit=1000, min_scaled_value=50)
             datasets.append(sound_in_data)
-            labels.append("Speaker %sB/s" % unit(sound_in_scale))
-        if SHOW_SOUND_STATS and self.sound_out_bytecount:
-            sound_out_scale, sound_out_data =  values_to_diff_scaled_values(list(self.sound_out_bytecount)[1:N_SAMPLES+3], scale_unit=1000, min_scaled_value=50)
+            labels.append("Speaker %sb/s" % unit(sound_in_scale))
+        if SHOW_SOUND_STATS and self.sound_out_bitcount:
+            sound_out_scale, sound_out_data =  values_to_diff_scaled_values(list(self.sound_out_bitcount)[1:N_SAMPLES+3], scale_unit=1000, min_scaled_value=50)
             datasets.append(sound_out_data)
-            labels.append("Mic %sB/s" % unit(sound_out_scale))
+            labels.append("Mic %sb/s" % unit(sound_out_scale))
 
         if labels and datasets:
             pixmap = make_graph_pixmap(datasets, labels=labels,
-                                       width=w, height=h/2,
+                                       width=w, height=h,
                                        title="Bandwidth", min_y_scale=10, rounding=10,
                                        start_x_offset=start_x_offset)
-            self.bandwidth_graph.set_size_request(*pixmap.get_size())
+            self.bandwidth_graph.set_size_request(W, H)
             self.bandwidth_graph.set_from_pixmap(pixmap, None)
-        if self.client.server_info_request:
-            pass
+
+        def norm_lists(items, size=N_SAMPLES):
+            #ensures we always have exactly 20 values,
+            #(and skip if we don't have any)
+            values, labels = [], []
+            for l, name in items:
+                if len(l)==0:
+                    continue
+                l = list(l)
+                if len(l)<size:
+                    for _ in range(size-len(l)):
+                        l.insert(0, None)
+                else:
+                    l = l[:size]
+                values.append(l)
+                labels.append(name)
+            return values, labels
+
         #latency graph:
-        latency_graph_items = (
+        latency_values, latency_labels = norm_lists((
                                 (self.avg_ping_latency, "network"),
                                 (self.avg_batch_delay, "batch delay"),
                                 (self.avg_damage_out_latency, "encode&send"),
                                 (self.avg_decoding_latency, "decoding"),
                                 (self.avg_total, "frame total"),
-                                )
-        latency_graph_values = []
-        labels = []
-        for l, name in latency_graph_items:
-            if len(l)==0:
-                continue
-            l = list(l)
-            if len(l)<20:
-                for _ in range(20-len(l)):
-                    l.insert(0, None)
-            latency_graph_values.append(l)
-            labels.append(name)
-        pixmap = make_graph_pixmap(latency_graph_values, labels=labels,
-                                    width=w, height=h/2,
+                                ))
+        pixmap = make_graph_pixmap(latency_values, labels=latency_labels,
+                                    width=w, height=h,
                                     title="Latency (ms)", min_y_scale=10, rounding=25,
                                     start_x_offset=start_x_offset)
-        self.latency_graph.set_size_request(*pixmap.get_size())
+        self.latency_graph.set_size_request(W, H)
         self.latency_graph.set_from_pixmap(pixmap, None)
+
+        if SHOW_SOUND_STATS and self.client.sound_sink:
+            #sound queue graph:
+            queue_values, queue_labels = norm_lists((
+                                 (self.sound_out_queue_max, "Max"),
+                                 (self.sound_out_queue_cur, "Level"),
+                                 (self.sound_out_queue_min, "Min"),
+                                 ), N_SAMPLES*10)
+            pixmap = make_graph_pixmap(queue_values, labels=queue_labels,
+                                        width=w, height=h,
+                                        title="Sound Buffer (ms)", min_y_scale=10, rounding=25,
+                                        start_x_offset=start_x_offset)
+            self.sound_queue_graph.set_size_request(W, H)
+            self.sound_queue_graph.set_from_pixmap(pixmap, None)
         return True
 
     def save_graphs(self, *args):
@@ -990,7 +1058,7 @@ class SessionInfo(gtk.Window):
         if response == gtk.RESPONSE_OK:
             if len(filenames)==1:
                 filename = filenames[0]
-                pixmaps = [image.get_pixmap()[0] for image in [self.bandwidth_graph, self.latency_graph]]
+                pixmaps = [image.get_pixmap()[0] for image in [self.bandwidth_graph, self.latency_graph, self.sound_queue_graph]]
                 log("saving pixmaps %s and %s to %s", pixmaps, filename)
                 w, h = 0, 0
                 for pixmap in pixmaps:
