@@ -29,16 +29,16 @@ log = Logger("util")
 
 
 _instance = None
-def getStartNewCommand(run_callback):
+def getStartNewCommand(run_callback, can_share=False):
     global _instance
     if _instance is None:
-        _instance = StartNewCommand(run_callback)
+        _instance = StartNewCommand(run_callback, can_share)
     return _instance
 
 
 class StartNewCommand(object):
 
-    def __init__(self, run_callback=None):
+    def __init__(self, run_callback=None, can_share=False):
         self.run_callback = run_callback
         self.window = gtk.Window()
         self.window.connect("destroy", self.close)
@@ -68,6 +68,14 @@ class StartNewCommand(object):
         self.entry.connect('activate', self.run_command)
         vbox.add(self.entry)
 
+        if can_share:
+            self.share = gtk.CheckButton("Shared", use_underline=False)
+            #Shared commands will also be shown to other clients
+            self.share.set_active(True)
+            vbox.add(self.share)
+        else:
+            self.share = None
+
         # Buttons:
         hbox = gtk.HBox(False, 20)
         vbox.pack_start(hbox)
@@ -134,37 +142,40 @@ class StartNewCommand(object):
         self.hide()
         command = self.entry.get_text()
         if self.run_callback:
-            self.run_callback(command)
+            self.run_callback(command, self.share is None or self.share.get_active())
 
 
 def main():
-    from xpra.platform import init as platform_init
+    from xpra.platform import init as platform_init, clean as platform_clean
     from xpra.platform.gui import ready as gui_ready
-    platform_init("Start-New-Command", "Start New Command")
-
-    #logging init:
-    if "-v" in sys.argv:
-        enable_debug_for("util")
-
-    from xpra.os_util import SIGNAMES
-    from xpra.gtk_common.quit import gtk_main_quit_on_fatal_exceptions_enable
-    gtk_main_quit_on_fatal_exceptions_enable()
-
-    app = StartNewCommand()
-    app.close = app.quit
-    def app_signal(signum, frame):
-        print("")
-        log.info("got signal %s", SIGNAMES.get(signum, signum))
-        app.quit()
-    signal.signal(signal.SIGINT, app_signal)
-    signal.signal(signal.SIGTERM, app_signal)
     try:
-        gui_ready()
-        app.show()
-        app.run()
-    except KeyboardInterrupt:
-        pass
-    return 0
+        platform_init("Start-New-Command", "Start New Command")
+
+        #logging init:
+        if "-v" in sys.argv:
+            enable_debug_for("util")
+
+        from xpra.os_util import SIGNAMES
+        from xpra.gtk_common.quit import gtk_main_quit_on_fatal_exceptions_enable
+        gtk_main_quit_on_fatal_exceptions_enable()
+
+        app = StartNewCommand()
+        app.close = app.quit
+        def app_signal(signum, frame):
+            print("")
+            log.info("got signal %s", SIGNAMES.get(signum, signum))
+            app.quit()
+        signal.signal(signal.SIGINT, app_signal)
+        signal.signal(signal.SIGTERM, app_signal)
+        try:
+            gui_ready()
+            app.show()
+            app.run()
+        except KeyboardInterrupt:
+            pass
+        return 0
+    finally:
+        platform_clean()
 
 
 if __name__ == "__main__":
