@@ -5,6 +5,7 @@
 # later version. See the file COPYING for details.
 
 import os
+import hashlib
 from xpra.log import Logger
 log = Logger("paint")
 deltalog = Logger("delta")
@@ -12,8 +13,7 @@ deltalog = Logger("delta")
 from threading import Lock
 from xpra.net.mmap_pipe import mmap_read
 from xpra.net import compression
-from xpra.util import typedict
-from xpra.codecs.codec_constants import get_PIL_decodings
+from xpra.util import typedict, csv
 from xpra.codecs.loader import get_codec
 from xpra.codecs.video_helper import getVideoHelper
 from xpra.os_util import BytesIOClass, bytestostr, _buffer
@@ -21,9 +21,8 @@ from xpra.codecs.xor.cyxor import xor_str   #@UnresolvedImport
 from xpra.codecs.argb.argb import unpremultiply_argb, unpremultiply_argb_in_place   #@UnresolvedImport
 
 DELTA_BUCKETS = int(os.environ.get("XPRA_DELTA_BUCKETS", "5"))
-
+INTEGRITY_HASH = os.environ.get("XPRA_INTEGRITY_HASH", "0")=="1"
 WEBP_PILLOW = os.environ.get("XPRA_WEBP_PILLOW", "0")=="1"
-PIL = get_codec("PIL")
 
 #ie:
 #CSC_OPTIONS = { "YUV420P" : {"RGBX" : [opencl.spec, swscale.spec], "BGRX" : ...} }
@@ -47,19 +46,19 @@ def load_video_decoders():
         for encoding in vh.get_decodings():
             specs = vh.get_decoder_specs(encoding)
             for colorspace, decoders in specs.items():
-                log("%s decoders for %s: %s", encoding, colorspace, decoders)
+                log("%-4s decoders for %7s: %s", encoding, colorspace, csv([d.get_type() for _,d in decoders]))
                 assert len(decoders)>0
                 #use the first one:
                 _, decoder_module = decoders[0]
                 VIDEO_DECODERS[encoding] = decoder_module
-        log("video decoders: %s", VIDEO_DECODERS)
+        log("video decoders: %s", dict((e,d.get_type()) for e,d in VIDEO_DECODERS.items()))
     return VIDEO_DECODERS
 
 
-def fire_paint_callbacks(callbacks, success):
+def fire_paint_callbacks(callbacks, success, message=""):
     for x in callbacks:
         try:
-            x(success)
+            x(success, message)
         except KeyboardInterrupt:
             raise
         except:
@@ -75,6 +74,7 @@ class WindowBackingBase(object):
         load_csc_options()
         load_video_decoders()
         self.wid = wid
+        self.size = 0, 0
         self.idle_add = idle_add
         self._alpha_enabled = window_alpha
         self._backing = None
@@ -82,7 +82,10 @@ class WindowBackingBase(object):
         self._video_decoder = None
         self._csc_decoder = None
         self._decoder_lock = Lock()
-        self._PIL_encodings = get_PIL_decodings(PIL)
+        self._PIL_encodings = []
+        PIL = get_codec("dec_pillow")
+        if PIL:
+            self._PIL_encodings = PIL.get_encodings()
         self.draw_needs_refresh = True
         self.mmap = None
         self.mmap_enabled = False
@@ -101,14 +104,15 @@ class WindowBackingBase(object):
         self.close_decoder(False)
 
     def close_decoder(self, blocking=False):
-        if self._decoder_lock is None or not self._decoder_lock.acquire(blocking):
+        dl = self._decoder_lock
+        if dl is None or not dl.acquire(blocking):
             return False
         try:
             self.do_clean_video_decoder()
             self.do_clean_csc_decoder()
             return True
         finally:
-            self._decoder_lock.release()
+            dl.release()
 
     def do_clean_video_decoder(self):
         if self._video_decoder:
@@ -122,12 +126,10 @@ class WindowBackingBase(object):
 
 
     def get_encoding_properties(self):
-        rgb_modes = self.RGB_MODES
         return {
-                 "encodings.rgb_formats"    : rgb_modes,
+                 "encodings.rgb_formats"    : self.RGB_MODES,
                  "encoding.transparency"    : self._alpha_enabled,
-                 "encoding.full_csc_modes"  : self._get_full_csc_modes(rgb_modes),
-                 "encoding.csc_modes"       : self._get_csc_modes(rgb_modes)
+                 "encoding.full_csc_modes"  : self._get_full_csc_modes(self.RGB_MODES),
                  }
 
     def _get_full_csc_modes(self, rgb_modes):
@@ -141,14 +143,6 @@ class WindowBackingBase(object):
         log("_get_full_csc_modes(%s)=%s (target_rgb_modes=%s)", rgb_modes, full_csc_modes, target_rgb_modes)
         return full_csc_modes
 
-    def _get_csc_modes(self, rgb_modes):
-        #as above, but for older servers: less detailed than "full" csc modes info
-        csc_modes = []
-        for modes in self._get_full_csc_modes(rgb_modes).values():
-            csc_modes += modes
-        csc_modes = list(set(csc_modes))
-        return csc_modes
-
 
     def unpremultiply(self, img_data):
         if type(img_data) not in (str, _buffer):
@@ -173,31 +167,35 @@ class WindowBackingBase(object):
                 assert len(comp)==1, "more than one compressor specified: %s" % str(comp)
                 img_data = compression.decompress_by_name(raw_data, algo=comp[0])
         if len(img_data)!=rowstride * height:
-            log.error("invalid img data %s: %s", type(img_data), str(img_data)[:256])
+            deltalog.error("invalid img data length: expected %s but got %s (%s: %s)", rowstride * height, len(img_data), type(img_data), str(img_data)[:256])
             raise Exception("expected %s bytes for %sx%s with rowstride=%s but received %s (%s compressed)" %
                                 (rowstride * height, width, height, rowstride, len(img_data), len(raw_data)))
         delta = options.intget("delta", -1)
         bucket = options.intget("bucket", 0)
+        rgb_format = options.strget("rgb_format")
         rgb_data = img_data
         if delta>=0:
             assert bucket>=0 and bucket<DELTA_BUCKETS, "invalid delta bucket number: %s" % bucket
             if self._delta_pixel_data[bucket] is None:
                 raise Exception("delta region bucket %s references pixmap data we do not have!" % bucket)
-            lwidth, lheight, seq, ldata = self._delta_pixel_data[bucket]
-            assert width==lwidth and height==lheight and delta==seq, "delta bucket %s data does not match: expected %s but got %s" % (bucket, (width, height, delta), (lwidth, lheight, seq))
+            lwidth, lheight, lrgb_format, seq, ldata = self._delta_pixel_data[bucket]
+            assert width==lwidth and height==lheight and delta==seq, \
+                "delta bucket %s data does not match: expected %s but got %s" % (bucket, (width, height, delta), (lwidth, lheight, seq))
+            assert lrgb_format==rgb_format, "delta region uses %s format, was expecting %s" % (rgb_format, lrgb_format)
             deltalog("delta: xoring with bucket %i", bucket)
             rgb_data = xor_str(img_data, ldata)
         #store new pixels for next delta:
         store = options.intget("store", -1)
         if store>=0:
             deltalog("delta: storing sequence %i in bucket %i", store, bucket)
-            self._delta_pixel_data[bucket] =  width, height, store, rgb_data
+            self._delta_pixel_data[bucket] =  width, height, rgb_format, store, rgb_data
         return rgb_data
 
 
     def paint_image(self, coding, img_data, x, y, width, height, options, callbacks):
         """ can be called from any thread """
         #log("paint_image(%s, %s bytes, %s, %s, %s, %s, %s, %s)", coding, len(img_data), x, y, width, height, options, callbacks)
+        PIL = get_codec("PIL")
         assert PIL, "PIL not found"
         buf = BytesIOClass(img_data)
         img = PIL.Image.open(buf)
@@ -229,7 +227,7 @@ class WindowBackingBase(object):
                 img = img.convert("RGB")
 
         #use tobytes() if present, fallback to tostring():
-        data_fn = getattr(img, "tobytes", getattr(img, "tostring"))
+        data_fn = getattr(img, "tobytes", getattr(img, "tostring", None))
         raw_data = data_fn("raw", img.mode)
         paint_options = typedict(options)
         if img.mode=="RGB":
@@ -277,16 +275,20 @@ class WindowBackingBase(object):
             the actual paint code is in _do_paint_rgb24
         """
         try:
-            success = (self._backing is not None) and self._do_paint_rgb24(img_data, x, y, width, height, rowstride, options)
+            if self._backing is None:
+                fire_paint_callbacks(callbacks, -1, "no backing")
+                return
+            success = self._do_paint_rgb24(img_data, x, y, width, height, rowstride, options)
             fire_paint_callbacks(callbacks, success)
         except KeyboardInterrupt:
             raise
-        except:
+        except Exception as e:
             if not self._backing:
-                log("paint error on closed backing ignored")
+                fire_paint_callbacks(callbacks, -1, "paint error on closed backing ignored")
             else:
                 log.error("do_paint_rgb24 error", exc_info=True)
-            fire_paint_callbacks(callbacks, False)
+                message = "do_paint_rgb24 error: %s" % e
+                fire_paint_callbacks(callbacks, False, message)
 
     def _do_paint_rgb24(self, img_data, x, y, width, height, rowstride, options):
         raise Exception("override me!")
@@ -306,13 +308,19 @@ class WindowBackingBase(object):
             the actual paint code is in _do_paint_rgb32
         """
         try:
-            success = (self._backing is not None) and self._do_paint_rgb32(img_data, x, y, width, height, rowstride, options)
+            if self._backing is None:
+                fire_paint_callbacks(callbacks, -1, "no backing")
+                return
+            success = self._do_paint_rgb32(img_data, x, y, width, height, rowstride, options)
             fire_paint_callbacks(callbacks, success)
         except KeyboardInterrupt:
             raise
-        except:
-            log.error("do_paint_rgb32 error", exc_info=True)
-            fire_paint_callbacks(callbacks, False)
+        except Exception as e:
+            if not self._backing:
+                fire_paint_callbacks(callbacks, -1, "paint error on closed backing ignored")
+            else:
+                log.error("do_paint_rgb32 error", exc_info=True)
+                fire_paint_callbacks(callbacks, False, "do_paint_rgb32 error: %s" % e)
 
     def _do_paint_rgb32(self, img_data, x, y, width, height, rowstride, options):
         raise Exception("override me!")
@@ -348,49 +356,58 @@ class WindowBackingBase(object):
     def paint_with_video_decoder(self, decoder_module, coding, img_data, x, y, width, height, options, callbacks):
         #log("paint_with_video_decoder%s", (decoder_module, coding, "%s bytes" % len(img_data), x, y, width, height, options, callbacks))
         assert decoder_module, "decoder module not found for %s" % coding
-        with self._decoder_lock:
+        dl = self._decoder_lock
+        if dl is None:
+            fire_paint_callbacks(callbacks, False, "no lock - retry")
+            return
+        with dl:
             if self._backing is None:
-                log("window %s is already gone!", self.wid)
-                fire_paint_callbacks(callbacks, False)
+                message = "window %s is already gone!" % self.wid
+                log(message)
+                fire_paint_callbacks(callbacks, -1, message)
                 return  False
             enc_width, enc_height = options.intpair("scaled_size", (width, height))
             input_colorspace = options.strget("csc")
             if not input_colorspace:
-                log.error("csc mode is missing from the video options!")
-                fire_paint_callbacks(callbacks, False)
+                message = "csc mode is missing from the video options!"
+                log.error(message)
+                fire_paint_callbacks(callbacks, False, message)
                 return  False
             #do we need a prep step for decoders that cannot handle the input_colorspace directly?
             decoder_colorspaces = decoder_module.get_input_colorspaces(coding)
             assert input_colorspace in decoder_colorspaces, "decoder does not support %s for %s" % (input_colorspace, coding)
 
-            if self._video_decoder:
-                if self._video_decoder.get_encoding()!=coding:
-                    log("paint_with_video_decoder: encoding changed from %s to %s", self._video_decoder.get_encoding(), coding)
+            vd = self._video_decoder
+            if vd:
+                if vd.get_encoding()!=coding:
+                    log("paint_with_video_decoder: encoding changed from %s to %s", vd.get_encoding(), coding)
                     self.do_clean_video_decoder()
-                elif self._video_decoder.get_width()!=enc_width or self._video_decoder.get_height()!=enc_height:
-                    log("paint_with_video_decoder: window dimensions have changed from %s to %s", (self._video_decoder.get_width(), self._video_decoder.get_height()), (enc_width, enc_height))
+                elif vd.get_width()!=enc_width or vd.get_height()!=enc_height:
+                    log("paint_with_video_decoder: window dimensions have changed from %s to %s", (vd.get_width(), vd.get_height()), (enc_width, enc_height))
                     self.do_clean_video_decoder()
-                elif self._video_decoder.get_colorspace()!=input_colorspace:
+                elif vd.get_colorspace()!=input_colorspace:
                     #this should only happen on encoder restart, which means this should be the first frame:
                     l = log
                     if options.get("frame", 0)>1:
                         l = log.warn
-                    l("paint_with_video_decoder: colorspace changed from %s to %s", self._video_decoder.get_colorspace(), input_colorspace)
+                    l("paint_with_video_decoder: colorspace changed from %s to %s", vd.get_colorspace(), input_colorspace)
                     self.do_clean_video_decoder()
                 elif options.get("frame")==0:
                     log("paint_with_video_decoder: first frame of new stream")
                     self.do_clean_video_decoder()
             if self._video_decoder is None:
                 log("paint_with_video_decoder: new %s(%s,%s,%s)", decoder_module.Decoder, width, height, input_colorspace)
-                self._video_decoder = decoder_module.Decoder()
-                self._video_decoder.init_context(coding, enc_width, enc_height, input_colorspace)
-                log("paint_with_video_decoder: info=%s", self._video_decoder.get_info())
+                vd = decoder_module.Decoder()
+                vd.init_context(coding, enc_width, enc_height, input_colorspace)
+                self._video_decoder = vd
+                log("paint_with_video_decoder: info=%s", vd.get_info())
 
-            img = self._video_decoder.decompress_image(img_data, options)
+            img = vd.decompress_image(img_data, options)
             if not img:
-                fire_paint_callbacks(callbacks, False)
-                log.error("paint_with_video_decoder: wid=%s, %s decompression error on %s bytes of picture data for %sx%s pixels using %s, options=%s",
-                      self.wid, coding, len(img_data), width, height, self._video_decoder, options)
+                fire_paint_callbacks(callbacks, False, "video decoder %s failed to decode %i bytes of %s data" % (vd.get_type(), len(img_data), coding))
+                log.error("Error: decode failed on %s bytes of %s data", len(img_data), coding)
+                log.error(" %sx%s pixels using %s", width, height, vd.get_type())
+                log.error(" decoding options=%s", options)
                 return False
             self.do_video_paint(img, x, y, enc_width, enc_height, width, height, options, callbacks)
         if self._backing is None:
@@ -399,24 +416,25 @@ class WindowBackingBase(object):
 
     def do_video_paint(self, img, x, y, enc_width, enc_height, width, height, options, callbacks):
         #try 24 bit first (paint_rgb24), then 32 bit (paint_rgb32):
-        target_rgb_formats = ["RGB", "RGBX"]
+        target_rgb_formats = self.RGB_MODES
         #as some video formats like vpx can forward transparency
         #also we could skip the csc step in some cases:
         pixel_format = img.get_pixel_format()
-        if self._csc_decoder is not None:
-            if self._csc_decoder.get_src_format()!=pixel_format:
-                log("do_video_paint csc: switching src format from %s to %s", self._csc_decoder.get_src_format(), pixel_format)
+        cd = self._csc_decoder
+        if cd is not None:
+            if cd.get_src_format()!=pixel_format:
+                log("do_video_paint csc: switching src format from %s to %s", cd.get_src_format(), pixel_format)
                 self.do_clean_csc_decoder()
-            elif self._csc_decoder.get_dst_format() not in target_rgb_formats:
-                log("do_video_paint csc: switching dst format from %s to %s", self._csc_decoder.get_dst_format(), target_rgb_formats)
+            elif cd.get_dst_format() not in target_rgb_formats:
+                log("do_video_paint csc: switching dst format from %s to %s", cd.get_dst_format(), target_rgb_formats)
                 self.do_clean_csc_decoder()
-            elif self._csc_decoder.get_src_width()!=enc_width or self._csc_decoder.get_src_height()!=enc_height:
+            elif cd.get_src_width()!=enc_width or cd.get_src_height()!=enc_height:
                 log("do_video_paint csc: switching src size from %sx%s to %sx%s",
-                         enc_width, enc_height, self._csc_decoder.get_src_width(), self._csc_decoder.get_src_height())
+                         enc_width, enc_height, cd.get_src_width(), cd.get_src_height())
                 self.do_clean_csc_decoder()
-            elif self._csc_decoder.get_dst_width()!=width or self._csc_decoder.get_dst_height()!=height:
+            elif cd.get_dst_width()!=width or cd.get_dst_height()!=height:
                 log("do_video_paint csc: switching src size from %sx%s to %sx%s",
-                         width, height, self._csc_decoder.get_dst_width(), self._csc_decoder.get_dst_height())
+                         width, height, cd.get_dst_width(), cd.get_dst_height())
                 self.do_clean_csc_decoder()
         if self._csc_decoder is None:
             #use higher quality csc to compensate for lower quality source
@@ -424,12 +442,13 @@ class WindowBackingBase(object):
             #or when upscaling the video:
             q = options.intget("quality", 50)
             csc_speed = int(min(100, 100-q, 100.0 * (enc_width*enc_height) / (width*height)))
-            self._csc_decoder = self.make_csc(enc_width, enc_height, pixel_format,
+            cd = self.make_csc(enc_width, enc_height, pixel_format,
                                            width, height, target_rgb_formats, csc_speed)
-            log("do_video_paint new csc decoder: %s", self._csc_decoder)
-        rgb_format = self._csc_decoder.get_dst_format()
-        rgb = self._csc_decoder.convert_image(img)
-        log("do_video_paint rgb using %s.convert_image(%s)=%s", self._csc_decoder, img, rgb)
+            log("do_video_paint new csc decoder: %s", cd)
+            self._csc_decoder = cd
+        rgb_format = cd.get_dst_format()
+        rgb = cd.convert_image(img)
+        log("do_video_paint rgb using %s.convert_image(%s)=%s", cd, img, rgb)
         img.free()
         assert rgb.get_planes()==0, "invalid number of planes for %s: %s" % (rgb_format, rgb.get_planes())
         #make a new options dict and set the rgb format:
@@ -468,34 +487,53 @@ class WindowBackingBase(object):
 
     def draw_region(self, x, y, width, height, coding, img_data, rowstride, options, callbacks):
         """ dispatches the paint to one of the paint_XXXX methods """
-        log("draw_region(%s, %s, %s, %s, %s, %s bytes, %s, %s, %s)", x, y, width, height, coding, len(img_data), rowstride, options, callbacks)
-        coding = bytestostr(coding)
-        options["encoding"] = coding            #used for choosing the color of the paint box
-        if coding == "mmap":
-            self.idle_add(self.paint_mmap, img_data, x, y, width, height, rowstride, options, callbacks)
-        elif coding == "rgb24" or coding == "rgb32":
-            #avoid confusion over how many bytes-per-pixel we may have:
-            rgb_format = options.get("rgb_format")
-            if rgb_format:
-                Bpp = len(rgb_format)
-            elif coding=="rgb24":
-                Bpp = 3
+        try:
+            assert self._backing is not None
+            log("draw_region(%s, %s, %s, %s, %s, %s bytes, %s, %s, %s)", x, y, width, height, coding, len(img_data), rowstride, options, callbacks)
+            coding = bytestostr(coding)
+            options["encoding"] = coding            #used for choosing the color of the paint box
+            if INTEGRITY_HASH:
+                l = options.get("z.len")
+                if l:
+                    assert l==len(img_data), "compressed pixel data failed length integrity check: expected %i bytes but got %i" % (l, len(img_data))
+                md5 = options.get("z.md5")
+                if md5:
+                    h = hashlib.md5(img_data)
+                    hd = h.hexdigest()
+                    assert md5==hd, "pixel data failed compressed md5 integrity check: expected %s but got %s" % (md5, hd)
+                deltalog("passed compressed data integrity checks: len=%s, md5=%s (type=%s)", l, md5, type(img_data))
+            if coding == "mmap":
+                self.idle_add(self.paint_mmap, img_data, x, y, width, height, rowstride, options, callbacks)
+            elif coding == "rgb24" or coding == "rgb32":
+                #avoid confusion over how many bytes-per-pixel we may have:
+                rgb_format = options.get("rgb_format")
+                if rgb_format:
+                    Bpp = len(rgb_format)
+                elif coding=="rgb24":
+                    Bpp = 3
+                else:
+                    Bpp = 4
+                if rowstride==0:
+                    rowstride = width * Bpp
+                if Bpp==3:
+                    self.paint_rgb24(img_data, x, y, width, height, rowstride, options, callbacks)
+                else:
+                    self.paint_rgb32(img_data, x, y, width, height, rowstride, options, callbacks)
+            elif coding in VIDEO_DECODERS:
+                self.paint_with_video_decoder(VIDEO_DECODERS.get(coding), coding, img_data, x, y, width, height, options, callbacks)
+            elif coding == "webp":
+                self.paint_webp(img_data, x, y, width, height, options, callbacks)
+            elif coding in self._PIL_encodings:
+                self.paint_image(coding, img_data, x, y, width, height, options, callbacks)
             else:
-                Bpp = 4
-            if rowstride==0:
-                rowstride = width * Bpp
-            if Bpp==3:
-                self.paint_rgb24(img_data, x, y, width, height, rowstride, options, callbacks)
+                self.do_draw_region(x, y, width, height, coding, img_data, rowstride, options, callbacks)
+        except Exception:
+            if self._backing is None:
+                fire_paint_callbacks(callbacks, -1, "this backing is closed - retry?")
             else:
-                self.paint_rgb32(img_data, x, y, width, height, rowstride, options, callbacks)
-        elif coding in VIDEO_DECODERS:
-            self.paint_with_video_decoder(VIDEO_DECODERS.get(coding), coding, img_data, x, y, width, height, options, callbacks)
-        elif coding == "webp":
-            self.paint_webp(img_data, x, y, width, height, options, callbacks)
-        elif coding in self._PIL_encodings:
-            self.paint_image(coding, img_data, x, y, width, height, options, callbacks)
-        else:
-            self.do_draw_region(x, y, width, height, coding, img_data, rowstride, options, callbacks)
+                raise
 
     def do_draw_region(self, x, y, width, height, coding, img_data, rowstride, options, callbacks):
-        raise Exception("invalid encoding: %s" % coding)
+        msg = "invalid encoding: %s" % coding
+        log.error("Error: %s", msg)
+        fire_paint_callbacks(callbacks, False, msg)
