@@ -10,8 +10,9 @@ import struct
 import re
 import binascii
 
-from xpra.gtk_common.gobject_compat import import_gobject, import_gtk, import_gdk
+from xpra.gtk_common.gobject_compat import import_gobject, import_gtk, import_gdk, import_glib
 gobject = import_gobject()
+glib = import_glib()
 gtk = import_gtk()
 gdk = import_gdk()
 
@@ -22,13 +23,15 @@ from xpra.gtk_common.gobject_util import n_arg_signal
 from xpra.gtk_common.gtk_util import GetClipboard, PROPERTY_CHANGE_MASK
 from xpra.gtk_common.nested_main import NestedMainLoop
 from xpra.net.compression import Uncompressed
+from xpra.util import csv
 
 
 MIN_CLIPBOARD_COMPRESSION_SIZE = 512
 MAX_CLIPBOARD_PACKET_SIZE = 4*1024*1024
 
-ALL_CLIPBOARDS = ["CLIPBOARD", "PRIMARY", "SECONDARY"]
-CLIPBOARDS = ALL_CLIPBOARDS
+from xpra.platform.features import CLIPBOARDS as PLATFORM_CLIPBOARDS
+ALL_CLIPBOARDS = PLATFORM_CLIPBOARDS
+CLIPBOARDS = PLATFORM_CLIPBOARDS
 CLIPBOARDS_ENV = os.environ.get("XPRA_CLIPBOARDS")
 if CLIPBOARDS_ENV is not None:
     CLIPBOARDS = CLIPBOARDS_ENV.split(",")
@@ -36,24 +39,33 @@ if CLIPBOARDS_ENV is not None:
 
 
 TEST_DROP_CLIPBOARD_REQUESTS = int(os.environ.get("XPRA_TEST_DROP_CLIPBOARD", "0"))
+STORE_ON_EXIT = os.environ.get("XPRA_CLIPBOARD_STORE_ON_EXIT", "1")=="1"
 
-_discard_target_strs_ = ("^SAVE_TARGETS$",
-        "^COMPOUND_TEXT$",
-        "^NeXT ",
+_discard_target_strs_ = os.environ.get("XPRA_DISCARD_TARGETS")
+if _discard_target_strs_ is not None:
+    DISCARD_TARGETS = _discard_target_strs_.split(",")
+else:
+    #default:
+    DISCARD_TARGETS = [
+        "^SAVE_TARGETS$",
+        "^COMPOUND_TEXT",
+        "^NeXT",
         "^com\.apple\.",
-        "^CorePasteboardFlavorType ",
-        "^dyn\.")
-DISCARD_TARGETS = [re.compile(x) for x in _discard_target_strs_]
+        "^CorePasteboardFlavorType",
+        "^dyn\."]
+log("DISCARD_TARGETS=%s", csv(DISCARD_TARGETS))
+DISCARD_TARGETS = [re.compile(x) for x in DISCARD_TARGETS]
 
 TEXT_TARGETS = ("UTF8_STRING", "TEXT", "STRING", "text/plain")
 
 
 class ClipboardProtocolHelperBase(object):
-    def __init__(self, send_packet_cb, progress_cb=None, clipboards=CLIPBOARDS, filter_res=None):
+    def __init__(self, send_packet_cb, progress_cb=None, **kwargs):
         self.send = send_packet_cb
         self.progress_cb = progress_cb
         self.max_clipboard_packet_size = MAX_CLIPBOARD_PACKET_SIZE
         self.filter_res = []
+        filter_res = kwargs.get("filters")
         if filter_res:
             for x in filter_res:
                 try:
@@ -64,9 +76,10 @@ class ClipboardProtocolHelperBase(object):
         self._clipboard_outstanding_requests = {}
         self._want_targets = False
         self.init_packet_handlers()
-        self.init_proxies(clipboards)
+        self.init_proxies(kwargs.get("clipboards.local", CLIPBOARDS))
+        self.remote_clipboards = kwargs.get("clipboards.remote", CLIPBOARDS)
 
-    def __str__(self):
+    def __repr__(self):
         return "ClipboardProtocolHelperBase"
 
     def get_info(self):
@@ -91,9 +104,10 @@ class ClipboardProtocolHelperBase(object):
         self._clipboard_proxies = {}
 
     def enable_selections(self, selections):
-        #when clients connect, they can tell us which
-        #clipboards they want enabled
-        #(ie: OSX and win32 only use "CLIPBOARD", and not "PRIMARY" or "SECONDARY")
+        #when clients first connect or later through the "clipboard-enable-selections" packet,
+        #they can tell us which clipboard selections they want enabled
+        #(ie: OSX and win32 only use "CLIPBOARD" by default, and not "PRIMARY" or "SECONDARY")
+        log("enabling selections: %s", csv(selections))
         for selection, proxy in self._clipboard_proxies.items():
             proxy.set_enabled(selection in selections)
 
@@ -112,6 +126,7 @@ class ClipboardProtocolHelperBase(object):
             "clipboard-contents":           self._process_clipboard_contents,
             "clipboard-contents-none":      self._process_clipboard_contents_none,
             "clipboard-pending-requests":   self._process_clipboard_pending_requests,
+            "clipboard-enable-selections":  self._process_clipboard_enable_selections,
             }
 
     def make_proxy(self, clipboard):
@@ -125,7 +140,7 @@ class ClipboardProtocolHelperBase(object):
             proxy.connect("get-clipboard-from-remote", self._get_clipboard_from_remote_handler)
             proxy.show()
             self._clipboard_proxies[clipboard] = proxy
-        log("%s.init_proxies : %s", type(self), self._clipboard_proxies)
+        log("%s.init_proxies : %s", self, self._clipboard_proxies)
 
     def local_to_remote(self, selection):
         #overriden in some subclasses (see: translated_clipboard)
@@ -144,7 +159,12 @@ class ClipboardProtocolHelperBase(object):
         name = self.remote_to_local(selection)
         proxy = self._clipboard_proxies.get(name)
         if proxy is None:
-            log.warn("ignoring token for clipboard proxy name '%s' (no proxy)", name)
+            #this can happen if the server has fewer clipboards than the client,
+            #ie: with win32 shadow servers
+            l = log
+            if name in ALL_CLIPBOARDS:
+                l = log.warn
+            l("ignoring token for clipboard proxy name '%s' (no proxy)", name)
             return
         if not proxy.is_enabled():
             log.warn("ignoring token for clipboard proxy name '%s' (disabled)", name)
@@ -346,7 +366,7 @@ class ClipboardProtocolHelperBase(object):
                 no_contents()
                 return
             wire_data = self._may_compress(dtype, dformat, wire_data)
-            if wire_data:
+            if wire_data is not None:
                 self.send("clipboard-contents", request_id, selection,
                        dtype, dformat, wire_encoding, wire_data)
         proxy.get_contents(target, got_contents)
@@ -377,6 +397,11 @@ class ClipboardProtocolHelperBase(object):
         if self.progress_cb:
             self.progress_cb(None, pending)
 
+    def _process_clipboard_enable_selections(self, packet):
+        selections = packet[1]
+        self.enable_selections(selections)
+
+
     def process_clipboard_packet(self, packet):
         packet_type = packet[0]
         log("process clipboard packet type=%s", packet_type)
@@ -449,7 +474,7 @@ class ClipboardProxy(gtk.Invisible):
         return info
 
     def cleanup(self):
-        if not self._have_token:
+        if not self._have_token and STORE_ON_EXIT:
             self._clipboard.store()
         self.destroy()
 
@@ -464,17 +489,18 @@ class ClipboardProxy(gtk.Invisible):
         log("%s.set_greedy_client(%s)", self, greedy)
         self._greedy_client = greedy
 
-    def __str__(self):
+    def __repr__(self):
         return  "ClipboardProxy(%s)" % self._selection
 
     def do_owner_changed(self, *args):
-        log("do_owner_changed(%s) greedy_client=%s, block_owner_change=%s", args, self._greedy_client, self._block_owner_change)
+        #log("do_owner_changed(%s) greedy_client=%s, block_owner_change=%s", args, self._greedy_client, self._block_owner_change)
+        log("clipboard: %s owner_changed greedy_client=%s, block_owner_change=%s", self._selection, self._greedy_client, self._block_owner_change)
         if self._enabled and self._greedy_client and not self._block_owner_change:
             self._block_owner_change = True
             self._have_token = False
             self.emit("send-clipboard-token", self._selection)
             self._sent_token_events += 1
-            gobject.idle_add(self.remove_block)
+            glib.idle_add(self.remove_block)
 
     def do_selection_request_event(self, event):
         log("do_selection_request_event(%s)", event)
@@ -576,7 +602,7 @@ class ClipboardProxy(gtk.Invisible):
                 self._block_owner_change = True
                 self.emit("send-clipboard-token", self._selection)
                 if boc is False:
-                    gobject.idle_add(self.remove_block)
+                    glib.idle_add(self.remove_block)
         gtk.Invisible.do_selection_clear_event(self, event)
 
     def got_token(self, targets, target_data):
@@ -592,10 +618,10 @@ class ClipboardProxy(gtk.Invisible):
         if self._block_owner_change:
             #re-enable the flag via idle_add so events like do_owner_changed
             #get a chance to run first.
-            gobject.idle_add(self.remove_block)
+            glib.idle_add(self.remove_block)
 
     def remove_block(self, *args):
-        log("remove_block(%s)", args)
+        log("remove_block: %s", self._selection)
         self._block_owner_change = False
 
     def claim(self):
