@@ -13,13 +13,14 @@ import numpy
 import pyopencl             #@UnresolvedImport
 from pyopencl import mem_flags  #@UnresolvedImport
 
-from xpra.os_util import memoryview_to_bytes, _memoryview
+from xpra.util import updict, engs
+from xpra.os_util import _memoryview
 
 PREFERRED_DEVICE_TYPE = os.environ.get("XPRA_OPENCL_DEVICE_TYPE", "GPU")
 PREFERRED_DEVICE_NAME = os.environ.get("XPRA_OPENCL_DEVICE_NAME", "")
 PREFERRED_DEVICE_PLATFORM = os.environ.get("XPRA_OPENCL_PLATFORM", "")
 
-NVIDIA_YUV2RGB = os.environ.get("XPRA_OPENCL_NVIDIA_YUV2RGB", "0")=="1"
+OPENCL_YUV2RGB = os.environ.get("XPRA_OPENCL_YUV2RGB", "0")=="1"
 
 AMD_WARNING_SHOWN = not os.environ.get("XPRA_AMD_WARNING", "1")=="1"
 
@@ -62,13 +63,13 @@ def log_device_info(device):
     log("max_work_item_sizes=%s", device.max_work_item_sizes)
 
 def log_platforms_info():
-    log("found %s OpenCL platforms:", len(opencl_platforms))
+    log("found %s OpenCL platform%s:", len(opencl_platforms), engs(opencl_platforms))
     for platform in opencl_platforms:
         devices = platform.get_devices()
         p = "*"
         if not is_supported(platform.name):
             p = "-"
-        log("%s %s - %s devices:", p, platform_info(platform), len(devices))
+        log("%s %s - %s device%s:", p, platform_info(platform), len(devices), engs(devices))
         for d in devices:
             p = "-"
             if d.available and d.compiler_available and d.get_info(pyopencl.device_info.IMAGE_SUPPORT) and is_supported(platform.name):
@@ -86,6 +87,7 @@ selected_platform = None
 selected_device_cpu_cost = 50
 selected_device_gpu_cost = 50
 selected_device_setup_cost = 50
+selected_device_max_size = 64, 64
 
 context = None
 def reselect_device():
@@ -95,7 +97,7 @@ def reselect_device():
     context = None
     select_device()
 def select_device():
-    global context, selected_device,selected_platform,AMD_WARNING_SHOWN
+    global context, selected_device,selected_platform,selected_device_max_size,AMD_WARNING_SHOWN
     if context is not None:
         return
     log_version_info()
@@ -139,13 +141,14 @@ def select_device():
                         score += 10
 
                 options.setdefault(score, []).append((d, platform))
-    log("best device/platform options: %s", options)
+    log("best device/platform option%s: %s", engs(options), options)
     for score in reversed(sorted(options.keys())):
         for d, p in options.get(score):
             try:
                 log("trying platform: %s", platform_info(p))
                 log("with %s device: %s", device_type(d), device_info(d))
                 context = pyopencl.Context([d])
+                selected_device_max_size = d.get_info(pyopencl.device_info.IMAGE2D_MAX_WIDTH), d.get_info(pyopencl.device_info.IMAGE2D_MAX_HEIGHT)
                 selected_platform = p
                 selected_device = d
                 log.info(" using platform: %s", platform_info(selected_platform))
@@ -161,6 +164,7 @@ def select_device():
                     selected_device_gpu_cost = 0
                     selected_device_setup_cost = 20
                 log("device is a %s, using CPU cost=%s, GPU cost=%s", device_type(d), selected_device_cpu_cost, selected_device_gpu_cost)
+                log(" max image 2d size: %s", selected_device_max_size)
                 return
             except Exception as e:
                 log.warn(" failed to use %s", platform_info(p))
@@ -170,7 +174,7 @@ def select_device():
     log.warn("OpenCL Error: failed to find a working platform and device combination... trying with pyopencl's 'create_some_context'")
     context = pyopencl.create_some_context(interactive=False)
     devices = context.get_info(pyopencl.context_info.DEVICES)
-    log.info("chosen context has %s device(s):", len(devices))
+    log.info("chosen context has %s device%s:", len(devices), engs(devices))
     for d in devices:
         log_device_info(d)
     assert len(devices)==1, "we only handle a single device at a time, sorry!"
@@ -194,7 +198,7 @@ for rgb_mode, channel_order_name in (
                   ("RGB" ,  "RGB"),    #pyopencl.channel_order.RGB
                   ):
     if not hasattr(pyopencl.channel_order, channel_order_name):
-        log("this build does not have support for %s", channel_order_name)
+        log.warn("this PyOpenCL build does not support %s", channel_order_name)
         continue
     channel_order = getattr(pyopencl.channel_order, channel_order_name)
     IN_CHANNEL_ORDER.append((rgb_mode, channel_order))
@@ -230,8 +234,8 @@ def gen_yuv_to_rgb():
 
     YUV_to_RGB_KERNELS = {}
 
-    if selected_platform and selected_platform.name and selected_platform.name.find("CUDA")>=0 and not NVIDIA_YUV2RGB:
-        log.warn("CUDA device detected, YUV to RGB disabled")
+    if not OPENCL_YUV2RGB:
+        log.warn("OpenCL YUV to RGB is disabled")
         return {}
 
     #for YUV to RGB support we need to be able to handle the channel_order in WRITE_ONLY mode
@@ -250,8 +254,7 @@ def gen_yuv_to_rgb():
             found_rgb.add(dst_rgb_mode)
 
     for rgb_mode, channel_order in IN_CHANNEL_ORDER:
-        #why do we discard RGBX download mode? because it doesn't work, don't ask me why
-        if not has_image_format(sif, channel_order, pyopencl.channel_type.UNSIGNED_INT8) or rgb_mode=="RGBX":
+        if not has_image_format(sif, channel_order, pyopencl.channel_type.UNSIGNED_INT8):
             log("YUV 2 RGB: channel order %s is not supported directly in WRITE_ONLY + UNSIGNED_INT8 mode", CHANNEL_ORDER_TO_STR.get(channel_order))
             missing.append((rgb_mode, channel_order))
             continue
@@ -444,7 +447,7 @@ def build_kernels():
 
 
 from xpra.codecs.image_wrapper import ImageWrapper
-from xpra.codecs.codec_constants import codec_spec, get_subsampling_divs
+from xpra.codecs.codec_constants import csc_spec, get_subsampling_divs
 
 
 def init_module():
@@ -469,31 +472,39 @@ def get_type():
 def get_version():
     return pyopencl.version.VERSION
 
-def get_info():
-    global selected_device, selected_platform, context, KERNELS_DEFS
+def get_pyopencl_info():
     info = {"version"               : pyopencl.version.VERSION,
             "version.text"          : pyopencl.VERSION_TEXT,
-            "version.status"        : pyopencl.VERSION_STATUS,
+            }
+    if pyopencl.VERSION_STATUS:
+            info["version.status"] = pyopencl.VERSION_STATUS
+    return info
+
+def get_info():
+    global selected_device, selected_platform, context, KERNELS_DEFS
+    info = {
             "version.cl_header"     : pyopencl.get_cl_header_version(),
             "opengl"                : pyopencl.have_gl(),
             #"kernels"               : KERNELS_DEFS.keys()
             }
+    updict(info, "pyopencl", get_pyopencl_info())
     if selected_platform:
-        info.update({
-            "platform.name"         : selected_platform.name,
-            "platform.vendor"       : selected_platform.vendor,
-            "platform.devices"      : len(selected_platform.get_devices()),
+        updict(info, "platform", {
+            "name"          : selected_platform.name,
+            "vendor"        : selected_platform.vendor,
+            "devices"       : len(selected_platform.get_devices()),
             })
     if selected_device:
         if hasattr(selected_device, "opencl_c_version"):
             info["device.opencl_c_version"] = getattr(selected_device, "opencl_c_version")
-        info.update({
-            "device.type"           : device_type(selected_device),
-            "device.name"           : selected_device.name.strip(),
-            "device.version"        : selected_device.version,
-            "device.max_work_group_size"        : selected_device.max_work_group_size,
-            "device.max_work_item_dimensions"   : selected_device.max_work_item_dimensions,
-            "device.max_work_item_sizes"        : selected_device.max_work_item_sizes})
+        updict(info, "device", {
+            "type"                      : device_type(selected_device),
+            "name"                      : selected_device.name.strip(),
+            "version"                   : selected_device.version,
+            "max_work_group_size"       : selected_device.max_work_group_size,
+            "max_work_item_dimensions"  : selected_device.max_work_item_dimensions,
+            "max_work_item_sizes"       : selected_device.max_work_item_sizes,
+            "max-size"                  : selected_device_max_size})
     return info
 
 
@@ -511,11 +522,12 @@ def validate_in_out(in_colorspace, out_colorspace):
 
 def get_spec(in_colorspace, out_colorspace):
     validate_in_out(in_colorspace, out_colorspace)
-    global selected_device_cpu_cost, selected_device_gpu_cost, selected_device_setup_cost
-    return codec_spec(ColorspaceConverter, codec_type=get_type(),
+    global selected_device_cpu_cost, selected_device_gpu_cost, selected_device_setup_cost, selected_device_max_size
+    max_w, max_h = selected_device_max_size
+    return csc_spec(ColorspaceConverter, codec_type=get_type(),
                       speed=100,
                       setup_cost=selected_device_setup_cost,
-                      cpu_cost=selected_device_cpu_cost, gpu_cost=selected_device_gpu_cost, min_w=128, min_h=128, can_scale=True)
+                      cpu_cost=selected_device_cpu_cost, gpu_cost=selected_device_gpu_cost, min_w=128, min_h=128, can_scale=True, max_w=max_w, max_h=max_h)
 
 
 class ColorspaceConverter(object):
@@ -561,7 +573,11 @@ class ColorspaceConverter(object):
         self.program = program
         self.queue = pyopencl.CommandQueue(self.context)
         fm = pyopencl.filter_mode.NEAREST
-        self.sampler = pyopencl.Sampler(self.context, False, pyopencl.addressing_mode.CLAMP_TO_EDGE, fm)
+        try:
+            self.sampler = pyopencl.Sampler(self.context, False, pyopencl.addressing_mode.CLAMP_TO_EDGE, fm)
+        except Exception as e:
+            #this fails with POCL:
+            raise Exception("failed to create CLAMP_TO_EDGE + NEAREST OpenCL Sampler: %s" % e)
         k_def = KERNELS_DEFS.get((self.src_format, self.dst_format))
         assert k_def, "no kernel found for %s to %s" % (self.src_format, self.dst_format)
         self.kernel_function_name, _, self.channel_order, src = k_def
@@ -678,7 +694,7 @@ class ColorspaceConverter(object):
             return self.do_convert_image(image)
         except pyopencl.LogicError as e:
             if retry>0:
-                raise e
+                raise
             log.warn("OpenCL error: %s", e)
             self.reinit()
             return self.convert_image(image, retry+1)
@@ -697,6 +713,7 @@ class ColorspaceConverter(object):
         height = image.get_height()
         strides = image.get_rowstride()
         pixels = image.get_pixels()
+        assert pixels, "failed to get pixels from %s" % image
         assert iplanes==ImageWrapper._3_PLANES, "we only handle planar data as input!"
         assert image.get_pixel_format()==self.src_format, "invalid source format: %s (expected %s)" % (image.get_pixel_format(), self.src_format)
         assert len(strides)==len(pixels)==3, "invalid number of planes or strides (should be 3)"
@@ -715,20 +732,20 @@ class ColorspaceConverter(object):
         input_images = []
         for i in range(3):
             _, y_div = divs[i]
+            shape = strides[i], self.src_height//y_div
             plane = pixels[i]
             if type(plane)==_memoryview:
-                plane = memoryview_to_bytes(plane)
+                plane = plane.tobytes()
             if type(plane)==str:
                 flags = mem_flags.READ_ONLY | mem_flags.COPY_HOST_PTR
             else:
                 flags = mem_flags.READ_ONLY | mem_flags.USE_HOST_PTR
-            shape = strides[i], self.src_height//y_div
             iimage = pyopencl.Image(self.context, flags, iformat, shape=shape, hostbuf=plane)
             input_images.append(iimage)
 
         #output image:
         oformat = pyopencl.ImageFormat(self.channel_order, pyopencl.channel_type.UNORM_INT8)
-        oimage = pyopencl.Image(self.context, mem_flags.WRITE_ONLY, oformat, shape=(self.dst_width, self.dst_height))
+        oimage = pyopencl.Image(self.context, mem_flags.WRITE_ONLY | mem_flags.ALLOC_HOST_PTR, oformat, shape=(self.dst_width, self.dst_height))
 
         kernelargs += input_images + [numpy.int32(self.src_width), numpy.int32(self.src_height),
                        numpy.int32(self.dst_width), numpy.int32(self.dst_height),
@@ -738,17 +755,18 @@ class ColorspaceConverter(object):
         log("convert_image(%s) calling %s%s after upload took %.1fms",
               image, self.kernel_function_name, tuple(kernelargs), 1000.0*(kstart-start))
         self.kernel_function(*kernelargs)
-        self.queue.finish()
-        #free input images:
-        for iimage in input_images:
-            iimage.release()
         kend = time.time()
         log("%s took %.1fms", self.kernel_function, 1000.0*(kend-kstart))
 
         out_array = numpy.empty(self.dst_width*self.dst_height*4, dtype=numpy.byte)
-        pyopencl.enqueue_read_image(self.queue, oimage, (0, 0), (self.dst_width, self.dst_height), out_array)
+        log("out array=%s", out_array)
+        pyopencl.enqueue_copy(self.queue, out_array, oimage, origin=(0,0), region=(self.dst_width,self.dst_height))
         self.queue.finish()
         log("readback using %s took %.1fms", CHANNEL_ORDER_TO_STR.get(self.channel_order), 1000.0*(time.time()-kend))
+        #free input images:
+        for iimage in input_images:
+            iimage.release()
+        oimage.release()
         self.time += time.time()-start
         self.frames += 1
         return ImageWrapper(0, 0, self.dst_width, self.dst_height, out_array.data, self.dst_format, 24, self.dst_width*4, planes=ImageWrapper.PACKED)
@@ -762,6 +780,7 @@ class ColorspaceConverter(object):
         stride = image.get_rowstride()
         pixels = image.get_pixels()
         #log("convert_image(%s) planes=%s, pixels=%s, size=%s", image, iplanes, type(pixels), len(pixels))
+        assert pixels, "failed to get pixels from %s" % image
         assert iplanes==ImageWrapper.PACKED, "we only handle packed data as input!"
         assert image.get_pixel_format()==self.src_format, "invalid source format: %s (expected %s)" % (image.get_pixel_format(), self.src_format)
         assert width>=self.src_width and height>=self.src_height, "expected source image with dimensions of at least %sx%s but got %sx%s" % (self.src_width, self.src_height, width, height)
@@ -779,7 +798,7 @@ class ColorspaceConverter(object):
         log("convert_image() type=%s, input image format=%s, shape=%s, work size: local=%s, global=%s", type(pixels), iformat, shape, localWorkSize, globalWorkSize)
         idata = pixels
         if type(idata)==_memoryview:
-            idata = memoryview_to_bytes(idata)
+            idata = idata.tobytes()
         if type(idata)==str:
             #str is not a buffer, so we have to copy the data
             #alternatively, we could copy it first ourselves using this:
@@ -814,9 +833,6 @@ class ColorspaceConverter(object):
         kstart = time.time()
         log("convert_image(%s) calling %s%s after %.1fms", image, self.kernel_function_name, tuple(kernelargs), 1000.0*(kstart-start))
         self.kernel_function(*kernelargs)
-        self.queue.finish()
-        #free input image:
-        iimage.release()
         kend = time.time()
         log("%s took %.1fms", self.kernel_function_name, 1000.0*(kend-kstart))
 
@@ -825,13 +841,22 @@ class ColorspaceConverter(object):
         for i in range(3):
             out_array = numpy.empty(out_sizes[i], dtype=numpy.byte)
             pixels.append(out_array.data)
-            pyopencl.enqueue_read_buffer(self.queue, out_buffers[i], out_array, is_blocking=False)
+            pyopencl.enqueue_copy(self.queue, out_array, out_buffers[i], is_blocking=False)
         readstart = time.time()
         log("queue read events took %.1fms (3 planes of size %s, with strides=%s)", 1000.0*(readstart-kend), out_sizes, strides)
         self.queue.finish()
         readend = time.time()
         log("wait for read events took %.1fms", 1000.0*(readend-readstart))
+        iimage.release()
         #free output buffers:
         for out_buf in out_buffers:
             out_buf.release()
         return ImageWrapper(0, 0, self.dst_width, self.dst_height, pixels, self.dst_format, 24, strides, planes=ImageWrapper._3_PLANES)
+
+
+def selftest(full=False):
+    if pyopencl.version.VERSION[0]<2015:
+        raise ImportError("PyOpenCL version %s is too old, use 2015 or later" % pyopencl.VERSION_TEXT)
+    from xpra.codecs.codec_checks import testcsc
+    from xpra.codecs.csc_opencl import colorspace_converter
+    testcsc(colorspace_converter, full)
