@@ -8,13 +8,28 @@ import sys
 import os.path
 from xpra.log import Logger
 log = Logger("codec", "loader")
+try:
+    #this makes py2exe do the right thing:
+    from xpra.codecs import codec_checks
+    assert codec_checks
+except:
+    pass
 
 if sys.version > '3':
     unicode = str           #@ReservedAssignment
 
 #these codecs may well not load because we
 #do not require the libraries to be installed
-NOWARN = ["nvenc3", "nvenc4", "nvenc5", "opencl"]
+NOWARN = ["nvenc4", "nvenc5", "opencl"]
+
+SELFTEST = os.environ.get("XPRA_CODEC_SELFTEST", "1")=="1"
+FULL_SELFTEST = os.environ.get("XPRA_CODEC_FULL_SELFTEST", "0")=="1"
+
+CODEC_FAIL_IMPORT = os.environ.get("XPRA_CODEC_FAIL_IMPORT", "").split(",")
+CODEC_FAIL_SELFTEST = os.environ.get("XPRA_CODEC_FAIL_SELFTEST", "").split(",")
+
+log("codec loader settings: SELFTEST=%s, FULL_SELFTEST=%s, CODEC_FAIL_IMPORT=%s, CODEC_FAIL_SELFTEST=%s",
+        SELFTEST, FULL_SELFTEST, CODEC_FAIL_IMPORT, CODEC_FAIL_SELFTEST)
 
 codec_errors = {}
 codecs = {}
@@ -23,9 +38,11 @@ def codec_import_check(name, description, top_module, class_module, *classnames)
     log(" codec_import_check%s", (name, description, top_module, class_module, classnames))
     try:
         try:
+            if name in CODEC_FAIL_IMPORT:
+                raise ImportError("codec found in fail import list")
             __import__(top_module, {}, {}, [])
         except ImportError as e:
-            log(" cannot import %s (%s): %s", name, description, e)
+            log("failed to import %s (%s)", description, name)
             log("", exc_info=True)
             codec_errors[name] = e
             return None
@@ -41,23 +58,25 @@ def codec_import_check(name, description, top_module, class_module, *classnames)
                 ic =  __import__(class_module, {}, {}, classname)
                 selftest = getattr(ic, "selftest", None)
                 log("%s.selftest=%s", name, selftest)
-                if selftest:
-                    selftest()
+                if SELFTEST and selftest:
+                    if name in CODEC_FAIL_SELFTEST:
+                        raise ImportError("codec found in fail selftest list")
+                    selftest(FULL_SELFTEST)
                 #log.warn("codec_import_check(%s, ..)=%s" % (name, ic))
                 log(" found %s : %s", name, ic)
                 codecs[name] = ic
                 return ic
         except ImportError as e:
             codec_errors[name] = e
-            l = log.warn
+            l = log.error
             if name in NOWARN:
                 l = log.debug
-            l(" cannot import %s (%s): %s", name, description, e)
+            l("Error importing %s (%s)", description, name)
+            l(" %s", e)
             log("", exc_info=True)
     except Exception as e:
         codec_errors[name] = e
-        log.warn(" cannot load %s (%s): %s missing from %s: %s", name, description, classname, class_module, e)
-        log.warn("", exc_info=True)
+        log.warn(" cannot load %s (%s): %s missing from %s", name, description, classname, class_module, exc_info=True)
     return None
 codec_versions = {}
 def add_codec_version(name, top_module, version="get_version()", alt_version="__version__"):
@@ -90,80 +109,75 @@ def add_codec_version(name, top_module, version="get_version()", alt_version="__
         log("", exc_info=True)
     except Exception as e:
         log.warn("error during codec import: %s", e)
-        log("", exc_info=True)
+        log.warn("", exc_info=True)
     return None
 
-def PIL_logging_workaround():
-    import logging
-    PIL_DEBUG = os.environ.get("XPRA_PIL_DEBUG", "0")=="1"
-    if PIL_DEBUG:
-        from xpra.log import Logger
-        log = Logger("util")
-        log.info("enabling PIL.DEBUG")
-        level = logging.DEBUG
-    else:
-        level = logging.INFO
 
-    #newer versions use this logger,
-    #we must initialize it before we load the class:
-    for x in ("Image", "PngImagePlugin", "WebPImagePlugin", "JpegImagePlugin"):
-        logger = logging.getLogger("PIL.%s" % x)
-        logger.setLevel(level)
-    import PIL
-    from PIL import Image
-    assert PIL and Image
-    if hasattr(Image, "DEBUG"):
-        #for older versions (pre 3.0), use Image.DEBUG flag:
-        Image.DEBUG = int(PIL_DEBUG)
-
-loaded = False
-def load_codecs():
+loaded = None
+def load_codecs(encoders=True, decoders=True, csc=True):
     global loaded
     if loaded:
         return
     loaded = True
+    show = []
     log("loading codecs")
-    try:
-        PIL_logging_workaround()
-    except:
-        log("error in PIL logging workaround", exc_info=True)
-    codec_import_check("PIL", "Python Imaging Library", "PIL", "PIL", "Image")
-    add_codec_version("PIL", "PIL.Image", "PILLOW_VERSION", "VERSION")
+    if encoders or decoders:
+        codec_import_check("PIL", "Python Imaging Library", "PIL", "PIL", "Image")
+        add_codec_version("PIL", "PIL.Image", "PILLOW_VERSION", "VERSION")
 
-    codec_import_check("enc_vpx", "vpx encoder", "xpra.codecs.vpx", "xpra.codecs.vpx.encoder", "Encoder")
-    codec_import_check("dec_vpx", "vpx decoder", "xpra.codecs.vpx", "xpra.codecs.vpx.decoder", "Decoder")
-    add_codec_version("vpx", "xpra.codecs.vpx.encoder")
+    if encoders:
+        show += list(ENCODER_CODECS)
+        codec_import_check("enc_pillow", "Pillow encoder", "xpra.codecs.pillow", "xpra.codecs.pillow.encode", "encode")
+        add_codec_version("enc_pillow", "xpra.codecs.pillow.encode")
 
-    codec_import_check("enc_x264", "x264 encoder", "xpra.codecs.enc_x264", "xpra.codecs.enc_x264.encoder", "Encoder")
-    add_codec_version("x264", "xpra.codecs.enc_x264.encoder")
+        codec_import_check("enc_webp", "webp encoder", "xpra.codecs.webp", "xpra.codecs.webp.encode", "compress")
+        add_codec_version("enc_webp", "xpra.codecs.webp.encode")
 
-    codec_import_check("enc_x265", "x265 encoder", "xpra.codecs.enc_x265", "xpra.codecs.enc_x265.encoder", "Encoder")
-    add_codec_version("x265", "xpra.codecs.enc_x265.encoder")
+        codec_import_check("enc_vpx", "vpx encoder", "xpra.codecs.vpx", "xpra.codecs.vpx.encoder", "Encoder")
+        add_codec_version("vpx", "xpra.codecs.vpx.decoder")
 
-    for v in (4, 3, 5):
-        codec_import_check("nvenc%s" % v, "nvenc encoder", "xpra.codecs.nvenc%s" % v, "xpra.codecs.nvenc%s.encoder" % v, "Encoder")
-        add_codec_version("nvenc%s" % v, "xpra.codecs.nvenc%s.encoder" % v)
+        codec_import_check("enc_x264", "x264 encoder", "xpra.codecs.enc_x264", "xpra.codecs.enc_x264.encoder", "Encoder")
+        add_codec_version("x264", "xpra.codecs.enc_x264.encoder")
 
-    codec_import_check("csc_swscale", "swscale colorspace conversion", "xpra.codecs.csc_swscale", "xpra.codecs.csc_swscale.colorspace_converter", "ColorspaceConverter")
-    add_codec_version("swscale", "xpra.codecs.csc_swscale.colorspace_converter")
+        codec_import_check("enc_x265", "x265 encoder", "xpra.codecs.enc_x265", "xpra.codecs.enc_x265.encoder", "Encoder")
+        add_codec_version("x265", "xpra.codecs.enc_x265.encoder")
 
-    codec_import_check("csc_cython", "cython colorspace conversion", "xpra.codecs.csc_cython", "xpra.codecs.csc_cython.colorspace_converter", "ColorspaceConverter")
-    add_codec_version("cython", "xpra.codecs.csc_cython.colorspace_converter")
+        for v in (4, 5):
+            codec_import_check("nvenc%s" % v, "nvenc encoder", "xpra.codecs.nvenc%s" % v, "xpra.codecs.nvenc%s.encoder" % v, "Encoder")
+            add_codec_version("nvenc%s" % v, "xpra.codecs.nvenc%s.encoder" % v)
 
-    codec_import_check("csc_opencl", "OpenCL colorspace conversion", "xpra.codecs.csc_opencl", "xpra.codecs.csc_opencl.colorspace_converter", "ColorspaceConverter")
-    add_codec_version("opencl", "xpra.codecs.csc_opencl.colorspace_converter")
+    if csc:
+        show += list(CSC_CODECS)
+        codec_import_check("csc_swscale", "swscale colorspace conversion", "xpra.codecs.csc_swscale", "xpra.codecs.csc_swscale.colorspace_converter", "ColorspaceConverter")
+        add_codec_version("swscale", "xpra.codecs.csc_swscale.colorspace_converter")
 
-    codec_import_check("dec_avcodec2", "avcodec2 decoder", "xpra.codecs.dec_avcodec2", "xpra.codecs.dec_avcodec2.decoder", "Decoder")
-    add_codec_version("avcodec2", "xpra.codecs.dec_avcodec2.decoder")
+        codec_import_check("csc_cython", "cython colorspace conversion", "xpra.codecs.csc_cython", "xpra.codecs.csc_cython.colorspace_converter", "ColorspaceConverter")
+        add_codec_version("cython", "xpra.codecs.csc_cython.colorspace_converter")
 
-    codec_import_check("enc_webp", "webp encoder", "xpra.codecs.webp", "xpra.codecs.webp.encode", "compress")
-    add_codec_version("enc_webp", "xpra.codecs.webp.encode")
+        codec_import_check("csc_opencl", "OpenCL colorspace conversion", "xpra.codecs.csc_opencl", "xpra.codecs.csc_opencl.colorspace_converter", "ColorspaceConverter")
+        add_codec_version("opencl", "xpra.codecs.csc_opencl.colorspace_converter")
 
-    codec_import_check("dec_webp", "webp decoder", "xpra.codecs.webp", "xpra.codecs.webp.decode", "decompress")
-    add_codec_version("dec_webp", "xpra.codecs.webp.decode")
+    if decoders:
+        show += list(DECODER_CODECS)
+        codec_import_check("dec_pillow", "Pillow decoder", "xpra.codecs.pillow", "xpra.codecs.pillow.decode", "decode")
+        add_codec_version("dec_pillow", "xpra.codecs.pillow.decode")
 
-    #not really a codec, but gets used by codecs:
+        codec_import_check("dec_webp", "webp decoder", "xpra.codecs.webp", "xpra.codecs.webp.decode", "decompress")
+        add_codec_version("dec_webp", "xpra.codecs.webp.decode")
+
+        codec_import_check("dec_vpx", "vpx decoder", "xpra.codecs.vpx", "xpra.codecs.vpx.decoder", "Decoder")
+        add_codec_version("vpx", "xpra.codecs.vpx.encoder")
+
+        codec_import_check("dec_avcodec2", "avcodec2 decoder", "xpra.codecs.dec_avcodec2", "xpra.codecs.dec_avcodec2.decoder", "Decoder")
+        add_codec_version("avcodec2", "xpra.codecs.dec_avcodec2.decoder")
+
+    #not really a codec, but gets used by codecs, so include version info:
     add_codec_version("numpy", "numpy")
+    try:
+        from xpra.codecs.argb.argb import buffer_api_version            #@UnresolvedImport
+        codec_versions["buffer_api"] = buffer_api_version()
+    except Exception as e:
+        log("unknown buffer api version: %s", e)
 
     log("done loading codecs")
     log("found:")
@@ -176,28 +190,27 @@ def load_codecs():
 
 
 def get_codec_error(name):
+    assert loaded
     return codec_errors.get(name)
 
 def get_codec(name):
-    load_codecs()
+    assert loaded
     return codecs.get(name)
 
 def get_codec_version(name):
-    load_codecs()
+    assert loaded
     return codec_versions.get(name)
 
 def has_codec(name):
-    load_codecs()
+    assert loaded
     return name in codecs
 
-OLD_ENCODING_NAMES_TO_NEW = {"x264" : "h264", "vpx" : "vp8"}
-ALL_OLD_ENCODING_NAMES_TO_NEW = {"x264" : "h264", "vpx" : "vp8", "rgb24" : "rgb"}
 
-ALL_CODECS = "PIL", "enc_vpx", "dec_vpx", "enc_x264", "enc_x265", "nvenc3", "nvenc4", "nvenc5", \
-            "csc_swscale", "csc_cython", "csc_opencl", \
-            "dec_avcodec2", \
-            "enc_webp", \
-            "dec_webp"
+CSC_CODECS = "csc_swscale", "csc_cython", "csc_opencl"
+ENCODER_CODECS = "enc_pillow", "enc_vpx", "enc_webp", "enc_x264", "enc_x265", "nvenc4", "nvenc5"
+DECODER_CODECS = "dec_pillow", "dec_vpx", "dec_webp", "dec_avcodec2"
+
+ALL_CODECS = tuple(set(CSC_CODECS + ENCODER_CODECS + DECODER_CODECS))
 
 #note: this is just for defining the order of encodings,
 #so we have both core encodings (rgb24/rgb32) and regular encodings (rgb) in here:
@@ -205,6 +218,7 @@ PREFERED_ENCODING_ORDER = ["h264", "vp9", "vp8", "png", "png/P", "png/L", "webp"
 #encoding order for edges (usually one pixel high or wide):
 EDGE_ENCODING_ORDER = ["rgb24", "rgb32", "jpeg", "png", "webp", "png/P", "png/L", "rgb"]
 
+
 from xpra.net import compression
 RGB_COMP_OPTIONS  = ["Raw RGB"]
 if compression.get_enabled_compressors():
@@ -257,8 +271,10 @@ def encoding_help(encoding):
 
 def main():
     from xpra.platform import init, clean
+    from xpra.log import enable_color
     try:
         init("Loader", "Encoding Info")
+        enable_color()
         verbose = "-v" in sys.argv or "--verbose" in sys.argv
         if verbose:
             log.enable_debug()
@@ -266,7 +282,7 @@ def main():
         load_codecs()
         print("codecs and csc modules found:")
         #print("codec_status=%s" % codecs)
-        for name in ALL_CODECS:
+        for name in sorted(ALL_CODECS):
             mod = codecs.get(name, "")
             f = mod
             if mod and hasattr(mod, "__file__"):
@@ -278,24 +294,18 @@ def main():
             print("* %s : %s" % (name.ljust(20), f))
             if mod and verbose:
                 try:
-                    if name=="PIL":
-                        #special case for PIL which can be used for both encoding and decoding:
-                        from xpra.codecs.codec_constants import get_PIL_encodings, get_PIL_decodings
-                        e = get_PIL_encodings(mod)
-                        print("                         ENCODE: %s" % ", ".join(e))
-                        d = get_PIL_decodings(mod)
-                        print("                         DECODE: %s" % ", ".join(d))
-                    elif name.find("csc")>=0:
+                    if name.find("csc")>=0:
                         cs = list(mod.get_input_colorspaces())
                         for c in list(cs):
                             cs += list(mod.get_output_colorspaces(c))
-                        print("                         %s" % ", ".join(list(set(cs))))
+                        print("                         colorspaces: %s" % ", ".join(list(set(cs))))
                     elif name.find("enc")>=0 or name.find("dec")>=0:
                         encodings = mod.get_encodings()
-                        print("                         %s" % ", ".join(encodings))
+                        print("                         encodings: %s" % ", ".join(encodings))
                     try:
                         i = mod.get_info()
-                        print("                         %s" % i)
+                        for k,v in sorted(i.items()):
+                            print("                         %s = %s" % (k,v))
                     except:
                         pass
                 except Exception as e:
