@@ -5,24 +5,26 @@
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-import sys
+import sys, os
 from threading import Lock
 from xpra.log import Logger
 log = Logger("codec", "video")
 
-from xpra.codecs.loader import get_codec, get_codec_error, load_codecs
+from xpra.codecs.loader import get_codec, get_codec_error
+from xpra.util import csv, engs
 
 
+NVENC_OPTIONS = [("nvenc%s" % v) for v in os.environ.get("XPRA_NVENC_VERSIONS", "4,5").split(",")]
+
 #the codec loader uses the names...
 #but we need the module name to be able to probe without loading the codec:
 CODEC_TO_MODULE = {"vpx"        : ["vpx"],
                    "x264"       : ["enc_x264"],
                    "x265"       : ["enc_x265"],
-                   "nvenc"      : ["nvenc4", "nvenc3", "nvenc5"],
+                   "nvenc"      : NVENC_OPTIONS,
                    "swscale"    : ["csc_swscale"],
                    "cython"     : ["csc_cython"],
                    "opencl"     : ["csc_opencl"],
-                   "avcodec"    : ["dec_avcodec"],
                    "avcodec2"   : ["dec_avcodec2"]}
 
 def has_codec_module(module_name):
@@ -50,10 +52,10 @@ def try_import_modules(codec_names):
 ALL_VIDEO_ENCODER_OPTIONS = try_import_modules(["x264", "vpx", "x265", "nvenc"])
 ALL_CSC_MODULE_OPTIONS = try_import_modules(["swscale", "cython", "opencl"])
 NO_GFX_CSC_OPTIONS = [x for x in ALL_CSC_MODULE_OPTIONS if x not in ("opencl", )]
-ALL_VIDEO_DECODER_OPTIONS = try_import_modules(["avcodec", "avcodec2", "vpx"])
+ALL_VIDEO_DECODER_OPTIONS = try_import_modules(["avcodec2", "vpx"])
 
 PREFERRED_ENCODER_ORDER = ["nvenc", "x264", "vpx", "x265"]
-PREFERRED_DECODER_ORDER = ["avcodec", "avcodec2", "vpx"]
+PREFERRED_DECODER_ORDER = ["avcodec2", "vpx"]
 log("video_helper: ALL_VIDEO_ENCODER_OPTIONS=%s", ALL_VIDEO_ENCODER_OPTIONS)
 log("video_helper: ALL_CSC_MODULE_OPTIONS=%s", ALL_CSC_MODULE_OPTIONS)
 log("video_helper: NO_GFX_CSC_OPTIONS=%s", NO_GFX_CSC_OPTIONS)
@@ -65,7 +67,7 @@ log("video_helper: ALL_VIDEO_DECODER_OPTIONS=%s", ALL_VIDEO_DECODER_OPTIONS)
 
 def get_encoder_module_names(x):
     if x=="nvenc":
-        return ["nvenc4", "nvenc3", "nvenc5"]
+        return NVENC_OPTIONS
     elif x.find("enc")>=0:
         return [x]              #ie: "nvenc" or "enc_vpx"
     return ["enc_"+x]           #ie: "enc_x264"
@@ -145,7 +147,7 @@ class VideoHelper(object):
         self.csc_modules    = filt("csc modules"    , csc_modules,      ALL_CSC_MODULE_OPTIONS)
         self.video_decoders = filt("video decoders" , video_decoders,   ALL_VIDEO_DECODER_OPTIONS)
         log("VideoHelper.set_modules(%s, %s, %s) video encoders=%s, csc=%s, video decoders=%s",
-            video_encoders, csc_modules, video_decoders, self.video_encoders, self.csc_modules, self.video_decoders)
+            csv(video_encoders), csv(csc_modules), csv(video_decoders), csv(self.video_encoders), csv(self.csc_modules), csv(self.video_decoders))
 
     def cleanup(self):
         with self._lock:
@@ -213,7 +215,6 @@ class VideoHelper(object):
 
     def init(self):
         log("VideoHelper.init()")
-        load_codecs()
         with self._lock:
             #check again with lock held (in case of race):
             log("VideoHelper.init() initialized=%s", self._initialized)
@@ -246,41 +247,44 @@ class VideoHelper(object):
 
 
     def init_video_encoders_options(self):
-        log("init_video_encoders_options() will try video encoders: %s", self.video_encoders)
+        log("init_video_encoders_options()")
+        log(" will try video encoders: %s", csv(self.video_encoders))
         for x in self.video_encoders:
             try:
                 mods = get_encoder_module_names(x)
-                log("init_video_encoders_options() modules for %s: %s", x, mods)
+                log(" modules for %s: %s", x, csv(mods))
                 for mod in mods:
                     try:
                         self.init_video_encoder_option(mod)
                         break
-                    except:
-                        log.warn("init_video_encoders_options() cannot add %s encoder", mod, exc_info=True)
+                    except Exception as e:
+                        log(" init_video_encoder_option(%s) error", mod, exc_info=True)
+                        log.warn("Warning: cannot load %s video encoder:", mod)
+                        log.warn(" %s", e)
             except Exception as e:
-                log.warn("init_video_encoders_options() cannot add %s encoder: %s", x, e)
-        log("init_video_encoders_options() video encoder specs: %s", self._video_encoder_specs)
+                log.warn("Warning: cannot add %s encoder: %s", x, e)
+        log("found %i video encoder%s: %s", len(self._video_encoder_specs), engs(self._video_encoder_specs), csv(self._video_encoder_specs))
 
     def init_video_encoder_option(self, encoder_name):
         encoder_module = get_codec(encoder_name)
-        log("init_video_encoder_option(%s) module=%s", encoder_name, encoder_module)
+        log("init_video_encoder_option(%s)", encoder_name)
+        log(" module=%s", encoder_module)
         if not encoder_module:
-            log.warn("video encoder '%s' could not be loaded:", encoder_name)
-            log.warn(" %s", get_codec_error(encoder_name))
+            log(" video encoder '%s' could not be loaded:", encoder_name)
+            log(" %s", get_codec_error(encoder_name))
             return
         encoder_type = encoder_module.get_type()
         try:
             encoder_module.init_module()
             self._cleanup_modules.append(encoder_module)
         except Exception as e:
-            log("exception in %s module %s initialization %s: %s", encoder_type, encoder_module.__name__, encoder_module.init_module, e, exc_info=True)
-            log.warn("Warning: %s video encoder failed: %s", encoder_type, e)
-            raise e
+            log(" exception in %s module %s initialization %s: %s", encoder_type, encoder_module.__name__, encoder_module.init_module, e, exc_info=True)
+            raise
         encodings = encoder_module.get_encodings()
-        log("init_video_encoder_option(%s) %s encodings=%s", encoder_module, encoder_type, encodings)
+        log(" %s encodings=%s", encoder_type, csv(encodings))
         for encoding in encodings:
             colorspaces = encoder_module.get_input_colorspaces(encoding)
-            log("init_video_encoder_option(%s) %s input colorspaces(%s)=%s", encoder_module, encoder_type, encoding, colorspaces)
+            log(" %s input colorspaces for %s: %s", encoder_type, encoding, csv(colorspaces))
             for colorspace in colorspaces:
                 spec = encoder_module.get_spec(encoding, colorspace)
                 self.add_encoder_spec(encoding, colorspace, spec)
@@ -290,24 +294,27 @@ class VideoHelper(object):
 
 
     def init_csc_options(self):
-        log("init_csc_options() will try csc modules: %s", self.csc_modules)
+        log("init_csc_options()")
+        log(" will try csc modules: %s", csv(self.csc_modules))
         for x in self.csc_modules:
             try:
                 mod = get_csc_module_name(x)
                 self.init_csc_option(mod)
             except:
                 log.warn("init_csc_options() cannot add %s csc", x, exc_info=True)
-        log("init_csc_options() csc specs: %s", self._csc_encoder_specs)
+        log(" csc specs: %s", csv(self._csc_encoder_specs))
         for src_format, d in sorted(self._csc_encoder_specs.items()):
-            log("%s - %s options:", src_format, len(d))
+            log(" %s - %s options:", src_format, len(d))
             for dst_format, specs in sorted(d.items()):
-                log(" * %s via: %s", dst_format, sorted(list(specs), key=lambda spec: spec.codec_type))
+                log("  * %s via: %s", dst_format, csv(sorted(spec.codec_type for spec in specs)))
 
     def init_csc_option(self, csc_name):
         csc_module = get_codec(csc_name)
-        log("init_csc_option(%s) module=%s", csc_name, csc_module)
+        log("init_csc_option(%s)", csc_name)
+        log(" module=%s", csc_module)
         if csc_module is None:
-            log.warn("csc module %s could not be loaded: %s", csc_name, get_codec_error(csc_name))
+            log(" csc module %s could not be loaded:", csc_name)
+            log(" %s", get_codec_error(csc_name))
             return
         csc_type = csc_module.get_type()
         try:
@@ -315,12 +322,12 @@ class VideoHelper(object):
             self._cleanup_modules.append(csc_module)
         except Exception as e:
             log("exception in %s module initialization %s: %s", csc_type, csc_module.init_module, e, exc_info=True)
-            log.warn("cannot use %s module %s: %s", csc_type, csc_module, e)
+            log.warn("Warning: cannot use %s module %s: %s", csc_type, csc_module, e)
             return
         in_cscs = csc_module.get_input_colorspaces()
         for in_csc in in_cscs:
             out_cscs = csc_module.get_output_colorspaces(in_csc)
-            log("init_csc_option(..) %s.get_output_colorspaces(%s)=%s", csc_module.get_type(), in_csc, out_cscs)
+            log("%s output colorspaces for %s: %s", csc_module.get_type(), in_csc, csv(out_cscs))
             for out_csc in out_cscs:
                 spec = csc_module.get_spec(in_csc, out_csc)
                 self.add_csc_spec(in_csc, out_csc, spec)
@@ -330,37 +337,40 @@ class VideoHelper(object):
 
 
     def init_video_decoders_options(self):
-        log("init_video_decoders_options() will try video decoders: %s", self.video_decoders)
+        log("init_video_decoders_options()")
+        log(" will try video decoders: %s", csv(self.video_decoders))
         for x in self.video_decoders:
             try:
                 mod = get_decoder_module_name(x)
                 self.init_video_decoder_option(mod)
             except:
-                log.warn("init_video_decoders_options() cannot add %s decoder", x, exc_info=True)
-        log("init_video_decoders_options() video decoder specs: %s", self._video_decoder_specs)
+                log.warn("Warning: cannot add %s decoder", x, exc_info=True)
+        log("found %s video decoder%s: %s", len(self._video_decoder_specs), engs(self._video_decoder_specs), csv(self._video_decoder_specs))
 
     def init_video_decoder_option(self, decoder_name):
         decoder_module = get_codec(decoder_name)
-        log("init_video_decoder_option(%s) module=%s", decoder_name, decoder_module)
+        log("init_video_decoder_option(%s)", decoder_name)
+        log(" module=%s", decoder_module)
         if not decoder_module:
-            log.warn("video decoder %s could not be loaded: %s", decoder_name, get_codec_error(decoder_name))
+            log(" video decoder %s could not be loaded:", decoder_name)
+            log(" %s", get_codec_error(decoder_name))
             return
-        encoder_type = decoder_module.get_type()
+        decoder_type = decoder_module.get_type()
         try:
             decoder_module.init_module()
             self._cleanup_modules.append(decoder_module)
         except Exception as e:
-            log("exception in %s module initialization %s: %s", encoder_type, decoder_module.init_module, e, exc_info=True)
-            log.warn("cannot use %s module %s: %s", encoder_type, decoder_module, e, exc_info=True)
+            log("exception in %s module initialization %s: %s", decoder_type, decoder_module.init_module, e, exc_info=True)
+            log.warn("Warning: cannot use %s module %s: %s", decoder_type, decoder_module, e, exc_info=True)
             return
         encodings = decoder_module.get_encodings()
-        log("init_video_decoder_option(%s) %s encodings=%s", decoder_module, encoder_type, encodings)
+        log(" %s encodings=%s", decoder_type, csv(encodings))
         for encoding in encodings:
             colorspaces = decoder_module.get_input_colorspaces(encoding)
-            log("init_video_decoder_option(%s) %s input colorspaces for %s: %s", decoder_module, encoder_type, encoding, colorspaces)
+            log(" %s input colorspaces for %s: %s", decoder_type, encoding, csv(colorspaces))
             for colorspace in colorspaces:
                 output_colorspace = decoder_module.get_output_colorspace(encoding, colorspace)
-                log("init_video_decoder_option(%s) get_output_colorspace(%s, %s)=%s", decoder_name, encoding, colorspace, output_colorspace)
+                log(" %s output colorspace for %s/%s: %s", decoder_type, encoding, colorspace, output_colorspace)
                 try:
                     assert decoder_module.Decoder
                     self.add_decoder(encoding, colorspace, decoder_name, decoder_module)
@@ -416,18 +426,28 @@ def getVideoHelper():
 
 
 def main():
-    from xpra.codecs.loader import log as loader_log
-    if "-v" in sys.argv or "--verbose" in sys.argv:
-        loader_log.enable_debug()
-        log.enable_debug()
-    vh = getVideoHelper()
-    vh.set_modules(ALL_VIDEO_ENCODER_OPTIONS, ALL_CSC_MODULE_OPTIONS, ALL_VIDEO_DECODER_OPTIONS)
-    vh.init()
-    log.info("VideoHelper.get_info():")
-    info = vh.get_info()
-    for k in sorted(info.keys()):
-        v = info.get(k)
-        log.info("%s=%s", k, v)
+    from xpra.codecs.loader import log as loader_log, load_codecs
+    from xpra.log import enable_color
+    from xpra.platform import init, clean
+    try:
+        init("Video Helper")
+        enable_color()
+        if "-v" in sys.argv or "--verbose" in sys.argv:
+            loader_log.enable_debug()
+            log.enable_debug()
+        load_codecs()
+        vh = getVideoHelper()
+        vh.set_modules(ALL_VIDEO_ENCODER_OPTIONS, ALL_CSC_MODULE_OPTIONS, ALL_VIDEO_DECODER_OPTIONS)
+        vh.init()
+        log.info("VideoHelper.get_info():")
+        info = vh.get_info()
+        for k in sorted(info.keys()):
+            v = info.get(k)
+            if type(v) in (list, tuple):
+                v = csv(v)
+            log.info("%s=%s", k, v)
+    finally:
+        clean()
 
 
 if __name__ == "__main__":
