@@ -9,7 +9,7 @@ with python2, we try to import pygtk (gobject/gtk/gdk) before trying gobject int
 Once we have imported something, stick to that version from then on for all other imports.
 """
 
-import sys
+import sys, os
 
 __all__ = ["is_gtk3", "get_xid", "import_gobject", "import_gtk", "import_gdk", "import_pango", "import_glib", "import_pixbufloader"]
 
@@ -22,20 +22,42 @@ def is_gtk3():
     global _is_gtk3
     return  _is_gtk3
 
+def want_gtk3(v):
+    global _is_gtk3
+    assert (_is_gtk3 is None or _is_gtk3==v), "cannot set gtk3=%s, already set to %s" % (v, _is_gtk3)
+    _is_gtk3 = v
+
+
 def _try_import(import_method_gtk3, import_method_gtk2):
     global _is_gtk3
+    if "gi" in sys.modules:
+        _is_gtk3 = True
     if _is_gtk3 is False:
         return  import_method_gtk2()
     if _is_gtk3 is True:
         return  import_method_gtk3()
+    #python3 sets _is_gtk3 early
+    assert sys.version_info[0]<3
     try:
         imported = import_method_gtk2()
         _is_gtk3 = False
     except:
-        _is_gtk3 = True
         imported = import_method_gtk3()
+        _is_gtk3 = True
     return imported
 
+def try_import_GdkX11():
+    if os.name=="posix" and not sys.platform.startswith("darwin"):
+        #try to ensure that we can call get_xid() on Gdk windows later,
+        #this is a workaround for this GTK bug:
+        #https://bugzilla.gnome.org/show_bug.cgi?id=656314
+        try:
+            import gi
+            gi.require_version('GdkX11', '3.0')
+            from gi.repository import GdkX11            #@UnresolvedImport @UnusedImport
+        except:
+            pass
+
 
 def get_xid(window):
     if is_gtk3():
@@ -68,7 +90,10 @@ def import_gtk2():
     import gtk
     return gtk
 def import_gtk3():
+    import gi
+    gi.require_version('Gtk', '3.0')
     from gi.repository import Gtk                   #@UnresolvedImport
+    try_import_GdkX11()
     return Gtk
 def import_gtk():
     return  _try_import(import_gtk3, import_gtk2)
@@ -78,6 +103,7 @@ def import_gdk2():
     return gdk
 def import_gdk3():
     from gi.repository import Gdk                   #@UnresolvedImport
+    try_import_GdkX11()
     return Gdk
 def import_gdk():
     return  _try_import(import_gdk3, import_gdk2)
