@@ -1,6 +1,6 @@
 # coding=utf8
 # This file is part of Xpra.
-# Copyright (C) 2011-2014 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2011-2015 Antoine Martin <antoine@devloop.org.uk>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -21,6 +21,8 @@ from xpra.log import Logger
 log = Logger("gtk", "util")
 
 
+WIN32 = sys.platform.startswith("win")
+
 GTK_VERSION_INFO = {}
 def get_gtk_version_info():
     #update props given:
@@ -64,12 +66,45 @@ def get_gtk_version_info():
 
 
 if is_gtk3():
+    def is_realized(widget):
+        return widget.get_realized()
+
     def color_parse(*args):
         ok, v = gdk.Color.parse(*args)
         if not ok:
             return None
         return v
-    default_Cursor          = gdk.Cursor.new(gdk.CursorType.X_CURSOR)
+    def get_xwindow(w):
+        return w.get_xid()
+    def get_default_root_window():
+        return gdk.Screen.get_default().get_root_window()
+    def get_root_size():
+        if WIN32:
+            #FIXME: hopefully, we can remove this code once GTK3 on win32 is fixed?
+            #we do it the hard way because the root window geometry is invalid on win32:
+            #and even just querying it causes this warning:
+            #"GetClientRect failed: Invalid window handle."
+            display = gdk.Display.get_default()
+            n = display.get_n_screens()
+            w, h = 0, 0
+            for i in range(n):
+                screen = display.get_screen(i)
+                w += screen.get_width()
+                h += screen.get_height()
+        else:
+            #the easy way for platforms that work out of the box:
+            root = get_default_root_window()
+            w, h = root.get_geometry()[2:4]
+        if w<=0 or h<=0 or w>32768 or h>32768:
+            log.warn("Warning: Gdk returned invalid root window dimensions: %ix%i", w, h)
+            w, h = 1920, 1080
+            log.warn(" using %ix%i instead", w, h)
+        return w, h
+
+    keymap_get_for_display  = gdk.Keymap.get_for_display
+
+    def get_default_cursor():
+        return gdk.Cursor.new(gdk.CursorType.X_CURSOR)
     new_Cursor_for_display  = gdk.Cursor.new_for_display
     new_Cursor_from_pixbuf  = gdk.Cursor.new_from_pixbuf
     from gi.repository import GdkPixbuf     #@UnresolvedImport
@@ -113,6 +148,9 @@ if is_gtk3():
     BUTTONS_CLOSE   = gtk.ButtonsType.CLOSE
     DIALOG_DESTROY_WITH_PARENT  = 0
 
+    WINDOW_POPUP    = gtk.WindowType.POPUP
+    WINDOW_TOPLEVEL = gtk.WindowType.TOPLEVEL
+
     LSB_FIRST       = gdk.ByteOrder.LSB_FIRST
     MSB_FIRST       = gdk.ByteOrder.MSB_FIRST
     STATIC_GRAY     = gdk.VisualType.STATIC_GRAY
@@ -122,15 +160,46 @@ if is_gtk3():
     TRUE_COLOR      = gdk.VisualType.TRUE_COLOR
     DIRECT_COLOR    = gdk.VisualType.DIRECT_COLOR
 
-    SHIFT_MASK      = gdk.ModifierType.SHIFT_MASK
-    LOCK_MASK       = gdk.ModifierType.LOCK_MASK
-    META_MASK       = gdk.ModifierType.META_MASK
-    CONTROL_MASK    = gdk.ModifierType.CONTROL_MASK
-    MOD1_MASK       = gdk.ModifierType.MOD1_MASK
-    MOD2_MASK       = gdk.ModifierType.MOD2_MASK
-    MOD3_MASK       = gdk.ModifierType.MOD3_MASK
-    MOD4_MASK       = gdk.ModifierType.MOD4_MASK
-    MOD5_MASK       = gdk.ModifierType.MOD5_MASK
+    mt = gdk.ModifierType
+    SHIFT_MASK      = mt.SHIFT_MASK
+    LOCK_MASK       = mt.LOCK_MASK
+    META_MASK       = mt.META_MASK
+    CONTROL_MASK    = mt.CONTROL_MASK
+    MOD1_MASK       = mt.MOD1_MASK
+    MOD2_MASK       = mt.MOD2_MASK
+    MOD3_MASK       = mt.MOD3_MASK
+    MOD4_MASK       = mt.MOD4_MASK
+    MOD5_MASK       = mt.MOD5_MASK
+
+    BUTTON_MASK = {mt.BUTTON1_MASK : 1,
+                   mt.BUTTON2_MASK : 2,
+                   mt.BUTTON3_MASK : 3,
+                   mt.BUTTON4_MASK : 4,
+                   mt.BUTTON5_MASK : 5}
+    del mt
+
+    em = gdk.EventMask
+    WINDOW_EVENT_MASK = em.STRUCTURE_MASK | em.KEY_PRESS_MASK | em.KEY_RELEASE_MASK \
+            | em.POINTER_MOTION_MASK | em.BUTTON_PRESS_MASK | em.BUTTON_RELEASE_MASK \
+            | em.PROPERTY_CHANGE_MASK | em.SCROLL_MASK
+    del em
+
+    WINDOW_NAME_TO_HINT = {
+                "NORMAL"        : gdk.WindowTypeHint.NORMAL,
+                "DIALOG"        : gdk.WindowTypeHint.DIALOG,
+                "MENU"          : gdk.WindowTypeHint.MENU,
+                "TOOLBAR"       : gdk.WindowTypeHint.TOOLBAR,
+                "SPLASH"        : gdk.WindowTypeHint.SPLASHSCREEN,
+                "UTILITY"       : gdk.WindowTypeHint.UTILITY,
+                "DOCK"          : gdk.WindowTypeHint.DOCK,
+                "DESKTOP"       : gdk.WindowTypeHint.DESKTOP,
+                "DROPDOWN_MENU" : gdk.WindowTypeHint.DROPDOWN_MENU,
+                "POPUP_MENU"    : gdk.WindowTypeHint.POPUP_MENU,
+                "TOOLTIP"       : gdk.WindowTypeHint.TOOLTIP,
+                "NOTIFICATION"  : gdk.WindowTypeHint.NOTIFICATION,
+                "COMBO"         : gdk.WindowTypeHint.COMBO,
+                "DND"           : gdk.WindowTypeHint.DND
+                }
 
     from gi.repository.Gtk import Clipboard     #@UnresolvedImport
     CLIPBOARD_SELECTION = {}
@@ -164,12 +233,27 @@ if is_gtk3():
             gdk.threads_leave()
 
 else:
+    def get_xwindow(w):
+        return w.xid
     #gtk2:
     if gtk.gtk_version<(2,18):
         raise Exception("your version of PyGTK is too old: %s" % str(gtk.pygtk_version))
 
+    if not hasattr(gtk.Widget, "get_realized"):
+        def is_realized(widget):
+            return widget.flags() & gtk.REALIZED
+    else:
+        def is_realized(widget):
+            return widget.get_realized()
+
+    get_default_root_window = gdk.get_default_root_window
+    def get_root_size():
+        return get_default_root_window().get_size()
+    keymap_get_for_display  = gdk.keymap_get_for_display
+
+    def get_default_cursor():
+        return gdk.Cursor(gdk.X_CURSOR)
     color_parse             = gdk.color_parse
-    default_Cursor          = gdk.Cursor(gdk.X_CURSOR)
     new_Cursor_for_display  = gdk.Cursor
     new_Cursor_from_pixbuf  = gdk.Cursor
     image_new_from_pixbuf   = gtk.image_new_from_pixbuf
@@ -185,8 +269,8 @@ else:
         return gdk.CairoContext(cairo_context)
     def cairo_set_source_pixbuf(cr, pixbuf, x, y):
         cr.set_source_pixbuf(pixbuf, x, y)
-    COLORSPACE_RGB          = gtk.gdk.COLORSPACE_RGB
-    INTERP_HYPER    = gtk.gdk.INTERP_HYPER
+    COLORSPACE_RGB          = gdk.COLORSPACE_RGB
+    INTERP_HYPER    = gdk.INTERP_HYPER
     INTERP_BILINEAR = gdk.INTERP_BILINEAR
     RELIEF_NONE     = gtk.RELIEF_NONE
     RELIEF_NORMAL   = gtk.RELIEF_NORMAL
@@ -211,24 +295,54 @@ else:
     BUTTONS_CLOSE   = gtk.BUTTONS_CLOSE
     DIALOG_DESTROY_WITH_PARENT = gtk.DIALOG_DESTROY_WITH_PARENT
 
-    LSB_FIRST       = gtk.gdk.LSB_FIRST
-    MSB_FIRST       = gtk.gdk.MSB_FIRST
-    STATIC_GRAY     = gtk.gdk.VISUAL_STATIC_GRAY
-    GRAYSCALE       = gtk.gdk.VISUAL_GRAYSCALE
-    STATIC_COLOR    = gtk.gdk.VISUAL_STATIC_COLOR
-    PSEUDO_COLOR    = gtk.gdk.VISUAL_PSEUDO_COLOR
-    TRUE_COLOR      = gtk.gdk.VISUAL_TRUE_COLOR
-    DIRECT_COLOR    = gtk.gdk.VISUAL_DIRECT_COLOR
+    WINDOW_POPUP    = gtk.WINDOW_POPUP
+    WINDOW_TOPLEVEL = gtk.WINDOW_TOPLEVEL
 
-    SHIFT_MASK      = gtk.gdk.SHIFT_MASK
-    LOCK_MASK       = gtk.gdk.LOCK_MASK
+    LSB_FIRST       = gdk.LSB_FIRST
+    MSB_FIRST       = gdk.MSB_FIRST
+    STATIC_GRAY     = gdk.VISUAL_STATIC_GRAY
+    GRAYSCALE       = gdk.VISUAL_GRAYSCALE
+    STATIC_COLOR    = gdk.VISUAL_STATIC_COLOR
+    PSEUDO_COLOR    = gdk.VISUAL_PSEUDO_COLOR
+    TRUE_COLOR      = gdk.VISUAL_TRUE_COLOR
+    DIRECT_COLOR    = gdk.VISUAL_DIRECT_COLOR
+
+    SHIFT_MASK      = gdk.SHIFT_MASK
+    LOCK_MASK       = gdk.LOCK_MASK
     META_MASK       = gdk.META_MASK
-    CONTROL_MASK    = gtk.gdk.CONTROL_MASK
-    MOD1_MASK       = gtk.gdk.MOD1_MASK
-    MOD2_MASK       = gtk.gdk.MOD2_MASK
-    MOD3_MASK       = gtk.gdk.MOD3_MASK
-    MOD4_MASK       = gtk.gdk.MOD4_MASK
-    MOD5_MASK       = gtk.gdk.MOD5_MASK
+    CONTROL_MASK    = gdk.CONTROL_MASK
+    MOD1_MASK       = gdk.MOD1_MASK
+    MOD2_MASK       = gdk.MOD2_MASK
+    MOD3_MASK       = gdk.MOD3_MASK
+    MOD4_MASK       = gdk.MOD4_MASK
+    MOD5_MASK       = gdk.MOD5_MASK
+
+    BUTTON_MASK = {gdk.BUTTON1_MASK : 1,
+                   gdk.BUTTON2_MASK : 2,
+                   gdk.BUTTON3_MASK : 3,
+                   gdk.BUTTON4_MASK : 4,
+                   gdk.BUTTON5_MASK : 5}
+
+    WINDOW_EVENT_MASK = gdk.STRUCTURE_MASK | gdk.KEY_PRESS_MASK | gdk.KEY_RELEASE_MASK \
+            | gdk.POINTER_MOTION_MASK | gdk.BUTTON_PRESS_MASK | gdk.BUTTON_RELEASE_MASK \
+            | gdk.PROPERTY_CHANGE_MASK
+
+    WINDOW_NAME_TO_HINT = {
+                "NORMAL"        : gdk.WINDOW_TYPE_HINT_NORMAL,
+                "DIALOG"        : gdk.WINDOW_TYPE_HINT_DIALOG,
+                "MENU"          : gdk.WINDOW_TYPE_HINT_MENU,
+                "TOOLBAR"       : gdk.WINDOW_TYPE_HINT_TOOLBAR,
+                "SPLASH"        : gdk.WINDOW_TYPE_HINT_SPLASHSCREEN,
+                "UTILITY"       : gdk.WINDOW_TYPE_HINT_UTILITY,
+                "DOCK"          : gdk.WINDOW_TYPE_HINT_DOCK,
+                "DESKTOP"       : gdk.WINDOW_TYPE_HINT_DESKTOP,
+                "DROPDOWN_MENU" : gdk.WINDOW_TYPE_HINT_DROPDOWN_MENU,
+                "POPUP_MENU"    : gdk.WINDOW_TYPE_HINT_POPUP_MENU,
+                "TOOLTIP"       : gdk.WINDOW_TYPE_HINT_TOOLTIP,
+                "NOTIFICATION"  : gdk.WINDOW_TYPE_HINT_NOTIFICATION,
+                "COMBO"         : gdk.WINDOW_TYPE_HINT_COMBO,
+                "DND"           : gdk.WINDOW_TYPE_HINT_DND
+                }
 
     OptionMenu  = gtk.OptionMenu
 
@@ -249,6 +363,8 @@ else:
 def get_display_info():
     display = display_get_default()
     info = {
+            "root"                  : get_default_root_window().get_geometry(),
+            "root-size"             : get_root_size(),
             "screens"               : display.get_n_screens(),
             "name"                  : display.get_name(),
             "pointer"               : display.get_pointer()[1:3],
@@ -265,6 +381,11 @@ def get_display_info():
     for i in range(display.get_n_screens()):
         screen = display.get_screen(i)
         sk = "screen[%s]" % i
+        try:
+            w = screen.get_root_window()
+            info[sk+".root"] = w.get_geometry()
+        except:
+            pass
         info[sk+".name"] = screen.make_display_name()
         for x in ("width", "height", "width_mm", "height_mm", "resolution", "primary_monitor"):
             fn = getattr(screen, "get_"+x)
@@ -376,11 +497,10 @@ def get_display_info():
             for i, d in enumerate(devices):
                 info[dk+"[%s]" % i] = d.get_name()
     else:
-        dk = "devices"
         devices = display.list_devices()
-        info[dk] = len(devices)
+        info["devices"] = len(devices)
         for i, d in enumerate(devices):
-            info[dk+"[%s]" % i] = d.get_name()
+            info["device[%s]" % i] = d.get_name()
     return info
 
 
@@ -618,25 +738,3 @@ def choose_file(parent_window, title, action, action_button, callback, file_filt
         return
     filename = filenames[0]
     callback(filename)
-
-
-def main():
-    from xpra.util import nonl, pver
-    def print_dict(d, vformat=pver):
-        for k in sorted(d.keys()):
-            v = d[k]
-            print("* %s : %s" % (str(k).replace(".version", "").ljust(12), nonl(vformat(v))))
-    from xpra.platform import init, clean
-    try:
-        init("GTK-Version-Info", "GTK Version Info")
-        print("GTK Version:")
-        print_dict(get_gtk_version_info())
-        print("Display:")
-        print_dict(get_display_info(), vformat=str)
-    finally:
-        clean()
-
-
-if __name__ == "__main__":
-    main()
-    sys.exit(0)
