@@ -56,13 +56,16 @@ def get_gtk_keymap(ignore_keys=[None, "VoidSymbol", "0xffffff"]):
 def main():
     import sys
     from xpra.platform import init, clean
+    from xpra.log import enable_color
     try:
         init("Keymap-Tool", "Keymap Information Tool")
-        if "-v" in sys.argv:
+        enable_color()
+        if "-v" in sys.argv or "--verbose" in sys.argv:
             log.enable_debug()
         gtk_keymap = get_gtk_keymap()
+        sizes = [16, 28, 8, 8, 8]
         def pkey(*entries):
-            print(("".join([str(x).ljust(18) for x in entries])).strip())
+            print(("".join([str(x).ljust(sizes[i]) for i,x in enumerate(entries)])).strip())
         pkey("keyval", "name", "keycode", "group", "level")
         for x in gtk_keymap:
             pkey(*x)
