@@ -12,11 +12,15 @@ import socket
 
 from xpra.log import Logger
 log = Logger("network", "protocol")
+from xpra.net import ConnectionClosedException
 
 #on some platforms (ie: OpenBSD), reading and writing from sockets
 #raises an IOError but we should continue if the error code is EINTR
 #this wrapper takes care of it.
-CONTINUE = {errno.EINTR : "EINTR"}
+#EWOULDBLOCK can also be hit with the proxy server when we handover the socket
+CONTINUE = {errno.EINTR         : "EINTR",
+            errno.EWOULDBLOCK   : "EWOULDBLOCK"
+            }
 ABORT = {
     errno.ECONNRESET    : "ECONNRESET",
     errno.EPIPE         : "EPIPE"}
@@ -43,6 +47,9 @@ if sys.platform.startswith("win"):
     WSAECONNRESET   = 10054
     WSAENOTCONN     = 10057
     WSAESHUTDOWN    = 10058
+    WSAETIMEDOUT    = 10060
+    WSAEHOSTUNREACH = 10065
+    WSAEDISCON      = 10101
     ABORT.update({
         WSAECONNABORTED     : "WSAECONNABORTED",
         WSAECONNRESET       : "WSAECONNRESET",
@@ -50,6 +57,9 @@ if sys.platform.startswith("win"):
         WSAENETUNREACH      : "WSAENETUNREACH",
         WSAENOTCONN         : "WSAENOTCONN",
         WSAESHUTDOWN        : "WSAESHUTDOWN",
+        WSAETIMEDOUT        : "WSAETIMEDOUT",
+        WSAEHOSTUNREACH     : "WSAEHOSTUNREACH",
+        WSAEDISCON          : "WSAEDISCON",
         })
     #on win32, we want to wait just a little while,
     #to prevent servers spinning wildly on non-blocking sockets:
@@ -74,6 +84,9 @@ def untilConcludes(is_active_cb, f, *a, **kw):
         except socket.timeout as e:
             log("untilConcludes(%s, %s, %s, %s) %s", is_active_cb, f, a, kw, e)
             continue
+        except TypeError:
+            log.warn("untilConcludes error calling %s with %s", f, a)
+            raise
         except (IOError, OSError) as e:
             code = e.args[0]
             can_continue = CONTINUE.get(code)
@@ -83,9 +96,19 @@ def untilConcludes(is_active_cb, f, *a, **kw):
                 if wait<continue_wait:
                     wait += 1
                 continue
-            log("untilConcludes(%s, %s, %s, %s) %s / %s (raised)", is_active_cb, f, a, kw, ABORT.get(code, code), e)
+            abort = ABORT.get(code, code)
+            if abort is not None:
+                raise ConnectionClosedException(e)
+            log("untilConcludes(%s, %s, %s, %s) %s / %s (raised)", is_active_cb, f, a, kw, abort, e)
             raise
 
+def pretty_socket(s):
+    try:
+        assert len(s)==2
+        return "%s:%s" % (s[0], s[1])
+    except:
+        return str(s)
+
 
 class Connection(object):
     def __init__(self, target, info):
@@ -168,22 +191,25 @@ class TwoFileConnection(Connection):
         return self._write(self._oswrite, self._write_fd, buf)
 
     def close(self):
+        log("%s.close() close callback=%s, readable=%s, writeable=%s", self, self._close_cb, self._readable, self._writeable)
         Connection.close(self)
-        try:
-            self._readable.close()
-        except:
-            pass
-        try:
-            self._writeable.close()
-        except:
-            pass
         cc = self._close_cb
         if cc:
             self._close_cb = None
+            log("%s.close() calling %s", self, cc)
             cc()
+        try:
+            self._readable.close()
+        except Exception as e:
+            log("%s.close() %s", self._readable, e)
+        try:
+            self._writeable.close()
+        except:
+            log("%s.close() %s", self._writeable, e)
+        log("%s.close() done", self)
 
     def __repr__(self):
-        return "TwoFileConnection(%s)" % str(self.target)
+        return "Pipe(%s)" % str(self.target)
 
     def get_info(self):
         d = Connection.get_info(self)
@@ -212,13 +238,16 @@ class SocketConnection(Connection):
         return self._write(self._socket.send, buf)
 
     def close(self):
+        log("%s.close() for socket=%s", self, self._socket)
         Connection.close(self)
+        self._socket.settimeout(0)
         self._socket.close()
+        log("%s.close() done", self)
 
     def __repr__(self):
         if self.remote:
-            return "SocketConnection(%s - %s)" % (self.local, self.remote)
-        return "SocketConnection(%s)" % self.local
+            return "%s socket: %s <- %s" % (self.info, pretty_socket(self.local), pretty_socket(self.remote))
+        return "%s socket:%s" % (self.info, pretty_socket(self.local))
 
     def get_info(self):
         d = Connection.get_info(self)
@@ -236,3 +265,9 @@ class SocketConnection(Connection):
         except:
             log.warn("failed to get socket information", exc_info=True)
         return d
+
+def set_socket_timeout(conn, timeout=None):
+    #FIXME: this is ugly, but less intrusive than the alternative?
+    log("set_socket_timeout(%s, %s)", conn, timeout)
+    if isinstance(conn, SocketConnection):
+        conn._socket.settimeout(timeout)
