@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 # This file is part of Xpra.
-# Copyright (C) 2011-2014 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2011-2015 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008, 2009, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -11,49 +11,49 @@ import zlib
 from xpra.log import Logger
 log = Logger("network", "protocol")
 from xpra.net.header import LZ4_FLAG, ZLIB_FLAG, LZO_FLAG
-from xpra.os_util import memoryview_to_bytes
+from xpra.os_util import _memoryview
 
 
+python_lz4_version = None
 lz4_version = None
 try:
     import lz4
-    from lz4 import LZ4_compress, LZ4_uncompress        #@UnresolvedImport
+    try:
+        from lz4 import VERSION as lz4_VERSION                          #@UnresolvedImport
+    except:
+        #older versions (0.7 and older):
+        import pkg_resources
+        lz4_VERSION = pkg_resources.get_distribution("lz4").version
+    from lz4 import LZ4_compress, compressHC, LZ4_uncompress        #@UnresolvedImport
     has_lz4 = True
-    def lz4_compress(packet, level):
-        return level | LZ4_FLAG, LZ4_compress(packet)
-    #try to figure out the version number:
-    if hasattr(lz4, "VERSION"):
-        lz4_version = str(lz4.VERSION)
+    if hasattr(lz4, "LZ4_compress_fast"):
+        from lz4 import LZ4_compress_fast                           #@UnresolvedImport
+        def lz4_compress(packet, level):
+            if level>=9:
+                return level | LZ4_FLAG, compressHC(packet)
+            #clamp it: 0->17, 1->12, 2->7, 3->2, >=4->1
+            accel = max(1, 17-level*5)
+            return level | LZ4_FLAG, LZ4_compress_fast(packet, accel)
     else:
-        #hack it..
-        try:
-           import pkg_resources
-           lz4_version = pkg_resources.get_distribution("lz4").version
-        except:
-           lz4_version = None
-    if not lz4_version and hasattr(lz4, "__file__"):
-        import os.path
-        f = lz4.__file__
-        #ie: /usr/lib/python2.7/site-packages/lz4-0.7.0-py2.7-linux-x86_64.egg/lz4.so
-        for x in f.split(os.path.sep):
-            #ie: lz4-0.7.0-py2.7-linux-x86_64.egg
-            if x.startswith("lz4-") and x.find("-py"):
-                tmp = x.split("-")[1]
-                #ie: "0.7.0"
-                tmpv = []
-                #stop if we hit non numeric chars
-                try:
-                    for x in tmp.split("."):
-                        tmpv.append(int(x))
-                except:
-                    pass
-                #we want at least two numbers first:
-                if len(tmpv)>=2:
-                    #ie: (0, 7, 0)
-                    lz4_version = ".".join([str(x) for x in tmpv])
-    assert lz4_version>="0.7", "versions older than 0.7.0 are vulnerable and should not be used, see CVE-2014-4715"
+        #v0.7.0 and earlier
+        def lz4_compress(packet, level):
+            if level>=9:
+                return level | LZ4_FLAG, compressHC(packet)
+            return level | LZ4_FLAG, LZ4_compress(packet)
+    #try to figure out the version number:
+    python_lz4_version = lz4_VERSION
+    assert python_lz4_version>="0.7", "python-lz4 version %s is older than 0.7.0 are vulnerable and should not be used, see CVE-2014-4715" % lz4_version
+    #now try to check the underlying "liblz4" version
+    #which is only available with python-lz4 0.8.0 onwards:
     if hasattr(lz4, "LZ4_VERSION"):
-        lz4_version += "."+str(lz4.LZ4_VERSION)
+        try:
+            from distutils.version import LooseVersion
+        except ImportError:
+            pass
+        else:
+            #last known security issue:
+            assert LooseVersion(lz4.LZ4_VERSION)>=LooseVersion("r119"), "lz4 version %s is vulnerable and should not be used, see CVE-2014-4715" % lz4.LZ4_VERSION
+        lz4_version = lz4.LZ4_VERSION
 except Exception as e:
     log("lz4 not found: %s", e)
     LZ4_uncompress = None
@@ -61,11 +61,12 @@ except Exception as e:
     def lz4_compress(packet, level):
         raise Exception("lz4 is not supported!")
 
-
+python_lzo_version = None
 lzo_version = None
 try:
     import lzo
     has_lzo = True
+    python_lzo_version = lzo.__version__
     lzo_version = lzo.LZO_VERSION_STRING
     def lzo_compress(packet, level):
         return level | LZO_FLAG, lzo.compress(packet)
@@ -123,8 +124,12 @@ def get_compression_caps():
            }
     if lzo_version:
         caps["lzo.version"] = lzo_version
+    if python_lzo_version:
+        caps["python-lzo.version"] = python_lzo_version
     if lz4_version:
         caps["lz4.version"] = lz4_version
+    if python_lz4_version:
+        caps["python-lz4.version"] = python_lz4_version
     return caps
 
 def get_enabled_compressors(order=ALL_COMPRESSORS):
@@ -166,7 +171,7 @@ class Compressed(object):
     def __len__(self):
         return len(self.data)
     def __repr__(self):
-        return  "Compressed(%s: %s bytes)" % (self.datatype, len(self.data))
+        return  "Compressed(%s: %i bytes)" % (self.datatype, len(self.data))
 
 
 class LevelCompressed(Compressed):
@@ -175,7 +180,7 @@ class LevelCompressed(Compressed):
         self.level = level
         self.algorithm = algo
     def __repr__(self):
-        return  "LevelCompressed(%s: %s bytes as %s/%s)" % (self.datatype, len(self.data), self.algorithm, self.level)
+        return  "LevelCompressed(%s: %i bytes as %s/%i)" % (self.datatype, len(self.data), self.algorithm, self.level)
 
 
 class Uncompressed(object):
@@ -185,24 +190,25 @@ class Uncompressed(object):
     def __len__(self):
         return len(self.data)
     def __repr__(self):
-        return  "Uncompressed(%s: %s bytes)" % (self.datatype, len(self.data))
+        return  "Uncompressed(%s: %i bytes)" % (self.datatype, len(self.data))
     def compress(self):
         raise Exception("compress() not defined on %s" % self)
 
 def compressed_wrapper(datatype, data, level=5, zlib=False, lz4=False, lzo=False, can_inline=True):
-    bdata = memoryview_to_bytes(data)
+    if _memoryview and isinstance(data, _memoryview):
+        data = data.tobytes()
     if lz4:
         assert use_lz4, "cannot use lz4"
         algo = "lz4"
-        cl, cdata = lz4_compress(bdata, level)
+        cl, cdata = lz4_compress(data, level)
     elif lzo:
         assert use_lzo, "cannot use lzo"
         algo = "lzo"
-        cl, cdata = lzo_compress(bdata, level)
+        cl, cdata = lzo_compress(data, level)
     else:
         assert use_zlib, "cannot use zlib"
         algo = "zlib"
-        cl, cdata = zcompress(bdata, level)
+        cl, cdata = zcompress(data, level)
     return LevelCompressed(datatype, cdata, cl, algo, can_inline=can_inline)
 
 
@@ -218,6 +224,8 @@ def get_compression_type(level):
     else:
         return "zlib"
 
+import struct
+LZ4_HEADER = struct.Struct('<L')
 def decompress(data, level):
     #log.info("decompress(%s bytes, %s) type=%s", len(data), get_compression_type(level))
     if level & LZ4_FLAG:
@@ -225,6 +233,10 @@ def decompress(data, level):
             raise InvalidCompressionException("lz4 is not available")
         if not use_lz4:
             raise InvalidCompressionException("lz4 is not enabled")
+        size = LZ4_HEADER.unpack_from(data[:4])[0]
+        #TODO: it would be better to use the max_size we have in protocol,
+        #but this hardcoded value will do for now
+        assert size<=(256*1024*1024)
         return LZ4_uncompress(data)
     elif level & LZO_FLAG:
         if not has_lzo:
