@@ -6,11 +6,10 @@
 # later version. See the file COPYING for details.
 
 import os
-import sys
 
 from xpra.log import Logger
 log = Logger("network", "protocol")
-from xpra.net.header import FLAGS_RENCODE, FLAGS_YAML   #, FLAGS_BENCODE
+from xpra.net.header import FLAGS_RENCODE, FLAGS_YAML, FLAGS_BENCODE
 
 
 rencode_dumps, rencode_loads, rencode_version = None, None, None
@@ -35,15 +34,13 @@ log("packet encoding: has_rencode=%s, use_rencode=%s, version=%s", has_rencode,
 
 
 bencode, bdecode, bencode_version = None, None, None
-if sys.version_info[0]<3:
-    #bencode needs porting to Python3..
+try:
     try:
-        try:
-            from xpra.net.bencode import bencode, bdecode, __version__ as bencode_version
-        except ImportError as e:
-            log.warn("bencode import error: %s", e, exc_info=True)
-    except Exception as e:
-        log.error("error loading bencoder", exc_info=True)
+        from xpra.net.bencode import bencode, bdecode, __version__ as bencode_version
+    except ImportError as e:
+        log.warn("bencode import error: %s", e, exc_info=True)
+except Exception as e:
+    log.error("error loading bencoder", exc_info=True)
 has_bencode = bencode is not None and bdecode is not None
 use_bencode = has_bencode and os.environ.get("XPRA_USE_BENCODER", "1")=="1"
 log("packet encoding: has_bencode=%s, use_bencode=%s, version=%s", has_bencode, use_bencode, bencode_version)
@@ -64,7 +61,7 @@ log("packet encoding: has_yaml=%s, use_yaml=%s, version=%s", has_yaml, use_yaml,
 
 
 def do_bencode(data):
-    return bencode(data), 0
+    return bencode(data), FLAGS_BENCODE
 
 def do_rencode(data):
     return  rencode_dumps(data), FLAGS_RENCODE
@@ -164,8 +161,6 @@ def decode(data, protocol_flags):
             raise InvalidPacketEncodingException("bencode is not available")
         if not use_bencode:
             raise InvalidPacketEncodingException("bencode is disabled")
-        #if sys.version>='3':
-        #    data = data.decode("latin1")
         packet, l = bdecode(data)
         assert l==len(data)
         return packet
