@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2011-2014 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2011-2015 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008, 2009, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -11,15 +11,17 @@
 import sys
 from socket import error as socket_error
 import os
-import threading
 import binascii
-from threading import Lock
+from threading import Lock, Event
 
 
 from xpra.log import Logger
 log = Logger("network", "protocol")
+cryptolog = Logger("network", "crypto")
+
 from xpra.os_util import Queue, strtobytes
-from xpra.util import repr_ellipsized, updict
+from xpra.util import repr_ellipsized, updict, csv
+from xpra.net import ConnectionClosedException
 from xpra.net.bytestreams import ABORT
 from xpra.net import compression
 from xpra.net import packet_encoding
@@ -27,7 +29,7 @@ from xpra.net.compression import get_compression_caps, decompress, sanity_checks
         InvalidCompressionException, Compressed, LevelCompressed, Uncompressed
 from xpra.net.packet_encoding import get_packet_encoding_caps, decode, sanity_checks as packet_encoding_sanity_checks, InvalidPacketEncodingException
 from xpra.net.header import unpack_header, pack_header, FLAGS_CIPHER, FLAGS_NOHEADER
-from xpra.net.crypto import get_crypto_caps, get_cipher
+from xpra.net.crypto import get_crypto_caps, get_cipher, pad, INITIAL_PADDING
 
 
 #stupid python version breakage:
@@ -38,18 +40,32 @@ if sys.version > '3':
     JOIN_TYPES = (bytes, )
 
 
+FAULT_RATE = int(os.environ.get("XPRA_PROTOCOL_FAULT_INJECTION_RATE", "0"))
+if FAULT_RATE>0:
+    _counter = 0
+    def INJECT_FAULT(p):
+        global _counter
+        _counter += 1
+        if (_counter % FAULT_RATE)==0:
+            log.warn("injecting fault in %s", p)
+            p.raw_write("Protocol JUNK! added by fault injection code")
+else:
+    def INJECT_FAULT(p):
+        pass
+
 USE_ALIASES = os.environ.get("XPRA_USE_ALIASES", "1")=="1"
 
 READ_BUFFER_SIZE = int(os.environ.get("XPRA_READ_BUFFER_SIZE", 65536))
 #merge header and packet if packet is smaller than:
 PACKET_JOIN_SIZE = int(os.environ.get("XPRA_PACKET_JOIN_SIZE", READ_BUFFER_SIZE))
-LARGE_PACKET_SIZE = 4096
+LARGE_PACKET_SIZE = int(os.environ.get("XPRA_LARGE_PACKET_SIZE", 4096))
 #inline compressed data in packet if smaller than:
 INLINE_SIZE = int(os.environ.get("XPRA_INLINE_SIZE", 32768))
 FAKE_JITTER = int(os.environ.get("XPRA_FAKE_JITTER", "0"))
+MIN_COMPRESS_SIZE = int(os.environ.get("XPRA_MIN_COMPRESS_SIZE", 378))
 
 
-def get_network_caps(legacy=True):
+def get_network_caps():
     try:
         from xpra.net.mmap_pipe import can_use_mmap
         mmap = can_use_mmap()
@@ -61,12 +77,6 @@ def get_network_caps(legacy=True):
                 "encoders"              : packet_encoding.get_enabled_encoders(),
                 "mmap"                  : mmap,
                }
-    if legacy:
-        #for backwards compatibility only:
-        caps.update({
-                "raw_packets"           : True,
-                "chunked_compression"   : True
-                })
     caps.update(get_crypto_caps())
     caps.update(get_compression_caps())
     caps.update(get_packet_encoding_caps())
@@ -78,10 +88,6 @@ def sanity_checks():
     packet_encoding_sanity_checks()
 
 
-class ConnectionClosedException(Exception):
-    pass
-
-
 class Protocol(object):
     CONNECTION_LOST = "connection-lost"
     GIBBERISH = "gibberish"
@@ -104,7 +110,7 @@ class Protocol(object):
             self._process_packet_cb = process_packet_cb
         self._write_queue = Queue(1)
         self._read_queue = Queue(20)
-        self._read_queue_put = self._read_queue.put
+        self._read_queue_put = self.read_queue_put
         # Invariant: if .source is None, then _source_has_more == False
         self._get_packet_cb = get_packet_cb
         #counters:
@@ -117,7 +123,7 @@ class Protocol(object):
         #initial value which may get increased by client/server after handshake:
         self.max_packet_size = 256*1024
         self.abs_max_packet_size = 256*1024*1024
-        self.large_packets = ["hello"]
+        self.large_packets = ["hello", "window-metadata"]
         self.send_aliases = {}
         self.receive_aliases = {}
         self._log_stats = None          #None here means auto-detect
@@ -130,20 +136,22 @@ class Protocol(object):
         self.cipher_in = None
         self.cipher_in_name = None
         self.cipher_in_block_size = 0
+        self.cipher_in_padding = INITIAL_PADDING
         self.cipher_out = None
         self.cipher_out_name = None
         self.cipher_out_block_size = 0
+        self.cipher_out_padding = INITIAL_PADDING
         self._write_lock = Lock()
-        from xpra.daemon_thread import make_daemon_thread
-        self._write_thread = make_daemon_thread(self._write_thread_loop, "write")
-        self._read_thread = make_daemon_thread(self._read_thread_loop, "read")
-        self._read_parser_thread = make_daemon_thread(self._read_parse_thread_loop, "parse")
-        self._write_format_thread = make_daemon_thread(self._write_format_thread_loop, "format")
-        self._source_has_more = threading.Event()
+        from xpra.make_thread import make_thread
+        self._write_thread = make_thread(self._write_thread_loop, "write", daemon=True)
+        self._read_thread = make_thread(self._read_thread_loop, "read", daemon=True)
+        self._read_parser_thread = None         #started when needed
+        self._write_format_thread = None        #started when needed
+        self._source_has_more = Event()
 
     STATE_FIELDS = ("max_packet_size", "large_packets", "send_aliases", "receive_aliases",
-                    "cipher_in", "cipher_in_name", "cipher_in_block_size",
-                    "cipher_out", "cipher_out_name", "cipher_out_block_size",
+                    "cipher_in", "cipher_in_name", "cipher_in_block_size", "cipher_in_padding",
+                    "cipher_out", "cipher_out_name", "cipher_out_block_size", "cipher_out_padding",
                     "compression_level", "encoder", "compressor")
     def save_state(self):
         state = {}
@@ -162,10 +170,11 @@ class Protocol(object):
 
     def wait_for_io_threads_exit(self, timeout=None):
         for t in (self._read_thread, self._write_thread):
-            t.join(timeout)
+            if t and t.isAlive():
+                t.join(timeout)
         exited = True
         for t in (self._read_thread, self._write_thread):
-            if t.isAlive():
+            if t and t.isAlive():
                 log.warn("%s thread of %s has not yet exited (timeout=%s)", t.name, self._conn, timeout)
                 exited = False
                 break
@@ -175,19 +184,21 @@ class Protocol(object):
         self._get_packet_cb = get_packet_cb
 
 
-    def set_cipher_in(self, ciphername, iv, password, key_salt, iterations):
+    def set_cipher_in(self, ciphername, iv, password, key_salt, iterations, padding):
         if self.cipher_in_name!=ciphername:
-            log.info("receiving data using %s encryption", ciphername)
+            cryptolog.info("receiving data using %s encryption", ciphername)
             self.cipher_in_name = ciphername
-        log("set_cipher_in%s", (ciphername, iv, password, key_salt, iterations))
+        cryptolog("set_cipher_in%s", (ciphername, iv, password, key_salt, iterations))
         self.cipher_in, self.cipher_in_block_size = get_cipher(ciphername, iv, password, key_salt, iterations)
+        self.cipher_in_padding = padding
 
-    def set_cipher_out(self, ciphername, iv, password, key_salt, iterations):
+    def set_cipher_out(self, ciphername, iv, password, key_salt, iterations, padding):
         if self.cipher_out_name!=ciphername:
-            log.info("sending data using %s encryption", ciphername)
+            cryptolog.info("sending data using %s encryption", ciphername)
             self.cipher_out_name = ciphername
-        log("set_cipher_out%s", (ciphername, iv, password, key_salt, iterations))
+        cryptolog("set_cipher_out%s", (ciphername, iv, password, key_salt, iterations, padding))
         self.cipher_out, self.cipher_out_block_size = get_cipher(ciphername, iv, password, key_salt, iterations)
+        self.cipher_out_padding = padding
 
 
     def __repr__(self):
@@ -202,9 +213,11 @@ class Protocol(object):
             "input.packetcount"     : self.input_packetcount,
             "input.raw_packetcount" : self.input_raw_packetcount,
             "input.cipher"          : self.cipher_in_name or "",
+            "input.cipher.padding"  : self.cipher_in_padding,
             "output.packetcount"    : self.output_packetcount,
             "output.raw_packetcount": self.output_raw_packetcount,
             "output.cipher"         : self.cipher_out_name or "",
+            "output.cipher.padding" : self.cipher_out_padding,
             "large_packets"         : self.large_packets,
             "compression_level"     : self.compression_level,
             "max_packet_size"       : self.max_packet_size}
@@ -244,8 +257,6 @@ class Protocol(object):
             if not self._closed:
                 self._write_thread.start()
                 self._read_thread.start()
-                self._read_parser_thread.start()
-                self._write_format_thread.start()
         self.idle_add(do_start)
 
     def send_now(self, packet):
@@ -262,18 +273,27 @@ class Protocol(object):
 
     def source_has_more(self):
         self._source_has_more.set()
+        #start the format thread:
+        if not self._write_format_thread and not self._closed:
+            from xpra.make_thread import make_thread
+            self._write_format_thread = make_thread(self._write_format_thread_loop, "format", daemon=True)
+            self._write_format_thread.start()
+        INJECT_FAULT(self)
 
     def _write_format_thread_loop(self):
         log("write_format_thread_loop starting")
         try:
             while not self._closed:
                 self._source_has_more.wait()
-                if self._closed:
+                gpc = self._get_packet_cb
+                if self._closed or not gpc:
                     return
                 self._source_has_more.clear()
-                self._add_packet_to_queue(*self._get_packet_cb())
-        except:
-            self._internal_error("error in network packet write/format", True)
+                self._add_packet_to_queue(*gpc())
+        except Exception as e:
+            if self._closed:
+                return
+            self._internal_error("error in network packet write/format", e, exc_info=True)
 
     def _add_packet_to_queue(self, packet, start_send_cb=None, end_send_cb=None, has_more=False):
         if has_more:
@@ -281,17 +301,17 @@ class Protocol(object):
         if packet is None:
             return
         log("add_packet_to_queue(%s ...)", packet[0])
-        chunks, proto_flags = self.encode(packet)
+        chunks = self.encode(packet)
         with self._write_lock:
             if self._closed:
                 return
-            self._add_chunks_to_queue(chunks, proto_flags, start_send_cb, end_send_cb)
+            self._add_chunks_to_queue(chunks, start_send_cb, end_send_cb)
 
-    def _add_chunks_to_queue(self, chunks, proto_flags, start_send_cb=None, end_send_cb=None):
+    def _add_chunks_to_queue(self, chunks, start_send_cb=None, end_send_cb=None):
         """ the write_lock must be held when calling this function """
         counter = 0
         items = []
-        for index,level,data in chunks:
+        for proto_flags,index,level,data in chunks:
             scb, ecb = None, None
             #fire the start_send_callback just before the first packet is processed:
             if counter==0:
@@ -304,18 +324,21 @@ class Protocol(object):
             if self.cipher_out:
                 proto_flags |= FLAGS_CIPHER
                 #note: since we are padding: l!=len(data)
-                padding = (self.cipher_out_block_size - len(data) % self.cipher_out_block_size) * " "
-                if len(padding)==0:
+                padding_size = self.cipher_out_block_size - (payload_size % self.cipher_out_block_size)
+                if padding_size==0:
                     padded = data
                 else:
-                    padded = data+padding
-                actual_size = payload_size + len(padding)
-                assert len(padded)==actual_size
+                    # pad byte value is number of padding bytes added
+                    padded = data + pad(self.cipher_out_padding, padding_size)
+                    actual_size += padding_size
+                assert len(padded)==actual_size, "expected padded size to be %i, but got %i" % (len(padded), actual_size)
                 data = self.cipher_out.encrypt(padded)
-                assert len(data)==actual_size
-                log("sending %s bytes encrypted with %s padding", payload_size, len(padding))
+                assert len(data)==actual_size, "expected encrypted size to be %i, but got %i" % (len(data), actual_size)
+                cryptolog("sending %s bytes %s encrypted with %s padding", payload_size, self.cipher_out_name, padding_size)
             if proto_flags & FLAGS_NOHEADER:
+                assert not self.cipher_out
                 #for plain/text packets (ie: gibberish response)
+                log("sending %s bytes without header", payload_size)
                 items.append((data, scb, ecb))
             elif actual_size<PACKET_JOIN_SIZE:
                 if type(data) not in JOIN_TYPES:
@@ -330,11 +353,15 @@ class Protocol(object):
         self._write_queue.put(items)
         self.output_packetcount += 1
 
+    def raw_write(self, contents, start_cb=None, end_cb=None):
+        """ Warning: this bypasses the compression and packet encoder! """
+        self._write_queue.put(((contents, start_cb, end_cb), ))
+
     def verify_packet(self, packet):
         """ look for None values which may have caused the packet to fail encoding """
         if type(packet)!=list:
             return
-        assert len(packet)>0
+        assert len(packet)>0, "invalid packet: %s" % packet
         tree = ["'%s' packet" % packet[0]]
         self.do_verify_packet(tree, packet)
 
@@ -435,13 +462,18 @@ class Protocol(object):
         packet = list(packet_in)
         level = self.compression_level
         size_check = LARGE_PACKET_SIZE
-        min_comp_size = 378
+        min_comp_size = MIN_COMPRESS_SIZE
         for i in range(1, len(packet)):
             item = packet[i]
+            if item is None:
+                raise TypeError("invalid None value in %s packet at index %s" % (packet[0], i))
             ti = type(item)
             if ti in (int, long, bool, dict, list, tuple):
                 continue
-            l = len(item)
+            try:
+                l = len(item)
+            except TypeError as e:
+                raise TypeError("invalid type %s in %s packet at index %s: %s" % (ti, packet[0], i, e))
             if ti==Uncompressed:
                 #this is a marker used to tell us we should compress it now
                 #(used by the client for clipboard data)
@@ -458,7 +490,7 @@ class Protocol(object):
                         #LevelCompressed is decompressed by the network layer
                         #so we must tell it how to do that and pass the level flag
                         il = item.level
-                    packets.append((i, il, item.data))
+                    packets.append((0, i, il, item.data))
                     packet[i] = ''
                 else:
                     #data is small enough, inline it:
@@ -469,7 +501,7 @@ class Protocol(object):
                 log.warn("found a large uncompressed item in packet '%s' at position %s: %s bytes", packet[0], i, len(item))
                 #add new binary packet with large item:
                 cl, cdata = self._compress(item, level)
-                packets.append((i, cl, cdata))
+                packets.append((0, i, cl, cdata))
                 #replace this item with an empty string placeholder:
                 packet[i] = ''
             elif ti not in (str, bytes):
@@ -481,25 +513,25 @@ class Protocol(object):
             #replace the packet type with the alias:
             packet[0] = self.send_aliases[packet_type]
         try:
-            main_packet, proto_version = self._encoder(packet)
-        except Exception as e:
+            main_packet, proto_flags = self._encoder(packet)
+        except Exception:
             if self._closed:
                 return [], 0
             log.error("failed to encode packet: %s", packet, exc_info=True)
             #make the error a bit nicer to parse: undo aliases:
             packet[0] = packet_type
             self.verify_packet(packet)
-            raise e
+            raise
         if len(main_packet)>size_check and packet_in[0] not in self.large_packets:
             log.warn("found large packet (%s bytes): %s, argument types:%s, sizes: %s, packet head=%s",
                      len(main_packet), packet_in[0], [type(x) for x in packet[1:]], [len(str(x)) for x in packet[1:]], repr_ellipsized(packet))
         #compress, but don't bother for small packets:
         if level>0 and len(main_packet)>min_comp_size:
             cl, cdata = self._compress(main_packet, level)
-            packets.append((0, cl, cdata))
+            packets.append((proto_flags, 0, cl, cdata))
         else:
-            packets.append((0, 0, main_packet))
-        return packets, proto_version
+            packets.append((proto_flags, 0, 0, main_packet))
+        return packets
 
     def set_compression_level(self, level):
         #this may be used next time encode() is called
@@ -514,10 +546,11 @@ class Protocol(object):
             log("io_thread_loop(%s, %s) loop ended, closed=%s", name, callback, self._closed)
         except ConnectionClosedException as e:
             if not self._closed:
-                self._internal_error("%s connection %s closed: %s" % (name, self._conn, e))
+                #ConnectionClosedException means the warning has been logged already
+                self._connection_lost("%s connection %s closed" % (name, self._conn))
         except (OSError, IOError, socket_error) as e:
             if not self._closed:
-                self._internal_error("%s connection %s reset: %s" % (name, self._conn, e), exc_info=e.args[0] not in ABORT)
+                self._internal_error("%s connection %s reset" % (name, self._conn), e, exc_info=e.args[0] not in ABORT)
         except:
             #can happen during close(), in which case we just ignore:
             if not self._closed:
@@ -568,8 +601,16 @@ class Protocol(object):
             return
         self.input_raw_packetcount += 1
 
-    def _internal_error(self, message="", exc_info=False):
-        log.error("internal error: %s", message, exc_info=exc_info)
+    def _internal_error(self, message="", exc=None, exc_info=False):
+        #log exception info with last log message
+        if self._closed:
+            return
+        ei = exc_info
+        if exc:
+            ei = None   #log it separately below
+        log.error("internal error: %s", message, exc_info=ei)
+        if exc:
+            log.error(" %s", exc, exc_info=exc_info)
         self.idle_add(self._connection_lost, message)
 
     def _connection_lost(self, message="", exc_info=False):
@@ -602,12 +643,22 @@ class Protocol(object):
         self.gibberish(err, data)
 
 
+    def read_queue_put(self, data):
+        #start the parse thread if needed:
+        if not self._read_parser_thread and not self._closed:
+            from xpra.make_thread import make_thread
+            self._read_parser_thread = make_thread(self._read_parse_thread_loop, "parse", daemon=True)
+            self._read_parser_thread.start()
+        self._read_queue.put(data)
+
     def _read_parse_thread_loop(self):
         log("read_parse_thread_loop starting")
         try:
             self.do_read_parse_thread_loop()
-        except:
-            self._internal_error("error in network packet reading/parsing", True)
+        except Exception as e:
+            if self._closed:
+                return
+            self._internal_error("error in network packet reading/parsing", e, exc_info=True)
 
     def do_read_parse_thread_loop(self):
         """
@@ -623,7 +674,7 @@ class Protocol(object):
         """
         read_buffer = None
         payload_size = -1
-        padding = None
+        padding_size = 0
         packet_index = 0
         compression_level = False
         packet = None
@@ -661,16 +712,16 @@ class Protocol(object):
                     bl = len(read_buffer)-8
                     if protocol_flags & FLAGS_CIPHER:
                         if self.cipher_in_block_size==0 or not self.cipher_in_name:
-                            log.warn("received cipher block but we don't have a cipher to decrypt it with, not an xpra client?")
+                            cryptolog.warn("received cipher block but we don't have a cipher to decrypt it with, not an xpra client?")
                             self._invalid_header(read_buffer)
                             return
-                        padding = (self.cipher_in_block_size - data_size % self.cipher_in_block_size) * " "
-                        payload_size = data_size + len(padding)
+                        padding_size = self.cipher_in_block_size - (data_size % self.cipher_in_block_size)
+                        payload_size = data_size + padding_size
                     else:
                         #no cipher, no padding:
-                        padding = None
+                        padding_size = 0
                         payload_size = data_size
-                    assert payload_size>0
+                    assert payload_size>0, "invalid payload size: %i" % payload_size
                     read_buffer = read_buffer[8:]
 
                     if payload_size>self.max_packet_size:
@@ -701,20 +752,26 @@ class Protocol(object):
                 #decrypt if needed:
                 data = raw_string
                 if self.cipher_in and protocol_flags & FLAGS_CIPHER:
-                    log("received %s encrypted bytes with %s padding", payload_size, len(padding))
+                    cryptolog("received %i %s encrypted bytes with %s padding", payload_size, self.cipher_in_name, padding_size)
                     data = self.cipher_in.decrypt(raw_string)
-                    if padding:
+                    if padding_size > 0:
                         def debug_str(s):
                             try:
-                                return list(bytearray(s))
+                                return binascii.hexlify(bytearray(s))
                             except:
-                                return list(str(s))
-                        if not data.endswith(padding):
-                            log("decryption failed: string does not end with '%s': %s (%s) -> %s (%s)",
-                            padding, debug_str(raw_string), type(raw_string), debug_str(data), type(data))
-                            self._internal_error("encryption error (wrong key?)")
-                            return
-                        data = data[:-len(padding)]
+                                return csv(list(str(s)))
+                        # pad byte value is number of padding bytes added
+                        padtext = pad(self.cipher_in_padding, padding_size)
+                        if data.endswith(padtext):
+                            cryptolog("found %s %s padding", self.cipher_in_padding, self.cipher_in_name)
+                        else:
+                            actual_padding = data[-padding_size:]
+                            cryptolog.warn("Warning: %s decryption failed: invalid padding", self.cipher_in_name)
+                            cryptolog(" data does not end with %s padding bytes %s", self.cipher_in_padding, debug_str(padtext))
+                            cryptolog(" but with %s (%s)", debug_str(actual_padding), type(data))
+                            cryptolog(" decrypted data: %s", debug_str(data[:128]))
+                            return self._internal_error("%s encryption padding error - wrong key?" % self.cipher_in_name)
+                        data = data[:-padding_size]
                 #uncompress if needed:
                 if compression_level>0:
                     try:
@@ -729,6 +786,8 @@ class Protocol(object):
                         if self.cipher_in:
                             msg += " (invalid encryption key?)"
                         else:
+                            #only include the exception text when not using encryption
+                            #as this may leak crypto information:
                             msg += " %s" % e
                         return self.gibberish(msg, data)
 
@@ -755,18 +814,21 @@ class Protocol(object):
                     return
                 except ValueError as e:
                     etype = packet_encoding.get_packet_encoding_type(protocol_flags)
-                    log.error("failed to parse %s packet: %s", etype, e, exc_info=not self._closed)
+                    log.error("Error parsing %s packet:", etype)
+                    log.error(" %s", e)
                     if self._closed:
                         return
-                    log("failed to parse %s packet: %s", etype, binascii.hexlify(data))
-                    msg = "packet index=%s, packet size=%s, buffer size=%s, error=%s" % (packet_index, payload_size, bl, e)
+                    log("failed to parse %s packet: %s", etype, binascii.hexlify(data[:128]))
+                    log(" %s", e)
+                    log(" data: %s", repr_ellipsized(data))
+                    log(" packet index=%i, packet size=%i, buffer size=%s", packet_index, payload_size, bl)
                     self.gibberish("failed to parse %s packet" % etype, data)
                     return
 
                 if self._closed:
                     return
                 payload_size = -1
-                padding = None
+                padding_size = 0
                 #add any raw packets back into it:
                 if raw_packets:
                     for index,raw_data in raw_packets.items():
@@ -784,6 +846,7 @@ class Protocol(object):
                 log("processing packet %s", packet_type)
                 self._process_packet_cb(self, packet)
                 packet = None
+                INJECT_FAULT(self)
 
     def flush_then_close(self, last_packet, done_callback=None):
         """ Note: this is best effort only
@@ -796,6 +859,7 @@ class Protocol(object):
             then no matter what, we close the connection and stop the threads.
         """
         def done():
+            log("flush_then_close: done, callback=%s", done_callback)
             if done_callback:
                 done_callback()
         if self._closed:
@@ -815,7 +879,7 @@ class Protocol(object):
                     self.timeout_add(100, wait_for_queue, timeout-1)
             else:
                 log("flush_then_close: queue is now empty, sending the last packet and closing")
-                chunks, proto_flags = self.encode(last_packet)
+                chunks = self.encode(last_packet)
                 def close_and_release():
                     log("flush_then_close: wait_for_packet_sent() close_and_release()")
                     self.close()
@@ -835,9 +899,9 @@ class Protocol(object):
                     #if we're here, we have the lock and the packet is in the write queue
                     log("flush_then_close: packet_queued() closed=%s", self._closed)
                     if wait_for_packet_sent():
-                        #check every 100ms
+                        #check again every 100ms
                         self.timeout_add(100, wait_for_packet_sent)
-                self._add_chunks_to_queue(chunks, proto_flags, start_send_cb=None, end_send_cb=packet_queued)
+                self._add_chunks_to_queue(chunks, start_send_cb=None, end_send_cb=packet_queued)
                 #just in case wait_for_packet_sent never fires:
                 self.timeout_add(5*1000, close_and_release)
 
@@ -865,14 +929,16 @@ class Protocol(object):
         wait_for_write_lock()
 
     def close(self):
-        log("close() closed=%s", self._closed)
+        log("Protocol.close() closed=%s, connection=%s", self._closed, self._conn)
         if self._closed:
             return
         self._closed = True
         self.idle_add(self._process_packet_cb, self, [Protocol.CONNECTION_LOST])
-        if self._conn:
+        c = self._conn
+        if c:
             try:
-                self._conn.close()
+                log("Protocol.close() calling %s", c.close)
+                c.close()
                 if self._log_stats is None and self._conn.input_bytecount==0 and self._conn.output_bytecount==0:
                     #no data sent or received, skip logging of stats:
                     self._log_stats = False
@@ -887,6 +953,7 @@ class Protocol(object):
             self._conn = None
         self.terminate_queue_threads()
         self.idle_add(self.clean)
+        log("Protocol.close() done")
 
     def steal_connection(self, read_callback=None):
         #so we can re-use this connection somewhere else
@@ -896,7 +963,7 @@ class Protocol(object):
         #to ensure that no packets get lost.
         #The caller must call wait_for_io_threads_exit() to ensure that this
         #class is no longer reading from the connection before it can re-use it
-        assert not self._closed
+        assert not self._closed, "cannot steal a closed connection"
         if read_callback:
             self._read_queue_put = read_callback
         conn = self._conn
@@ -920,9 +987,10 @@ class Protocol(object):
 
     def terminate_queue_threads(self):
         log("terminate_queue_threads()")
-        #the format thread will exit since closed is set too:
+        #the format thread will exit:
+        self._get_packet_cb = None
         self._source_has_more.set()
-        #make the threads exit by adding the empty marker:
+        #make all the queue based threads exit by adding the empty marker:
         exit_queue = Queue()
         for _ in range(10):     #just 2 should be enough!
             exit_queue.put(None)
@@ -938,4 +1006,3 @@ class Protocol(object):
             orq.put_nowait(None)
         except:
             pass
-
