@@ -9,11 +9,11 @@ import signal
 import subprocess
 import binascii
 
-from xpra.gtk_common.gobject_compat import import_gobject
-gobject = import_gobject()
-gobject.threads_init()
+from xpra.gtk_common.gobject_compat import import_glib
+from xpra.util import repr_ellipsized
 
 from xpra.net.bytestreams import TwoFileConnection
+from xpra.net import ConnectionClosedException
 from xpra.net.protocol import Protocol
 from xpra.os_util import Queue, setbinarymode, SIGNAMES, bytestostr
 from xpra.child_reaper import getChildReaper
@@ -36,16 +36,33 @@ HEXLIFY_PACKETS = os.environ.get("XPRA_HEXLIFY_PACKETS", "0")=="1"
 #avoids showing a new console window on win32:
 WIN32_SHOWWINDOW = os.environ.get("XPRA_WIN32_SHOWWINDOW", "0")=="1"
 
+FAULT_RATE = int(os.environ.get("XPRA_WRAPPER_FAULT_INJECTION_RATE", "0"))
+if FAULT_RATE>0:
+    _counter = 0
+    def INJECT_FAULT(p):
+        global _counter
+        _counter += 1
+        if (_counter % FAULT_RATE)==0:
+            log.warn("injecting fault in %s", p)
+            p.raw_write("Wrapper JUNK! added by fault injection code")
+else:
+    def INJECT_FAULT(p):
+        pass
 
-#this allows us to use the gtk main loop instead:
-#import gtk
-#gtk.threads_init()
-#class mainloop(object):
-#    def run(self):
-#        gtk.main()
-#    def quit(self):
-#        gtk.main_quit()
-mainloop = gobject.MainLoop
+
+def setup_fastencoder_nocompression(protocol):
+    from xpra.net.packet_encoding import get_enabled_encoders, PERFORMANCE_ORDER
+    encoders = get_enabled_encoders(PERFORMANCE_ORDER)
+    assert len(encoders)>0, "no packet encoders available!?"
+    for encoder in encoders:
+        try:
+            protocol.enable_encoder(encoder)
+            log("protocol using %s", encoder)
+            break
+        except Exception as e:
+            log("failed to enable %s: %s", encoder, e)
+    #we assume this is local, so no compression:
+    protocol.enable_compressor("none")
 
 
 class subprocess_callee(object):
@@ -58,7 +75,6 @@ class subprocess_callee(object):
     (there is no validation of which signals are valid or not)
     """
     def __init__(self, input_filename="-", output_filename="-", wrapped_object=None, method_whitelist=None):
-        self.mainloop = mainloop()
         self.name = ""
         self.input_filename = input_filename
         self.output_filename = output_filename
@@ -68,8 +84,18 @@ class subprocess_callee(object):
         self.wrapped_object = wrapped_object
         self.send_queue = Queue()
         self.protocol = None
-        signal.signal(signal.SIGINT, self.handle_signal)
+        if sys.version_info[0]<3:
+            #this breaks gobject3!
+            signal.signal(signal.SIGINT, self.handle_signal)
         signal.signal(signal.SIGTERM, self.handle_signal)
+        self.setup_mainloop()
+
+    def setup_mainloop(self):
+        glib = import_glib()
+        self.mainloop = glib.MainLoop()
+        self.idle_add = glib.idle_add
+        self.timeout_add = glib.timeout_add
+        self.source_remove = glib.source_remove
 
 
     def connect_export(self, signal_name, *user_data):
@@ -95,12 +121,14 @@ class subprocess_callee(object):
             self.run()
             return 0
         except KeyboardInterrupt as e:
-            log.warn("%s", e)
+            if str(e):
+                log.warn("%s", e)
             return 0
         except Exception:
             log.error("error in main loop", exc_info=True)
             return 1
         finally:
+            self.cleanup()
             if self.protocol:
                 self.protocol.close()
                 self.protocol = None
@@ -132,13 +160,8 @@ class subprocess_callee(object):
         #stdin and stdout wrapper:
         conn = TwoFileConnection(self._output, self._input, abort_test=None, target=self.name, info=self.name, close_cb=self.net_stop)
         conn.timeout = 0
-        protocol = Protocol(gobject, conn, self.process_packet, get_packet_cb=self.get_packet)
-        try:
-            protocol.enable_encoder("rencode")
-        except Exception as e:
-            log.warn("failed to enable rencode: %s", e)
-            protocol.enable_encoder("bencode")
-        protocol.enable_compressor("none")
+        protocol = Protocol(self, conn, self.process_packet, get_packet_cb=self.get_packet)
+        setup_fastencoder_nocompression(protocol)
         protocol.large_packets = self.large_packets
         return protocol
 
@@ -151,7 +174,7 @@ class subprocess_callee(object):
         #this is called from the network thread,
         #we use idle add to ensure we clean things up from the main thread
         log("net_stop() will call stop from main thread")
-        gobject.idle_add(self.stop)
+        self.idle_add(self.stop)
 
 
     def cleanup(self):
@@ -177,7 +200,7 @@ class subprocess_callee(object):
         self.send("signal", signame)
         self.cleanup()
         #give time for the network layer to send the signal message
-        gobject.timeout_add(150, self.stop)
+        self.timeout_add(150, self.stop)
 
     def signal_stop(self, sig, frame):
         """ This time we really want to exit without waiting """
@@ -194,6 +217,7 @@ class subprocess_callee(object):
         p = self.protocol
         if p:
             p.source_has_more()
+        INJECT_FAULT(p)
 
     def get_packet(self):
         try:
@@ -209,7 +233,13 @@ class subprocess_callee(object):
             self.net_stop()
             return
         elif command==Protocol.GIBBERISH:
-            log("gibberish received: %s", repr_ellipsized(packet))
+            log.warn("gibberish received:")
+            log.warn(" %s", repr_ellipsized(packet[1], limit=80))
+            log.warn(" stopping")
+            self.net_stop()
+            return
+        elif command=="stop":
+            log("received stop message")
             self.net_stop()
             return
         #make it easier to hookup signals to methods:
@@ -217,14 +247,45 @@ class subprocess_callee(object):
         if self.method_whitelist is not None and attr not in self.method_whitelist:
             log.warn("invalid command: %s (not in whitelist: %s)", attr, self.method_whitelist)
             return
-        method = getattr(self.wrapped_object, attr, None)
+        wo = self.wrapped_object
+        if not wo:
+            log("wrapped object is no more, ignoring method call '%s'", attr)
+            return
+        method = getattr(wo, attr, None)
         if not method:
-            if self.wrapped_object is not None:
-                log.warn("unknown command: %s", command)
+            log.warn("unknown command: '%s'", attr)
+            log.warn(" packet: '%s'", repr_ellipsized(str(packet)))
             return
         if DEBUG_WRAPPER:
-            log("calling %s.%s%s", self.wrapped_object, attr, str(tuple(packet[1:]))[:128])
-        gobject.idle_add(method, *packet[1:])
+            log("calling %s.%s%s", wo, attr, str(tuple(packet[1:]))[:128])
+        self.idle_add(method, *packet[1:])
+        INJECT_FAULT(proto)
+
+
+def exec_kwargs():
+    if os.name=="posix":
+        return {"close_fds" : True}
+    elif sys.platform.startswith("win"):
+        if not WIN32_SHOWWINDOW:
+            startupinfo = subprocess.STARTUPINFO()
+            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
+            return {"startupinfo" : startupinfo}
+    return {}
+
+def exec_env(blacklist=["LS_COLORS", ]):
+    env = os.environ.copy()
+    env["XPRA_SKIP_UI"] = "1"
+    env["XPRA_FORCE_COLOR_LOG"] = "1"
+    #let's make things more complicated than they should be:
+    #on win32, the environment can end up containing unicode, and subprocess chokes on it
+    for k,v in env.items():
+        if k in blacklist:
+            continue
+        try:
+            env[k] = bytestostr(v.encode("utf8"))
+        except:
+            env[k] = bytestostr(v)
+    return env
 
 
 class subprocess_caller(object):
@@ -248,6 +309,10 @@ class subprocess_caller(object):
         #hook a default packet handlers:
         self.connect(Protocol.CONNECTION_LOST, self.connection_lost)
         self.connect(Protocol.GIBBERISH, self.gibberish)
+        glib = import_glib()
+        self.idle_add = glib.idle_add
+        self.timeout_add = glib.timeout_add
+        self.source_remove = glib.source_remove
 
 
     def connect(self, signal, cb, *args):
@@ -261,69 +326,51 @@ class subprocess_caller(object):
         self._fire_callback("exit")
 
     def start(self):
+        self.start = self.fail_start
         self.process = self.exec_subprocess()
         self.protocol = self.make_protocol()
         self.protocol.start()
 
+    def fail_start(self):
+        raise Exception("this wrapper has already been started")
+
+    def abort_test(self, action):
+        p = self.process
+        if p is None or p.poll():
+            raise ConnectionClosedException("cannot %s: subprocess has terminated" % action)
+
     def make_protocol(self):
         #make a connection using the process stdin / stdout
-        conn = TwoFileConnection(self.process.stdin, self.process.stdout, abort_test=None, target=self.description, info=self.description, close_cb=self.subprocess_exit)
+        conn = TwoFileConnection(self.process.stdin, self.process.stdout, abort_test=self.abort_test, target=self.description, info=self.description, close_cb=self.subprocess_exit)
         conn.timeout = 0
-        protocol = Protocol(gobject, conn, self.process_packet, get_packet_cb=self.get_packet)
-        #we assume the other end has the same encoders (which is reasonable):
-        #TODO: fallback to bencoder
-        try:
-            protocol.enable_encoder("rencode")
-        except Exception as e:
-            log.warn("failed to enable rencode: %s", e)
-            protocol.enable_encoder("bencode")
-        #we assume this is local, so no compression:
-        protocol.enable_compressor("none")
+        protocol = Protocol(self, conn, self.process_packet, get_packet_cb=self.get_packet)
+        setup_fastencoder_nocompression(protocol)
         protocol.large_packets = self.large_packets
         return protocol
 
 
     def exec_subprocess(self):
-        kwargs = self.exec_kwargs()
-        log("exec_subprocess() command=%s, kwargs=%s", self.command, kwargs)
-        proc = subprocess.Popen(self.command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=sys.stderr.fileno(), env=self.get_env(), **kwargs)
+        kwargs = exec_kwargs()
+        env = self.get_env()
+        log("exec_subprocess() command=%s, env=%s, kwargs=%s", self.command, env, kwargs)
+        proc = subprocess.Popen(self.command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=sys.stderr.fileno(), env=env, **kwargs)
         getChildReaper().add_process(proc, self.description, self.command, True, True, callback=self.subprocess_exit)
         return proc
 
     def get_env(self):
-        env = os.environ.copy()
-        env["XPRA_SKIP_UI"] = "1"
+        env = exec_env()
         env["XPRA_LOG_PREFIX"] = "%s " % self.description
-        #let's make things more complicated than they should be:
-        #on win32, the environment can end up containing unicode, and subprocess chokes on it
-        for k,v in env.items():
-            try:
-                env[k] = bytestostr(v.encode("utf8"))
-            except:
-                env[k] = bytestostr(v)
         return env
 
-    def exec_kwargs(self):
-        if os.name=="posix":
-            return {"close_fds" : True}
-        elif sys.platform.startswith("win"):
-            if not WIN32_SHOWWINDOW:
-                startupinfo = subprocess.STARTUPINFO()
-                startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
-                return {"startupinfo" : startupinfo}
-        return {}
-
-
     def cleanup(self):
         self.stop()
 
     def stop(self):
         self.stop_process()
-        #call via idle_add to prevent deadlocks on win32!
-        gobject.idle_add(self.stop_protocol)
+        self.stop_protocol()
 
     def stop_process(self):
-        log("stop() sending stop request to %s", self.description)
+        log("%s.stop_process() sending stop request to %s", self, self.description)
         proc = self.process
         if proc and proc.poll() is None:
             try:
@@ -336,7 +383,7 @@ class subprocess_caller(object):
         p = self.protocol
         if p:
             self.protocol = None
-            log("%s.stop() calling %s", self, p.close)
+            log("%s.stop_protocol() calling %s", self, p.close)
             try:
                 p.close()
             except Exception as e:
@@ -348,7 +395,8 @@ class subprocess_caller(object):
         self.stop()
 
     def gibberish(self, *args):
-        log("gibberish%s", args)
+        log.warn("%s stopping on gibberish:", self.description)
+        log.warn(" %s", repr_ellipsized(args[1], limit=80))
         self.stop()
 
 
@@ -364,20 +412,22 @@ class subprocess_caller(object):
         p = self.protocol
         if p:
             p.source_has_more()
+        INJECT_FAULT(p)
 
     def process_packet(self, proto, packet):
         if DEBUG_WRAPPER:
             log("process_packet(%s, %s)", proto, [str(x)[:32] for x in packet])
         signal_name = bytestostr(packet[0])
         self._fire_callback(signal_name, packet[1:])
+        INJECT_FAULT(proto)
 
     def _fire_callback(self, signal_name, extra_args=[]):
         callbacks = self.signal_callbacks.get(signal_name)
-        log("firing callback for %s: %s", signal_name, callbacks)
+        log("firing callback for '%s': %s", signal_name, callbacks)
         if callbacks:
             for cb, args in callbacks:
                 try:
                     all_args = list(args) + extra_args
-                    gobject.idle_add(cb, self, *all_args)
+                    self.idle_add(cb, self, *all_args)
                 except Exception:
                     log.error("error processing callback %s for %s packet", cb, signal_name, exc_info=True)
