@@ -3,23 +3,34 @@
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+import os
+
 from xpra.gtk_common.gobject_compat import import_gtk
 gtk = import_gtk()
 from xpra.gtk_common.gtk_util import scaled_image
 from xpra.client.gtk_base.about import about
 from xpra.client.gtk_base.gtk_tray_menu_base import GTKTrayMenuBase, populate_encodingsmenu
 from xpra.platform.paths import get_icon
+from xpra.platform.darwin.gui import get_OSXApplication
 
 from xpra.log import Logger
-log = Logger("tray", "osx")
+log = Logger("osx", "tray", "menu")
 
 
 #control which menus are shown in the OSX global menu:
+SHOW_FEATURES_MENU = True
 SHOW_SOUND_MENU = True
 SHOW_ENCODINGS_MENU = True
 SHOW_ACTIONS_MENU = True
+SHOW_INFO_MENU = True
+
 SHOW_ABOUT_XPRA = True
 
+SINGLE_MENU = os.environ.get("XPRA_OSX_SINGLE_MENU", "1")=="1"
+
+
+SEPARATOR = "SEPARATOR"
+
 
 _OSXMenuHelper = None
 def getOSXMenuHelper(client=None):
@@ -45,7 +56,8 @@ class OSXMenuHelper(GTKTrayMenuBase):
         self.menu_bar = None
         self.hidden_window = None
         self.keyboard = None
-        self.menus = {}
+        self.menus = {}             #all the "top-level" menu items we manage
+        self.app_menus = {}         #the ones added to the app_menu via insert_app_menu_item (which cannot be removed!)
         self.full = False
         self.set_client(client)
 
@@ -53,44 +65,89 @@ class OSXMenuHelper(GTKTrayMenuBase):
         self.client = client
         if client and client.keyboard_helper:
             self.keyboard = client.keyboard_helper.keyboard
+        #if we call add_about before the main loop is ready,
+        #things don't work...
+        if client and SHOW_ABOUT_XPRA:
+            self.client.after_handshake(self.add_about)
 
     def build(self):
         log("OSXMenuHelper.build()")
         if self.menu_bar is None:
             self.menu_bar = gtk.MenuBar()
-            self.build_menu_bar()
+            self.menu_bar.show_all()
         return self.menu_bar
 
     def rebuild(self):
         log("OSXMenuHelper.rebuild()")
-        if self.menu_bar:
-            self.remove_all_menus()
-            self.build_menu_bar()
-        return self.build()
+        if not self.menu_bar:
+            return self.build()
+        self.remove_all_menus()
+        return self.menu_bar
 
     def remove_all_menus(self):
         log("OSXMenuHelper.remove_all_menus()")
         if self.menu_bar:
             for x in self.menus.values():
-                self.menu_bar.remove(x)
-                x.hide()
+                if x in self.menu_bar.get_children():
+                    self.menu_bar.remove(x)
+                    x.hide()
         self.menus = {}
+        self.full = False
 
-    def make_osxmenu(self, name):
-        item = gtk.MenuItem(name)
-        submenu = gtk.Menu()
-        item.set_submenu(submenu)
+
+    def add_top_level_menu(self, label, submenu):
+        """ Adds the item to the app-menu or to the top bar,
+            but only if it has not been added yet.
+            (depending on the SINGLE_MENU flag)
+        """
+        if SINGLE_MENU:
+            #add or re-use menu item in the app-menu:
+            self.add_to_app_menu(label, submenu)
+        else:
+            self.add_to_menu_bar(label, submenu)
+
+    def add_to_app_menu(self, label, submenu):
+        item = self.app_menus.get(label)
+        if item:
+            log("application menu already has a '%s' entry", label)
+            if submenu is not None:
+                item.set_submenu(submenu)
+            item.show_all()
+        else:
+            if label.startswith(SEPARATOR):
+                item = gtk.SeparatorMenuItem()
+            else:
+                item = self.menuitem(label)
+                item.set_submenu(submenu)
+            item.show_all()
+            macapp = get_OSXApplication()
+            macapp.insert_app_menu_item(item, 1)
+            self.app_menus[label] = item
+
+    def add_to_menu_bar(self, label, submenu):
+        if label.startswith(SEPARATOR):
+            return      #not relevant
+        item = self.menus.get(label)
+        if item is None:
+            item = self.menuitem(label)
+            item.set_submenu(submenu)
+            item.show_all()
+            self.menu_bar.add(item)
+            self.menus[label] = item
+        else:
+            item.set_submenu(submenu)
+            item.show_all()
+
+
+    def add_about(self):
+        if "About" in self.app_menus:
+            return
+        item = self.menuitem("About", cb=about)
         item.show_all()
-        self.menu_bar.add(item)
-        self.menus[name] = (item, submenu)
-        return submenu
+        macapp = get_OSXApplication()
+        macapp.insert_app_menu_item(item, 0)
+        self.app_menus["About"] = item
 
-    def build_menu_bar(self):
-        log("OSXMenuHelper.build_menu_bar()")
-        if SHOW_ABOUT_XPRA:
-            info_menu = self.make_osxmenu("Info")
-            info_menu.add(self.menuitem("About Xpra", "information.png", None, about))
-        self.menu_bar.show_all()
 
     def add_full_menu(self):
         log("OSXMenuHelper.add_full_menu()")
@@ -98,41 +155,58 @@ class OSXMenuHelper(GTKTrayMenuBase):
             return
         self.full = True
         assert self.client
-        if SHOW_ABOUT_XPRA:
-            _, info_menu = self.menus.get("Info")
+        menus = self.get_extra_menus()
+        for label, submenu in reversed(menus):
+            self.add_top_level_menu(label, submenu)
+        self.menu_bar.show_all()
+
+    def get_extra_menus(self):
+        menus = []
+        if SHOW_INFO_MENU:
+            info_menu = self.make_menu()
             info_menu.append(self.make_sessioninfomenuitem())
             info_menu.append(self.make_bugreportmenuitem())
-        features_menu = self.make_osxmenu("Features")
-        features_menu.add(self.make_bellmenuitem())
-        features_menu.add(self.make_cursorsmenuitem())
-        features_menu.add(self.make_notificationsmenuitem())
-        features_menu.add(self.make_swapkeysmenuitem())
-        features_menu.add(self.make_numlockmenuitem())
-        features_menu.add(self.make_openglmenuitem())
+            menus.append(("Info", info_menu))
+        if SHOW_FEATURES_MENU:
+            features_menu = self.make_menu()
+            menus.append(("Features", features_menu))
+            features_menu.add(self.make_bellmenuitem())
+            features_menu.add(self.make_cursorsmenuitem())
+            features_menu.add(self.make_notificationsmenuitem())
+            features_menu.add(self.make_swapkeysmenuitem())
+            features_menu.add(self.make_numlockmenuitem())
+            features_menu.add(self.make_openglmenuitem())
+            features_menu.add(self.make_scalingmenuitem())
         if SHOW_SOUND_MENU:
-            sound_menu = self.make_osxmenu("Sound")
+            sound_menu = self.make_menu()
             if self.client.speaker_allowed and len(self.client.speaker_codecs)>0:
                 sound_menu.add(self.make_speakermenuitem())
             if self.client.microphone_allowed and len(self.client.microphone_codecs)>0:
                 sound_menu.add(self.make_microphonemenuitem())
+            menus.append(("Sound", sound_menu))
         if SHOW_ENCODINGS_MENU:
-            encodings_menu = self.make_osxmenu("Encoding")
+            encodings_menu = self.make_menu()
             def set_encodings_menu(*args):
                 from xpra.codecs.loader import PREFERED_ENCODING_ORDER
                 encodings = [x for x in PREFERED_ENCODING_ORDER if x in self.client.get_encodings()]
                 populate_encodingsmenu(encodings_menu, self.get_current_encoding, self.set_current_encoding, encodings, self.client.server_encodings)
-            self.client.connect("handshake-complete", set_encodings_menu)
+            self.client.after_handshake(set_encodings_menu)
+            menus.append(("Encoding", encodings_menu))
         if SHOW_ACTIONS_MENU:
-            actions_menu = self.make_osxmenu("Actions")
+            actions_menu = self.make_menu()
             actions_menu.add(self.make_refreshmenuitem())
             actions_menu.add(self.make_raisewindowsmenuitem())
             #set_sensitive(bool) does not work on OSX,
             #so we only add the menu item if it does something
-            def addsnc(*args):
+            def add_ah(*args):
                 if self.client.start_new_commands:
-                    actions_menu.add(self.make_startnewcommandmenuitem(True))
-            self.client.connect("handshake-complete", addsnc)
-        self.menu_bar.show_all()
+                    actions_menu.add(self.make_startnewcommandmenuitem())
+                if self.client.server_file_transfer:
+                    actions_menu.add(self.make_uploadmenuitem())
+            self.client.after_handshake(add_ah)
+            menus.append(("Actions", actions_menu))
+        menus.append((SEPARATOR+"-EXTRAS", None))
+        return menus
 
     #these methods are called by the superclass
     #but we don't have a quality or speed menu, so override and ignore
@@ -155,7 +229,7 @@ class OSXMenuHelper(GTKTrayMenuBase):
             else:
                 log("set_swapkeys_menuitem(%s) no keyboard!", args)
                 self.swapkeys_menuitem.set_sensitive(False)
-        self.client.connect("handshake-complete", set_swapkeys_menuitem)
+        self.client.after_handshake(set_swapkeys_menuitem)
         return  self.swapkeys_menuitem
 
     def make_numlockmenuitem(self):
@@ -173,7 +247,7 @@ class OSXMenuHelper(GTKTrayMenuBase):
             else:
                 log("set_numlock_menuitem(%s) no keyboard!", args)
                 self.swapkeys_menuitem.set_sensitive(False)
-        self.client.connect("handshake-complete", set_numlock_menuitem)
+        self.client.after_handshake(set_numlock_menuitem)
         return self.numlock_menuitem
 
     def update_numlock(self, on):
@@ -182,7 +256,7 @@ class OSXMenuHelper(GTKTrayMenuBase):
 
     def build_dock_menu(self):
         log("OSXMenuHelper.build_dock_menu()")
-        self.dock_menu = gtk.Menu()
+        self.dock_menu = self.make_menu()
         if SHOW_ABOUT_XPRA:
             self.dock_menu.add(self.menuitem("About Xpra", "information.png", None, about))
         self.dock_menu.show_all()
