@@ -49,8 +49,6 @@ def get_native_notifier_classes():
     return []
 
 
-def get_dpi():
-    return -1
 
 def get_xdpi():
     return -1
@@ -60,7 +58,13 @@ def get_ydpi():
 
 
 def get_icon_size():
-    dpi = get_dpi()
+    xdpi = get_xdpi()
+    ydpi = get_ydpi()
+    if xdpi>0 and ydpi>0:
+        from xpra.util import iround
+        dpi = iround((xdpi + ydpi)/2.0)
+    else:
+        dpi = 96
     if dpi > 144:
         return 48
     elif dpi > 120:
@@ -99,6 +103,12 @@ def get_double_click_distance():
 def get_fixed_cursor_size():
     return -1, -1
 
+def get_cursor_size():
+    return -1
+
+def get_window_frame_size(x, y, w, h):
+    return None
+
 def get_window_frame_sizes():
     return {}
 
@@ -113,7 +123,7 @@ def remove_window_hooks(window):
 def show_desktop(show):
     pass
 
-def set_fullscreen_monitors(window, fsm):
+def set_fullscreen_monitors(window, fsm, source_indication=0):
     pass
 
 def set_shaded(window, shaded):
@@ -123,6 +133,10 @@ def set_shaded(window, shaded):
 def gl_check():
     return None     #no problem
 
+def get_menu_support_function():
+    return None
+
+
 take_screenshot = None
 ClientExtras = None
 
@@ -149,7 +163,7 @@ def get_info_base():
             "double_click.time"             : get_double_click_time(),
             "double_click.distance"         : get_double_click_distance(),
             "fixed_cursor_size"             : get_fixed_cursor_size(),
-            "dpi"                           : get_dpi(),
+            "cursor_size"                   : get_cursor_size(),
             "dpi.x"                         : get_xdpi(),
             "dpi.y"                         : get_ydpi(),
             "icon_size"                     : get_icon_size(),
@@ -170,15 +184,16 @@ platform_import(globals(), "gui", False,
                 "show_desktop", "set_fullscreen_monitors", "set_shaded",
                 "ClientExtras",
                 "take_screenshot",
+                "get_menu_support_function",
                 "get_native_tray_menu_helper_classes",
                 "get_native_tray_classes",
                 "get_native_system_tray_classes",
                 "get_native_notifier_classes",
                 "get_vrefresh", "get_workarea", "get_workareas",
                 "get_number_of_desktops", "get_desktop_names",
-                "get_antialias_info", "get_icon_size", "get_dpi", "get_xdpi", "get_ydpi",
+                "get_antialias_info", "get_icon_size", "get_xdpi", "get_ydpi",
                 "get_double_click_time", "get_double_click_distance",
-                "get_fixed_cursor_size", "get_window_frame_sizes",
+                "get_fixed_cursor_size", "get_cursor_size", "get_window_frame_sizes",
                 "add_window_hooks", "remove_window_hooks",
                 "system_bell",
                 "get_info")
@@ -187,8 +202,10 @@ platform_import(globals(), "gui", False,
 def main():
     from xpra.platform import init as platform_init,clean
     from xpra.util import nonl
+    from xpra.log import enable_color
     try:
         platform_init("GUI-Properties")
+        enable_color()
         init()
         verbose = "-v" in sys.argv or "--verbose" in sys.argv
         if verbose:
@@ -199,7 +216,10 @@ def main():
         #naughty, but how else can I hook this up?
         import os
         if os.name=="posix":
-            from xpra.x11.bindings import posix_display_source      #@UnusedImport
+            try:
+                from xpra.x11.bindings import posix_display_source      #@UnusedImport
+            except:
+                pass    #maybe running on OSX? hope for the best..
         i = get_info()
         for k in sorted(i.keys()):
             v = i[k]
