@@ -1,12 +1,14 @@
 # This file is part of Xpra.
 # Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011-2014 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2011-2015 Antoine Martin <antoine@devloop.org.uk>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 # Platform-specific code for Win32 -- the parts that may import gtk.
 
-import os, sys
+import sys
+import os
+import types
 from xpra.log import Logger
 log = Logger("win32")
 grablog = Logger("win32", "grab")
@@ -14,7 +16,7 @@ screenlog = Logger("win32", "screen")
 
 from xpra.platform.win32.win32_events import get_win32_event_listener
 from xpra.platform.win32.window_hooks import Win32Hooks
-from xpra.util import AdHocStruct
+from xpra.util import AdHocStruct, csv
 import ctypes
 from ctypes import windll, byref
 
@@ -25,6 +27,10 @@ UNDECORATED_STYLE = WINDOW_HOOKS and os.environ.get("XPRA_WIN32_UNDECORATED_STYL
 DEFAULT_MAX_SIZE_HINT = sys.version_info[0]<3
 MAX_SIZE_HINT = WINDOW_HOOKS and os.environ.get("XPRA_WIN32_MAX_SIZE_HINT", str(int(DEFAULT_MAX_SIZE_HINT)))=="1"
 GEOMETRY = WINDOW_HOOKS and os.environ.get("XPRA_WIN32_GEOMETRY", "1")=="1"
+LANGCHANGE = WINDOW_HOOKS and os.environ.get("XPRA_WIN32_LANGCHANGE", "1")=="1"
+
+DPI_AWARE = os.environ.get("XPRA_DPI_AWARE", "1")=="1"
+DPI_AWARENESS = int(os.environ.get("XPRA_DPI_AWARENESS", "1"))
 
 
 KNOWN_EVENTS = {}
@@ -46,18 +52,31 @@ except Exception as e:
 
 def do_init():
     #tell win32 we handle dpi
+    if not DPI_AWARE:
+        screenlog.warn("SetProcessDPIAware not set due to environment override")
+        return
     try:
-        if os.environ.get("XPRA_DPI_AWARE", "1")!="1":
-            log.warn("SetProcessDPIAware not set due to environment override")
-            return
-        from ctypes import WINFUNCTYPE
-        from ctypes.wintypes import BOOL
-        prototype = WINFUNCTYPE(BOOL)
-        SetProcessDPIAware = prototype(("SetProcessDPIAware", windll.user32))
-        dpi_set = SetProcessDPIAware()
-        log("SetProcessDPIAware()=%s", dpi_set)
+        SetProcessDPIAware = windll.user32.SetProcessDPIAware
+        dpiaware = SetProcessDPIAware()
+        screenlog("SetProcessDPIAware: %s()=%s", SetProcessDPIAware, dpiaware)
+        assert dpiaware!=0
     except Exception as e:
-        log("failed to set DPI: %s", e)
+        screenlog("SetProcessDPIAware() failed: %s", e)
+    if DPI_AWARENESS<=0:
+        screenlog.warn("SetProcessDPIAwareness not set due to environment override")
+        return
+    try:
+        Process_System_DPI_Aware        = 1
+        Process_DPI_Unaware             = 0
+        Process_Per_Monitor_DPI_Aware   = 2
+        assert DPI_AWARENESS in (Process_System_DPI_Aware, Process_DPI_Unaware, Process_Per_Monitor_DPI_Aware)
+        SetProcessDpiAwarenessInternal = windll.user32.SetProcessDpiAwarenessInternal
+        dpiawareness = SetProcessDpiAwarenessInternal(DPI_AWARENESS)
+        screenlog("SetProcessDPIAwareness: %s(%s)=%s", SetProcessDpiAwarenessInternal, DPI_AWARENESS, dpiawareness)
+        assert dpiawareness==0
+    except Exception as e:
+        screenlog("SetProcessDpiAwarenessInternal(%s) failed: %s", DPI_AWARENESS, e)
+        screenlog(" (not available on MS Windows before version 8.1)")
 
 
 def get_native_notifier_classes():
@@ -130,86 +149,193 @@ def get_propsys():
     return _propsys
 
 
+def get_window_handle(window):
+    """ returns the win32 hwnd from a gtk.Window or gdk.Window """
+    gdk_window = window
+    try:
+        gdk_window = window.get_window()
+    except:
+        pass
+    try:
+        return gdk_window.handle
+    except:
+        return None
+
+
+def win32_propsys_set_group_leader(self, leader):
+    """ implements set group leader using propsys """
+    hwnd = get_window_handle(self)
+    if not hwnd:
+        return
+    try:
+        lhandle = leader.handle
+    except:
+        return
+    if not lhandle:
+        return
+    #returns the setter method we can use
+    try:
+        log("win32 hooks: set_group(%#x)", lhandle)
+        propsys = get_propsys()
+        ps = propsys.SHGetPropertyStoreForWindow(hwnd)
+        key = propsys.PSGetPropertyKeyFromName("System.AppUserModel.ID")
+        value = propsys.PROPVARIANTType(lhandle)
+        log("win32 hooks: calling %s(%s, %s)", ps.SetValue, key, value)
+        ps.SetValue(key, value)
+    except Exception as e:
+        log.error("failed to set group leader: %s", e)
+
+WS_NAMES = {
+            win32con.WS_BORDER              : "BORDER",
+            win32con.WS_CAPTION             : "CAPTION",
+            win32con.WS_CHILD               : "CHILD",
+            win32con.WS_CHILDWINDOW         : "CHILDWINDOW",
+            win32con.WS_CLIPCHILDREN        : "CLIPCHILDREN",
+            win32con.WS_CLIPSIBLINGS        : "CLIPSIBLINGS",
+            win32con.WS_DISABLED            : "DISABLED",
+            win32con.WS_DLGFRAME            : "DLGFRAME",
+            win32con.WS_GROUP               : "GROUP",
+            win32con.WS_HSCROLL             : "HSCROLL",
+            win32con.WS_ICONIC              : "ICONIC",
+            win32con.WS_MAXIMIZE            : "MAXIMIZE",
+            win32con.WS_MAXIMIZEBOX         : "MAXIMIZEBOX",
+            win32con.WS_MINIMIZE            : "MINIMIZE",
+            win32con.WS_MINIMIZEBOX         : "MINIMIZEBOX",
+            win32con.WS_OVERLAPPED          : "OVERLAPPED",
+            win32con.WS_POPUP               : "POPUP",
+            win32con.WS_SIZEBOX             : "SIZEBOX",
+            win32con.WS_SYSMENU             : "SYSMENU",
+            win32con.WS_TABSTOP             : "TABSTOP",
+            win32con.WS_THICKFRAME          : "THICKFRAME",
+            win32con.WS_TILED               : "TILED",
+            win32con.WS_VISIBLE             : "VISIBLE",
+            win32con.WS_VSCROLL             : "VSCROLL",
+            }
+
+def style_str(style):
+    return csv([s for c,s in WS_NAMES.items() if (c & style)==c])
+
+def fixup_window_style(self, *args):
+    """ a fixup function we want to call from other places """
+    hwnd = get_window_handle(self)
+    if not hwnd:
+        return
+    try:
+        #warning: accessing "_metadata" on the client window class is fugly..
+        metadata = getattr(self, "_metadata", {})
+        if metadata.get("modal", False):
+            #window is not / no longer meant to be decorated
+            #(this is what GTK does for modal windows - keep it consistent)
+            return
+        cur_style = win32api.GetWindowLong(hwnd, win32con.GWL_STYLE)
+        #re-add taskbar menu:
+        style = cur_style
+        if cur_style & win32con.WS_CAPTION:
+            style |= win32con.WS_SYSMENU
+            style |= win32con.WS_MAXIMIZEBOX
+            style |= win32con.WS_MINIMIZEBOX
+            if style!=cur_style:
+                log("fixup_window_style() using %s (%#x) instead of %s (%#x) on window %#x with metadata=%s", style_str(style), style, style_str(cur_style), cur_style, hwnd, metadata)
+                win32gui.SetWindowLong(hwnd, win32con.GWL_STYLE, style)
+            else:
+                log("fixup_window_style() unchanged style %s (%#x) on window %#x", style_str(style), style, hwnd)
+    except:
+        log.warn("failed to fixup window style", exc_info=True)
+
+def set_decorated(self, decorated):
+    """ override method which ensures that we call
+        fixup_window_style whenever decorations are toggled """
+    self.__set_decorated(decorated)         #call the original saved method
+    self.fixup_window_style()
+
+def window_state_updated(window):
+    """ fixup_window_style whenever the window state changes """
+    log("window_state_updated(%s)", window)
+    fixup_window_style(window)
+
+def apply_maxsize_hints(window, hints):
+    """ extracts the max-size hints from the hints,
+        and passes it to the win32hooks class which can implement it
+        (as GTK2 does not honour it properly on win32)
+    """
+    workw, workh = 0, 0
+    handle = get_window_handle(window)
+    if not window.get_decorated():
+        workarea = get_monitor_workarea_for_window(handle)
+        log("using workarea as window size limit for undecorated window: %s", workarea)
+        if workarea:
+            workw, workh = workarea[2:4]
+    maxw = hints.get("max_width", 0)
+    maxh = hints.get("max_height", 0)
+    if workw>0 and workh>0:
+        #clamp to workspace for undecorated windows:
+        if maxw>0 and maxh>0:
+            maxw = min(workw, maxw)
+            maxh = min(workh, maxh)
+        else:
+            maxw, maxh = workw, workh
+    log("apply_maxsize_hints(%s, %s) found max: %sx%s", window, hints, maxw, maxh)
+    if (maxw>0 and maxw<32767) or (maxh>0 and maxh<32767):
+        window.win32hooks.max_size = (maxw or 32000), (maxh or 32000)
+    elif window.win32hooks.max_size:
+        #was set, clear it
+        window.win32hooks.max_size = None
+    #remove them so GTK doesn't try to set attributes,
+    #which would remove the maximize button:
+    for x in ("max_width", "max_height"):
+        if x in hints:
+            del hints[x]
+
+def apply_geometry_hints(self, hints):
+    log("apply_geometry_hints(%s)", hints)
+    apply_maxsize_hints(self, hints)
+    return self.__apply_geometry_hints(hints)   #call the original saved method
+
 def add_window_hooks(window):
+    log("add_window_hooks(%s) WINDOW_HOOKS=%s, GROUP_LEADER=%s, UNDECORATED_STYLE=%s, MAX_SIZE_HINT=%s, GEOMETRY=%s",
+            window, WINDOW_HOOKS, GROUP_LEADER, UNDECORATED_STYLE, MAX_SIZE_HINT, GEOMETRY)
     if not WINDOW_HOOKS:
         #allows us to disable the win32 hooks for testing
         return
-    #win32 cannot use set_group by default:
     try:
-        window.get_window().set_group = noop
+        gdk_window = window.get_window()
+        #win32 cannot use set_group by default:
+        gdk_window.set_group = noop
     except:
-        pass
-    #gtk2 to window handle:
+        gdk_window = None
+    #window handle:
     try:
-        handle = window.get_window().handle
+        handle = gdk_window.handle
     except:
+        handle = None
+    if not handle:
+        log.warn("cannot add window hooks without a window handle!")
         return
+    log("add_window_hooks(%s) gdk window=%s, hwnd=%#x", window, gdk_window, handle)
 
     if GROUP_LEADER:
         #windows 7 onwards can use AppUserModel to emulate the group leader stuff:
         propsys = get_propsys()
         log("win32 hooks: propsys=%s", propsys)
         if propsys:
-            def set_group(leader):
-                try:
-                    log("win32 hooks: set_group(%#x)", leader.handle)
-                    ps = propsys.SHGetPropertyStoreForWindow(handle)
-                    key = propsys.PSGetPropertyKeyFromName("System.AppUserModel.ID")
-                    value = propsys.PROPVARIANTType(leader.handle)
-                    log("win32 hooks: calling %s(%s, %s)", ps.SetValue, key, value)
-                    ps.SetValue(key, value)
-                except Exception as e:
-                    log.error("failed to set group leader: %s", e)
-            window.get_window().set_group = set_group
+            gdk_window.set_group = types.MethodType(win32_propsys_set_group_leader, gdk_window)
             log("hooked group leader override using %s", propsys)
 
     if UNDECORATED_STYLE:
         #OR windows never have any decorations or taskbar menu
-        readd_window_options = None
         if not window._override_redirect:
+            #the method to call to fix things up:
+            window.fixup_window_style = types.MethodType(fixup_window_style, window)
             #override set_decorated so we can preserve the taskbar menu for undecorated windows
             window.__set_decorated = window.set_decorated
-            def readd_window_options():
-                try:
-                    cur_style = win32api.GetWindowLong(handle, win32con.GWL_STYLE)
-                    #re-add taskbar menu:
-                    style = cur_style
-                    style |= win32con.WS_SYSMENU
-                    style |= win32con.WS_MAXIMIZEBOX
-                    #can always minimize:
-                    style |= win32con.WS_MINIMIZEBOX
-                    log("readd_window_options() using %s style=%#x on window %#x", ["unchanged", "new"][int(style!=cur_style)], style, handle)
-                    if style!=cur_style:
-                        win32gui.SetWindowLong(handle, win32con.GWL_STYLE, style)
-                except:
-                    log.warn("failed to override window style", exc_info=True)
-            def set_decorated(b):
-                window.__set_decorated(b)
-                readd_window_options()
-            window.set_decorated = set_decorated
-            readd_window_options()
+            window.set_decorated = types.MethodType(set_decorated, window, type(window))
             #override after_window_state_updated so we can re-add the missing style options
             #(somehow doing it from on_realize which calls add_window_hooks is not enough)
-            window.__after_window_state_updated = window.after_window_state_updated
-            def after_window_state_updated(*args):
-                log("after_window_state_updated%s", args)
-                window.__after_window_state_updated()
-                readd_window_options()
-            window.after_window_state_updated = after_window_state_updated
+            window.connect("state-updated", window_state_updated)
+            #call it at least once:
+            window.fixup_window_style()
 
-            try:
-                el = get_win32_event_listener(True)
-                log("win32_event_listener=%s", el)
-                if el:
-                    def activate_cb(*args):
-                        log("activate_cb%s", args)
-                        readd_window_options()
-                    el.add_event_callback(win32con.WM_ACTIVATEAPP, activate_cb)
-                    window.activate_app_hook = activate_cb
-            except:
-                log.warn("failed to add activateapp callback", exc_info=True)
-
-    if MAX_SIZE_HINT:
+    if MAX_SIZE_HINT or LANGCHANGE:
         #glue code for gtk to win32 APIs:
         #add event hook class:
         win32hooks = Win32Hooks(handle)
@@ -221,57 +347,25 @@ def add_window_hooks(window):
         if GEOMETRY:
             #save original geometry function:
             window.__apply_geometry_hints = window.apply_geometry_hints
-            #our function for taking gdk window hints and passing them to the win32 hooks class:
-            def apply_maxsize_hints(hints):
-                workw, workh = 0, 0
-                if not window.get_decorated():
-                    workarea = get_monitor_workarea_for_window(handle)
-                    log("using workarea as window size limit for undecorated window: %s", workarea)
-                    if workarea:
-                        workw, workh = workarea[2:4]
-                maxw = hints.get("max_width", 0)
-                maxh = hints.get("max_height", 0)
-                if workw>0 and workh>0:
-                    #clamp to workspace for undecorated windows:
-                    if maxw>0 and maxh>0:
-                        maxw = min(workw, maxw)
-                        maxh = min(workh, maxh)
-                    else:
-                        maxw, maxh = workw, workh
-                log("apply_maxsize_hints(%s) for window %s, found max: %sx%s", hints, window, maxw, maxh)
-                if (maxw>0 and maxw<32767) or (maxh>0 and maxh<32767):
-                    window.win32hooks.max_size = (maxw or 32000), (maxh or 32000)
-                elif window.win32hooks.max_size:
-                    #was set, clear it
-                    window.win32hooks.max_size = None
-                #remove them so GTK doesn't try to set attributes,
-                #which would remove the maximize button:
-                for x in ("max_width", "max_height"):
-                    if x in hints:
-                        del hints[x]
-            #our monkey patching method, which calls the function above:
-            def apply_geometry_hints(hints):
-                apply_maxsize_hints(hints)
-                return window.__apply_geometry_hints(hints)
-            window.apply_geometry_hints = apply_geometry_hints
-            #apply current geometry hints, if any:
+            window.apply_geometry_hints = types.MethodType(apply_geometry_hints, window)
+            #apply current max-size from hints, if any:
             if window.geometry_hints:
-                apply_maxsize_hints(window.geometry_hints)
+                apply_maxsize_hints(window, window.geometry_hints)
+
+        if LANGCHANGE:
+            def inputlangchange(hwnd, event, wParam, lParam):
+                log("WM_INPUTLANGCHANGE: character set: %i, input locale identifier: %i", wParam, lParam)
+                window.keyboard_layout_changed("WM_INPUTLANGCHANGE", wParam, lParam)
+            win32hooks.add_window_event_handler(win32con.WM_INPUTLANGCHANGE, inputlangchange)
+
 
 def remove_window_hooks(window):
     try:
-        if hasattr(window, "win32hooks"):
-            win32hooks = window.win32hooks
+        win32hooks = getattr(window, "win32hooks", None)
+        if win32hooks:
             log("remove_window_hooks(%s) found %s", window, win32hooks)
-            if win32hooks:
-                win32hooks.cleanup()
-                window.win32hooks = None
-        if hasattr(window, "activate_app_hook"):
-            activate_app_cb = window.activate_app_hook
-            if activate_app_cb:
-                el = get_win32_event_listener(True)
-                el.remove_event_callback(win32con.WM_ACTIVATEAPP, activate_app_cb)
-            window.activate_app_hook = None
+            win32hooks.cleanup()
+            window.win32hooks = None
     except:
         log.error("remove_window_hooks(%s)", exc_info=True)
 
@@ -290,13 +384,6 @@ def get_ydpi():
         log.warn("failed to get ydpi: %s", e)
     return -1
 
-def get_dpi():
-    try:
-        return (get_xdpi() + get_ydpi())//2
-    except Exception as e:
-        log.warn("failed to get dpi: %s", e)
-    return -1
-
 #those constants aren't found in win32con:
 SPI_GETFONTSMOOTHING            = 0x004A
 SPI_GETFONTSMOOTHINGCONTRAST    = 0x200C
@@ -384,6 +471,7 @@ def get_workarea():
 
 #ie: for a 60 pixel bottom bar on the second monitor at 1280x800:
 # [(0,0,1920,1080), (0,0,1280,740)]
+MONITORINFOF_PRIMARY = 1
 def get_workareas():
     try:
         workareas = []
@@ -400,7 +488,13 @@ def get_workareas():
             rx2 = max(0, min(mx2-mx1, wx2-mx1))
             ry2 = max(0, min(my2-my1, wy2-my1))
             assert rx1<rx2 and ry1<ry2, "invalid relative workarea coordinates"
-            workareas.append((rx1, ry1, rx2-rx1, ry2-ry1))
+            geom = rx1, ry1, rx2-rx1, ry2-ry1
+            #GTK will return the PRIMARY monitor first,
+            #so we have to do the same thing:
+            if mi['Flags'] & MONITORINFOF_PRIMARY:
+                workareas.insert(0, geom)
+            else:
+                workareas.append(geom)
         assert len(workareas)>0
         screenlog("get_workareas()=%s", workareas)
         return workareas
@@ -451,6 +545,11 @@ def get_fixed_cursor_size():
         #best to try to use a limit anyway:
         return 32, 32
 
+def get_cursor_size():
+    w,h = get_fixed_cursor_size()
+    return (w+h)//2
+
+
 def get_window_frame_sizes():
     try:
         #normal resizable windows:
@@ -475,12 +574,23 @@ def get_window_frame_sizes():
                 "menu-bar"  : m,
                 "border"    : b,
                 "caption"   : c,
+                "offset"    : (rx, ry+c),
+                #left, right, top, bottom:
+                "frame"     : (rx, rx, ry+c, ry),
                 }
     except Exception as e:
         log.warn("failed to get window frame size information: %s", e)
         return None
 
+def get_virtualscreenmetrics():
+    dx = win32api.GetSystemMetrics(win32con.SM_XVIRTUALSCREEN)
+    dy = win32api.GetSystemMetrics(win32con.SM_YVIRTUALSCREEN)
+    dw = win32api.GetSystemMetrics(win32con.SM_CXVIRTUALSCREEN)
+    dh = win32api.GetSystemMetrics(win32con.SM_CYVIRTUALSCREEN)
+    return dx, dy, dw, dh
+
 def take_screenshot():
+    #would be better to refactor the code..
     from xpra.platform.win32.shadow_server import Win32RootWindowModel
     rwm = Win32RootWindowModel(object())
     return rwm.take_screenshot()
@@ -502,55 +612,131 @@ def show_desktop(b):
         log.warn("failed to call show_desktop(%s): %s", b, e)
 
 
+WM_WTSSESSION_CHANGE = 0x02b1
+WTS_CONSOLE_CONNECT         = 0x1
+WTS_CONSOLE_DISCONNECT      = 0x2
+WTS_REMOTE_CONNECT          = 0x3
+WTS_REMOTE_DISCONNECT       = 0x4
+WTS_SESSION_LOGON           = 0x5
+WTS_SESSION_LOGOFF          = 0x6
+WTS_SESSION_LOCK            = 0x7
+WTS_SESSION_UNLOCK          = 0x8
+WTS_SESSION_REMOTE_CONTROL  = 0x9
+WTS_SESSION_EVENTS = {
+                      WTS_CONSOLE_CONNECT       : "CONSOLE CONNECT",
+                      WTS_CONSOLE_DISCONNECT    : "CONSOLE_DISCONNECT",
+                      WTS_REMOTE_CONNECT        : "REMOTE_CONNECT",
+                      WTS_REMOTE_DISCONNECT     : "REMOTE_DISCONNECT",
+                      WTS_SESSION_LOGON         : "SESSION_LOGON",
+                      WTS_SESSION_LOGOFF        : "SESSION_LOGOFF",
+                      WTS_SESSION_LOCK          : "SESSION_LOCK",
+                      WTS_SESSION_UNLOCK        : "SESSION_UNLOCK",
+                      WTS_SESSION_REMOTE_CONTROL: "SESSION_REMOTE_CONTROL",
+                      }
+
 class ClientExtras(object):
     def __init__(self, client, opts):
         self.client = client
         self._kh_warning = False
-        self.setup_console_event_listener()
+        self._console_handler_registered = self.setup_console_event_listener(True)
         try:
             import win32con                 #@Reimport @UnresolvedImport
             el = get_win32_event_listener(True)
             if el:
-                el.add_event_callback(win32con.WM_ACTIVATEAPP, self.activateapp)
-                el.add_event_callback(win32con.WM_POWERBROADCAST, self.power_broadcast_event)
+                el.add_event_callback(win32con.WM_ACTIVATEAPP,      self.activateapp)
+                el.add_event_callback(win32con.WM_POWERBROADCAST,   self.power_broadcast_event)
+                el.add_event_callback(win32con.WM_MOVE,             self.wm_move)
+                el.add_event_callback(WM_WTSSESSION_CHANGE,         self.session_change_event)
+                el.add_event_callback(win32con.WM_INPUTLANGCHANGE,  self.inputlangchange)
+                el.add_event_callback(win32con.WM_WININICHANGE,     self.inichange)
         except Exception as e:
             log.error("cannot register focus and power callbacks: %s", e)
 
     def cleanup(self):
-        self.setup_console_event_listener(False)
-        log("ClientExtras.cleanup() ended")
+        log("ClientExtras.cleanup()")
+        if self._console_handler_registered:
+            self._console_handler_registered = False
+            self.setup_console_event_listener(False)
         el = get_win32_event_listener(False)
         if el:
             el.cleanup()
-        self.client = None
+        log("ClientExtras.cleanup() ended")
+        #self.client = None
+
+    def wm_move(self, wParam, lParam):
+        c = self.client
+        log("WM_MOVE: %s/%s client=%s", wParam, lParam, c)
+        if c:
+            #this is not really a screen size change event,
+            #but we do want to process it as such (see window reinit code)
+            c.screen_size_changed()
+
+    def session_change_event(self, event, session):
+        event_name = WTS_SESSION_EVENTS.get(event, event)
+        log("WM_WTSSESSION_CHANGE: %s on session %#x", event_name, session)
+        c = self.client
+        if c and event in (WTS_SESSION_LOGOFF, WTS_SESSION_LOCK):
+            log("will freeze all the windows")
+            c.freeze()
+
+
+    def inputlangchange(self, wParam, lParam):
+        log("WM_INPUTLANGCHANGE: %i, %i", wParam, lParam)
+
+    def inichange(self, wParam, lParam):
+        if lParam:
+            from ctypes import c_char_p
+            log("WM_WININICHANGE: %#x=%s", lParam, c_char_p(lParam).value)
+        else:
+            log("WM_WININICHANGE: %i, %i", wParam, lParam)
+
 
     def activateapp(self, wParam, lParam):
-        log("WM_ACTIVATEAPP: %s/%s client=%s", wParam, lParam, self.client)
-        if wParam==0 and self.client:
+        c = self.client
+        log("WM_ACTIVATEAPP: %s/%s client=%s", wParam, lParam, c)
+        if not c:
+            return
+        if wParam==0:
             #our app has lost focus
-            self.client.update_focus(0, False)
+            c.update_focus(0, False)
+        #workaround for windows losing their style:
+        for window in c._id_to_window.values():
+            fixup_window_style = getattr(window, "fixup_window_style", None)
+            if fixup_window_style:
+                fixup_window_style()
+
 
     def power_broadcast_event(self, wParam, lParam):
-        log("WM_POWERBROADCAST: %s/%s client=%s", POWER_EVENTS.get(wParam, wParam), lParam, self.client)
+        c = self.client
+        log("WM_POWERBROADCAST: %s/%s client=%s", POWER_EVENTS.get(wParam, wParam), lParam, c)
         #maybe also "PBT_APMQUERYSUSPEND" and "PBT_APMQUERYSTANDBY"?
-        if wParam==win32con.PBT_APMSUSPEND and self.client:
-            self.client.suspend()
+        if wParam==win32con.PBT_APMSUSPEND and c:
+            c.suspend()
         #According to the documentation:
         #The system always sends a PBT_APMRESUMEAUTOMATIC message whenever the system resumes.
-        elif wParam==win32con.PBT_APMRESUMEAUTOMATIC and self.client:
-            self.client.resume()
+        elif wParam==win32con.PBT_APMRESUMEAUTOMATIC and c:
+            c.resume()
 
-    def setup_console_event_listener(self, enable=1):
+    def setup_console_event_listener(self, enable=True):
         try:
-            result = win32api.SetConsoleCtrlHandler(self.handle_console_event, enable)
+            v = self.handle_console_event
+            if not enable:
+                v = None
+            log("calling win32api.SetConsoleCtrlHandler(%s, %s)", v, enable)
+            result = win32api.SetConsoleCtrlHandler(v, int(enable))
             if result == 0:
                 log.error("could not SetConsoleCtrlHandler (error %r)", win32api.GetLastError())
-        except:
-            pass
+                return False
+            return True
+        except Exception as e:
+            log.error("SetConsoleCtrlHandler error: %s", e)
+            return False
+
 
     def handle_console_event(self, event):
-        log("handle_console_event(%s)", event)
+        c = self.client
         event_name = KNOWN_EVENTS.get(event, event)
+        log("handle_console_event(%s) client=%s, event_name=%s", event, c, event_name)
         info_events = [win32con.CTRL_C_EVENT,
                        win32con.CTRL_LOGOFF_EVENT,
                        win32con.CTRL_BREAK_EVENT,
@@ -560,9 +746,9 @@ class ClientExtras(object):
             log.info("received console event %s", str(event_name).replace("_EVENT", ""))
         else:
             log.warn("unknown console event: %s", event_name)
-        c = self.client
         if event==win32con.CTRL_C_EVENT:
             if c:
+                log("calling=%s", c.signal_disconnect_and_quit)
                 c.signal_disconnect_and_quit(0, "CTRL_C")
                 return 1
         if event==win32con.CTRL_CLOSE_EVENT:
@@ -576,12 +762,17 @@ def main():
     from xpra.platform import init, clean
     try:
         init("Platform-Events", "Platform Events Test")
-        import sys
         if "-v" in sys.argv or "--verbose" in sys.argv:
             from xpra.platform.win32.win32_events import log as win32_event_logger
             log.enable_debug()
             win32_event_logger.enable_debug()
 
+        import gobject
+        gobject.threads_init()      #@UndefinedVariable
+
+        log.info("Event loop is running")
+        loop = gobject.MainLoop()
+
         def suspend():
             log.info("suspend event")
         def resume():
@@ -591,13 +782,11 @@ def main():
         fake_client.suspend = suspend
         fake_client.resume = resume
         fake_client.keyboard_helper = None
+        def signal_quit(*args):
+            loop.quit()
+        fake_client.signal_disconnect_and_quit = signal_quit
         ClientExtras(fake_client, None)
 
-        import gobject
-        gobject.threads_init()
-
-        log.info("Event loop is running")
-        loop = gobject.MainLoop()
         try:
             loop.run()
         except KeyboardInterrupt:
