@@ -4,6 +4,10 @@
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+import win32api         #@UnresolvedImport
+import win32con         #@UnresolvedImport
+import win32gui         #@UnresolvedImport
+import os
 
 from xpra.platform.keyboard_base import KeyboardBase
 from xpra.keyboard.layouts import WIN32_LAYOUTS
@@ -11,6 +15,13 @@ from xpra.gtk_common.keymap import KEY_TRANSLATIONS
 from xpra.log import Logger
 log = Logger("keyboard")
 
+EMULATE_ALTGR = os.environ.get("XPRA_EMULATE_ALTGR", "1")=="1"
+EMULATE_ALTGR_CONTROL_KEY_DELAY = int(os.environ.get("XPRA_EMULATE_ALTGR_CONTROL_KEY_DELAY", "50"))
+if EMULATE_ALTGR:
+    #needed for altgr emulation timeouts:
+    from xpra.gtk_common.gobject_compat import import_glib
+    glib = import_glib()
+
 
 class Keyboard(KeyboardBase):
     """ This is for getting keys from the keyboard on the client side.
@@ -21,9 +32,9 @@ class Keyboard(KeyboardBase):
 
     def __init__(self):
         KeyboardBase.__init__(self)
-        self.emulate_altgr = False
         self.num_lock_modifier = None
-        self.last_key_event_sent = None
+        self.altgr_modifier = None
+        self.delayed_event = None
         #workaround for "period" vs "KP_Decimal" with gtk2 (see ticket #586):
         #translate "period" with keyval=46 and keycode=110 to KP_Decimal:
         KEY_TRANSLATIONS[("period",     46,     110)]   = "KP_Decimal"
@@ -34,17 +45,22 @@ class Keyboard(KeyboardBase):
     def set_modifier_mappings(self, mappings):
         KeyboardBase.set_modifier_mappings(self, mappings)
         self.num_lock_modifier = self.modifier_keys.get("Num_Lock")
-        log("set_modifier_mappings found 'Num_Lock' modifier value: %s", self.num_lock_modifier)
+        log("set_modifier_mappings found 'Num_Lock' with modifier value: %s", self.num_lock_modifier)
+        for x in ("ISO_Level3_Shift", "Mode_switch"):
+            mod = self.modifier_keys.get(x)
+            if mod:
+                self.altgr_modifier = mod
+                log("set_modifier_mappings found 'AltGr'='%s' with modifier value: %s", x, self.altgr_modifier)
+                break
 
     def mask_to_names(self, mask):
         """ Patch NUMLOCK and AltGr """
         names = KeyboardBase.mask_to_names(self, mask)
-        if self.emulate_altgr:
+        altgr = win32api.GetKeyState(win32con.VK_RMENU) not in (0, 1)
+        if altgr:
             self.AltGr_modifiers(names)
         if self.num_lock_modifier:
             try:
-                import win32api         #@UnresolvedImport
-                import win32con         #@UnresolvedImport
                 numlock = win32api.GetKeyState(win32con.VK_NUMLOCK)
                 if numlock and self.num_lock_modifier not in names:
                     names.append(self.num_lock_modifier)
@@ -60,11 +76,12 @@ class Keyboard(KeyboardBase):
     def AltGr_modifiers(self, modifiers, pressed=True):
         add = []
         clear = ["mod1", "mod2", "control"]
-        if pressed:
-            add.append("mod5")
-        else:
-            clear.append("mod5")
-        log("AltGr_modifiers(%s, %s) add=%s, clear=%s", modifiers, pressed, add, clear)
+        if self.altgr_modifier:
+            if pressed:
+                add.append(self.altgr_modifier)
+            else:
+                clear.append(self.altgr_modifier)
+        log("AltGr_modifiers(%s, %s) AltGr=%s, add=%s, clear=%s", modifiers, pressed, self.altgr_modifier, add, clear)
         for x in add:
             if x not in modifiers:
                 modifiers.append(x)
@@ -85,7 +102,6 @@ class Keyboard(KeyboardBase):
         variant = None
         variants = None
         try:
-            import win32api         #@UnresolvedImport
             kbid = win32api.GetKeyboardLayout(0) & 0xffff
             if kbid in WIN32_LAYOUTS:
                 code, _, _, _, layout, variants = WIN32_LAYOUTS.get(kbid)
@@ -100,8 +116,6 @@ class Keyboard(KeyboardBase):
 
     def get_keyboard_repeat(self):
         try:
-            import win32con         #@UnresolvedImport
-            import win32gui         #@UnresolvedImport
             _delay = win32gui.SystemParametersInfo(win32con.SPI_GETKEYBOARDDELAY)
             _speed = win32gui.SystemParametersInfo(win32con.SPI_GETKEYBOARDSPEED)
             #now we need to normalize those weird win32 values:
@@ -131,19 +145,33 @@ class Keyboard(KeyboardBase):
         #self.modifier_keycodes = {"ISO_Level3_Shift": [108]}
         #we can only deal with 'Alt_R' and simulate AltGr (ISO_Level3_Shift)
         #if we have modifier_mappings
-        if key_event.keyname=="Alt_R" and len(self.modifier_mappings)>0:
-            keyname = "ISO_Level3_Shift"
-            altgr_keycodes = self.modifier_keycodes.get(keyname, [])
-            if len(altgr_keycodes)>0:
-                self.emulate_altgr = key_event.pressed
-                if key_event.pressed and self.last_key_event_sent:
-                    #check for spurious control and undo it
-                    last_wid, last_key_event = self.last_key_event_sent
-                    if last_wid==wid and last_key_event.keyname=="Control_L" and last_key_event.pressed==True:
-                        #undo it:
-                        last_key_event.pressed = False
-                        KeyboardBase.process_key_event(self, send_key_action_cb, last_wid, last_key_event)
-                self.AltGr_modifiers(key_event.modifiers, not key_event.pressed)
-        self.last_key_event_sent = (wid, key_event)
-        #now fallback to default behaviour:
+        if EMULATE_ALTGR and self.altgr_modifier and len(self.modifier_mappings)>0:
+            if key_event.keyname=="Control_L":
+                altgr = win32api.GetKeyState(win32con.VK_RMENU) not in (0, 1)
+                #AltGr is often preceded by a spurious "Control_L" event
+                #delay this one a little bit so we can skip it if an "AltGr" does come through next:
+                if key_event.pressed:
+                    if not altgr:
+                        self.delayed_event = (send_key_action_cb, wid, key_event)
+                        glib.timeout_add(EMULATE_ALTGR_CONTROL_KEY_DELAY, self.send_delayed_key)
+                    return
+                if not key_event.pressed and altgr:
+                    #unpressed: could just skip it?
+                    #(but maybe the real one got pressed.. and this would get it stuck)
+                    pass
+            if key_event.keyname=="Alt_R" :
+                #cancel "Control_L" if one was due:
+                self.delayed_event = None
+                return
+        self.send_delayed_key()
         KeyboardBase.process_key_event(self, send_key_action_cb, wid, key_event)
+
+    def send_delayed_key(self):
+        #timeout: this must be a real one, send it now
+        dk = self.delayed_event
+        if dk:
+            self.delayed_event = None
+            altgr = win32api.GetKeyState(win32con.VK_RMENU) not in (0, 1)
+            if altgr:
+                send_key_action_cb, wid, key_event = dk
+                KeyboardBase.process_key_event(self, send_key_action_cb, wid, key_event)
