@@ -4,19 +4,25 @@
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+import os
 import time
 import win32api         #@UnresolvedImport
 import win32con         #@UnresolvedImport
 import win32ui          #@UnresolvedImport
 import win32gui         #@UnresolvedImport
+import win32process     #@UnresolvedImport
+
 from xpra.log import Logger
+from xpra.util import AdHocStruct
 log = Logger("shadow", "win32")
-keylog = Logger("keyboard", "shadow", "win32")
+shapelog = Logger("shape")
 
 from xpra.os_util import StringIOClass
 from xpra.server.gtk_server_base import GTKServerBase
-from xpra.server.shadow_server_base import ShadowServerBase, RootWindowModel
+from xpra.server.shadow.shadow_server_base import ShadowServerBase
+from xpra.server.shadow.root_window_model import RootWindowModel
 from xpra.platform.win32.keyboard_config import KeyboardConfig, fake_key
+from xpra.platform.win32.gui import get_virtualscreenmetrics
 from xpra.codecs.image_wrapper import ImageWrapper
 
 NOEVENT = object()
@@ -34,6 +40,8 @@ BUTTON_EVENTS = {
                  (5, False) : NOEVENT,
                  }
 
+SEAMLESS = os.environ.get("XPRA_WIN32_SEAMLESS", "0")=="1"
+
 
 class Win32RootWindowModel(RootWindowModel):
 
@@ -41,23 +49,123 @@ class Win32RootWindowModel(RootWindowModel):
         RootWindowModel.__init__(self, root)
         self.metrics = None
         self.ddc, self.cdc, self.memdc, self.bitmap = None, None, None, None
+        if SEAMLESS:
+            self.property_names.append("shape")
+            self.dynamic_property_names.append("shape")
+            self.rectangles = self.get_shape_rectangles(logit=True)
+            self.shape_notify = []
+
+    def refresh_shape(self):
+        rectangles = self.get_shape_rectangles()
+        if rectangles==self.rectangles:
+            return  #unchanged
+        self.rectangles = rectangles
+        shapelog("refresh_shape() sending notify for updated rectangles: %s", rectangles)
+        #notify listeners:
+        pspec = AdHocStruct()
+        pspec.name = "shape"
+        for cb, args in self.shape_notify:
+            shapelog("refresh_shape() notifying: %s", cb)
+            try:
+                cb(self, pspec, *args)
+            except:
+                shapelog.error("error in shape notify callback %s", cb, exc_info=True)
+
+    def connect(self, signal, cb, *args):
+        if signal=="notify::shape":
+            self.shape_notify.append((cb, args))
+        else:
+            RootWindowModel.connect(self, signal, cb, *args)
+
+    def get_shape_rectangles(self, logit=False):
+        #get the list of windows
+        l = log
+        if logit or os.environ.get("XPRA_SHAPE_DEBUG", "0")=="1":
+            l = shapelog
+        taskbar = win32gui.FindWindow("Shell_TrayWnd", None)
+        l("taskbar window=%#x", taskbar)
+        ourpid = os.getpid()
+        l("our pid=%i", ourpid)
+        def enum_windows_cb(hwnd, rects):
+            if not win32gui.IsWindowVisible(hwnd):
+                l("skipped invisible window %#x", hwnd)
+                return True
+            _, pid = win32process.GetWindowThreadProcessId(hwnd)
+            if pid==ourpid:
+                l("skipped our own window %#x", hwnd)
+                return True
+            #skipping IsWindowEnabled check
+            window_title = win32gui.GetWindowText(hwnd)
+            l("get_shape_rectangles() found window '%s' with pid=%s", window_title, pid)
+            left, top, right, bottom = win32gui.GetWindowRect(hwnd)
+            if right<0 or bottom<0:
+                l("skipped offscreen window at %ix%i", right, bottom)
+                return True
+            if hwnd==taskbar:
+                l("skipped taskbar")
+                return True
+            #dirty way:
+            if window_title=='Program Manager':
+                return True
+            #this should be the proper way using GetTitleBarInfo (but does not seem to work)
+            #import ctypes
+            #from ctypes.windll.user32 import GetTitleBarInfo        #@UnresolvedImport
+            #from ctypes.wintypes import (DWORD, RECT)
+            #class TITLEBARINFO(ctypes.Structure):
+            #    pass
+            #TITLEBARINFO._fields_ = [
+            #    ('cbSize', DWORD),
+            #    ('rcTitleBar', RECT),
+            #    ('rgstate', DWORD * 6),
+            #]
+            #ti = TITLEBARINFO()
+            #ti.cbSize = ctypes.sizeof(ti)
+            #GetTitleBarInfo(hwnd, ctypes.byref(ti))
+            #if ti.rgstate[0] & win32con.STATE_SYSTEM_INVISIBLE:
+            #    log("skipped system invisible window")
+            #    return True
+            w = right-left
+            h = bottom-top
+            l("shape(%s - %#x)=%s", window_title, hwnd, (left, top, w, h))
+            if w<=0 and h<=0:
+                l("skipped invalid window size: %ix%i", w, h)
+                return True
+            if left==-32000 and top==-32000:
+                #there must be a better way of skipping those - I haven't found it
+                l("skipped special window")
+                return True
+            #now clip rectangle:
+            if left<0:
+                left = 0
+                w = right
+            if top<0:
+                top = 0
+                h = bottom
+            rects.append((left, top, w, h))
+            return True
+        rectangles = []
+        win32gui.EnumWindows(enum_windows_cb, rectangles)
+        l("get_shape_rectangles()=%s", rectangles)
+        return sorted(rectangles)
+
+    def get_property(self, prop):
+        if prop=="shape":
+            assert SEAMLESS
+            shape = {"Bounding.rectangles" : self.rectangles}
+            #provide clip rectangle? (based on workspace area?)
+            return shape
+        return RootWindowModel.get_property(self, prop)
+
 
     def get_root_window_size(self):
         w = win32api.GetSystemMetrics(win32con.SM_CXVIRTUALSCREEN)
         h = win32api.GetSystemMetrics(win32con.SM_CYVIRTUALSCREEN)
         return w, h
 
-    def get_metrics(self):
-        dx = win32api.GetSystemMetrics(win32con.SM_XVIRTUALSCREEN)
-        dy = win32api.GetSystemMetrics(win32con.SM_YVIRTUALSCREEN)
-        dw = win32api.GetSystemMetrics(win32con.SM_CXVIRTUALSCREEN)
-        dh = win32api.GetSystemMetrics(win32con.SM_CYVIRTUALSCREEN)
-        return dx, dy, dw, dh
-
     def get_image(self, x, y, width, height, logger=None):
         start = time.time()
         desktop_wnd = win32gui.GetDesktopWindow()
-        metrics = self.get_metrics()
+        metrics = get_virtualscreenmetrics()
         if self.metrics is None or self.metrics!=metrics:
             #new metrics, start from scratch:
             self.metrics = metrics
@@ -101,8 +209,8 @@ class Win32RootWindowModel(RootWindowModel):
         return v
 
     def take_screenshot(self):
-        from PIL import Image
-        x, y, w, h = self.get_metrics()
+        from PIL import Image               #@UnresolvedImport
+        x, y, w, h = get_virtualscreenmetrics()
         image = self.get_image(x, y, w, h)
         assert image.get_width()==w and image.get_height()==h
         assert image.get_pixel_format()=="BGRX"
@@ -113,9 +221,12 @@ class Win32RootWindowModel(RootWindowModel):
         out.close()
         return screenshot
 
+
 class ShadowServer(ShadowServerBase, GTKServerBase):
 
     def __init__(self):
+        #TODO: root should be a wrapper for the win32 system metrics bits?
+        #(or even not bother passing root to ShadowServerBase?
         import gtk.gdk
         ShadowServerBase.__init__(self, gtk.gdk.get_default_root_window())
         GTKServerBase.__init__(self)
@@ -124,6 +235,13 @@ class ShadowServer(ShadowServerBase, GTKServerBase):
     def makeRootWindowModel(self):
         return Win32RootWindowModel(self.root)
 
+    def refresh(self):
+        v = ShadowServerBase.refresh(self)
+        if v and SEAMLESS:
+            self.root_window_model.refresh_shape()
+        log("refresh()=%s", v)
+        return v
+
     def _process_mouse_common(self, proto, wid, pointer, modifiers):
         #adjust pointer position for offset in client:
         x, y = pointer
@@ -164,7 +282,6 @@ class ShadowServer(ShadowServerBase, GTKServerBase):
         return info
 
 
-
 def main():
     from xpra.platform import init, clean
     try:
@@ -181,4 +298,3 @@ def main():
 
 if __name__ == "__main__":
     main()
-
