@@ -13,7 +13,8 @@ from xpra.platform.win32.wndproc_events import WNDPROC_EVENT_NAMES
 
 #no idea where we're supposed to get those from:
 WM_WTSSESSION_CHANGE        = 0x02b1
-WM_DWMNCRENDERINGCHANGED    = 0x31F
+WM_DWMNCRENDERINGCHANGED    = 0x031F
+WM_DWMCOMPOSITIONCHANGED    = 0x031E
 IGNORE_EVENTS = {
             win32con.WM_DESTROY             : "WM_DESTROY",
             win32con.WM_COMMAND             : "WM_COMMAND",
@@ -23,7 +24,8 @@ IGNORE_EVENTS = {
             win32con.WM_WINDOWPOSCHANGED    : "WM_WINDOWPOSCHANGED",    #happens after resume too?
             win32con.WM_WININICHANGE        : "WM_WININICHANGE",        #happens after resume too?
             win32con.WM_WINDOWPOSCHANGING   : "WM_WINDOWPOSCHANGING",
-            win32con.WM_GETMINMAXINFO       : "WM_GETMINMAXINFO",       #could be used to limit window size?
+            win32con.WM_GETMINMAXINFO       : "WM_GETMINMAXINFO",
+            win32con.WM_SYSCOLORCHANGE      : "WM_SYSCOLORCHANGE",
             WM_WTSSESSION_CHANGE            : "WM_WTSSESSION_CHANGE",
             WM_DWMNCRENDERINGCHANGED        : "WM_DWMNCRENDERINGCHANGED",
             800                             : "screen background changed",  #I can't find this definition anywhere
@@ -31,6 +33,8 @@ IGNORE_EVENTS = {
 LOG_EVENTS = {
             win32con.WM_POWERBROADCAST      : "WM_POWERBROADCAST: power management event",
             win32con.WM_TIMECHANGE          : "WM_TIMECHANGE: time change event",
+            win32con.WM_INPUTLANGCHANGE     : "WM_INPUTLANGCHANGE: input language changed",
+            WM_DWMCOMPOSITIONCHANGED        : "WM_DWMCOMPOSITIONCHANGED: Desktop Window Manager composition has been enabled or disabled",
             }
 KNOWN_WM_EVENTS = IGNORE_EVENTS.copy()
 KNOWN_WM_EVENTS.update(WNDPROC_EVENT_NAMES)
@@ -90,17 +94,22 @@ class Win32EventListener(object):
             l.remove(callback)
 
     def cleanup(self):
+        log("Win32EventListener.cleanup()")
         self.event_callback = {}
         self.stop_win32_session_events()
         if self.hwnd:
             try:
                 win32gui.DestroyWindow(self.hwnd)
             except Exception as e:
-                log.warn("error during win32 events cleanup: %s", e)
+                log.warn("error during win32 events cleanup of event window instance: %s", e)
             self.hwnd = None
-            win32gui.UnregisterClass(self.wc.lpszClassName, None)
+            try:
+                win32gui.UnregisterClass(self.wc.lpszClassName, None)
+            except Exception as e:
+                log.warn("error during win32 events cleanup of event window class: %s", e)
 
     def stop_win32_session_events(self):
+        log("stop_win32_session_events() old win32 proc=%s", self.old_win32_proc)
         if not self.old_win32_proc:
             return
         try:
@@ -135,7 +144,7 @@ class Win32EventListener(object):
 
     def MyWndProc(self, hWnd, msg, wParam, lParam):
         callbacks = self.event_callbacks.get(msg)
-        event_name = KNOWN_WM_EVENTS.get(msg, msg)
+        event_name = KNOWN_WM_EVENTS.get(msg, hex(msg))
         log("callbacks for event %s: %s", event_name, callbacks)
         if hWnd==self.hwnd:
             if callbacks:
@@ -151,7 +160,19 @@ class Win32EventListener(object):
             #elif msg==win32con.WM_ACTIVATEAPP:
             #    log("WM_ACTIVATEAPP focus changed: %s / %s", wParam, lParam)
             else:
-                log.warn("unexpected message: %s / %s / %s", event_name, wParam, lParam)
+                l = log.warn
+                if (msg>=0 and msg<=win32con.WM_USER) or msg>0xFFFF:
+                    ut = "reserved system"
+                elif msg>=win32con.WM_USER and msg<=0x7FFF:
+                    ut = "WM_USER"
+                elif msg>=0x8000 and msg<=0xBFFF:
+                    ut = "WM_APP"
+                elif msg>=0xC000 and msg<=0xFFFF:
+                    ut = "string"
+                    l = log.info
+                else:
+                    ut = "/ unexpected"
+                l("unknown %s message: %s / %s / %s", ut, event_name, wParam, lParam)
         else:
             log.warn("invalid hwnd: %s (expected %s)", hWnd, self.hwnd)
         # Pass all messages to the original WndProc
