@@ -48,9 +48,9 @@ class Win32Hooks(object):
     def __init__(self, hwnd):
         self._hwnd = hwnd
         self._message_map = {}
-        if HOOK_MINMAXINFO:
-            self._message_map[win32con.WM_GETMINMAXINFO] = self.on_getminmaxinfo
         self.max_size = None
+        if HOOK_MINMAXINFO:
+            self.add_window_event_handler(win32con.WM_GETMINMAXINFO, self.on_getminmaxinfo)
         try:
             #we only use this code for resizable windows, so use SM_C?SIZEFRAME:
             self.frame_width = win32api.GetSystemMetrics(win32con.SM_CXSIZEFRAME)
@@ -64,13 +64,16 @@ class Win32Hooks(object):
         log("Win32Hooks: message_map=%s", self._message_map)
         self._oldwndproc = None
 
+    def add_window_event_handler(self, event, handler):
+        self._message_map[event] = handler
+
     def setup(self):
         assert self._oldwndproc is None
         self._newwndproc = WndProcType(self._wndproc)
         self._oldwndproc = SetWindowLong(self._hwnd, win32con.GWL_WNDPROC, self._newwndproc)
 
     def on_getminmaxinfo(self, hwnd, msg, wparam, lparam):
-        if self.max_size:
+        if self.max_size and lparam:
             info = ctypes.cast(lparam, ctypes.POINTER(MINMAXINFO)).contents
             width, height = self.max_size
             style = win32api.GetWindowLong(hwnd, win32con.GWL_STYLE)
@@ -83,9 +86,9 @@ class Win32Hooks(object):
             point  = POINT(w, h)
             info.ptMaxSize       = point
             info.ptMaxTrackSize  = point
-            log("on_getminmaxinfo%s max_size=%s, frame=%sx%s, minmaxinfo size=%sx%s", (hwnd, msg, wparam, lparam), self.max_size, fw, fh, w, h)
+            log("on_getminmaxinfo window=%#x max_size=%s, frame=%sx%s, minmaxinfo size=%sx%s", hwnd, self.max_size, fw, fh, w, h)
         else:
-            log("on_getminmaxinfo%s max_size=%s", (hwnd, msg, wparam, lparam), self.max_size)
+            log("on_getminmaxinfo window=%#x max_size=%s", hwnd, self.max_size)
 
     def cleanup(self, *args):
         log("cleanup%s", args)
