@@ -6,13 +6,12 @@
 
 # Ubuntu re-invents the wheel, and it's a broken one
 
-import sys
 import os
 
 from xpra.log import Logger
 log = Logger("tray", "posix")
 
-from xpra.util import is_unity
+from xpra.os_util import is_unity
 from xpra.client.tray_base import TrayBase
 from xpra.platform.paths import get_icon_dir
 
@@ -22,12 +21,13 @@ def get_appindicator():
     global _appindicator
     if _appindicator is False:
         try:
-            if sys.version<'3':
-                import appindicator                     #@UnresolvedImport
-                _appindicator = appindicator
-            else:
+            import sys
+            if "gi" in sys.modules:
                 from gi.repository import AppIndicator3 #@UnresolvedImport @Reimport
                 _appindicator = AppIndicator3
+            else:
+                import appindicator                     #@UnresolvedImport
+                _appindicator = appindicator
         except ImportError:
             _appindicator = None
     return _appindicator
@@ -117,8 +117,8 @@ def main():
     if not can_use_appindicator():
         log("appindicator may not be shown...")
 
-    from xpra.gtk_common.gobject_compat import import_gobject, import_gtk
-    gobject = import_gobject()
+    from xpra.gtk_common.gobject_compat import import_glib, import_gtk
+    glib = import_glib()
     gtk = import_gtk()
 
     menu = gtk.Menu()
@@ -127,7 +127,7 @@ def main():
     menu.show_all()
     a = AppindicatorTray(menu, "test", "xpra.png", None, None, None, gtk.main_quit)
     a.show()
-    gobject.timeout_add(1000*10, gtk.main_quit)
+    glib.timeout_add(1000*10, gtk.main_quit)
     gtk.main()
 
 
