@@ -20,17 +20,24 @@ def debug(*args):
     #can be overriden
     pass
 
+
+class InitException(Exception):
+    pass
+class InitInfo(Exception):
+    pass
+class InitExit(Exception):
+    def __init__(self, status, msg):
+        self.status = status
+        Exception.__init__(self, msg)
+
+
 DEFAULT_XPRA_CONF_FILENAME = os.environ.get("XPRA_CONF_FILENAME", 'xpra.conf')
 DEFAULT_NET_WM_NAME = os.environ.get("XPRA_NET_WM_NAME", "Xpra")
 
-
-ENCRYPTION_CIPHERS = []
-try:
-    from Crypto.Cipher import AES
-    assert AES
-    ENCRYPTION_CIPHERS.append("AES")
-except:
-    pass
+POSIX = os.name=="posix"
+WIN32 = sys.platform.startswith("win")
+OSX = sys.platform.startswith("darwin")
+PYTHON3 = sys.version_info[0]>=3
 
 try:
     import xpra.sound
@@ -39,19 +46,40 @@ except:
     has_sound_support = False
 
 
+def get_Xdummy_command(use_wrapper, log_dir="${HOME}/.xpra", xorg_conf="/etc/xpra/xorg.conf"):
+    if use_wrapper:
+        Xorg = "xpra_Xdummy"
+    else:
+        Xorg = "Xorg"
+    return [Xorg,
+          "-noreset",
+          "-nolisten", "tcp",
+          "+extension", "GLX",
+          "+extension", "RANDR",
+          "+extension", "RENDER",
+          "-auth", "$XAUTHORITY",
+          "-logfile", "%s/Xorg.${DISPLAY}.log" % log_dir,
+          #must be specified with some Xorg versions (ie: arch linux)
+          #this directory can store xorg config files, it does not need to be created:
+          "-configdir", "${HOME}/.xpra/xorg.conf.d",
+          "-config", xorg_conf]
+
+def get_Xvfb_command():
+    return ["Xvfb",
+           "+extension", "Composite",
+           "-screen", "0", "5760x2560x24+32",
+           "-nolisten", "tcp",
+           "-noreset",
+           "-auth", "$XAUTHORITY"]
+
+
 def OpenGL_safety_check():
     #Ubuntu 12.04 will just crash on you if you try:
-    distro = ""
-    if hasattr(python_platform, "linux_distribution"):
-        distro = python_platform.linux_distribution()
-    if distro and len(distro)==3 and distro[0]=="Ubuntu":
-        ur = distro[1]  #ie: "12.04"
-        try:
-            rnum = [int(x) for x in ur.split(".")]  #ie: [12, 4]
-            if rnum<=[12, 4]:
-                return "Ubuntu %s is too buggy" % ur
-        except:
-            pass
+    from xpra.os_util import is_Ubuntu, getUbuntuVersion
+    if is_Ubuntu():
+        rnum = getUbuntuVersion()
+        if rnum<=[12, 4]:
+            return "Ubuntu %s is too buggy" % rnum
     #try to detect VirtualBox:
     #based on the code found here:
     #http://spth.virii.lu/eof2/articles/WarGame/vboxdetect.html
@@ -211,7 +239,7 @@ def read_xpra_defaults():
         returns a dict with values as strings and arrays of strings.
         If the <conf_dir> is not specified, we figure out its location.
     """
-    from xpra.platform.paths import get_default_conf_dir, get_system_conf_dir, get_user_conf_dir
+    from xpra.platform.paths import get_default_conf_dirs, get_system_conf_dirs, get_user_conf_dirs
     # load config files in this order (the later ones override earlier ones):
     # * application defaults   (ie: "/Volumes/Xpra/Xpra.app/Contents/Resources/" on OSX)
     #                          (ie: "C:\Program Files\Xpra\" on win32)
@@ -223,7 +251,7 @@ def read_xpra_defaults():
     # * user config            (ie: "~/.xpra/" on all Posix, including OSX)
     #                          (ie: "C:\Documents and Settings\Username\Application Data\Xpra" with XP)
     #                          (ie: "C:\Users\<user name>\AppData\Roaming" with Visa onwards)
-    dirs = [get_default_conf_dir(), get_system_conf_dir(), get_user_conf_dir()]
+    dirs = get_default_conf_dirs() + get_system_conf_dirs() + get_user_conf_dirs()
     defaults = {}
     for d in dirs:
         if not d:
@@ -253,23 +281,32 @@ OPTION_TYPES = {
                     "tray-icon"         : str,
                     "window-icon"       : str,
                     "password-file"     : str,
+                    "clipboard"         : str,
                     "clipboard-filter-file" : str,
+                    "remote-clipboard"  : str,
+                    "local-clipboard"   : str,
                     "pulseaudio-command": str,
+                    "tcp-encryption"    : str,
+                    "tcp-encryption-keyfile": str,
                     "encryption"        : str,
-                    "encryption_keyfile": str,
+                    "encryption-keyfile": str,
                     "mode"              : str,
                     "ssh"               : str,
                     "xvfb"              : str,
                     "socket-dir"        : str,
+                    "log-dir"           : str,
                     "log-file"          : str,
                     "border"            : str,
                     "max-size"          : str,
-                    "window-layout"     : str,
+                    "desktop-scaling"   : str,
                     "display"           : str,
                     "tcp-proxy"         : str,
                     "download-path"     : str,
                     "open-command"      : str,
                     "lpadmin"           : str,
+                    "lpinfo"            : str,
+                    "pdf-printer"       : str,
+                    "postscript-printer": str,
                     "debug"             : str,
                     "input-method"      : str,
                     "microphone"        : str,
@@ -283,9 +320,11 @@ OPTION_TYPES = {
                     "min-speed"         : int,
                     "compression_level" : int,
                     "dpi"               : int,
-                    "scaling"           : int,
+                    "video-scaling"     : int,
                     "file-size-limit"   : int,
                     "idle-timeout"      : int,
+                    "server-idle-timeout" : int,
+                    "sync-xvfb"         : int,
                     #float options:
                     "auto-refresh-delay": float,
                     #boolean options:
@@ -294,7 +333,6 @@ OPTION_TYPES = {
                     "displayfd"         : bool,
                     "fake-xinerama"     : bool,
                     "tray"              : bool,
-                    "clipboard"         : bool,
                     "pulseaudio"        : bool,
                     "dbus-proxy"        : bool,
                     "mmap"              : bool,
@@ -313,6 +351,8 @@ OPTION_TYPES = {
                     "exit-with-children": bool,
                     "exit-with-client"  : bool,
                     "exit-ssh"          : bool,
+                    "dbus-control"      : bool,
+                    "av-sync"           : bool,
                     "opengl"            : bool,
                     "mdns"              : bool,
                     "file-transfer"     : bool,
@@ -321,7 +361,10 @@ OPTION_TYPES = {
                     "swap-keys"         : bool,
                     "start-new-commands": bool,
                     "remote-logging"    : bool,
+                    "shadow-fullscreen" : bool,
+                    "global-menus"      : bool,
                     #arrays of strings:
+                    "socket-dirs"       : list,
                     "encodings"         : list,
                     "video-encoders"    : list,
                     "csc-modules"       : list,
@@ -344,12 +387,20 @@ def get_defaults():
     global GLOBAL_DEFAULTS
     if GLOBAL_DEFAULTS is not None:
         return GLOBAL_DEFAULTS
-    from xpra.platform.features import DEFAULT_SSH_CMD, DOWNLOAD_PATH, OPEN_COMMAND, DEFAULT_PULSEAUDIO_COMMAND, DEFAULT_XVFB_COMMAND
+    from xpra.platform.features import DEFAULT_SSH_COMMAND, OPEN_COMMAND, DEFAULT_PULSEAUDIO_COMMAND, XDUMMY, XDUMMY_WRAPPER, DISPLAYFD, DEFAULT_ENV, CAN_DAEMONIZE
+    from xpra.platform.paths import get_download_dir, get_default_log_dir
     try:
         from xpra.platform.info import get_username
         username = get_username()
     except:
         username = ""
+    if WIN32 or OSX or PYTHON3:
+        xvfb = ""
+    elif XDUMMY:
+        xvfb = get_Xdummy_command(use_wrapper=XDUMMY_WRAPPER)
+    else:
+        xvfb = get_Xvfb_command()
+
     GLOBAL_DEFAULTS = {
                     "encoding"          : "",
                     "title"             : "@title@ on @client-machine@",
@@ -363,22 +414,31 @@ def get_defaults():
                     "tray-icon"         : "",
                     "window-icon"       : "",
                     "password-file"     : "",
+                    "clipboard"         : "yes",
                     "clipboard-filter-file" : "",
+                    "remote-clipboard"  : "CLIPBOARD",
+                    "local-clipboard"   : "CLIPBOARD",
                     "pulseaudio-command": DEFAULT_PULSEAUDIO_COMMAND,
                     "encryption"        : "",
-                    "encryption_keyfile": "",
-                    "ssh"               : DEFAULT_SSH_CMD,
-                    "xvfb"              : DEFAULT_XVFB_COMMAND,
+                    "tcp-encryption"    : "",
+                    "encryption-keyfile": "",
+                    "tcp-encryption-keyfile": "",
+                    "ssh"               : DEFAULT_SSH_COMMAND,
+                    "xvfb"              : " ".join(xvfb),
                     "socket-dir"        : "",
+                    "log-dir"           : get_default_log_dir(),
                     "log-file"          : "$DISPLAY.log",
                     "border"            : "auto,0",
                     "max-size"          : "",
-                    "window-layout"     : "",
+                    "desktop-scaling"   : "auto",
                     "display"           : "",
                     "tcp-proxy"         : "",
-                    "download-path"     : DOWNLOAD_PATH,
+                    "download-path"     : get_download_dir(),
                     "open-command"      : OPEN_COMMAND,
-                    "lpadmin"           : "lpadmin",
+                    "lpadmin"           : "/usr/sbin/lpadmin",
+                    "lpinfo"            : "/usr/sbin/lpinfo",
+                    "pdf-printer"       : "",
+                    "postscript-printer": "",
                     "debug"             : "",
                     "input-method"      : "none",
                     "sound-source"      : "",
@@ -390,19 +450,20 @@ def get_defaults():
                     "min-speed"         : 0,
                     "compression_level" : 1,
                     "dpi"               : 0,
-                    "scaling"           : 1,
+                    "video-scaling"     : 1,
                     "file-size-limit"   : 10,
                     "idle-timeout"      : 0,
-                    "auto-refresh-delay": 0.25,
-                    "daemon"            : True,
+                    "server-idle-timeout" : 0,
+                    "sync-xvfb"         : 0,
+                    "auto-refresh-delay": 0.15,
+                    "daemon"            : CAN_DAEMONIZE,
                     "use-display"       : False,
-                    "displayfd"         : False,
-                    "fake-xinerama"     : True,
+                    "displayfd"         : DISPLAYFD,
+                    "fake-xinerama"     : not OSX and not WIN32,
                     "tray"              : True,
-                    "clipboard"         : True,
-                    "pulseaudio"        : True,
-                    "dbus-proxy"        : os.name=="posix" and not sys.platform.startswith("darwin"),
-                    "mmap"              : True,
+                    "pulseaudio"        : not OSX and not WIN32,
+                    "dbus-proxy"        : not OSX and not WIN32,
+                    "mmap"              : not OSX and not WIN32,
                     "mmap-group"        : False,
                     "speaker"           : ["disabled", "on"][has_sound_support],
                     "microphone"        : ["disabled", "off"][has_sound_support],
@@ -412,7 +473,7 @@ def get_defaults():
                     "cursors"           : True,
                     "bell"              : True,
                     "notifications"     : True,
-                    "xsettings"         : os.name=="posix",
+                    "xsettings"         : not OSX and not WIN32,
                     "system-tray"       : True,
                     "sharing"           : False,
                     "delay-tray"        : False,
@@ -420,14 +481,19 @@ def get_defaults():
                     "exit-with-children": False,
                     "exit-with-client"  : False,
                     "start-new-commands": False,
-                    "remote-logging"    : sys.platform.startswith("win") or sys.platform.startswith("darwin"),
+                    "remote-logging"    : WIN32 or OSX,
+                    "av-sync"           : True,
                     "exit-ssh"          : True,
+                    "dbus-control"      : not WIN32 and not OSX,
                     "opengl"            : OPENGL_DEFAULT,
-                    "mdns"              : False,
+                    "mdns"              : not WIN32,
                     "file-transfer"     : True,
                     "printing"          : True,
                     "open-files"        : False,
-                    "swap-keys"         : sys.platform.startswith("darwin"),    #only used on osx
+                    "swap-keys"         : OSX,  #only used on osx
+                    "shadow-fullscreen" : False,
+                    "global-menus"      : True,
+                    "socket-dirs"       : [],
                     "encodings"         : ["all"],
                     "video-encoders"    : ["all"],
                     "csc-modules"       : ["all"],
@@ -437,15 +503,29 @@ def get_defaults():
                     "compressors"       : ["all"],
                     "packet-encoders"   : ["all"],
                     "key-shortcut"      : [
+                                           "Meta+Shift+F1:show_menu",
                                            "Meta+Shift+F2:show_start_new_command",
                                            "Meta+Shift+F4:quit",
-                                           "Meta+Shift+F8:magic_key",
-                                           "Meta+Shift+F11:show_session_info"
+                                           "Meta+Shift+F5:increase_quality",
+                                           "Meta+Shift+F6:decrease_quality",
+                                           "Meta+Shift+F7:increase_speed",
+                                           "Meta+Shift+F8:decrease_speed",
+                                           "Meta+Shift+F10:magic_key",
+                                           "Meta+Shift+F11:show_session_info",
+                                           "Meta+Shift+plus:scaleup",
+                                           "Meta+Shift+plusminus:scaleup",          #the keyname on OSX...
+                                           "Meta+Shift+minus:scaledown",
+                                           "Meta+Shift+underscore:scaledown",
+                                           "Meta+Shift+emdash:scaledown",           #OSX
+                                           "Meta+Shift+KP_Add:scaleup",
+                                           "Meta+Shift+KP_Subtract:scaledown",
+                                           "Meta+Shift+KP_Multiply:scalereset",
+                                           "Meta+Shift+degree:scalereset",          #OSX
                                            ],
-                    "bind-tcp"          : None,
-                    "start"             : None,
-                    "start-child"       : None,
-                    "env"               : None,
+                    "bind-tcp"          : [],
+                    "start"             : [],
+                    "start-child"       : [],
+                    "env"               : DEFAULT_ENV,
                     }
     return GLOBAL_DEFAULTS
 #fields that got renamed:
@@ -572,7 +652,7 @@ def make_defaults_struct(extras_defaults={}, extras_types={}, extras_validation=
     defaults = read_xpra_defaults()
     return dict_to_validated_config(defaults, extras_defaults, extras_types, extras_validation)
 
-def dict_to_validated_config(d, extras_defaults={}, extras_types={}, extras_validation={}):
+def dict_to_validated_config(d={}, extras_defaults={}, extras_types={}, extras_validation={}):
     options = get_defaults().copy()
     options.update(extras_defaults)
     #parse config:
@@ -587,6 +667,113 @@ def dict_to_validated_config(d, extras_defaults={}, extras_types={}, extras_vali
     return config
 
 
+def fixup_debug_option(value):
+    """ backwards compatible parsing of the debug option, which used to be a boolean """
+    if not value:
+        return ""
+    value = str(value)
+    if value.strip().lower() in ("yes", "true", "on", "1"):
+        return "all"
+    if value.strip().lower() in ("no", "false", "off", "0"):
+        return ""
+    #if we're here, the value should be a CSV list of categories
+    return value
+
+def _csvstr(value):
+    if type(value) in (tuple, list):
+        return ",".join(str(x).lower().strip() for x in value)
+    elif type(value)==str:
+        return value.strip().lower()
+    raise Exception("don't know how to convert %s to a csv list!" % type(value))
+
+def _nodupes(s):
+    from xpra.util import remove_dupes
+    return remove_dupes(x.strip().lower() for x in s.split(","))
+
+def fixup_video_all_or_none(options):
+    from xpra.codecs.video_helper import ALL_VIDEO_ENCODER_OPTIONS as aveco
+    from xpra.codecs.video_helper import ALL_CSC_MODULE_OPTIONS as acsco
+    from xpra.codecs.video_helper import ALL_VIDEO_DECODER_OPTIONS as avedo
+    vestr   = _csvstr(options.video_encoders)
+    cscstr  = _csvstr(options.csc_modules)
+    vdstr   = _csvstr(options.video_decoders)
+    def getlist(strarg, help_txt, all_list):
+        if strarg=="help":
+            raise InitInfo("the following %s may be available: %s" % (help_txt, ", ".join(all_list)))
+        elif strarg=="none":
+            return []
+        elif strarg=="all":
+            return all_list
+        else:
+            return _nodupes(strarg)
+    options.video_encoders  = getlist(vestr,    "video encoders",   aveco)
+    options.csc_modules     = getlist(cscstr,   "csc modules",      acsco)
+    options.video_decoders  = getlist(vdstr,    "video decoders",   avedo)
+
+def fixup_socketdirs(options):
+    if not options.socket_dirs:
+        from xpra.platform.paths import get_socket_dirs
+        options.socket_dirs = get_socket_dirs()
+    elif type(options.socket_dirs)==str:
+        options.socket_dirs = options.socket_dirs.split(os.path.pathsep)
+    else:
+        assert type(options.socket_dirs) in (list, tuple)
+        options.socket_dirs = [v for x in options.socket_dirs for v in x.split(os.path.pathsep)]
+
+def fixup_encodings(options):
+    from xpra.codecs.loader import PREFERED_ENCODING_ORDER
+    RENAME = {"jpg" : "jpeg"}
+    if options.encoding:
+        options.encoding = RENAME.get(options.encoding, options.encoding)
+    estr = _csvstr(options.encodings)
+    if estr=="all":
+        #replace with an actual list
+        options.encodings = PREFERED_ENCODING_ORDER
+        return
+    encodings = [RENAME.get(x, x) for x in _nodupes(estr)]
+    if "rgb" in encodings:
+        if "rgb24" not in encodings:
+            encodings.append("rgb24")
+        if "rgb32" not in encodings:
+            encodings.append("rgb32")
+    options.encodings = encodings
+
+def fixup_compression(options):
+    #packet compression:
+    from xpra.net import compression
+    cstr = _csvstr(options.compressors)
+    if cstr=="none":
+        compressors = []
+    elif cstr=="all":
+        compressors = compression.PERFORMANCE_ORDER
+    else:
+        compressors = _nodupes(cstr)
+        unknown = [x for x in compressors if x and x not in compression.ALL_COMPRESSORS]
+        if unknown:
+            warn("warning: invalid compressor(s) specified: %s" % (", ".join(unknown)))
+    options.compressors = compressors
+
+def fixup_packetencoding(options):
+    #packet encoding
+    from xpra.net import packet_encoding
+    pestr = _csvstr(options.packet_encoders)
+    if pestr=="all":
+        packet_encoders = packet_encoding.PERFORMANCE_ORDER
+    else:
+        packet_encoders = _nodupes(pestr)
+        unknown = [x for x in packet_encoders if x and x not in packet_encoding.ALL_ENCODERS]
+        if unknown:
+            warn("warning: invalid packet encoder(s) specified: %s" % (", ".join(unknown)))
+    options.packet_encoders = packet_encoders
+
+def fixup_options(options):
+    fixup_encodings(options)
+    fixup_compression(options)
+    fixup_packetencoding(options)
+    fixup_video_all_or_none(options)
+    fixup_socketdirs(options)
+
+
 def main():
     from xpra.util import nonl
     def print_options(o):
@@ -598,10 +785,12 @@ def main():
                 v = ", ".join(str(x) for x in v)
             print("* %-32s : %s" % (k, nonl(v)))
     from xpra.platform import init, clean
+    from xpra.log import enable_color
     try:
         init("Config-Info", "Config Info")
+        enable_color()
         args = list(sys.argv[1:])
-        if "-v" in args:
+        if "-v" in args or "--verbose" in sys.argv:
             global debug
             def debug(*args):
                 print(args[0] % args[1:])
