@@ -18,8 +18,10 @@ import select
 import time
 import traceback
 
-from xpra.scripts.main import TCP_NODELAY, warn
-from xpra.dotxpra import DotXpra, ServerSockInUse
+from xpra.scripts.main import TCP_NODELAY, warn, no_gtk
+from xpra.scripts.config import InitException
+from xpra.os_util import SIGNAMES
+from xpra.dotxpra import DotXpra, norm_makepath
 
 
 # use process polling with python versions older than 2.7 and 3.0, (because SIGCHLD support is broken)
@@ -49,7 +51,7 @@ def add_cleanup(f):
 
 
 def deadly_signal(signum, frame):
-    sys.stdout.write("got deadly signal %s, exiting\n" % {signal.SIGINT:"SIGINT", signal.SIGTERM:"SIGTERM"}.get(signum, signum))
+    sys.stdout.write("got deadly signal %s, exiting\n" % SIGNAMES.get(signum, signum))
     sys.stdout.flush()
     run_cleanups()
     # This works fine in tests, but for some reason if I use it here, then I
@@ -136,7 +138,7 @@ fi
 """)
     return "".join(script)
 
-def write_runner_shell_script(dotxpra, contents, overwrite=True):
+def write_runner_shell_script(contents, overwrite=True):
     # This used to be given a display-specific name, but now we give it a
     # single fixed name and if multiple servers are started then the last one
     # will clobber the rest.  This isn't great, but the tradeoff is that it
@@ -145,7 +147,11 @@ def write_runner_shell_script(dotxpra, contents, overwrite=True):
     # is running on the remote host.  Might need to revisit this later if
     # people run into problems or autodiscovery turns out to be less useful
     # than expected.
-    scriptpath = os.path.join(dotxpra.confdir(), "run-xpra")
+    from xpra.platform.paths import get_script_bin_dir
+    scriptdir = os.path.expanduser(get_script_bin_dir())
+    if not os.path.exists(scriptdir):
+        os.mkdir(scriptdir, 0o700)
+    scriptpath = os.path.join(scriptdir, "run-xpra")
     if os.path.exists(scriptpath) and not overwrite:
         return
     # Write out a shell-script so that we can start our proxy in a clean
@@ -190,8 +196,8 @@ def mdns_publish(display_name, mode, listen_on, text_dict={}):
             MDNS_WARNING = True
             from xpra.log import Logger
             log = Logger("mdns")
-            log.error("failed to load the mdns avahi publisher: %s", e)
-            log.error("either fix your installation or use the '--no-mdns' flag")
+            log.warn("Warning: failed to load the mdns avahi publisher: %s", e)
+            log.warn(" either fix your installation or use the 'mdns=no' option")
         return
     d = text_dict.copy()
     d["mode"] = mode
@@ -242,7 +248,10 @@ def create_tcp_socket(host, iport):
 def setup_tcp_socket(host, iport):
     from xpra.log import Logger
     log = Logger("network")
-    tcp_socket = create_tcp_socket(host, iport)
+    try:
+        tcp_socket = create_tcp_socket(host, iport)
+    except Exception as e:
+        raise InitException("failed to setup TCP socket on %s:%s %s" % (host, iport, e))
     def cleanup_tcp_socket():
         log.info("closing tcp socket %s:%s", host, iport)
         try:
@@ -250,40 +259,73 @@ def setup_tcp_socket(host, iport):
         except:
             pass
     _cleanups.append(cleanup_tcp_socket)
-    return "tcp", tcp_socket
+    return "tcp", tcp_socket, (host, iport)
 
 def parse_bind_tcp(bind_tcp):
     tcp_sockets = set()
     if bind_tcp:
         for spec in bind_tcp:
             if ":" not in spec:
-                raise Exception("TCP port must be specified as [HOST]:PORT")
+                raise InitException("TCP port must be specified as [HOST]:PORT")
             host, port = spec.rsplit(":", 1)
             if host == "":
                 host = "127.0.0.1"
             try:
                 iport = int(port)
             except:
-                raise Exception("invalid port number: %s" % port)
+                raise InitException("invalid port number: %s" % port)
             tcp_sockets.add((host, iport))
     return tcp_sockets
 
 
-def setup_local_socket(dotxpra, display_name, clobber, mmap_group, socket_permissions):
+def normalize_local_display_name(local_display_name):
+    if not local_display_name.startswith(":"):
+        local_display_name = ":" + local_display_name
+    if "." in local_display_name:
+        local_display_name = local_display_name[:local_display_name.rindex(".")]
+    assert local_display_name.startswith(":")
+    for char in local_display_name[1:]:
+        assert char in "0123456789", "invalid character in display name: %s" % char
+    return local_display_name
+
+# Same as socket_path, but preps for the server:
+def setup_server_socket_path(dotxpra, sockpath, local_display_name, clobber, wait_for_unknown=0):
+    if not clobber:
+        state = dotxpra.get_server_state(sockpath, 1)
+        counter = 0
+        while state==dotxpra.UNKNOWN and counter<wait_for_unknown:
+            if counter==0:
+                sys.stdout.write("%s is not responding, waiting for it to timeout before clearing it" % sockpath)
+            sys.stdout.write(".")
+            sys.stdout.flush()
+            counter += 1
+            time.sleep(1)
+            state = dotxpra.get_server_state(sockpath)
+        if counter>0:
+            sys.stdout.write("\n")
+            sys.stdout.flush()
+        if state not in (dotxpra.DEAD, dotxpra.UNKNOWN):
+            raise InitException("You already have an xpra server running at %s\n"
+                     "  (did you want 'xpra upgrade'?)"
+                     % (local_display_name,))
+    if os.path.exists(sockpath):
+        os.unlink(sockpath)
+    return sockpath
+
+
+def setup_local_socket(socket_dir, socket_dirs, display_name, clobber, mmap_group, socket_permissions):
     if sys.platform.startswith("win"):
         return None, None
     from xpra.log import Logger
     log = Logger("network")
-    #print("creating server socket %s" % sockpath)
+    dotxpra = DotXpra(socket_dir or socket_dirs[0])
     try:
-        display_name = dotxpra.normalize_local_display_name(display_name)
-        sockpath = dotxpra.server_socket_path(display_name, clobber, wait_for_unknown=5)
-    except ServerSockInUse:
-        raise Exception("You already have an xpra server running at %s\n"
-                     "  (did you want 'xpra upgrade'?)"
-                     % (display_name,))
+        dotxpra.mksockdir()
+        display_name = normalize_local_display_name(display_name)
+        sockpath = dotxpra.socket_path(display_name)
     except Exception as e:
-        raise Exception("socket path error: %s" % e)
+        raise InitException("socket path error: %s" % e)
+    setup_server_socket_path(dotxpra, sockpath, display_name, clobber, wait_for_unknown=5)
     sock = create_unix_domain_socket(sockpath, mmap_group, socket_permissions)
     def cleanup_socket():
         log.info("removing socket %s", sockpath)
@@ -292,7 +334,7 @@ def setup_local_socket(dotxpra, display_name, clobber, mmap_group, socket_permis
         except:
             pass
     _cleanups.append(cleanup_socket)
-    return ("unix-domain", sock), cleanup_socket
+    return ("unix-domain", sock, sockpath), cleanup_socket
 
 
 def get_free_tcp_port():
@@ -314,12 +356,12 @@ def start_websockify(child_reaper, opts, tcp_sockets):
         #html disabled
         return
     if opts.tcp_proxy:
-        raise Exception("cannot use tcp-proxy mode with html, use one or the other")
+        raise InitException("cannot use tcp-proxy mode with html, use one or the other")
         return 1
     from xpra.platform.paths import get_resources_dir
     www_dir = os.path.abspath(os.path.join(get_resources_dir(), "www"))
     if not os.path.exists(www_dir):
-        raise Exception("cannot find xpra's html directory (not found in '%s')" % www_dir)
+        raise InitException("cannot find xpra's html directory (not found in '%s')" % www_dir)
     import websockify           #@UnresolvedImport
     assert websockify
     html_port = -1
@@ -331,17 +373,19 @@ def start_websockify(child_reaper, opts, tcp_sockets):
         try:
             html_port = int(html)
         except Exception as e:
-            raise Exception("invalid html port: %s" % e)
+            raise InitException("invalid html port: %s" % e)
         #we now have the host and port (port may be -1 to mean auto..)
+    from xpra.log import Logger
+    log = Logger("server")
     if html_port==-1:
         #try to find a free port and hope that websockify can then use it..
         html_port = get_free_tcp_port()
+    elif os.name=="posix" and html_port<1024 and os.geteuid()!=0:
+        log.warn("Warning: the html port specified may require special privileges (%s:%s)", html_host, html_port)
     if len(tcp_sockets)<1:
-        raise Exception("html web server requires at least one tcp socket, see 'bind-tcp'")
+        raise InitException("html web server requires at least one tcp socket, see 'bind-tcp'")
     #use the first tcp socket for websockify to talk back to us:
     _, xpra_tcp_port = list(tcp_sockets)[0]
-    from xpra.log import Logger
-    log = Logger("server")
     websockify_command = ["websockify", "--web", www_dir, "%s:%s" % (html_host, html_port), "127.0.0.1:%s" % xpra_tcp_port]
     log("websockify_command: %s", websockify_command)
     websockify_proc = subprocess.Popen(websockify_command, close_fds=True)
@@ -400,7 +444,7 @@ def open_log_file(logpath):
         os.rename(logpath, logpath + ".old")
     return os.open(logpath, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o666)
 
-def select_log_file(dotxpra, log_file, display_name):
+def select_log_file(log_dir, log_file, display_name):
     """ returns the log file path we should be using given the parameters,
         this may return a temporary logpath if display_name is not available.
     """
@@ -408,12 +452,12 @@ def select_log_file(dotxpra, log_file, display_name):
         if os.path.isabs(log_file):
             logpath = log_file
         else:
-            logpath = os.path.join(dotxpra.sockdir(), log_file)
+            logpath = os.path.join(log_dir, log_file)
         logpath = shellsub(logpath, {"DISPLAY" : display_name})
     elif display_name:
-        logpath = dotxpra.log_path(display_name) + ".log"
+        logpath = norm_makepath(log_dir, display_name) + ".log"
     else:
-        logpath = os.path.join(dotxpra.sockdir(), "tmp_%d.log" % os.getpid())
+        logpath = os.path.join(log_dir, "tmp_%d.log" % os.getpid())
     return logpath
 
 
@@ -457,13 +501,20 @@ def sanitize_env():
     #TODO: simply whitelisting the env would be safer/better
     unsetenv("DESKTOP_SESSION",
              "GDMSESSION",
+             "GNOME_DESKTOP_SESSION_ID",
              "SESSION_MANAGER",
              "XDG_VTNR",
              "XDG_MENU_PREFIX",
+             "XDG_CURRENT_DESKTOP",
+             "XDG_SESSION_DESKTOP",
+             "XDG_SESSION_TYPE",
+             "XDG_SESSION_ID",
              "XDG_SEAT",
+             "XDG_VTNR",
              #"XDG_RUNTIME_DIR",
              "QT_GRAPHICSSYSTEM_CHECKED",
              )
+    os.environ["XDG_SESSION_TYPE"] = "x11"
 
 def configure_imsettings_env(input_method):
     im = (input_method or "").lower()
@@ -513,6 +564,8 @@ def start_Xvfb(xvfb_str, display_name):
             #trying to continue anyway!
             sys.stderr.write("Error trying to create XAUTHORITY file %s: %s\n" % (xauthority, e))
 
+    if not xvfb_str:
+        raise InitException("the 'xvfb' command is not defined")
     #identify logfile argument if it exists and if we may have to rename it:
     xvfb_cmd = xvfb_str.split()
     if '-logfile' in xvfb_cmd and display_name[0]=='S':
@@ -537,7 +590,7 @@ def start_Xvfb(xvfb_str, display_name):
 
     xvfb_cmd = xvfb_str.split()
     if not xvfb_cmd:
-        raise Exception("cannot start Xvfb, command definition is missing!")
+        raise InitException("cannot start Xvfb, the command definition is missing!")
     xvfb_executable = xvfb_cmd[0]
     if display_name[0]=='S':
         # 'S' means that we allocate the display automatically
@@ -622,6 +675,8 @@ def check_xvfb_process(xvfb=None):
     log = Logger("server")
     log.error("")
     log.error("Xvfb command has terminated! xpra cannot continue")
+    log.error(" if the display is already running, try a different one,")
+    log.error(" or use the --use-display flag")
     log.error("")
     return False
 
@@ -637,7 +692,7 @@ def verify_display_ready(xvfb, display_name, shadowing):
         sys.stderr.write("%s\n" % e)
         return  None
     # Now we can safely load gtk and connect:
-    assert "gtk" not in sys.modules
+    no_gtk()
     import gtk.gdk          #@Reimport
     try:
         import glib
@@ -661,8 +716,23 @@ def verify_display_ready(xvfb, display_name, shadowing):
         return  None
     return display
 
+def guess_xpra_display(socket_dir, socket_dirs):
+    dotxpra = DotXpra(socket_dir, socket_dirs)
+    results = dotxpra.sockets()
+    live = [display for state, display in results if state==DotXpra.LIVE]
+    if len(live)==0:
+        raise InitException("no existing xpra servers found")
+    if len(live)>1:
+        raise InitException("too many existing xpra servers found, cannot guess which one to use")
+    return live[0]
+
 
 def run_server(error_cb, opts, mode, xpra_file, extra_args):
+    try:
+        cwd = os.getcwd()
+    except:
+        cwd = os.path.expanduser("~")
+        sys.stderr.write("current working directory does not exist, using '%s'\n" % cwd)
     if opts.encoding and opts.encoding=="help":
         #avoid errors and warnings:
         opts.encoding = ""
@@ -696,6 +766,7 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
     shadowing = mode == "shadow"
     proxying  = mode == "proxy"
     clobber   = upgrading or opts.use_display
+    start_vfb = not shadowing and not proxying and not clobber
 
     if upgrading or shadowing:
         #there should already be one running
@@ -708,10 +779,9 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
             display_name = ":0"
         else:
             from xpra.scripts.main import guess_X11_display
-            display_name = guess_X11_display(opts.socket_dir)
+            display_name = guess_X11_display(opts.socket_dir, opts.socket_dirs)
     elif upgrading and len(extra_args)==0:
-        from xpra.scripts.main import guess_xpra_display
-        display_name = guess_xpra_display(opts.socket_dir)
+        display_name = guess_xpra_display(opts.socket_dir, opts.socket_dirs)
     else:
         if len(extra_args) > 1:
             error_cb("too many extra arguments: only expected a display number")
@@ -726,8 +796,7 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
                 error_cb("displayfd support is not enabled on this system, you must specify the display to use")
             if opts.use_display:
                 #only use automatic guess for xpra displays and not X11 displays:
-                from xpra.scripts.main import guess_xpra_display     #@Reimport
-                display_name = guess_xpra_display(opts.socket_dir)
+                display_name = guess_xpra_display(opts.socket_dir, opts.socket_dirs)
             else:
                 # We will try to find one automaticaly
                 # Use the temporary magic value 'S' as marker:
@@ -738,15 +807,21 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
 
     atexit.register(run_cleanups)
     #the server class will usually override those:
+    #SIGINT breaks GTK3.. (but there are no py3k servers!)
     signal.signal(signal.SIGINT, deadly_signal)
     signal.signal(signal.SIGTERM, deadly_signal)
 
-    dotxpra = DotXpra(opts.socket_dir)
-
     # Generate the script text now, because os.getcwd() will
     # change if/when we daemonize:
     script = xpra_runner_shell_script(xpra_file, os.getcwd(), opts.socket_dir)
 
+    if start_vfb or opts.daemon:
+        #we will probably need a log dir
+        #either for the vfb, or for our own log file
+        log_dir = os.path.expanduser(opts.log_dir)
+        if not os.path.exists(log_dir):
+            os.mkdir(log_dir, 0o700)
+
     stdout = sys.stdout
     stderr = sys.stderr
     # Daemonize:
@@ -756,7 +831,7 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
             opts.password_file = os.path.abspath(opts.password_file)
         # At this point we may not know the display name,
         # so log_filename0 may point to a temporary file which we will rename later
-        log_filename0 = select_log_file(dotxpra, opts.log_file, display_name)
+        log_filename0 = select_log_file(log_dir, opts.log_file, display_name)
         logfd = open_log_file(log_filename0)
         assert logfd > 2
         stdout, stderr = daemonize(logfd)
@@ -771,26 +846,30 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
     if os.name=="posix":
         # Write out a shell-script so that we can start our proxy in a clean
         # environment:
-        write_runner_shell_script(dotxpra, script)
+        write_runner_shell_script(script)
 
     from xpra.log import Logger
     log = Logger("server")
 
+    #warn early about this:
+    if starting:
+        de = os.environ.get("XDG_SESSION_DESKTOP") or os.environ.get("SESSION_DESKTOP")
+        if de and (opts.pulseaudio or opts.notifications):
+            log.warn("Warning: xpra start from an existing '%s' desktop session", de)
+            log.warn(" pulseaudio and notifications forwarding may not work")
+            log.warn(" try using a clean environment, a dedicated user, or turn off those options")
+
     mdns_recs = []
     sockets = []
-    try:
-        # Initialize the TCP sockets before the display,
-        # That way, errors won't make us kill the Xvfb
-        # (which may not be ours to kill at that point)
-        bind_tcp = parse_bind_tcp(opts.bind_tcp)
-        for host, iport in bind_tcp:
-            socket = setup_tcp_socket(host, iport)
-            sockets.append(socket)
+    # Initialize the TCP sockets before the display,
+    # That way, errors won't make us kill the Xvfb
+    # (which may not be ours to kill at that point)
+    bind_tcp = parse_bind_tcp(opts.bind_tcp)
+    for host, iport in bind_tcp:
+        socket = setup_tcp_socket(host, iport)
+        sockets.append(socket)
         if opts.mdns:
             mdns_recs.append(("tcp", bind_tcp))
-    except Exception as e:
-        log.error("cannot start server: failed to setup sockets: %s", e)
-        return 1
 
     # Do this after writing out the shell script:
     if display_name[0] != 'S':
@@ -801,7 +880,7 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
     # Start the Xvfb server first to get the display_name if needed
     xvfb = None
     xvfb_pid = None
-    if not shadowing and not proxying and not clobber:
+    if start_vfb:
         try:
             xvfb, display_name = start_Xvfb(opts.xvfb, display_name)
         except OSError as e:
@@ -812,7 +891,7 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
         os.environ["DISPLAY"] = display_name
 
     if opts.daemon:
-        log_filename1 = select_log_file(dotxpra, opts.log_file, display_name)
+        log_filename1 = select_log_file(log_dir, opts.log_file, display_name)
         if log_filename0 != log_filename1:
             # we now have the correct log filename, so use it:
             os.rename(log_filename0, log_filename1)
@@ -825,7 +904,7 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
         return  1
 
     #setup unix domain socket:
-    socket, cleanup_socket = setup_local_socket(dotxpra, display_name, clobber, opts.mmap_group, opts.socket_permissions)
+    socket, cleanup_socket = setup_local_socket(opts.socket_dir, opts.socket_dirs, display_name, clobber, opts.mmap_group, opts.socket_permissions)
     if socket:      #win32 returns None!
         sockets.append(socket)
         if opts.mdns:
@@ -853,7 +932,7 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
         if not display:
             return 1
     elif not proxying:
-        assert "gtk" not in sys.modules
+        no_gtk()
         import gtk          #@Reimport
         assert gtk
 
@@ -862,12 +941,12 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
         app = ShadowServer()
         info = "shadow"
     elif proxying:
-        from xpra.server.proxy_server import ProxyServer
+        from xpra.server.proxy.proxy_server import ProxyServer
         app = ProxyServer()
         info = "proxy"
     else:
         assert starting or upgrading
-        from xpra.x11.gtk_x11 import gdk_display_source
+        from xpra.x11.gtk2 import gdk_display_source
         assert gdk_display_source
         #(now we can access the X11 server)
 
@@ -879,7 +958,7 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
             save_xvfb_pid(xvfb_pid)
 
         #check for an existing window manager:
-        from xpra.x11.gtk_x11.wm import wm_check
+        from xpra.x11.gtk2.wm import wm_check
         if not wm_check(display, opts.wm_name, upgrading):
             return 1
         try:
@@ -893,11 +972,30 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
         if not X11Window.displayHasXComposite():
             log.error("Xpra is a compositing manager, it cannot use a display which lacks the XComposite extension!")
             return 1
+        log("XShape=%s", X11Window.displayHasXShape())
         app = XpraServer(clobber)
         info = "xpra"
 
+    #we got this far so the sockets have initialized and
+    #the server should be able to manage the display
+    #from now on, if we exit without upgrading we will also kill the Xvfb
+    def kill_xvfb():
+        # Close our display(s) first, so the server dying won't kill us.
+        log.info("killing xvfb with pid %s" % xvfb_pid)
+        import gtk  #@Reimport
+        for display in gtk.gdk.display_manager_get().list_displays():
+            display.close()
+        os.kill(xvfb_pid, signal.SIGTERM)
+    if xvfb_pid is not None and not opts.use_display and not shadowing:
+        _cleanups.append(kill_xvfb)
+
     try:
+        app.exec_cwd = cwd
         app.init(opts)
+    except InitException as e:
+        log.error("xpra server initialization error:")
+        log.error(" %s", e)
+        return 1
     except Exception as e:
         log.error("Error: cannot start the %s server", info, exc_info=True)
         log.error(str(e))
@@ -929,19 +1027,6 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
     log("%s(%s)", app.init_when_ready, _when_ready)
     app.init_when_ready(_when_ready)
 
-    #we got this far so the sockets have initialized and
-    #the server should be able to manage the display
-    #from now on, if we exit without upgrading we will also kill the Xvfb
-    def kill_xvfb():
-        # Close our display(s) first, so the server dying won't kill us.
-        log.info("killing xvfb with pid %s" % xvfb_pid)
-        import gtk  #@Reimport
-        for display in gtk.gdk.display_manager_get().list_displays():
-            display.close()
-        os.kill(xvfb_pid, signal.SIGTERM)
-    if xvfb_pid is not None and not opts.use_display and not shadowing:
-        _cleanups.append(kill_xvfb)
-
     try:
         log("running %s", app.run)
         e = app.run()
@@ -959,7 +1044,7 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args):
         from xpra.server.server_core import ServerCore
         if e==ServerCore.EXITING_CODE:
             log.info("exiting: not cleaning up Xvfb")
-        else:
+        elif cleanup_socket in _cleanups:
             log.info("upgrading: not cleaning up Xvfb or socket")
             # don't delete the new socket (not ours)
             _cleanups.remove(cleanup_socket)
