@@ -14,34 +14,38 @@ import time
 import socket
 import signal
 import threading
-import thread
+import traceback
 
 from xpra.log import Logger
 log = Logger("server")
 netlog = Logger("network")
 proxylog = Logger("proxy")
 commandlog = Logger("command")
+authlog = Logger("auth")
+timeoutlog = Logger("timeout")
 
 import xpra
 from xpra.server import ClientException
 from xpra.scripts.main import SOCKET_TIMEOUT, _socket_connect
-from xpra.scripts.config import ENCRYPTION_CIPHERS
 from xpra.scripts.server import deadly_signal
-from xpra.net.bytestreams import SocketConnection
-from xpra.platform import set_application_name
-from xpra.os_util import load_binary_file, get_machine_id, get_user_uuid, SIGNAMES, Queue
-from xpra.version_util import version_compat_check, get_version_info, get_platform_info, get_host_info, local_version
+from xpra.net.bytestreams import SocketConnection, pretty_socket, set_socket_timeout
+from xpra.platform import set_name
+from xpra.os_util import load_binary_file, get_machine_id, get_user_uuid, platform_name, SIGNAMES, Queue
+from xpra.version_util import version_compat_check, get_version_info_full, get_platform_info, get_host_info, local_version
 from xpra.net.protocol import Protocol, get_network_caps, sanity_checks
-from xpra.net.crypto import new_cipher_caps
+from xpra.net.crypto import new_cipher_caps, ENCRYPTION_CIPHERS, ENCRYPT_FIRST_PACKET, DEFAULT_IV, DEFAULT_SALT, DEFAULT_ITERATIONS, INITIAL_PADDING, DEFAULT_PADDING,\
+    ALL_PADDING_OPTIONS
 from xpra.server.background_worker import stop_worker, get_worker
-from xpra.daemon_thread import make_daemon_thread
-from xpra.server.proxy import XpraProxy
+from xpra.make_thread import make_thread
+from xpra.scripts.fdproxy import XpraProxy
+from xpra.server.control_command import ControlError, HelloCommand, HelpCommand, DebugControl
 from xpra.util import typedict, updict, repr_ellipsized, \
-        SERVER_SHUTDOWN, SERVER_EXIT, LOGIN_TIMEOUT, DONE, PROTOCOL_ERROR, SERVER_ERROR, VERSION_ERROR
+        SERVER_SHUTDOWN, SERVER_UPGRADE, LOGIN_TIMEOUT, DONE, PROTOCOL_ERROR, SERVER_ERROR, VERSION_ERROR, CLIENT_REQUEST
 
 main_thread = threading.current_thread()
 
 MAX_CONCURRENT_CONNECTIONS = 20
+SIMULATE_SERVER_HELLO_ERROR = os.environ.get("XPRA_SIMULATE_SERVER_HELLO_ERROR", "0")=="1"
 
 
 def get_server_info():
@@ -51,7 +55,7 @@ def get_server_info():
     def up(prefix, d):
         updict(info, prefix, d)
     up("platform",  get_platform_info())
-    up("build",     get_version_info())
+    up("build",     get_version_info_full())
     return info
 
 def get_thread_info(proto=None, protocols=[]):
@@ -95,7 +99,6 @@ def get_thread_info(proto=None, protocols=[]):
             if x is None:
                 return ""
             return x
-        import traceback
         frames = sys._current_frames()
         for i,frame_pair in enumerate(frames.items()):
             stack = traceback.extract_stack(frame_pair[1])
@@ -128,6 +131,7 @@ class ServerCore(object):
         self._when_ready = []
         self.child_reaper = None
 
+        self._closing = False
         self._upgrading = False
         #networking bits:
         self._potential_protocols = []
@@ -139,19 +143,23 @@ class ServerCore(object):
         self._max_connections = MAX_CONCURRENT_CONNECTIONS
         self._socket_timeout = 0.1
         self._socket_dir = None
+        self.main_socket_path = ""
 
         self.session_name = ""
 
         #Features:
         self.digest_modes = ("hmac", )
+        self.encryption = None
         self.encryption_keyfile = None
+        self.tcp_encryption = None
+        self.tcp_encryption_keyfile = None
         self.password_file = None
         self.compression_level = 1
         self.exit_with_client = False
+        self.server_idle_timeout = 0
+        self.server_idle_timer = None
 
-        #control mode:
-        self.control_commands = ["hello"]
-
+        self.init_control_commands()
         self.init_packet_handlers()
         self.init_aliases()
         sanity_checks()
@@ -168,28 +176,36 @@ class ServerCore(object):
     def init(self, opts):
         log("ServerCore.init(%s)", opts)
         self.session_name = opts.session_name
-        set_application_name(self.session_name or "Xpra")
+        set_name("Xpra", self.session_name or "Xpra")
 
-        self._socket_dir = opts.socket_dir
+        self.main_socket_path = ""
+        self._socket_dir = opts.socket_dir or opts.socket_dirs[0]
         self._tcp_proxy = opts.tcp_proxy
+        self.encryption = opts.encryption
         self.encryption_keyfile = opts.encryption_keyfile
+        self.tcp_encryption = opts.tcp_encryption
+        self.tcp_encryption_keyfile = opts.tcp_encryption_keyfile
         self.password_file = opts.password_file
         self.compression_level = opts.compression_level
         self.exit_with_client = opts.exit_with_client
+        self.server_idle_timeout = opts.server_idle_timeout
 
         self.init_auth(opts)
 
     def init_auth(self, opts):
-        self.auth_class = self.get_auth_module(opts.auth, opts)
-        self.tcp_auth_class = self.get_auth_module(opts.tcp_auth or opts.auth, opts)
-        log("init_auth(%s) auth class=%s, tcp auth class=%s", opts, self.auth_class, self.tcp_auth_class)
+        self.auth_class = self.get_auth_module("unix-domain-socket", opts.auth, opts)
+        self.tcp_auth_class = self.get_auth_module("tcp-socket", opts.tcp_auth or opts.auth, opts)
+        authlog("init_auth(%s) auth class=%s, tcp auth class=%s", opts, self.auth_class, self.tcp_auth_class)
 
-    def get_auth_module(self, auth, opts):
+    def get_auth_module(self, socket_type, auth, opts):
+        authlog("get_auth_module(%s, %s, %s)", socket_type, auth, opts)
         if not auth and opts.password_file:
-            log.warn("no authentication module specified with 'password_file', using 'file' based authentication")
+            authlog.warn("Warning: 'password-file' used without an authentication module for %ss", socket_type)
+            authlog.warn(" using 'file' based authentication")
             auth = "file"
         if not auth and os.environ.get('XPRA_PASSWORD'):
-            log.warn("no authentication module specified with 'XPRA_PASSWORD', using 'file' based authentication")
+            authlog.warn("no authentication module specified for %ss with 'XPRA_PASSWORD'", socket_type)
+            authlog.warn(" using 'file' based authentication")
             auth = "file"
         if auth=="":
             return None
@@ -199,7 +215,7 @@ class ServerCore(object):
                 auth = "win32"
             else:
                 auth = "pam"
-            log("will try to use sys auth module '%s' for %s", auth, sys.platform)
+            authlog("will try to use sys auth module '%s' for %s", auth, sys.platform)
         from xpra.server.auth import fail_auth, reject_auth, allow_auth, none_auth, file_auth
         AUTH_MODULES = {
                         "fail"      : fail_auth,
@@ -212,37 +228,47 @@ class ServerCore(object):
             from xpra.server.auth import pam_auth
             AUTH_MODULES["pam"] = pam_auth
         except Exception as e:
-            log("cannot load pam auth: %s", e)
-        try:
-            from xpra.server.auth import win32_auth
-            AUTH_MODULES["win32"] = win32_auth
-        except Exception as e:
-            log("cannot load win32 auth: %s", e)
+            authlog("cannot load pam auth: %s", e)
+        from xpra.scripts.config import InitException
+        if sys.platform.startswith("win"):
+            try:
+                from xpra.server.auth import win32_auth
+                AUTH_MODULES["win32"] = win32_auth
+            except Exception as e:
+                authlog.error("Error: cannot load the MS Windows authentication module:")
+                authlog.error(" %s", e)
         auth_module = AUTH_MODULES.get(auth.lower())
         if not auth_module:
-            raise Exception("cannot find authentication module '%s' (supported: %s)", auth, AUTH_MODULES.keys())
+            raise InitException("cannot find authentication module '%s' (supported: %s)" % (auth, AUTH_MODULES.keys()))
         try:
             auth_module.init(opts)
             auth_class = getattr(auth_module, "Authenticator")
             return auth_class
         except Exception as e:
-            raise Exception("Authenticator class not found in %s" % auth_module)
+            raise InitException("Authenticator class not found in %s" % auth_module)
 
 
     def init_sockets(self, sockets):
         ### All right, we're ready to accept customers:
-        for socktype, sock in sockets:
-            log("init_sockets(%s) will add %s socket %s", sockets, socktype, sock)
+        for socktype, sock, info in sockets:
+            netlog("init_sockets(%s) will add %s socket %s (%s)", sockets, socktype, sock, info)
             self.idle_add(self.add_listen_socket, socktype, sock)
+            if socktype=="unix-domain" and info:
+                try:
+                    self.main_socket_path = os.path.abspath(info)
+                    netlog("main socket path: %s", self.main_socket_path)
+                except Exception as e:
+                    log.error("failed to set socket path to %s: %s", info, e)
 
     def init_when_ready(self, callbacks):
         self._when_ready = callbacks
 
 
     def init_packet_handlers(self):
-        log("initializing packet handlers")
+        netlog("initializing packet handlers")
         self._default_packet_handlers = {
             "hello":                                self._process_hello,
+            "disconnect":                           self._process_disconnect,
             Protocol.CONNECTION_LOST:               self._process_connection_lost,
             Protocol.GIBBERISH:                     self._process_gibberish,
             Protocol.INVALID:                       self._process_invalid,
@@ -258,12 +284,20 @@ class ServerCore(object):
             self._reverse_aliases[key] = i
             i += 1
 
+    def init_control_commands(self):
+        self.control_commands = {"hello"    : HelloCommand(),
+                                 "debug"    : DebugControl()}
+        help_command = HelpCommand(self.control_commands)
+        self.control_commands["help"] = help_command
 
-    def reaper_quit(self):
+
+    def reaper_exit(self):
         self.clean_quit()
 
     def signal_quit(self, signum, frame):
-        log.info("")
+        sys.stdout.write("\n")
+        sys.stdout.flush()
+        self._closing = True
         log.info("got signal %s, exiting", SIGNAMES.get(signum, signum))
         signal.signal(signal.SIGINT, deadly_signal)
         signal.signal(signal.SIGTERM, deadly_signal)
@@ -272,15 +306,18 @@ class ServerCore(object):
 
     def clean_quit(self, upgrading=False):
         log("clean_quit(%s)", upgrading)
+        self._upgrading = upgrading
+        self._closing = True
         #ensure the reaper doesn't call us again:
         if self.child_reaper:
             def noop():
                 pass
-            self.reaper_quit = noop
-            log("clean_quit: reaper_quit=%s", self.reaper_quit)
+            self.reaper_exit = noop
+            log("clean_quit: reaper_exit=%s", self.reaper_exit)
         self.cleanup()
         def quit_timer(*args):
             log.debug("quit_timer()")
+            stop_worker(True)
             self.quit(upgrading)
         #if from a signal, just force quit:
         stop_worker()
@@ -293,6 +330,16 @@ class ServerCore(object):
             os_util.force_quit()
         self.timeout_add(5000, force_quit)
         log("clean_quit(..) quit timers scheduled")
+        self.log_exit_state()
+
+    def log_exit_state(self):
+        frames = sys._current_frames()
+        log("clean_quit() after cleanup, found %s frames:", len(frames))
+        for i,(fid,frame) in enumerate(frames.items()):
+            log("%i: %s - %s:", i, fid, frame)
+            for x in traceback.format_stack(frame):
+                for l in x.splitlines():
+                    log("%s", l)
 
     def quit(self, upgrading):
         log("quit(%s)", upgrading)
@@ -306,16 +353,11 @@ class ServerCore(object):
         raise NotImplementedError()
 
     def get_server_mode(self):
-        return "server"
+        return "core"
 
     def run(self):
-        try:
-            from xpra.src_info import REVISION
-            rev_info = " (r%s)" % REVISION
-        except:
-            rev_info = ""
-        log.info("xpra %s version %s%s", self.get_server_mode(), local_version, rev_info)
-        log.info("running with pid %s", os.getpid())
+        self.print_run_info()
+        self.print_screen_info()
         signal.signal(signal.SIGTERM, self.signal_quit)
         signal.signal(signal.SIGINT, self.signal_quit)
         def start_ready_callbacks():
@@ -325,13 +367,36 @@ class ServerCore(object):
                 except Exception as e:
                     log.error("error on %s: %s", x, e)
         self.idle_add(start_ready_callbacks)
-        def print_ready():
-            log.info("xpra is ready.")
-            sys.stdout.flush()
-        self.idle_add(print_ready)
+        self.idle_add(self.reset_server_timeout)
+        self.idle_add(self.server_is_ready)
         self.do_run()
         return self._upgrading
 
+    def print_run_info(self):
+        try:
+            from xpra.src_info import REVISION
+            rev_info = "-r%s" % REVISION
+        except:
+            rev_info = ""
+        log.info("xpra %s version %s%s", self.get_server_mode(), local_version, rev_info)
+        try:
+            pinfo = get_platform_info()
+            osinfo = " on %s" % platform_name(sys.platform, pinfo.get("linux_distribution") or pinfo.get("release", ""))
+        except:
+            log("platform name error:", exc_info=True)
+            osinfo = ""
+        log.info(" running with pid %s%s", os.getpid(), osinfo)
+
+    def print_screen_info(self):
+        display = os.environ.get("DISPLAY")
+        if display and display.startswith(":"):
+            log.info(" on display %s", display)
+
+
+    def server_is_ready(self):
+        log.info("xpra is ready.")
+        sys.stdout.flush()
+
     def do_run(self):
         raise NotImplementedError()
 
@@ -341,24 +406,50 @@ class ServerCore(object):
             p.quit()
         netlog("cleanup will disconnect: %s", self._potential_protocols)
         if self._upgrading:
-            reason = SERVER_EXIT
+            reason = SERVER_UPGRADE
         else:
             reason = SERVER_SHUTDOWN
-        for proto in list(self._potential_protocols):
-            self.disconnect_client(proto, reason)
+        protocols = self.get_all_protocols()
+        self.cleanup_protocols(protocols, reason)
+        self.do_cleanup()
+        self.cleanup_protocols(protocols, reason, True)
         self._potential_protocols = []
 
+    def do_cleanup(self):
+        #allow just a bit of time for the protocol packet flush
+        time.sleep(0.1)
+
+
+    def cleanup_all_protocols(self, reason):
+        protocols = self.get_all_protocols()
+        self.cleanup_protocols(protocols, reason)
+
+    def get_all_protocols(self):
+        return list(self._potential_protocols)
+
+    def cleanup_protocols(self, protocols, reason, force=False):
+        netlog("do_cleanup_all_protocols(%s, %s, %s)", protocols, reason, force)
+        for protocol in protocols:
+            if force:
+                self.force_disconnect(protocol)
+            else:
+                self.disconnect_protocol(protocol, reason)
+
+
     def add_listen_socket(self, socktype, socket):
         raise NotImplementedError()
 
     def _new_connection(self, listener, *args):
+        if self._closing:
+            netlog.warn("ignoring new connection during shutdown")
+            return False
         socktype = self.socket_types.get(listener)
         assert socktype, "cannot find socket type for %s" % listener
         sock, address = listener.accept()
         if len(self._potential_protocols)>=self._max_connections:
             netlog.error("too many connections (%s), ignoring new one", len(self._potential_protocols))
             sock.close()
-            return  True
+            return True
         try:
             peername = sock.getpeername()
         except:
@@ -371,31 +462,42 @@ class ServerCore(object):
         netlog("socket connection: %s", sc)
         frominfo = ""
         if peername:
-            frominfo = " from %s" % str(peername)
+            frominfo = " from %s" % pretty_socket(peername)
+        elif socktype=="unix-domain":
+            frominfo = " on %s" % sockname
         netlog.info("New %s connection received%s", socktype, frominfo)
         protocol = Protocol(self, sc, self.process_packet)
         self._potential_protocols.append(protocol)
         protocol.large_packets.append("info-response")
+        protocol.challenge_sent = False
         protocol.authenticator = None
         if socktype=="tcp":
             protocol.auth_class = self.tcp_auth_class
+            protocol.encryption = self.tcp_encryption
+            protocol.keyfile = self.tcp_encryption_keyfile
         else:
             protocol.auth_class = self.auth_class
+            protocol.encryption = self.encryption
+            protocol.keyfile = self.encryption_keyfile
         protocol.socket_type = socktype
         protocol.invalid_header = self.invalid_header
         protocol.receive_aliases.update(self._aliases)
+        authlog("socktype=%s, auth class=%s, encryption=%s, keyfile=%s", socktype, protocol.auth_class, protocol.encryption, protocol.keyfile)
+        if protocol.encryption and ENCRYPT_FIRST_PACKET:
+            password = self.get_encryption_key(None, protocol.keyfile)
+            protocol.set_cipher_in(protocol.encryption, DEFAULT_IV, password, DEFAULT_SALT, DEFAULT_ITERATIONS, INITIAL_PADDING)
         protocol.start()
         self.timeout_add(SOCKET_TIMEOUT*1000, self.verify_connection_accepted, protocol)
         return True
 
+
     def invalid_header(self, proto, data):
         netlog("invalid_header(%s, %s bytes: '%s') input_packetcount=%s, tcp_proxy=%s", proto, len(data or ""), repr_ellipsized(data), proto.input_packetcount, self._tcp_proxy)
         if proto.input_packetcount==0 and self._tcp_proxy and not proto._closed:
             #start a new proxy in a thread
             def run_proxy():
                 self.start_tcp_proxy(proto, data)
-            t = make_daemon_thread(run_proxy, "web-proxy-for-%s" % proto)
-            t.start()
+            make_thread(run_proxy, "web-proxy-for-%s" % proto).start()
             return
         err = "invalid packet format, not an xpra client?"
         proto.gibberish(err, data)
@@ -431,8 +533,9 @@ class ServerCore(object):
         #now that we own it, we can start it again:
         client_connection.set_active(True)
         #and we can use blocking sockets:
-        self.set_socket_timeout(client_connection, None)
-        sock.settimeout(None)
+        set_socket_timeout(client_connection, None)
+        #prevent deadlocks on exit:
+        sock.settimeout(1)
 
         proxylog("pushing initial buffer to its new destination: %s", repr_ellipsized(data))
         web_server_connection.write(data)
@@ -486,10 +589,24 @@ class ServerCore(object):
             proto_info = " '%s'" % protocol._conn.get_info().get("endpoint")
         except:
             proto_info = " %s" % protocol
-        netlog.info("Disconnecting client%s: %s", proto_info, i)
+        netlog.info("Disconnecting client%s:", proto_info)
+        netlog.info(" %s", i)
         protocol.flush_then_close(["disconnect", reason]+list(extra))
 
 
+    def _process_disconnect(self, proto, packet):
+        info = packet[1]
+        if len(packet)>2:
+            info += " (%s)" % (", ".join(packet[2:]))
+        #only log protocol info if there is more than one client:
+        proto_info = self._disconnect_proto_info(proto)
+        netlog.info("client%s has requested disconnection: %s", proto_info, info)
+        self.disconnect_protocol(proto, CLIENT_REQUEST)
+
+    def _disconnect_proto_info(self, proto):
+        #overriden in server_base in case there is more than one protocol
+        return ""
+
     def _process_connection_lost(self, proto, packet):
         netlog("process_connection_lost(%s, %s)", proto, packet)
         if proto in self._potential_protocols:
@@ -540,24 +657,23 @@ class ServerCore(object):
             command_req = c.strlistget("command_request")
             if len(command_req)>0:
                 #call from UI thread:
-                self.idle_add(self.handle_command_request, proto, command_req)
+                self.idle_add(self.handle_command_request, proto, *command_req)
                 return
             #continue processing hello packet:
             try:
+                if SIMULATE_SERVER_HELLO_ERROR:
+                    raise Exception("Simulating a server error")
                 self.hello_oked(proto, packet, c, auth_caps)
             except ClientException as e:
-                log.error("error setting up connection for %s: %s", proto, e)
+                log.error("Error setting up new connection for")
+                log.error(" %s:", proto)
+                log.error(" %s", e)
                 self.disconnect_client(proto, SERVER_ERROR, str(e))
             except Exception as e:
                 #log exception but don't disclose internal details to the client
                 log.error("server error processing new connection from %s: %s", proto, e, exc_info=True)
                 self.disconnect_client(proto, SERVER_ERROR, "error accepting new connection")
 
-    def set_socket_timeout(self, conn, timeout=None):
-        #FIXME: this is ugly, but less intrusive than the alternative?
-        if isinstance(conn, SocketConnection):
-            conn._socket.settimeout(timeout)
-
 
     def verify_hello(self, proto, c):
         remote_version = c.strget("version")
@@ -568,16 +684,17 @@ class ServerCore(object):
             return  False
 
         def auth_failed(msg):
-            log.info("authentication failed: %s", msg)
+            authlog.warn("Warning: authentication failed: %s", msg)
             self.timeout_add(1000, self.disconnect_client, proto, msg)
 
         #authenticator:
         username = c.strget("username")
         if proto.authenticator is None and proto.auth_class:
+            authlog("creating authenticator %s", proto.auth_class)
             try:
                 proto.authenticator = proto.auth_class(username)
             except Exception as e:
-                log.warn("error instantiating %s: %s", proto.auth_class, e)
+                authlog.error("Error instantiating %s: %s", proto.auth_class, e)
                 auth_failed("authentication failed")
                 return False
         self.digest_modes = c.get("digest", ("hmac", ))
@@ -587,37 +704,50 @@ class ServerCore(object):
         cipher_iv = c.strget("cipher.iv")
         key_salt = c.strget("cipher.key_salt")
         iterations = c.intget("cipher.key_stretch_iterations")
+        padding = c.strget("cipher.padding", DEFAULT_PADDING)
+        padding_options = c.strlistget("cipher.padding.options", [DEFAULT_PADDING])
         auth_caps = {}
         if cipher and cipher_iv:
             if cipher not in ENCRYPTION_CIPHERS:
-                log.warn("unsupported cipher: %s", cipher)
+                authlog.warn("unsupported cipher: %s", cipher)
                 auth_failed("unsupported cipher")
                 return False
-            encryption_key = self.get_encryption_key(proto.authenticator)
+            encryption_key = self.get_encryption_key(proto.authenticator, proto.keyfile)
             if encryption_key is None:
                 auth_failed("encryption key is missing")
                 return False
-            proto.set_cipher_out(cipher, cipher_iv, encryption_key, key_salt, iterations)
+            if padding not in ALL_PADDING_OPTIONS:
+                auth_failed("unsupported padding: %s" % padding)
+                return False
+            authlog("set output cipher using encryption key %s", encryption_key)
+            proto.set_cipher_out(cipher, cipher_iv, encryption_key, key_salt, iterations, padding)
             #use the same cipher as used by the client:
-            auth_caps = new_cipher_caps(proto, cipher, encryption_key)
-            log("server cipher=%s", auth_caps)
+            auth_caps = new_cipher_caps(proto, cipher, encryption_key, padding_options)
+            authlog("server cipher=%s", auth_caps)
         else:
+            if proto.encryption:
+                authlog.warn("Warning: client does not provide encryption tokens")
+                auth_failed("missing encryption")
+                return False
             auth_caps = None
 
         #verify authentication if required:
         if (proto.authenticator and proto.authenticator.requires_challenge()) or c.get("challenge") is not None:
             challenge_response = c.strget("challenge_response")
             client_salt = c.strget("challenge_client_salt")
-            log("processing authentication with %s, response=%s, client_salt=%s", proto.authenticator, challenge_response, binascii.hexlify(client_salt or ""))
+            authlog("processing authentication with %s, response=%s, client_salt=%s, challenge_sent=%s", proto.authenticator, challenge_response, binascii.hexlify(client_salt or ""), proto.challenge_sent)
             #send challenge if this is not a response:
             if not challenge_response:
+                if proto.challenge_sent:
+                    auth_failed("invalid state, challenge already sent - no response!")
+                    return False                
                 if proto.authenticator:
                     challenge = proto.authenticator.get_challenge()
                     if challenge is None:
-                        auth_failed("invalid authentication state: unexpected challenge response")
+                        auth_failed("invalid state, unexpected challenge response")
                         return False
                     salt, digest = challenge
-                    log.info("Authentication required, %s sending challenge for '%s' using digest %s", proto.authenticator, username, digest)
+                    authlog.info("Authentication required, %s sending challenge for '%s' using digest %s", proto.authenticator, username, digest)
                     if digest not in self.digest_modes:
                         auth_failed("cannot proceed without %s digest support" % digest)
                         return False
@@ -627,13 +757,14 @@ class ServerCore(object):
                     from xpra.os_util import get_hex_uuid
                     salt = get_hex_uuid()+get_hex_uuid()
                     digest = "hmac"
+                proto.challenge_sent = True
                 proto.send_now(("challenge", salt, auth_caps or "", digest))
                 return False
 
             if not proto.authenticator.authenticate(challenge_response, client_salt):
                 auth_failed("invalid challenge response")
                 return False
-            log("authentication challenge passed")
+            authlog("authentication challenge passed")
         else:
             #did the client expect a challenge?
             if c.boolget("challenge"):
@@ -647,22 +778,25 @@ class ServerCore(object):
             return None
         return v.strip("\n\r")
 
-    def get_encryption_key(self, authenticator=None):
+    def get_encryption_key(self, authenticator=None, keyfile=None):
         #if we have a keyfile specified, use that:
-        if self.encryption_keyfile:
-            netlog("trying to load encryption key from keyfile: %s", self.encryption_keyfile)
-            return self.filedata_nocrlf(self.encryption_keyfile)
+        authlog("get_encryption_key(%s, %s)", authenticator, keyfile)
+        if keyfile:
+            authlog("loading encryption key from keyfile: %s", keyfile)
+            return self.filedata_nocrlf(keyfile)
         env_key = os.environ.get('XPRA_ENCRYPTION_KEY')
         if env_key:
+            authlog("using encryption key from %s environment variable", 'XPRA_ENCRYPTION_KEY')
             return env_key
         v = None
         if authenticator:
-            netlog("trying to get encryption key from: %s", authenticator)
+            authlog("trying to get encryption key from: %s", authenticator)
             v = authenticator.get_password()
         if v is None and self.password_file:
-            netlog("trying to load encryption key from password file: %s", self.password_file)
+            authlog("trying to load encryption key from password file: %s", self.password_file)
             v = self.filedata_nocrlf(self.password_file)
         if v is None and os.environ.get('XPRA_PASSWORD'):
+            authlog("using encryption key from %s environment variable", 'XPRA_PASSWORD')
             v = os.environ.get('XPRA_PASSWORD')
         return v
 
@@ -670,45 +804,63 @@ class ServerCore(object):
         pass
 
 
-    def handle_command_request(self, proto, args):
+    def handle_command_request(self, proto, *args):
         """ client sent a command request as part of the hello packet """
-        try:
-            assert len(args)>0
-            command = args[0]
-            error = 0
-            if command not in self.control_commands:
-                commandlog.warn("invalid command: '%s' (must be one of: %s)", command, ", ".join(self.control_commands))
-                error = 6
-                response = "invalid command"
-            else:
-                error, response = self.do_handle_command_request(command, args[1:])
-        except Exception as e:
-            commandlog.error("error processing command %s", command, exc_info=True)
-            error = 127
-            response = "error processing command: %s" % e
-        hello = {"command_response"  : (error, response)}
+        assert len(args)>0
+        code, response = self.process_control_command(*args)
+        hello = {"command_response"  : (code, response)}
         proto.send_now(("hello", hello))
 
-    def do_handle_command_request(self, command, args):
-        """ this may get called by:
-            * handle_command_request from a hello packet
-            * _process_command_request from a dedicated packet
-            it is overriden in subclasses.
-        """
-        if command=="hello":
-            return 0, "hello"
-        elif command=="help":
-            return 0, "control supports: %s" % (", ".join(self.control_commands))
-        return 9, "invalid command '%s'" % command
+    def process_control_command(self, *args):
+        assert len(args)>0
+        name = args[0]
+        try:
+            command = self.control_commands.get(name)
+            commandlog("process_control_command control_commands[%s]=%s", name, command)
+            if not command:
+                commandlog.warn("invalid command: '%s' (must be one of: %s)", name, ", ".join(self.control_commands))
+                return 6, "invalid command"
+            commandlog("process_control_command calling %s%s", command.run, args[1:])
+            v = command.run(*args[1:])
+            return 0, v
+        except ControlError as e:
+            commandlog.error("error %s processing command %s", e.code, name)
+            msgs = [" %s" % e]
+            if e.help:
+                msgs.append(" '%s': %s" % (name, e.help))
+            for msg in msgs:
+                commandlog.error(msg)
+            return e.code, "\n".join(msgs)
+        except Exception as e:
+            commandlog.error("error processing command %s", name, exc_info=True)
+            return 127, "error processing command: %s" % e
 
 
     def accept_client(self, proto, c):
         #max packet size from client (the biggest we can get are clipboard packets)
         netlog("accept_client(%s, %s)", proto, c)
+        #TODO: when we add the code to upload files,
+        #this will need to be increased up to file-size-limit
         proto.max_packet_size = 1024*1024  #1MB
         proto.send_aliases = c.dictget("aliases")
         if proto in self._potential_protocols:
             self._potential_protocols.remove(proto)
+        self.reset_server_timeout(False)
+
+    def reset_server_timeout(self, reschedule=True):
+        timeoutlog("reset_server_timeout(%s) server_idle_timeout=%s, server_idle_timer=%s", reschedule, self.server_idle_timeout, self.server_idle_timer)
+        if self.server_idle_timeout<=0:
+            return
+        if self.server_idle_timer:
+            self.source_remove(self.server_idle_timer)
+            self.server_idle_timer = None
+        if reschedule:
+            self.server_idle_timer = self.timeout_add(self.server_idle_timeout*1000, self.server_idle_timedout)
+
+    def server_idle_timedout(self, *args):
+        timeoutlog.info("No valid client connections for %s seconds, exiting the server", self.server_idle_timeout)
+        self.quit(False)
+
 
     def make_hello(self, source):
         now = time.time()
@@ -721,6 +873,7 @@ class ServerCore(object):
                         "current_time"          : int(now),
                         "elapsed_time"          : int(now - self.start_time),
                         "server_type"           : "core",
+                        "server.mode"           : self.get_server_mode(),
                         })
         if source.wants_features:
             capabilities["info-request"] = True
@@ -753,7 +906,7 @@ class ServerCore(object):
             except Exception as e:
                 log.error("error during info collection: %s", e, exc_info=True)
             callback(proto, ui_info)
-        thread.start_new_thread(in_thread, ())
+        make_thread(in_thread, "Info", daemon=True).start()
 
     def get_ui_info(self, proto, *args):
         #this function is for info which MUST be collected from the UI thread
@@ -774,6 +927,9 @@ class ServerCore(object):
             filtered_env['XPRA_ENCRYPTION_KEY'] = "*****"
 
         up("network",   get_network_caps())
+        up("network", {"encryption"     : self.encryption or "",
+                       "tcp-encryption" : self.tcp_encryption or "",
+                       })
         up("server",    get_server_info())
         up("threads",   self.get_thread_info(proto))
         up("env",       filtered_env)
@@ -781,6 +937,7 @@ class ServerCore(object):
                 "mode"              : self.get_server_mode(),
                 "type"              : "Python",
                 "start_time"        : int(self.start_time),
+                "idle-timeout"      : int(self.server_idle_timeout),
                 "authenticator"     : str((self.auth_class or str)("")),
                 "argv"              : sys.argv,
                 "path"              : sys.path,
@@ -803,7 +960,8 @@ class ServerCore(object):
                 netlog("process packet %s", packet_type)
                 handler(proto, packet)
                 return
-            netlog.error("unknown or invalid packet type: %s from %s", packet_type, proto)
+            if not self._closing:
+                netlog.error("unknown or invalid packet type: '%s' from %s", packet_type, proto)
             proto.close()
         except KeyboardInterrupt:
             raise
