@@ -8,6 +8,7 @@
 from xpra.scripts.config import python_platform as pp
 import sys
 import os
+from xpra.util import updict
 from xpra import __version__ as local_version
 from xpra.log import Logger
 log = Logger("util")
@@ -45,9 +46,14 @@ def get_host_info():
         info.update({
                     "pid"                   : os.getpid(),
                     "byteorder"             : sys.byteorder,
-                    "hostname"              : socket.gethostname(),
-                    "python.full_version"   : sys.version,
-                    "python.version"        : sys.version_info[:3],
+                    "hostname"              : socket.gethostname()
+                    })
+        import struct
+        bits = struct.calcsize("P") * 8
+        updict(info, "python", {
+                    "bits"                  : bits,
+                    "full_version"          : sys.version,
+                    "version"               : ".".join(str(x) for x in sys.version_info[:3]),
                     })
     except:
         pass
@@ -65,29 +71,38 @@ def get_version_info():
              }
     try:
         from xpra.src_info import LOCAL_MODIFICATIONS, REVISION
-        props.update({
-                    "local_modifications"  : LOCAL_MODIFICATIONS,
-                    "revision"             : REVISION,
-                  })
-    except Exception as e:
+        props["local_modifications"]    = LOCAL_MODIFICATIONS
+        props["revision"]               = REVISION
+    except ImportError as e:
         log.warn("missing some source information: %s", e)
+    return props
+
+def get_version_info_full():
+    props = get_version_info()
     try:
-        from xpra.build_info import BUILD_DATE, BUILT_BY, BUILT_ON, BUILD_BIT, BUILD_CPU, \
-                                    COMPILER_VERSION, LINKER_VERSION, BUILD_TIME, PYTHON_VERSION, CYTHON_VERSION
-        props.update({
-                    "date"                 : BUILD_DATE,
-                    "time"                 : BUILD_TIME,
-                    "by"                   : BUILT_BY,
-                    "on"                   : BUILT_ON,
-                    "bit"                  : BUILD_BIT,
-                    "cpu"                  : BUILD_CPU,
-                    "compiler"             : COMPILER_VERSION,
-                    "linker"               : LINKER_VERSION,
-                    "python"               : PYTHON_VERSION,
-                    "cython"               : CYTHON_VERSION,
-                  })
+        from xpra import build_info
+        #rename these build info properties:
+        for k,bk in {
+                    "date"                 : "BUILD_DATE",
+                    "time"                 : "BUILD_TIME",
+                    "by"                   : "BUILT_ON",
+                    "bit"                  : "BUILD_BIT",
+                    "cpu"                  : "BUILD_CPU",
+                    "compiler"             : "COMPILER_VERSION",
+                    "nvcc"                 : "NVCC_VERSION",
+                    "linker"               : "LINKER_VERSION",
+                    "python"               : "PYTHON_VERSION",
+                    "cython"               : "CYTHON_VERSION",
+                  }.items():
+            v = getattr(build_info, bk, None)
+            if v:
+                props[k] = v
+        #record library versions:
+        d = dict((k.lstrip("lib_"), getattr(build_info, k)) for k in dir(build_info) if k.startswith("lib_"))
+        updict(props, "lib", d)
     except Exception as e:
         log.warn("missing some build information: %s", e)
+    log("get_version_info_full()=%s", props)
     return props
 
 def do_get_platform_info():
@@ -108,21 +123,22 @@ def do_get_platform_info():
                 if "model name" in line:
                     return re.sub(".*model name.*:", "", line,1).strip()
         assert False
-    info = {
+    info = {}
+    if sys.platform.startswith("linux") and hasattr(pp, "linux_distribution"):
+        info["linux_distribution"] = pp.linux_distribution()
+    info.update({
             ""          : sys.platform,
-            "name"      : platform_name(sys.platform, pp.release()),
+            "name"      : platform_name(sys.platform, info.get("linux_distribution") or pp.release()),
             "release"   : pp.release(),
             "platform"  : pp.platform(),
             "machine"   : pp.machine(),
             "processor" : pp.processor(),
             "architecture" : pp.architecture(),
-            }
+            })
     try:
         info["processor"] = get_processor_name()
     except:
         info["processor"] = pp.processor()
-    if sys.platform.startswith("linux") and hasattr(pp, "linux_distribution"):
-        info["linux_distribution"] = pp.linux_distribution()
     return info
 #cache the output:
 platform_info_cache = None
