@@ -9,16 +9,32 @@
 Everyone else should just use prop_set/prop_get with nice clean Python calling
 conventions, and if you need more (un)marshalling smarts, add them here."""
 
+import sys
 import binascii
 import struct
-import gtk.gdk
 import cairo
 
-from xpra.x11.gtk_x11.gdk_bindings import (
-                get_xatom, get_pyatom,      #@UnresolvedImport
-                get_xwindow, get_pywindow,  #@UnresolvedImport
-                get_xvisual,                #@UnresolvedImport
-               )
+from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, is_gtk3
+gtk = import_gtk()
+gdk = import_gdk()
+from xpra.gtk_common.gtk_util import get_xwindow
+
+from xpra.log import Logger
+log = Logger("x11", "window")
+
+try:
+    from xpra.x11.gtk2.gdk_bindings import (
+                    get_pywindow,               #@UnresolvedImport
+                    get_xvisual,                #@UnresolvedImport
+                   )
+except ImportError as e:
+    #we should only ever be missing the gdk_bindings with GTK3 builds:
+    log("cannot import gdk bindings", exc_info=True)
+    assert is_gtk3()
+    def missing_fn(*args):
+        raise NotImplementedError()
+    get_pywindow, get_xvisual = missing_fn, missing_fn
+
 from xpra.x11.bindings.window_bindings import (
                 constants,                      #@UnresolvedImport
                 X11WindowBindings,          #@UnresolvedImport
@@ -26,18 +42,10 @@ from xpra.x11.bindings.window_bindings import (
 X11Window = X11WindowBindings()
 
 from xpra.os_util import StringIOClass
-from xpra.x11.xsettings_prop import set_settings, get_settings
-from xpra.gtk_common.error import trap, XError
-from xpra.log import Logger
-log = Logger("x11", "window")
+from xpra.gtk_common.error import xsync, XError
+from xpra.codecs.argb.argb import premultiply_argb_in_place #@UnresolvedImport
 
-try:
-    from xpra.codecs.argb.argb import premultiply_argb_in_place #@UnresolvedImport
-except:
-    log.warn("argb module is missing, icon transparency will be incorrect")
-    premultiply_argb_in_place = None
 
-import sys
 if sys.version > '3':
     long = int              #@ReservedAssignment
     unicode = str           #@ReservedAssignment
@@ -71,135 +79,29 @@ def _force_length(name, data, length, noerror_length=None):
     data += "\0" * length
     return data[:length]
 
-class WMSizeHints(object):
-    def __init__(self, disp, data):
-        # pre-ICCCM size is 15
-        data = _force_length("WM_SIZE_HINTS", data, 18*4, noerror_length=15*4)
-        def normint(v):
-            if v==(2**32-1):
-                return -1
-            return v
-        (flags,
-         pad1, pad2, pad3, pad4,            #@UnusedVariable
-         min_width, min_height,
-         max_width, max_height,
-         width_inc, height_inc,
-         min_aspect_num, min_aspect_denom,
-         max_aspect_num, max_aspect_denom,
-         base_width, base_height,
-         win_gravity) = [normint(x) for x in struct.unpack("=" + "I" * 18, data)] #@UnusedVariable
-        # We only extract the pieces we care about:
-        if flags & PMaxSize:
-            self.max_size = (max_width, max_height)
-        else:
-            self.max_size = None
-        if flags & PMinSize:
-            self.min_size = (min_width, min_height)
-        else:
-            self.min_size = None
-        if flags & PBaseSize:
-            self.base_size = (base_width, base_height)
-        else:
-            self.base_size = None
-        if flags & PResizeInc:
-            self.resize_inc = (width_inc, height_inc)
-        else:
-            self.resize_inc = None
-        if flags & PAspect:
-            self.min_aspect = min_aspect_num * 1.0 / min_aspect_denom
-            self.min_aspect_ratio = (min_aspect_num, min_aspect_denom)
-            self.max_aspect = max_aspect_num * 1.0 / max_aspect_denom
-            self.max_aspect_ratio = (max_aspect_num,  max_aspect_denom)
-        else:
-            self.min_aspect, self.max_aspect = (None, None)
-            self.min_aspect_ratio, self.max_aspect_ratio = (None, None)
-        if flags & PWinGravity:
-            self.win_gravity = win_gravity
-        else:
-            self.win_gravity = -1
-        self.set_initial_position = bool((flags & USPosition) or (flags & PPosition))
-
-    def to_dict(self, include_none=False):
-        d = {}
-        for attr, metakey in [
-            ("max_size", "maximum-size"),
-            ("min_size", "minimum-size"),
-            ("base_size", "base-size"),
-            ("resize_inc", "increment"),
-            ("min_aspect_ratio", "minimum-aspect-ratio"),
-            ("max_aspect_ratio", "maximum-aspect-ratio"),
-            ("set_initial_position", "set-initial-position")
-            ]:
-            v = getattr(self, attr)
-            if v is not None or include_none:
-                d[metakey] = v
-        return d
-
-    def __str__(self):
-        return "WMSizeHints(%s)" % {"max_size"  : self.max_size,
-                                    "min_size"  : self.min_size,
-                                    "base_size" : self.base_size,
-                                    "resize_inc": self.resize_inc,
-                                    "min_aspect": self.min_aspect,
-                                    "max_aspect": self.max_aspect,
-                                    "min_aspect_ratio"  : self.min_aspect_ratio,
-                                    "max_aspect_ratio"  : self.max_aspect_ratio,
-                                    "win_gravity"       : self.win_gravity,
-                                    "set_initial_position"  : self.set_initial_position}
-
-
-class WMHints(object):
-    def __init__(self, disp, data):
-        data = _force_length("WM_HINTS", data, 9 * 4)
-        (flags, _input, initial_state,  #@UnusedVariable
-         icon_pixmap, icon_window,      #@UnusedVariable
-         icon_x, icon_y, icon_mask,     #@UnusedVariable
-         window_group) = struct.unpack("=" + "i" * 9, data)
-        # NB the last field is missing from at least some ICCCM 2.0's (typo).
-        # FIXME: extract icon stuff too
-        self.urgency = bool(flags & XUrgencyHint)
-        if flags & WindowGroupHint:
-            try:
-                pywin = get_pywindow(disp, window_group)
-            except:
-                pywin = None
-            self.group_leader = (window_group, pywin)
-        else:
-            self.group_leader = None
-        if flags & StateHint:
-            self.start_iconic = (initial_state == IconicState)
-        else:
-            self.start_iconic = None
-        if flags & InputHint:
-            self.input = _input
-        else:
-            self.input = None
-
-    def __str__(self):
-        return "WMHints(%s)" % {"group_leader"  : self.group_leader,
-                                "start_iconic"  : self.start_iconic,
-                                "input"         : self.input}
 
 class NetWMStrut(object):
     def __init__(self, disp, data):
         # This eats both _NET_WM_STRUT and _NET_WM_STRUT_PARTIAL.  If we are
         # given a _NET_WM_STRUT instead of a _NET_WM_STRUT_PARTIAL, then it
-        # will be only length 4 instead of 12, but _force_length will zero-pad
-        # and _NET_WM_STRUT is *defined* as a _NET_WM_STRUT_PARTIAL where the
-        # extra fields are zero... so it all works out.
+        # will be only length 4 instead of 12, we just don't define the other values
+        # and let the client deal with it appropriately
         if len(data)==16:
             self.left, self.right, self.top, self.bottom = struct.unpack("=IIII", data)
-            return
-        data = _force_length("_NET_WM_STRUT or _NET_WM_STRUT_PARTIAL", data, 4 * 12)
-        (self.left, self.right, self.top, self.bottom,
-         self.left_start_y, self.left_end_y,
-         self.right_start_y, self.right_end_y,
-         self.top_start_x, self.top_end_x,
-         self.bottom_start_x, self.bottom_stop_x,
-         ) = struct.unpack("=" + "I" * 12, data)
+        else:
+            data = _force_length("_NET_WM_STRUT or _NET_WM_STRUT_PARTIAL", data, 4 * 12)
+            (self.left, self.right, self.top, self.bottom,
+             self.left_start_y, self.left_end_y,
+             self.right_start_y, self.right_end_y,
+             self.top_start_x, self.top_end_x,
+             self.bottom_start_x, self.bottom_stop_x,
+             ) = struct.unpack("=" + "I" * 12, data)
+
+    def todict(self):
+        return self.__dict__
 
     def __str__(self):
-        return "NetWMStrut(%s)" % self.__dict__
+        return "NetWMStrut(%s)" % self.todict()
 
 
 class MotifWMHints(object):
@@ -210,6 +112,44 @@ class MotifWMHints(object):
             struct.unpack("=IIIiI", pdata)
         log("MotifWMHints(%s)=%s", binascii.hexlify(data), self)
 
+    #found in mwmh.h:
+    # "flags":
+    FUNCTIONS_BIT   = 0
+    DECORATIONS_BIT = 1
+    INPUT_MODE_BIT  = 2
+    STATUS_BIT      = 3
+    # "functions":
+    ALL_BIT         = 0
+    RESIZE_BIT      = 1
+    MOVE_BIT        = 2      # like _NET_WM_ACTION_MOVE
+    MINIMIZE_BIT    = 3      # like _NET_WM_ACTION_MINIMIZE
+    MAXIMIZE_BIT    = 4      # like _NET_WM_ACTION_(FULLSCREEN|MAXIMIZE_(HORZ|VERT))
+    CLOSE_BIT       = 5      # like _NET_WM_ACTION_CLOSE
+    SHADE_BIT       = 6      # like _NET_WM_ACTION_SHADE
+    STICK_BIT       = 7      # like _NET_WM_ACTION_STICK
+    FULLSCREEN_BIT  = 8      # like _NET_WM_ACTION_FULLSCREEN
+    ABOVE_BIT       = 9      # like _NET_WM_ACTION_ABOVE
+    BELOW_BIT       = 10     # like _NET_WM_ACTION_BELOW
+    MAXIMUS_BIT     = 11     # like _NET_WM_ACTION_MAXIMUS_(LEFT|RIGHT|TOP|BOTTOM)
+    # "decorations":
+    ALL_BIT         = 0
+    BORDER_BIT      = 1
+    RESIZEH_BIT     = 2
+    TITLE_BIT       = 3
+    MENU_BIT        = 4
+    MINIMIZE_BIT    = 5
+    MAXIMIZE_BIT    = 6
+    #CLOSE_BIT                # non-standard close button
+    #RESIZE_BIT               # non-standard resize button
+    #SHADE_BIT,               # non-standard shade button
+    #STICK_BIT,               # non-standard stick button
+    #MAXIMUS_BIT              # non-standard maxim
+    # "input":
+    MODELESS        = 0
+    PRIMARY_APPLICATION_MODAL = 1
+    SYSTEM_MODAL    = 2
+    FULL_APPLICATION_MODAL = 3
+
     def __str__(self):
         return "MotifWMHints(%s)" % {"flags"        : self.flags,
                                      "functions"    : self.functions,
@@ -228,7 +168,7 @@ def _read_image(disp, stream):
         if len(data) < width * height * 4:
             log.warn("Corrupt _NET_WM_ICON")
             return None
-    except Exception, e:
+    except Exception as e:
         log.warn("Weird corruption in _NET_WM_ICON: %s", e)
         return None
     # Cairo wants a native-endian array here, and since the icon is
@@ -247,8 +187,7 @@ def _read_image(disp, stream):
     # Cairo uses premultiplied alpha. EWMH actually doesn't specify what it
     # uses, but apparently the de-facto standard is non-premultiplied. (At
     # least that's what Compiz's sources say.)
-    if premultiply_argb_in_place:
-        premultiply_argb_in_place(surf.get_data())
+    premultiply_argb_in_place(surf.get_data())
     return (width * height, surf)
 
 # This returns a cairo ImageSurface which contains the largest icon defined in
@@ -268,11 +207,19 @@ def NetWMIcons(disp, data):
 
 def _get_atom(disp, d):
     unpacked = struct.unpack("@I", d)[0]
-    pyatom = get_pyatom(disp, unpacked)
+    with xsync:
+        pyatom = X11Window.XGetAtomName(unpacked)
     if not pyatom:
         log.error("invalid atom: %s - %s", repr(d), repr(unpacked))
         return  None
-    return str(pyatom)
+    if type(pyatom)!=str:
+        #py3k:
+        return pyatom.decode()
+    return pyatom
+
+def _get_xatom(str_or_int):
+    with xsync:
+        return X11Window.get_xatom(str_or_int)
 
 def _get_multiple(disp, d):
     uint_struct = struct.Struct("@I")
@@ -282,49 +229,77 @@ def _get_multiple(disp, d):
         return  str(d)
     return _get_atom(disp, d)
 
+def _to_latin1(disp, v):
+    return v.encode("latin1")
+
+def _from_latin1(disp, v):
+    return v.decode("latin1")
+
+def _to_utf8(disp, v):
+    return v.encode("UTF-8")
+
+def _from_utf8(disp, v):
+    return v.decode("UTF-8")
+
+
+def _get_display_name(disp):
+    try:
+        return disp.get_display().get_name()
+    except:
+        return None
+
+def set_xsettings(disp, v):
+    from xpra.x11.xsettings_prop import set_settings
+    return set_settings(_get_display_name(disp), v)
+
+def get_xsettings(disp, v):
+    from xpra.x11.xsettings_prop import get_settings
+    return get_settings(_get_display_name(disp), v)
+
+
+def get_python_type(scalar_type):
+    #ie: get_python_type("STRING") = "latin1"
+    return {"UTF8_STRING"  : "utf8",
+            "STRING"       : "latin1",
+            "ATOM"         : "atom",
+            "CARDINAL"     : "u32",
+            "INTEGER"      : "integer",
+            "VISUALID"     : "visual",
+            "WINDOW"       : "window",
+            }.get(scalar_type)
 
 _prop_types = {
     # Python type, X type Atom, formatbits, serializer, deserializer, list
     # terminator
-    "utf8": (unicode, "UTF8_STRING", 8,
-             lambda disp, u: u.encode("UTF-8"),
-             lambda disp, d: d.decode("UTF-8"),
-             "\0"),
+    "utf8": (unicode, "UTF8_STRING", 8, _to_utf8, _from_utf8, b"\0"),
     # In theory, there should be something clever about COMPOUND_TEXT here.  I
     # am not sufficiently clever to deal with COMPOUNT_TEXT.  Even knowing
     # that Xutf8TextPropertyToTextList exists.
-    "latin1": (unicode, "STRING", 8,
-               lambda disp, u: u.encode("latin1"),
-               lambda disp, d: d.decode("latin1"),
-               "\0"),
+    "latin1": (unicode, "STRING", 8, _to_latin1, _from_latin1, b"\0"),
     "atom": (str, "ATOM", 32,
-             lambda disp, a: struct.pack("@I", get_xatom(a)),
+             lambda disp, a: struct.pack("@I", _get_xatom(a)),
               _get_atom,
-             ""),
+             b""),
     "state": ((int, long), "WM_STATE", 32,
             lambda disp, c: struct.pack("=I", c),
             lambda disp, d: struct.unpack("=I", d)[0],
-            ""),
+            b""),
     "u32": ((int, long), "CARDINAL", 32,
             lambda disp, c: struct.pack("=I", c),
             lambda disp, d: struct.unpack("=I", d)[0],
-            ""),
-    "visual": (gtk.gdk.Visual, "VISUALID", 32,
+            b""),
+    "integer": ((int, long), "INTEGER", 32,
+            lambda disp, c: struct.pack("=I", c),
+            lambda disp, d: struct.unpack("=I", d)[0],
+            b""),
+    "visual": (gdk.Visual, "VISUALID", 32,
                lambda disp, c: struct.pack("=I", get_xvisual(c)),
                unsupported,
-               ""),
-    "window": (gtk.gdk.Window, "WINDOW", 32,
+               b""),
+    "window": (gdk.Window, "WINDOW", 32,
                lambda disp, c: struct.pack("=I", get_xwindow(c)),
                lambda disp, d: get_pywindow(disp, struct.unpack("=I", d)[0]),
-               ""),
-    "wm-size-hints": (WMSizeHints, "WM_SIZE_HINTS", 32,
-                      unsupported,
-                      WMSizeHints,
-                      None),
-    "wm-hints": (WMHints, "WM_HINTS", 32,
-                 unsupported,
-                 WMHints,
-                 None),
+               b""),
     "strut": (NetWMStrut, "CARDINAL", 32,
               unsupported, NetWMStrut, None),
     "strut-partial": (NetWMStrut, "CARDINAL", 32,
@@ -334,8 +309,8 @@ _prop_types = {
     "icon": (cairo.ImageSurface, "CARDINAL", 32,
              unsupported, NetWMIcons, None),
     "xsettings-settings": (tuple, "_XSETTINGS_SETTINGS", 8,
-                           set_settings,
-                           get_settings,
+                           set_xsettings,
+                           get_xsettings,
                            None),
     # For uploading ad-hoc instances of the above complex structures to the
     # server, so we can test reading them out again:
@@ -373,7 +348,8 @@ def _prop_encode_list(disp, etype, value):
 
 
 def prop_set(target, key, etype, value):
-    trap.call_synced(X11Window.XChangeProperty, get_xwindow(target), key,
+    with xsync:
+        X11Window.XChangeProperty(get_xwindow(target), key,
                        _prop_encode(target, etype, value))
 
 def _prop_decode(disp, etype, data):
@@ -410,7 +386,8 @@ def prop_get(target, key, etype, ignore_errors=False, raise_xerrors=False):
         scalar_type = etype
     (_, atom, _, _, _, _) = _prop_types[scalar_type]
     try:
-        data = trap.call_synced(X11Window.XGetWindowProperty, get_xwindow(target), key, atom, etype)
+        with xsync:
+            data = X11Window.XGetWindowProperty(get_xwindow(target), key, atom, etype)
         if data is None:
             if not ignore_errors:
                 log("Missing property %s (%s)", key, etype)
@@ -433,3 +410,7 @@ def prop_get(target, key, etype, ignore_errors=False, raise_xerrors=False):
                      + "  Data: %r[...?]",
                      key, etype, data[:160], exc_info=True)
         raise
+
+def prop_del(target, key):
+    with xsync:
+        X11Window.XDeleteProperty(get_xwindow(target), key)
