@@ -12,6 +12,7 @@ from xpra.log import Logger
 log = Logger("keyboard")
 
 
+from xpra.util import csv
 from xpra.gtk_common.keymap import get_gtk_keymap
 from xpra.x11.gtk_x11.keys import grok_modifier_map
 from xpra.keyboard.mask import DEFAULT_MODIFIER_NUISANCE, DEFAULT_MODIFIER_NUISANCE_KEYNAMES, mask_to_names
@@ -108,7 +109,7 @@ class KeyboardConfig(KeyboardConfigBase):
             v = getattr(self, "xkbmap_mod_"+x)
             if v:
                 info["modifiers."+x] = v
-        log("keyboard info: %s", "\n".join(["%s=%s" % (k,v) for k,v in info.items()]))
+        log("keyboard info: %s", info)
         return info
 
 
@@ -351,34 +352,35 @@ class KeyboardConfig(KeyboardConfigBase):
             #this is not a keyboard event, ignore modifiers in "mod_pointermissing"
             def is_ignored(modifier, modifier_keynames):
                 m = modifier in (self.xkbmap_mod_pointermissing or [])
-                log("is_ignored(%s, %s)=%s", modifier, modifier_keynames, m)
                 return m
         else:
             #keyboard event: ignore the keynames specified
             #(usually the modifier key being pressed/unpressed)
             def is_ignored(modifier, modifier_keynames):
                 m = set(modifier_keynames) & set(ignored_modifier_keynames)
-                log("is_ignored(%s, %s)=%s", modifier, modifier_keynames, m)
                 return bool(m)
 
-        current = set(self.get_current_mask())
-        wanted = set(modifier_list or [])
-        if current==wanted:
-            return
-        log("make_keymask_match(%s) current mask: %s, wanted: %s, ignoring=%s/%s, keys_pressed=%s", modifier_list, current, wanted, ignored_modifier_keycode, ignored_modifier_keynames, self.keys_pressed)
-
-        def change_mask(modifiers, press, info):
+        def filtered_modifiers_set(modifiers):
+            m = set()
             for modifier in modifiers:
                 if self.xkbmap_mod_managed and modifier in self.xkbmap_mod_managed:
                     log("modifier is server managed: %s", modifier)
                     continue
                 keynames = self.keynames_for_mod.get(modifier)
-                if not keynames:
-                    log.error("unknown modifier: %s", modifier)
-                    continue
                 if is_ignored(modifier, keynames):
                     log("modifier %s ignored (in ignored keynames=%s)", modifier, keynames)
                     continue
+                m.add(modifier)
+            log("filtered_modifiers_set(%s)=%s", modifiers, csv(list(m)))
+            return m
+
+        def change_mask(modifiers, press, info):
+            failed = []
+            for modifier in modifiers:
+                keynames = self.keynames_for_mod.get(modifier)
+                if not keynames:
+                    log.error("unknown modifier: %s", modifier)
+                    continue
                 #find the keycodes that match the keynames for this modifier
                 keycodes = []
                 #log.info("keynames(%s)=%s", modifier, keynames)
@@ -399,8 +401,16 @@ class KeyboardConfig(KeyboardConfigBase):
                     continue
                 #nuisance keys (lock, num, scroll) are toggled by a
                 #full key press + key release (so act accordingly in the loop below)
-                nuisance = modifier in DEFAULT_MODIFIER_NUISANCE
+                nuisance = modifier in self.xkbmap_mod_nuisance
                 log("keynames(%s)=%s, keycodes=%s, nuisance=%s", modifier, keynames, keycodes, nuisance)
+                modkeycode = None
+                if not press:
+                    #since we want to unpress something,
+                    #let's try the keycodes we know are pressed first:
+                    kdown = X11Keyboard.get_keycodes_down()
+                    pressed = [x for x in keycodes if x in kdown]
+                    others = [x for x in keycodes if x not in kdown]
+                    keycodes = pressed+others
                 for keycode in keycodes:
                     if nuisance:
                         X11Keyboard.xtest_fake_key(keycode, True)
@@ -409,16 +419,43 @@ class KeyboardConfig(KeyboardConfigBase):
                         X11Keyboard.xtest_fake_key(keycode, press)
                     new_mask = self.get_current_mask()
                     success = (modifier in new_mask)==press
-                    log("make_keymask_match(%s) %s modifier %s using %s, success: %s", info, modifier_list, modifier, keycode, success)
                     if success:
-                        break
-                    elif not nuisance:
-                        log("%s %s with keycode %s did not work - trying to undo it!", info, modifier, keycode)
-                        X11Keyboard.xtest_fake_key(keycode, not press)
+                        modkeycode = keycode
+                        log("change_mask(%s) %s modifier %s using %s", info, modifier_list, modifier, keycode)
+                        break   #we're done for this modifier
+                    log("%s %s with keycode %s did not work", info, modifier, keycode)
+                    if press and not nuisance:
+                        log(" trying to unpress it!", info, modifier, keycode)
+                        X11Keyboard.xtest_fake_key(keycode, False)
+                        #maybe doing the full keypress (down+up) worked:
                         new_mask = self.get_current_mask()
-                        #maybe doing the full keypress (down+up or u+down) worked:
                         if (modifier in new_mask)==press:
                             break
+                    log("change_mask(%s) %s modifier %s using %s, success: %s", info, modifier_list, modifier, keycode, success)
+                if not modkeycode:
+                    failed.append(modifier)
+            log("change_mask(%s, %s, %s) failed=%s", modifiers, press, info, failed)
+            return failed
 
+        current = filtered_modifiers_set(self.get_current_mask())
+        wanted = filtered_modifiers_set(modifier_list or [])
+        if current==wanted:
+            return
+        log("make_keymask_match(%s) current mask: %s, wanted: %s, ignoring=%s/%s, keys_pressed=%s", modifier_list, current, wanted, ignored_modifier_keycode, ignored_modifier_keynames, self.keys_pressed)
+        fr = change_mask(current.difference(wanted), False, "remove")
+        fa = change_mask(wanted.difference(current), True, "add")
+        if fr:
+            log.warn("Warning: failed to remove the following modifiers: %s", csv(fr))
+        elif fa:
+            log.warn("Warning: failed to add the following modifiers: %s", csv(fa))
+        else:
+            return  #all good!
+        #this should never happen.. but if it does?
+        #something didn't work, use the big hammer and start again from scratch:
+        log.warn(" keys still pressed=%s", X11Keyboard.get_keycodes_down())
+        X11Keyboard.unpress_all_keys()
+        log.warn(" doing a full keyboard reset, keys now pressed=%s", X11Keyboard.get_keycodes_down())
+        #and try to set the modifiers one last time:
+        current = filtered_modifiers_set(self.get_current_mask())
         change_mask(current.difference(wanted), False, "remove")
         change_mask(wanted.difference(current), True, "add")
