@@ -6,13 +6,11 @@
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+import os
 import gtk.gdk
-import gobject
-import os.path
-import sys
 
 #ensure that we use gtk as display source:
-from xpra.x11.gtk_x11 import gdk_display_source
+from xpra.x11.gtk2 import gdk_display_source
 assert gdk_display_source
 
 from xpra.x11.bindings.randr_bindings import RandRBindings  #@UnresolvedImport
@@ -23,6 +21,7 @@ from xpra.x11.bindings.core_bindings import X11CoreBindings #@UnresolvedImport
 X11Core = X11CoreBindings()
 from xpra.gtk_common.error import XError, xswallow, xsync
 from xpra.server.server_uuid import save_uuid, get_uuid
+from xpra.x11.fakeXinerama import find_libfakeXinerama, save_fakeXinerama_config, cleanup_fakeXinerama
 
 from xpra.log import Logger
 log = Logger("x11", "server")
@@ -30,14 +29,15 @@ keylog = Logger("x11", "server", "keyboard")
 mouselog = Logger("x11", "server", "mouse")
 grablog = Logger("server", "grab")
 cursorlog = Logger("server", "cursor")
+screenlog = Logger("server", "screen")
 
-from xpra.util import prettify_plug_name
-from xpra.os_util import find_lib, find_lib_ldconfig
+from xpra.util import iround
 from xpra.server.gtk_server_base import GTKServerBase
 from xpra.x11.xkbhelper import clean_keyboard_state
 from xpra.x11.server_keyboard_config import KeyboardConfig
 
 MAX_CONCURRENT_CONNECTIONS = 20
+RANDR = os.environ.get("XPRA_RANDR", "1")=="1"
 
 
 def window_name(window):
@@ -50,7 +50,7 @@ def window_info(window):
     return "%s %s (%s / %s)" % (net_wm_name, window, window.get_geometry(), window.is_visible())
 
 
-from xpra.x11.gtk_x11.gdk_bindings import get_children #@UnresolvedImport
+from xpra.x11.gtk2.gdk_bindings import get_children #@UnresolvedImport
 def dump_windows():
     root = gtk.gdk.get_default_root_window()
     log("root window: %s" % root)
@@ -69,19 +69,30 @@ class X11ServerBase(GTKServerBase):
 
     def __init__(self, clobber):
         self.clobber = clobber
+        self.screen_number = gtk.gdk.display_get_default().get_default_screen().get_number()
+        self.root_window = gtk.gdk.get_default_root_window()
         GTKServerBase.__init__(self)
 
     def init(self, opts):
         self.fake_xinerama = opts.fake_xinerama
         self.current_xinerama_config = None
+        if self.fake_xinerama:
+            self.libfakeXinerama_so = find_libfakeXinerama()
+        else:
+            self.libfakeXinerama_so = None
         self.x11_init()
         GTKServerBase.init(self, opts)
 
     def x11_init(self):
+        if not X11Keyboard.hasXFixes() and self.cursors:
+            log.error("Error: cursor forwarding support disabled")
+        if not X11Keyboard.hasXTest():
+            log.error("Error: keyboard and mouse disabled")
+        elif not X11Keyboard.hasXkb():
+            log.error("Error: limited keyboard support")
         self.init_x11_atoms()
-        self.randr = RandR.has_randr()
+        self.randr = RANDR and RandR.has_randr()
         if self.randr and len(RandR.get_screen_sizes())<=1:
-            log.info("no RandR support")
             #disable randr when we are dealing with a Xvfb
             #with only one resolution available
             #since we don't support adding them on the fly yet
@@ -93,29 +104,31 @@ class X11ServerBase(GTKServerBase):
                 screen = display.get_screen(i)
                 screen.connect("size-changed", self._screen_size_changed)
                 i += 1
+        else:
+            log.warn("Warning: no X11 RandR support on %s", os.environ.get("DISPLAY"))
         log("randr enabled: %s", self.randr)
 
     def init_x11_atoms(self):
         #some applications (like openoffice), do not work properly
         #if some x11 atoms aren't defined, so we define them in advance:
-        for atom_name in ["_NET_WM_WINDOW_TYPE",
-                          "_NET_WM_WINDOW_TYPE_NORMAL",
-                          "_NET_WM_WINDOW_TYPE_DESKTOP",
-                          "_NET_WM_WINDOW_TYPE_DOCK",
-                          "_NET_WM_WINDOW_TYPE_TOOLBAR",
-                          "_NET_WM_WINDOW_TYPE_MENU",
-                          "_NET_WM_WINDOW_TYPE_UTILITY",
-                          "_NET_WM_WINDOW_TYPE_SPLASH",
-                          "_NET_WM_WINDOW_TYPE_DIALOG",
-                          "_NET_WM_WINDOW_TYPE_DROPDOWN_MENU",
-                          "_NET_WM_WINDOW_TYPE_POPUP_MENU",
-                          "_NET_WM_WINDOW_TYPE_TOOLTIP",
-                          "_NET_WM_WINDOW_TYPE_NOTIFICATION",
-                          "_NET_WM_WINDOW_TYPE_COMBO",
-                          "_NET_WM_WINDOW_TYPE_DND",
-                          "_NET_WM_WINDOW_TYPE_NORMAL"
-                          ]:
-            X11Core.get_xatom(atom_name)
+        for wtype in ["",
+                      "_NORMAL",
+                      "_DESKTOP",
+                      "_DOCK",
+                      "_TOOLBAR",
+                      "_MENU",
+                      "_UTILITY",
+                      "_SPLASH",
+                      "_DIALOG",
+                      "_DROPDOWN_MENU",
+                      "_POPUP_MENU",
+                      "_TOOLTIP",
+                      "_NOTIFICATION",
+                      "_COMBO",
+                      "_DND",
+                      "_NORMAL"
+                      ]:
+            X11Core.get_xatom("_NET_WM_WINDOW_TYPE"+wtype)
 
     def init_keyboard(self):
         GTKServerBase.init_keyboard(self)
@@ -128,24 +141,17 @@ class X11ServerBase(GTKServerBase):
         self._authenticated_ui_packet_handlers["force-ungrab"] = self._process_force_ungrab
 
 
-    def get_child_env(self):
-        #adds fakexinerama:
-        env = GTKServerBase.get_child_env(self)
-        if self.fake_xinerama:
-            libfakeXinerama_so = self.find_fakeXinerama()
-            if libfakeXinerama_so:
-                env["LD_PRELOAD"] = libfakeXinerama_so
-        return env
+    def get_server_source_class(self):
+        from xpra.x11.x11_source import X11ServerSource
+        return X11ServerSource
 
-    def find_fakeXinerama(self):
-        if sys.platform.startswith("linux"):
-            try:
-                libpath = find_lib_ldconfig("fakeXinerama")
-                if libpath:
-                    return libpath
-            except:
-                log.warn("Failed to launch ldconfig -p")
-        return find_lib("libfakeXinerama.so.1")
+
+    def get_child_env(self):
+        #adds fakeXinerama:
+        env = GTKServerBase.get_child_env(self)
+        if self.fake_xinerama and self.libfakeXinerama_so:
+            env["LD_PRELOAD"] = self.libfakeXinerama_so
+        return env
 
 
     def get_uuid(self):
@@ -172,15 +178,18 @@ class X11ServerBase(GTKServerBase):
         capabilities = GTKServerBase.make_hello(self, source)
         capabilities["server_type"] = "Python/gtk/x11"
         if source.wants_features:
-            capabilities["resize_screen"] = self.randr
-            capabilities["force_ungrab"] = True
+            capabilities.update({
+                    "resize_screen"             : self.randr,
+                    "force_ungrab"              : True,
+                    "keyboard.fast-switching"   : True,
+                    })
         return capabilities
 
     def do_get_info(self, proto, server_sources, window_ids):
         info = GTKServerBase.do_get_info(self, proto, server_sources, window_ids)
         info["server.type"] = "Python/gtk/x11"
         try:
-            from xpra.x11.gtk_x11.composite import CompositeHelper
+            from xpra.x11.gtk2.composite import CompositeHelper
             info["server.XShm"] = CompositeHelper.XShmEnabled
         except:
             pass
@@ -191,14 +200,13 @@ class X11ServerBase(GTKServerBase):
                 info["server.randr.options"] = list(reversed(sorted(sizes)))
         except:
             pass
-        try:
-            fx = self.find_fakeXinerama()
-        except:
-            fx = None
-        info["server.fakeXinerama"] = self.fake_xinerama and bool(fx)
-        info["server.libfakeXinerama"] = fx or ""
+        info["server.fakeXinerama"] = self.fake_xinerama and bool(self.libfakeXinerama_so)
+        info["server.libfakeXinerama"] = self.libfakeXinerama_so or ""
         #this is added here because the server keyboard config doesn't know about "keys_pressed"..
         info["keyboard.state.keys_pressed"] = list(self.keys_pressed.keys())
+        info["keyboard.fast-switching"] = True
+        info["server.Xkb"] = X11Keyboard.hasXkb()
+        info["server.XTest"] = X11Keyboard.hasXTest()
         return info
 
     def get_window_info(self, window):
@@ -231,7 +239,7 @@ class X11ServerBase(GTKServerBase):
                 keylog("reenable_keymap_changes(%s)", args)
                 self.keymap_changing = False
                 self._keys_changed()
-            gobject.idle_add(reenable_keymap_changes)
+            self.idle_add(reenable_keymap_changes)
 
 
     def _clear_keys_pressed(self):
@@ -255,7 +263,8 @@ class X11ServerBase(GTKServerBase):
             with xsync:
                 cursor_data = X11Keyboard.get_cursor_image()
         except Exception as e:
-            cursorlog.error("Error getting cursor data: %s", e)
+            cursorlog.error("Error getting cursor data:")
+            cursorlog.error(" %s", e)
             return None, []
         if cursor_data is None:
             cursorlog("get_cursor_data() failed to get cursor image")
@@ -272,48 +281,55 @@ class X11ServerBase(GTKServerBase):
 
 
     def get_max_screen_size(self):
-        from xpra.x11.gtk_x11.window import MAX_WINDOW_SIZE
-        max_w, max_h = gtk.gdk.get_default_root_window().get_size()
+        from xpra.x11.gtk2.window import MAX_WINDOW_SIZE
+        max_w, max_h = self.root_window.get_size()
         sizes = RandR.get_screen_sizes()
         if self.randr and len(sizes)>=1:
             for w,h in sizes:
                 max_w = max(max_w, w)
                 max_h = max(max_h, h)
         if max_w>MAX_WINDOW_SIZE or max_h>MAX_WINDOW_SIZE:
-            log.warn("maximum size is very large: %sx%s, you may encounter window sizing problems", max_w, max_h)
+            screenlog.warn("maximum size is very large: %sx%s, you may encounter window sizing problems", max_w, max_h)
+        screenlog("get_max_screen_size()=%s", (max_w, max_h))
         return max_w, max_h
 
 
     def set_best_screen_size(self):
         #return ServerBase.set_best_screen_size(self)
         """ sets the screen size to use the largest width and height used by any of the clients """
-        root_w, root_h = gtk.gdk.get_default_root_window().get_size()
+        root_w, root_h = self.root_window.get_size()
         if not self.randr:
             return root_w, root_h
         max_w, max_h = 0, 0
         sss = self._server_sources.values()
+        if len(sss)>1:
+            screenlog.info("screen used by %i clients:", len(sss))
         for ss in sss:
             client_size = ss.desktop_size
             if not client_size:
-                continue
-            if ss.screen_sizes and len(sss)>1:
-                log.info("* %s:", ss.uuid)
-            w, h = client_size
-            max_w = max(max_w, w)
-            max_h = max(max_h, h)
-        log("maximum client resolution is %sx%s (current server resolution is %sx%s)", max_w, max_h, root_w, root_h)
+                size = "unknown"
+            else:
+                w, h = client_size
+                size = "%ix%i" % (w, h)
+                max_w = max(max_w, w)
+                max_h = max(max_h, h)
+            if len(sss)>1:
+                screenlog.info("* %s: %s", ss.uuid, size)
+        screenlog("maximum client resolution is %sx%s (current server resolution is %sx%s)", max_w, max_h, root_w, root_h)
         if max_w>0 and max_h>0:
             return self.set_screen_size(max_w, max_h)
         return  root_w, root_h
 
     def set_screen_size(self, desired_w, desired_h):
-        root_w, root_h = gtk.gdk.get_default_root_window().get_size()
+        root_w, root_h = self.root_window.get_size()
+        if not self.randr:
+            return root_w,root_h
         if desired_w==root_w and desired_h==root_h and not self.fake_xinerama:
-            return    root_w,root_h    #unlikely: perfect match already!
+            return root_w,root_h    #unlikely: perfect match already!
         #clients may supply "xdpi" and "ydpi" (v0.15 onwards), or just "dpi", or nothing...
         xdpi = self.xdpi or self.dpi
         ydpi = self.ydpi or self.dpi
-        log("set_screen_size(%s, %s) xdpi=%s, ydpi=%s", desired_w, desired_h, xdpi, ydpi)
+        screenlog("set_screen_size(%s, %s) xdpi=%s, ydpi=%s", desired_w, desired_h, xdpi, ydpi)
         if xdpi<=0 or ydpi<=0:
             #use some sane defaults: either the command line option, or fallback to 96
             #(96 is better than nothing, because we do want to set the dpi
@@ -335,15 +351,18 @@ class X11ServerBase(GTKServerBase):
                         hmm = max(hmm, s[4])
             if wmm>0 and hmm>0 and client_w>0 and client_h>0:
                 #calculate "real" dpi:
-                xdpi = int(client_w * 25.4 / wmm + 0.5)
-                ydpi = int(client_h * 25.4 / hmm + 0.5)
-                log("calculated DPI: %s x %s (from w: %s / %s, h: %s / %s)", xdpi, ydpi, client_w, wmm, client_h, hmm)
+                xdpi = iround(client_w * 25.4 / wmm)
+                ydpi = iround(client_h * 25.4 / hmm)
+                screenlog("calculated DPI: %s x %s (from w: %s / %s, h: %s / %s)", xdpi, ydpi, client_w, wmm, client_h, hmm)
         self.set_dpi(xdpi, ydpi)
 
         #try to find the best screen size to resize to:
         new_size = None
+        closest = {}
         for w,h in RandR.get_screen_sizes():
             if w<desired_w or h<desired_h:
+                distance = abs(w-desired_w)*abs(h-desired_h)
+                closest[distance] = (w, h)
                 continue            #size is too small for client
             if new_size:
                 ew,eh = new_size
@@ -351,16 +370,29 @@ class X11ServerBase(GTKServerBase):
                     continue        #we found a better (smaller) candidate already
             new_size = w,h
         if not new_size:
-            log.warn("resolution not found for %sx%s", desired_w, desired_h)
-            return  root_w, root_h
-        log("best resolution for client(%sx%s) is: %s", desired_w, desired_h, new_size)
+            screenlog.warn("Warning: no matching resolution found for %sx%s", desired_w, desired_h)
+            if len(closest)>0:
+                new_size = sorted(closest.items())[0]
+                screenlog.warn(" using %sx%s instead", *new_size)
+            else:
+                return  root_w, root_h
+        screenlog("best resolution for client(%sx%s) is: %s", desired_w, desired_h, new_size)
         #now actually apply the new settings:
         w, h = new_size
-        xinerama_changed = self.save_fakexinerama_config()
+        #fakeXinerama:
+        ui_clients = [s for s in self._server_sources.values() if s.ui_client]
+        source = None
+        screen_sizes = []
+        if len(ui_clients)==1:
+            source = ui_clients[0]
+            screen_sizes = source.screen_sizes
+        else:
+            screenlog("fakeXinerama can only be enabled for a single client (found %s)" % len(ui_clients))
+        xinerama_changed = save_fakeXinerama_config(self.fake_xinerama and len(ui_clients)==1, source, screen_sizes)
         #we can only keep things unchanged if xinerama was also unchanged
         #(many apps will only query xinerama again if they get a randr notification)
         if (w==root_w and h==root_h) and not xinerama_changed:
-            log.info("best resolution matching %sx%s is unchanged: %sx%s", desired_w, desired_h, w, h)
+            screenlog.info("best resolution matching %sx%s is unchanged: %sx%s", desired_w, desired_h, w, h)
             return  root_w, root_h
         try:
             if (w==root_w and h==root_h) and xinerama_changed:
@@ -375,131 +407,54 @@ class X11ServerBase(GTKServerBase):
                         #(so we can choose the closest resolution)
                         temp[abs((tw*th) - (w*h))] = (tw, th)
                 if len(temp)==0:
-                    log.warn("cannot find a temporary resolution for Xinerama workaround!")
+                    screenlog.warn("cannot find a temporary resolution for Xinerama workaround!")
                 else:
                     k = sorted(temp.keys())[0]
                     tw, th = temp[k]
-                    log.info("temporarily switching to %sx%s as a Xinerama workaround", tw, th)
+                    screenlog.info("temporarily switching to %sx%s as a Xinerama workaround", tw, th)
                     RandR.set_screen_size(tw, th)
-            log("calling RandR.set_screen_size(%s, %s)", w, h)
+            screenlog("calling RandR.set_screen_size(%s, %s)", w, h)
             with xsync:
                 RandR.set_screen_size(w, h)
-            log("calling RandR.get_screen_size()")
+            screenlog("calling RandR.get_screen_size()")
             root_w, root_h = RandR.get_screen_size()
-            log("RandR.get_screen_size()=%s,%s", root_w, root_h)
-            log("RandR.get_vrefresh()=%s", RandR.get_vrefresh())
+            screenlog("RandR.get_screen_size()=%s,%s", root_w, root_h)
+            screenlog("RandR.get_vrefresh()=%s", RandR.get_vrefresh())
             if root_w!=w or root_h!=h:
-                log.error("odd, failed to set the new resolution, "
-                          "tried to set it to %sx%s and ended up with %sx%s", w, h, root_w, root_h)
+                screenlog.error("odd, failed to set the new resolution, "
+                                "tried to set it to %sx%s and ended up with %sx%s", w, h, root_w, root_h)
             else:
                 msg = "server virtual display now set to %sx%s" % (root_w, root_h)
                 if desired_w!=root_w or desired_h!=root_h:
                     msg += " (best match for %sx%s)" % (desired_w, desired_h)
-                log.info(msg)
+                screenlog.info(msg)
             def show_dpi():
                 wmm, hmm = RandR.get_screen_size_mm()      #ie: (1280, 1024)
-                actual_xdpi = int(root_w * 25.4 / wmm + 0.5)
-                actual_ydpi = int(root_h * 25.4 / hmm + 0.5)
-                if actual_xdpi==xdpi and actual_ydpi==ydpi:
-                    log.info("DPI set to %s x %s", xdpi, ydpi)
+                screenlog("RandR.get_screen_size_mm=%s,%s", wmm, hmm)
+                actual_xdpi = iround(root_w * 25.4 / wmm)
+                actual_ydpi = iround(root_h * 25.4 / hmm)
+                if abs(actual_xdpi-xdpi)<=1 and abs(actual_ydpi-ydpi)<=1:
+                    screenlog.info("DPI set to %s x %s", xdpi, ydpi)
                 else:
                     #should this be a warning:
-                    l = log.info
+                    l = screenlog.info
                     maxdelta = max(abs(actual_xdpi-xdpi), abs(actual_ydpi-ydpi))
                     if maxdelta>=10:
                         l = log.warn
                     l("DPI set to %s x %s (wanted %s x %s)", actual_xdpi, actual_ydpi, xdpi, ydpi)
                     if maxdelta>=10:
                         l(" you may experience scaling problems, such as huge or small fonts, etc")
-                        l(" to fix this issue, try the dpi switch, or use a patched Xdummy driver")
+                        l(" to fix this issue, try the dpi switch, or use a patched Xorg dummy driver")
             #show dpi via idle_add so server has time to change the screen size (mm)
             self.idle_add(show_dpi)
         except Exception as e:
-            log.error("ouch, failed to set new resolution: %s", e, exc_info=True)
+            screenlog.error("ouch, failed to set new resolution: %s", e, exc_info=True)
         return  root_w, root_h
 
-    def save_fakexinerama_config(self):
-        """ returns True if the fakexinerama config was modified """
-        xinerama_files = self.get_xinerama_filenames()
-        def delfile(msg):
-            if msg:
-                log.warn(msg)
-            self.cleanup_fakexinerama()
-            oldconf = self.current_xinerama_config
-            self.current_xinerama_config = None
-            return oldconf is not None
-        if not self.fake_xinerama:
-            return delfile(None)
-        ui_client_count = len([True for s in self._server_sources.values() if s.ui_client])
-        if ui_client_count!=1:
-            return delfile("fakeXinerama can only be enabled for a single client (found %s)" % ui_client_count)
-        source = self._server_sources.values()[0]
-        ss = source.screen_sizes
-        if len(ss)==0:
-            return delfile("cannot save fake xinerama settings: no display found")
-        if len(ss)>1:
-            return delfile("cannot save fake xinerama settings: more than one display found")
-        if len(ss)==2 and type(ss[0])==int and type(ss[1])==int:
-            #just WxH, not enough display information
-            return delfile("cannot save fake xinerama settings: missing display data from client %s" % source)
-        display_info = ss[0]
-        if len(display_info)<10:
-            return delfile("cannot save fake xinerama settings: incomplete display data from client %s" % source)
-        #display_name, width, height, width_mm, height_mm, \
-        #monitors, work_x, work_y, work_width, work_height = s[:11]
-        monitors = display_info[5]
-        if len(monitors)>=10:
-            return delfile("cannot save fake xinerama settings: too many monitors! (%s)" % len(monitors))
-        #generate the file data:
-        from xpra import __version__
-        data = ["# file generated by xpra %s for display %s" % (__version__, os.environ.get("DISPLAY")),
-                "# %s monitors:" % len(monitors),
-                "%s" % len(monitors)]
-        #the new config (numeric values only)
-        config = [len(monitors)]
-        for i, m in enumerate(monitors):
-            if len(m)<7:
-                return delfile("cannot save fake xinerama settings: incomplete monitor data for monitor: %s" % m)
-            plug_name, x, y, width, height, wmm, hmm = m[:7]
-            data.append("# %s (%smm x %smm)" % (prettify_plug_name(plug_name, "monitor %s" % i), wmm, hmm))
-            data.append("%s %s %s %s" % (x, y, width, height))
-            config.append((x, y, width, height))
-        if self.current_xinerama_config==config:
-            #we assume that no other process is going to overwrite the deprecated .fakexinerama
-            log("fake xinerama config unchanged")
-            return False
-        self.current_xinerama_config = config
-        data.append("")
-        contents = "\n".join(data)
-        for filename in xinerama_files:
-            try:
-                with open(filename, 'wb') as f:
-                    f.write(contents)
-            except Exception as e:
-                log.warn("error writing fake xinerama file %s: %s", filename, e)
-        log("saved %s monitors to fake xinerama files: %s", len(monitors), xinerama_files)
-        return True
 
-    def get_xinerama_filenames(self):
-        return [
-                #the new fakexinerama file:
-                os.path.expanduser("~/.%s-fakexinerama" % os.environ.get("DISPLAY")),
-                #compat file for "old" version found on github:
-                os.path.expanduser("~/.fakexinerama"),
-               ]
-
-    def cleanup_fakexinerama(self):
-        for f in self.get_xinerama_filenames():
-            try:
-                if os.path.exists(f):
-                    log=("cleanup_fakexinerama() deleting fake xinerama file '%s'", f)
-                    os.unlink(f)
-            except Exception as e:
-                log.warn("cleanup_fakexinerama() error deleting fake xinerama file '%s': %s", f, e)
-
-    def cleanup(self, *args):
-        GTKServerBase.cleanup(self)
-        self.cleanup_fakexinerama()
+    def do_cleanup(self, *args):
+        GTKServerBase.do_cleanup(self)
+        cleanup_fakeXinerama()
 
 
     def _process_server_settings(self, proto, packet):
@@ -535,20 +490,18 @@ class X11ServerBase(GTKServerBase):
         #(this is called within an xswallow context)
         mouselog("move_pointer(%s, %s)", wid, pos)
         x, y = pos
-        n = gtk.gdk.display_get_default().get_default_screen().get_number()
-        X11Keyboard.xtest_fake_motion(n, x, y)
+        X11Keyboard.xtest_fake_motion(self.screen_number, x, y)
 
     def _process_mouse_common(self, proto, wid, pointer, modifiers):
-        ss = self._server_sources.get(proto)
-        if ss is None:
-            return
-        pos = gtk.gdk.get_default_root_window().get_pointer()[:2]
+        pos = self.root_window.get_pointer()[:2]
         if pos!=pointer:
             with xswallow:
                 self._move_pointer(wid, pointer)
-        ss.make_keymask_match(modifiers)
-        if wid==self.get_focus():
-            ss.user_event()
+        ss = self._server_sources.get(proto)
+        if ss:
+            ss.make_keymask_match(modifiers)
+            if wid==self.get_focus():
+                ss.user_event()
 
     def _process_button_action(self, proto, packet):
         ss = self._server_sources.get(proto)
