@@ -14,12 +14,15 @@ and
 It is used by xpra.x11.gtk_x11.prop
 """
 
-import sys
+import sys, os
 import struct
 from xpra.log import Logger
 log = Logger("x11", "xsettings")
 
 
+DEBUG_XSETTINGS = os.environ.get("XPRA_XSETTINGS_DEBUG", "0")=="1"
+
+
 if sys.version > '3':
     unicode = str           #@ReservedAssignment
 
@@ -48,13 +51,20 @@ if sys.version>='3':
     long = int              #@ReservedAssignment
 
 
+XSETTINGS_CACHE = {}
 def get_settings(disp, d):
+    global XSETTINGS_CACHE
     #parse xsettings according to
     #http://standards.freedesktop.org/xsettings-spec/xsettings-spec-0.5.html
     assert len(d)>=12, "_XSETTINGS_SETTINGS property is too small: %s" % len(d)
-    log("get_settings(%s)", list(d))
+    if DEBUG_XSETTINGS:
+        log("get_settings(%s)", list(d))
     byte_order, _, _, _, serial, n_settings = struct.unpack("=BBBBII", d[:12])
-    log("get_settings(..) found byte_order=%s (local is %s), serial=%s, n_settings=%s", byte_order, get_local_byteorder(), serial, n_settings)
+    cache = XSETTINGS_CACHE.get(disp)
+    log("get_settings(..) found byte_order=%s (local is %s), serial=%s, n_settings=%s, cache(%s)=%s", byte_order, get_local_byteorder(), serial, n_settings, disp, cache)
+    if cache and cache[0]==serial:
+        log("get_settings(..) returning value from cache")
+        return cache
     settings = []
     pos = 12
     while n_settings>len(settings) and len(d)>0:
@@ -69,7 +79,8 @@ def get_settings(disp, d):
         assert len(d)>=pos+4, "not enough data (%s bytes) to extract serial (4 bytes needed)" % (len(d)-pos)
         last_change_serial = struct.unpack("=I", d[pos:pos+4])[0]
         pos += 4
-        log("get_settings(..) found property %s of type %s, serial=%s", prop_name, XSettingsNames.get(setting_type, "INVALID!"), last_change_serial)
+        if DEBUG_XSETTINGS:
+            log("get_settings(..) found property %s of type %s, serial=%s", prop_name, XSettingsNames.get(setting_type, "INVALID!"), last_change_serial)
         #extract value:
         if setting_type==XSettingsTypeInteger:
             assert len(d)>=pos+4, "not enough data (%s bytes) to extract int (4 bytes needed)" % (len(d)-pos)
@@ -90,9 +101,12 @@ def get_settings(disp, d):
             log.error("invalid setting type: %s, cannot continue parsing XSETTINGS!", setting_type)
             break
         setting = setting_type, prop_name, value, last_change_serial
-        log("get_settings(..) %s -> %s", list(d[istart:pos]), setting)
+        if DEBUG_XSETTINGS:
+            log("get_settings(..) %s -> %s", list(d[istart:pos]), setting)
         settings.append(setting)
     log("get_settings(..) settings=%s", settings)
+    if disp:
+        XSETTINGS_CACHE[disp] = (serial, settings)
     return  serial, settings
 
 def set_settings(disp, d):
@@ -155,7 +169,6 @@ def main():
                 x.enable_debug()
 
         #naughty, but how else can I hook this up?
-        import os
         if os.name!="posix":
             print("xsettings require a posix OS")
             return 1
